//
//  Graph.cpp
//  QDMGraphsDouble
//
//  Created by Dominik Ixert on 31.01.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#include "Graph.h"
#include "functors.h"
#include "functions.h"
#include "mathFunctions.h"
#include "orthogonalization.h"
#include "nr3.h"
#include "gaussj.h"
#include "eigen_sym.h"
#include "svd.h"
#include "stepper.h"
#include "stepperdopr5.h"
#include <map>

using namespace std;

StateDouble Graph::hubbardSinglet=(StateDouble(4,"1001")-StateDouble(4,"0110"))/Doub(sqrt(2));
StateDouble Graph::hubbardTriplet0=(StateDouble(4,"1001")+StateDouble(4,"0110"))/Doub(sqrt(2));

void Graph::createEigenstatesFromED(unsigned int matrixSize, MatDoub &evec, StateVector &vs, StateVector &eigenstates){
	eigenstates.clear();
	for (unsigned int i = 0; i < matrixSize; i++) {
		StateDouble tempState = evec[0][i] * vs[0];
		for (unsigned int j = 1; j < matrixSize; j++) {
			tempState += evec[j][i] * vs[j];
		}
		eigenstates.push_back(tempState);
	}
    /*cout << "overlap of hubbard ground state with start singlets:\n";
    for(unsigned int i=0;i<basis.size();i++){
        cout << basis.get(i)*eigenstates[0] << "\n";
    }*/
}

bool Graph::performCUTQP(Doub stepEnd, Doub stepWidth, Doub epsAbs, Doub epsRel, bool createResFile, stringstream &ssResFile, bool abortAtMin){
    // performs the CUT for the QP-Generator
	// stepEnd, stepWidth, epsAbs and epsRel are parameters for the RK4-Method (see NR3)
	if(H.size()==0){
		cerr << "ERROR: performCutQP - H is empty\n";
		return false;
	}
	unsigned int basisSize=basis.size();
	unsigned int nn=basisSize*basisSize;
	bool converged=true;
    
    // set QP-Number for CUT if q is not already specified
	vector<unsigned int> q;
	if(QPNumbers.size()==0){
		for(unsigned int i=0;i<lls;i++){
			q.push_back(0);
		}
		for(unsigned int i=lls;i<basisSize;i++){
			q.push_back(1);
		}
	}
	else{
		q=QPNumbers;
	}
    
    unsigned int nn2=nn;
	if(unitaryCUT==1){
		nn2*=2;
        if((UCUT.size()!=nn)){
            UCUT.assign(nn,0);
            for(unsigned int i=0;i<basisSize;i++){
                UCUT[i*basisSize+i]=1;
            }
        }
	}
    
	VecDoub y(nn2);
	for(unsigned int i=0;i<basisSize;i++){
		for(unsigned int j=0;j<basisSize;j++){
			y[i*basisSize+j]=H[i*basisSize+j];
			if(unitaryCUT==1){
				y[nn+i*basisSize+j]=UCUT[i*basisSize+j];
			}
		}
	}
    
	// solve CUT    
	Doub stepStart=0;
	funcQP f(basisSize,q,unitaryCUT);
	VecDoub dydx(nn);
	StepperDopr5<funcQP> stepper(y,dydx,stepStart,epsAbs,epsRel,false);
	f(stepStart,y,dydx);
	stepper.step(stepWidth,f);
	Doub res=f.offBlockDiagonalElements(y);
    fstream resFile;
    if(createResFile){
        resFile.open(ssResFile.str().c_str(),fstream::out);
        resFile << setprecision(16);
        resFile << stepStart << " " << res << "\n";
        resFile.close();
    }
	while (res > zeroResEPS) {
		stepper.step(stepper.hnext,f);
		//cout << "stepper.hnext=" << stepper.hnext << "\n";
		Doub oldRes=res;
		res = f.offBlockDiagonalElements(y);
        if(createResFile){
            resFile.open(ssResFile.str().c_str(),fstream::out | fstream::app);
            resFile << stepStart << " " << res << "\n";
            resFile.close();
        }
        if((res>oldRes) && abortAtMin){
        	cout << "ROD is rising!\n";
        	cout << "Aborting!\n";
        	break;
        }
		if (stepStart == stepEnd) {
			cout << "stepStart==stepEnd\n";
			converged = false;
			break;
		}
	}
    
    // copy matrices
	Heff.resize(nn);
    
	for(unsigned int i=0;i<basisSize;i++){
		for(unsigned int j=0;j<basisSize;j++){
			Heff[i*basisSize+j]=y[i*basisSize+j];
			if(unitaryCUT==1){
				UCUT[i*basisSize+j]=y[nn+i*basisSize+j];
			}
		}
	}
    
	return converged;
}

bool Graph::performCUTWegner(Doub stepEnd, Doub stepWidth, Doub epsAbs, Doub epsRel, bool createResFile, stringstream &ssResFile){
    // performs the CUT for the Wegner-Generator
	// stepEnd, stepWidth, epsAbs and epsRel are parameters for the RK4-Method (see NR3)
    // calculates U if unitaryCUT==true
	if(H.size()==0){
		cerr << "ERROR: performCutWegner - H is empty\n";
		return false;
	}
	unsigned int basisSize=basis.size();
	unsigned int nn=basisSize*basisSize;
	bool converged=true;
    
    unsigned int nn2=nn;
	if(unitaryCUT==1){
		nn2*=2;
        if((UCUT.size()!=nn)){
            UCUT.assign(nn,0);
            for(unsigned int i=0;i<basisSize;i++){
                    UCUT[i*basisSize+i]=1;
            }
        }
	}

	VecDoub y(nn2);
	for(unsigned int i=0;i<basisSize;i++){
		for(unsigned int j=0;j<basisSize;j++){
			y[i*basisSize+j]=H[i*basisSize+j];
			if(unitaryCUT==1){
				y[nn+i*basisSize+j]=UCUT[i*basisSize+j];
			}
		}
	}

	// solve CUT
	Doub stepStart=0;
	funcWegner f(basisSize,lls,unitaryCUT);
	VecDoub dydx(nn2);
	StepperDopr5<funcWegner> stepper(y,dydx,stepStart,epsAbs,epsRel,false);
	f(stepStart,y,dydx);
	stepper.step(stepWidth,f);
	Doub res=f.offBlockDiagonalElements(y);
    fstream resFile;
    if(createResFile){
        resFile.open(ssResFile.str().c_str(),fstream::out);
        resFile << stepStart << " " << res << "\n";
        resFile.close();
    }
	while (res > zeroResEPS) {
		stepper.step(stepper.hnext,f);
		res = f.offBlockDiagonalElements(y);
        if(createResFile){
            resFile.open(ssResFile.str().c_str(),fstream::out | fstream::app);
            resFile << stepStart << " " << res << "\n";
            resFile.close();
        }
		if (stepStart == stepEnd) {
			cout << "stepStart==stepEnd\n";
			converged = false;
			break;
		}
	}
    
	// copy matrices
	Heff.resize(nn);
    
	for(unsigned int i=0;i<basisSize;i++){
		for(unsigned int j=0;j<basisSize;j++){
			Heff[i*basisSize+j]=y[i*basisSize+j];
			if(unitaryCUT==1){
				UCUT[i*basisSize+j]=y[nn+i*basisSize+j];
			}
		}
	}
    
	return converged;
}


Graph::Graph():n(1),order(1),graphNumber(1),unitaryCUT(0),withoutSqrt(false),lls(0){
}
Graph::Graph(unsigned int nVal, unsigned int orderVal, unsigned int graphNumberVal, vector<Doub> const &tVal, unsigned int unitaryCUTVal,bool withoutSqrtVal):n(nVal),order(orderVal),graphNumber(graphNumberVal),t(tVal),unitaryCUT(unitaryCUTVal),withoutSqrt(withoutSqrtVal),lls(0){
}
Graph::Graph(unsigned int nVal, unsigned int orderVal, unsigned int graphNumberVal, vector<Doub> const &tVal, list<CouplingStruct> const &couplingsContVal, unsigned int unitaryCUTVal, bool withoutSqrtVal):n(nVal),order(orderVal),graphNumber(graphNumberVal),t(tVal),couplingsCont(couplingsContVal),unitaryCUT(unitaryCUTVal),withoutSqrt(withoutSqrtVal),lls(0){
}
Graph::Graph(const Graph &g){
	n=g.n;
	order=g.order;
	graphNumber=g.graphNumber;
	t=g.t;
    couplingsCont=g.couplingsCont;
    basis=g.basis;
	tempBasis=g.tempBasis;
	lls=g.lls;
    initialStates=g.initialStates;
    withoutSqrt=g.withoutSqrt;
    QPNumbers=g.QPNumbers;
    unitaryCUT=g.unitaryCUT;
    H=g.H;
    Heff=g.Heff;
    UCUT=g.UCUT;
    Sorth=g.Sorth;
}
Graph& Graph::operator=(const Graph &g){
    if(this!=&g){
    	n=g.n;
    	order=g.order;
    	graphNumber=g.graphNumber;
        t=g.t;
        couplingsCont=g.couplingsCont;
        basis=g.basis;
        tempBasis=g.tempBasis;
        lls=g.lls;
        initialStates=g.initialStates;
        withoutSqrt=g.withoutSqrt;
        QPNumbers=g.QPNumbers;
        unitaryCUT=g.unitaryCUT;
        H=g.H;
        Heff=g.Heff;
        UCUT=g.UCUT;
        Sorth=g.Sorth;
    }
    return *this;
}
Graph::~Graph(){
}

StateDouble Graph::hubbardHamiltonian(StateDouble const &a, Doub const &U){
    return hubbardHamiltonian(a,U,t);
}

StateDouble Graph::hubbardHamiltonian(StateDouble const &a, Doub const &U, vector<Doub> const &tVal){
	// applies the Hamiltonian to the State a and returns the resulting state
	StateDouble c(2*n);
	// run over all coupled sites
	list<CouplingStruct>::iterator it;
	for(it=couplingsCont.begin();it!=couplingsCont.end();it++){
		unsigned int f=it->from;
		unsigned int g=it->to;
		int coupling=it->coupling;
		for(unsigned int sigma=0;sigma<=1;sigma++){
			// f->g
			c-=tVal[coupling]*(a.ccDagger(2*f+sigma,2*g+sigma));
			// g->f
			c-=tVal[coupling]*(a.ccDagger(2*g+sigma,2*f+sigma));
		}
	}
	// run over all sites for double occupancies
	for(unsigned int k=0;k<n;k++){
		c += U*(a.nnState(2*k));
	}

	return c;
}

void Graph::createSingletCovering(list<list<pair<unsigned int,unsigned int> > > const &dimerCont, unsigned int orthAlg, bool unorthInitialStates){
	// creates the singlet covering specified in dimerCont
    // the singlets are orthonormnalized with the algorithm specified in orthAlg
	vector<bool> vb(2*n,false);
	basis.clear();

	// create all states defined in dimerCovering_o1g1.config
	unsigned int pos[4];
	for(list<list<pair<unsigned int,unsigned int> > >::const_iterator it1=dimerCont.begin();it1!=dimerCont.end();it1++){
		StateDouble sTemp(vb); //create vacuum state
		for(list<pair<unsigned int,unsigned int> >::const_iterator it2=it1->begin();it2!=it1->end();it2++){
			pos[0]=2*(it2->first);
			pos[1]=pos[0]+1;
			pos[2]=2*(it2->second);
			pos[3]=pos[2]+1;
			sTemp=insertState(sTemp, hubbardSinglet, pos);
		}
		basis.push_back(sTemp);
	}
	lls=basis.size();

	// save initialStates if unorthInitialStates is true
    if(unorthInitialStates){
    	initialStates=basis;
    }

	// orthonormalize
    switch(orthAlg){
        case 1:{
            // gram-schmidt
            for(unsigned int i=1;i<lls;i++){
            	//cout << "i=" << i << "\n";
                for(unsigned int j = 0; j < i; j++) {
                    basis[i] -=(basis[i] * basis[j]) * basis[j];
                }
                basis[i]/=sqrt(basis[i].normQuad());
            }
            break;
        }
        case 2:{
            // overlap matrix
        	unsigned int SoverSize=lls;
        	if(SoverSize>0){
        		MatDoub Sover(SoverSize,SoverSize);
        		for(unsigned int i=0;i<SoverSize;i++){
        			Sover[i][i]=basis[i]*basis[i];
        			for(unsigned int j=0;j<i;j++){
        				Sover[i][j]=basis[i]*basis[j];
        			}
        		}
            
        		// calculate eigenvalues
        		Symmeig eigSys(Sover);
        		invertSorting(eigSys.d,eigSys.z);
            
        		// calculate square root of inverse overlap matrix
        		for(unsigned int i=0;i<SoverSize;i++){
        			for(unsigned int j=0;j<SoverSize;j++){
        				Doub temp=0;
        				for(unsigned int k=0;k<SoverSize;k++){
        					temp+=eigSys.z[i][k]*eigSys.z[j][k]/sqrt(eigSys.d[k]);
        				}
        				Sover[i][j]=temp;
        			}
        		}
            
                // save Sover in Sorth
                Sorth.assign(lls*lls,0.0);
                for(unsigned int i=0;i<lls;i++){
                	for(unsigned int j=0;j<lls;j++){
                		Sorth[i*lls+j]=Sover[i][j];
                	}
                }
                StateVector basis2;
                for(unsigned int i=0;i<SoverSize;i++){
                	StateDouble sTemp=Sover[i][0]*basis[0];
                	for(unsigned int j=1;j<SoverSize;j++){
                		sTemp+=Sover[i][j]*basis[j];
                	}
                	basis2.push_back(sTemp);
                }
                basis.swap(basis2);
            }
            break;
        }
    }
}

void Graph::createDimerCovering(list<list<pair<unsigned int,unsigned int> > > const &dimerCont, unsigned int qp, unsigned int orthAlg, bool unorthInitialStates){
	// creates the dimer covering specified in dimerCont where all combinations with the specified qp are used TODO: at the moment only qp=1 works! qp>1 must be implemented!
    // the coverings are orthonormnalized with the algorithm specified in orthAlg
	vector<bool> vb(2*n,false);
	basis.clear();

	// create all states defined in dimerCovering_o1g1.config
	unsigned int pos[4];
	for(list<list<pair<unsigned int,unsigned int> > >::const_iterator it1=dimerCont.begin();it1!=dimerCont.end();it1++){
		unsigned int listSize=it1->size();
		for(unsigned int qpPosition=0;qpPosition<listSize;qpPosition++){
			StateDouble sTemp(vb); //create vacuum state
			unsigned int actualPosition=0;
			for(list<pair<unsigned int,unsigned int> >::const_iterator it2=it1->begin();it2!=it1->end();it2++){
				pos[0]=2*(it2->first);
				pos[1]=pos[0]+1;
				pos[2]=2*(it2->second);
				pos[3]=pos[2]+1;
				if(qpPosition==actualPosition){
					sTemp=insertState(sTemp, hubbardTriplet0, pos);
				}
				else{
					sTemp=insertState(sTemp, hubbardSinglet, pos);
				}
				actualPosition++;
			}
			basis.push_back(sTemp);
		}
	}
	lls=basis.size();

	// save initialStates if unorthInitialStates is true
    if(unorthInitialStates){
    	initialStates=basis;
    }

	// orthonormalize
    switch(orthAlg){
        case 1:{
            // gram-schmidt
            for(unsigned int i=1;i<lls;i++){
            	//cout << "i=" << i << "\n";
                for(unsigned int j = 0; j < i; j++) {
                    basis[i] -=(basis[i] * basis[j]) * basis[j];
                }
                basis[i]/=sqrt(basis[i].normQuad());
            }
            break;
        }
        case 2:{
            // overlap matrix
        	unsigned int SoverSize=lls;
        	if(SoverSize>0){
        		MatDoub Sover(SoverSize,SoverSize);
        		for(unsigned int i=0;i<SoverSize;i++){
        			Sover[i][i]=basis[i]*basis[i];
        			for(unsigned int j=0;j<i;j++){
        				Sover[i][j]=basis[i]*basis[j];
        			}
        		}

        		// calculate eigenvalues
        		Symmeig eigSys(Sover);
        		invertSorting(eigSys.d,eigSys.z);

        		// calculate square root of inverse overlap matrix
        		for(unsigned int i=0;i<SoverSize;i++){
        			for(unsigned int j=0;j<SoverSize;j++){
        				Doub temp=0;
        				for(unsigned int k=0;k<SoverSize;k++){
        					temp+=eigSys.z[i][k]*eigSys.z[j][k]/sqrt(eigSys.d[k]);
        				}
        				Sover[i][j]=temp;
        			}
        		}

                // save Sover in Sorth
                Sorth.assign(lls*lls,0.0);
                for(unsigned int i=0;i<lls;i++){
                	for(unsigned int j=0;j<lls;j++){
                		Sorth[i*lls+j]=Sover[i][j];
                	}
                }
                StateVector basis2;
                for(unsigned int i=0;i<SoverSize;i++){
                	StateDouble sTemp=Sover[i][0]*basis[0];
                	for(unsigned int j=1;j<SoverSize;j++){
                		sTemp+=Sover[i][j]*basis[j];
                	}
                	basis2.push_back(sTemp);
                }
                basis.swap(basis2);
            }
            break;
        }
    }
}

void Graph::createUpDownCovering(mpq_class const &SzVal, bool const &halfFilling){
	vector<vector<bool> > states;
    vector<bool> vbEmpty;
    states.push_back(vbEmpty);

	for(unsigned int i=0;i<n;i++){
		unsigned int size=states.size();
		unsigned int j=0;
		vector<vector<bool> > states2;
		for(unsigned int j=0;j<size;j++){
			vector<bool> temp=states[j];
			temp.push_back(true);
			temp.push_back(false);
			states2.push_back(temp);
			temp=states[j];
			temp.push_back(false);
			temp.push_back(true);
			states2.push_back(temp);
		}
		states.swap(states2);
	}

	for(unsigned int i=0;i<states.size();i++){
		unsigned int up=0;
		unsigned int down=0;
		mpq_class Sz=0;
		mpq_class oneHalf(1,2);
		for(unsigned int j=0;j<n;j++){
			if(states[i][2*j]){
				Sz+=oneHalf;
				up++;
	        }
	        if(states[i][2*j+1]){
	        	Sz-=oneHalf;
	        	down++;
	        }
		}
		bool SzProof=(SzVal == Sz);
		bool halfFillingProof=(halfFilling && (up+down==n)) || !halfFilling;
	    if(SzProof && halfFillingProof){
	    	StateDouble tempState(states[i]);
	        basis.push_back(tempState);
	    }
	}
	lls=basis.size();
}

void Graph::createCoveringFromBasis(StateVector const &basisVal){
	basis.clear();
	for(unsigned int i=0;i<basisVal.size();i++){
		basis.push_back(basisVal[i]);
	}
	lls=basis.size();
}

StateDouble Graph::getState(unsigned int i){
	// returns the i-th basis state
	if(i<basis.size()){
		return basis[i];
	}
	else{
		cerr << "Graph:getState: index out of range\n";
		return StateDouble();
	}
}
StateDouble Graph::getInitialState(unsigned int i){
	// returns the i-th initial basis state
	if(i<initialStates.size()){
		return initialStates[i];
	}
	else{
		cerr << "Graph:getInitialState: index out of range\n";
		return StateDouble();
	}
}

unsigned int Graph::size(){
	// returns the number of sites
	return n;
}
unsigned int Graph::basisSize(){
	// returns the basis size
	return basis.size();
}
unsigned int Graph::llsSize(){
	// returns the number of low-lying-singlets, i.e. the number of singlets before createBasis
	return lls;
}

Doub Graph::getH(unsigned int i, unsigned int j) const{
	// returns H_ij
    unsigned int basisSize=basis.size();
	if((i<basisSize) && (j<basisSize)){
		return H[i*basisSize+j];
	}
	else{
		cerr << "index out of range\n";
		return 0;
	}
}
Doub Graph::getHeff(unsigned int i, unsigned int j) const{
	// return Heff_ij after the CUT is performed
    unsigned int basisSize=basis.size();
	if((i<basisSize) && (j<basisSize)){
		return Heff[i*basisSize+j];
	}
	else{
		cerr << "index out of range\n";
		return 0;
	}
}
Doub Graph::getU(unsigned int i, unsigned int j) const{
    // return U_ij
    unsigned int basisSize=basis.size();
    if((i<basisSize) && (j<basisSize) && unitaryCUT){
        return UCUT[i*basisSize+j];
    }
    else{
        cerr << "index out of range\n";
        return 0;
    }
}
Doub Graph::getSorth(unsigned int i, unsigned int j) const{
	// return Sorth_ij
	unsigned int basisSize=basis.size();
	if((i<basisSize) && (j<basisSize) && Sorth.size()>0){
		return Sorth[i*basisSize+j];
	}
	else{
		cerr << "index out of range\n";
		return 0;
	}
}
void Graph::addHeff(unsigned int i, unsigned int j, Doub HeffVal){
	unsigned int basisSize=basis.size();
	if((i<basisSize) && (j<basisSize)){
		Heff[i*basisSize+j]+=HeffVal;
	}
	else{
		cerr << "index out of range\n";
	}
}

bool Graph::performCUT(Doub stepEnd, Doub stepWidth, Doub epsAbs, Doub epsRel, bool createResFile, stringstream &ssResFile, unsigned int generator, bool abortAtMin){
	// performs the CUT
	// stepEnd, stepWidth, epsAbs and epsRel are parameters for the RK4-Method (see NR3)
    // if generator=1 the QP-Generator is used; if generator=2 the Wegner-Generator is used
    switch(generator){
        case 1: 
            return performCUTQP(stepEnd, stepWidth, epsAbs, epsRel, createResFile, ssResFile, abortAtMin);
            break;
        case 2:
            return performCUTWegner(stepEnd,stepWidth,epsAbs,epsRel, createResFile, ssResFile);
            break;
        default:
            return false;
            break;
    }
}

unsigned int Graph::createBasis(unsigned int cutoff,unsigned int i,unsigned int appliedH){
	// calculates a singlet-basis starting with the specified dimer coverings
	// cutoff is a truncation parameter which specifies how often H is applied to generate the singlets; if cutoff=0 no truncation is performed
	// this function can used for iterative generation of singlets (i.e: convergence in dependence of cutoff); to this end one needs to specify i and appliedH of the last step
    // basisOrthAlg specifies the orthogonalization algorithm used
	unsigned int size = basis.size();
	unsigned int actualCutoff = size;

	while(i<basis.size()){
		StateDouble s1=hubbardHamiltonian(basis[i],1);
        if(withoutSqrt){
        	MGSWithoutSqrt(s1, basis);
        }
        else{
        	MGS(s1,basis);
        }
		i++;
		if (i == actualCutoff) {
			actualCutoff = basis.size();
			appliedH++;
			if (appliedH == cutoff) {
				break;
			}
		}
	}
	size=basis.size();

	/*
	cout << "test orthogonality:\n";
	for(unsigned int j=0;j<size;j++){
		StateDouble state_j=basis.get(j);
		for(unsigned int k=0;k<size;k++){
			StateDouble state_k=basis.get(k);
			cout << state_j*state_k << " ";
		}
		cout << "\n";
	}
	*/
    
	return i;
}

unsigned int Graph::createBasisForUpDownStates(unsigned int cutoff,unsigned int i,unsigned int appliedH){
	set<vector<bool> > svb;
	unsigned int actualCutoff = basis.size();
	for(unsigned int j=0;j<basis.size();j++){
		svb.insert((basis[j].getSingleStates())[0]);
	}
	while(i<basis.size()){
		vector<Doub> tVal;
		tVal.push_back(1.0);
		tVal.push_back(1.0);
		StateDouble s1=hubbardHamiltonian(basis[i],0,tVal);
		vector<vector<bool> > vvbTemp=s1.getSingleStates();
		for(unsigned int j=0;j<vvbTemp.size();j++){
			if(svb.find(vvbTemp[j])==svb.end()){
				svb.insert(vvbTemp[j]);
				basis.push_back(StateDouble(vvbTemp[j]));
			}
		}
		i++;
		if (i == actualCutoff) {
			actualCutoff = basis.size();
			appliedH++;
			if (appliedH == cutoff) {
				break;
			}
		}
	}
	return i;
}

bool Graph::calculateLanczosGS(unsigned int maxCutoff, Doub epsAbs){
	bool converged=false;

	unsigned int cutoff=1;
	unsigned int i=basis.size();
	if(i==0){
		return false;
	}
	else{
		i--;
	}

	vector<Doub> a;
	vector<Doub> b;

    b.push_back(0);
	// first step outside of loop
	StateDouble w=hubbardHamiltonian(basis[i],1);
	a.push_back(w*basis[i]);
	w=w-a[i]*basis[i];
	Doub bTemp=w.normQuad();
	if(bTemp>zeroNormQuadEPS){
		b.push_back(sqrt(bTemp));
		basis.push_back(w/b.back());
		i++;
	}
	else{
		converged=true;
	}
	StateDouble s(2*n);
    Doub E0Conv=a[0];
	while((!converged) && (cutoff<=maxCutoff)){
		w=hubbardHamiltonian(basis[i],1);
		//cout << "w=" << w << "\n";
		Doub aTemp=w*basis[i];
		a.push_back(aTemp);
		w=w-aTemp*basis[i]-b[i]*basis[i-1];
		bTemp=w.normQuad();
        cout << "bTemp=" << bTemp << "\n";
		if(bTemp>zeroNormQuadEPS){
			unsigned int size=basis.size();
			Doub norm=0;
			for (unsigned int j = 0; j < size; j++) {
				w -=(w * basis[j]) * basis[j];
			}
			norm = w.normQuad();
		    cout << "normQuad=" << norm << "\n";
			if (norm > zeroNormQuadEPS) {
				w/=sqrt(norm);
		        bool isOrtho=true;
		        for(unsigned int j=0;j<size;j++){
		        	Doub orthValue=w*basis[j];
		        	cout << "<s|" << j << ">=" << orthValue << "\n";
		        	if(orthValue>zeroOrthoEPS){
		        		isOrtho=false;
		        		cout << "State not orthogonal and thus not added to basis!\n";
		            }
		        }
		        if(isOrtho){
		        	basis.push_back(w);
		        	b.push_back(sqrt(bTemp));
		        }
                else{
                    converged=true;
                }
			}
            else{
                converged=true;
            }
			//MGS(w,basis); // re-orthogonalize with MGS
		}
		else{
			converged=true;
		}

		// proof for convergence
        VecDoub aNR(a.size(),0.0);
        VecDoub bNR(a.size(),0.0);
        for(unsigned int j=0;j<a.size();j++){
            aNR[j]=a[j];
            bNR[j]=b[j];
            //cout << "aNR[j]=" << aNR[j] << "\n";
            //cout << "bNR[j]=" << bNR[j] << "\n";
        }
        // diagonalize matrix
        Symmeig TSymmEig(aNR,bNR); // use tridiagonality!
        if(fabs(E0Conv-TSymmEig.d[a.size()-1])<=epsAbs){
            converged=true;
        }

        // calculate ground-state eigenvector if converged
        if(converged){
            s=TSymmEig.z[0][a.size()-1]*basis[0];
            for(unsigned int j=1;j<TSymmEig.d.size();j++){
            	s+=TSymmEig.z[j][a.size()-1]*basis[j];
            }
            
        }
        E0Conv=TSymmEig.d[a.size()-1];
        
        i++;
	}
	cout << "E0Conv=" << E0Conv << "\n";
    H.assign(1,E0Conv);
    Heff.assign(1,E0Conv);
    basis.clear();
    basis.push_back(s);
    lls=1;

	return converged;
}

void Graph::createHMatrix(bool iterative, bool unitaryTransformed){
    createHMatrix(t,iterative,unitaryTransformed);
}
void Graph::createHMatrix(vector<Doub> const &tVal, bool iterative, bool unitaryTransformed){
    // this functions changes the t and s parameters to tVal and sVal and calculates the Hamilton-Matrix with these new parameters
    t=tVal;
    
    unsigned int size=basis.size();
    unsigned int initialSize=0;
    
    //TODO: implement iterative
    /*
    unsigned int initialSize=sqrt(H.size());
    if(iterative && initialSize>0){
        VecDoub H2(size*size);
        for (unsigned int k = 0; k < initialSize; k++) {
            for (unsigned int j = 0; j < initialSize; j++) {
                H2[k * size + j] = H[k * initialSize + j];
            }
        }
        H = H2;
    }
    else{
        H.resize(size*size);
        initialSize=0;
    }
     */
    H.resize(size*size);
    if(withoutSqrt){
        for(unsigned int k=initialSize;k<size;k++){
            Doub kNorm=sqrt(basis[k].normQuad());
            StateDouble HState=hubbardHamiltonian(basis[k],1)/kNorm;
            H[k * size + k] = (basis[k] * HState)/kNorm;
            for(unsigned int j=0;j<k;j++){
                H[k*size+j]=(HState*basis[j])/(sqrt(basis[j].normQuad()));
                H[j*size+k]=H[k*size+j];
            }
        }
    }
    else{
        for(unsigned int k=initialSize;k<size;k++){
            StateDouble HState=hubbardHamiltonian(basis[k],1);
            H[k * size + k] = basis[k] * HState;
            for(unsigned int j=0;j<k;j++){
                H[k*size+j]=HState*basis[j];
                H[j*size+k]=H[k*size+j];
            }
        }
    }

    if(unitaryTransformed){
    	if(unitaryCUT==1){
    		MatDoub U2(size,size,0.0);
    		unsigned int oldSize=sqrt(UCUT.size());
    		for(unsigned int i=0;i<oldSize;i++){
    			for(unsigned int j=0;j<oldSize;j++){
    				U2[i][j]=UCUT[i*oldSize+j];
    			}
    		}
    		for(unsigned int i=oldSize;i<size;i++){
    			U2[i][i]=1;
    		}
    		MatDoub H2(size,size,0.0);
    		for(unsigned int i=0;i<size;i++){
    			for(unsigned int j=0;j<size;j++){
    				for(unsigned int k=0;k<size;k++){
    					H2[i][j]+=U2[k][i]*H[k*size+j];
    				}
    			}
    		}
    		for(unsigned int i=0;i<size;i++){
    			for(unsigned int j=0;j<size;j++){
    				H[i*size+j]=0;
    				for(unsigned int k=0;k<size;k++){
    					H[i*size+j]+=H2[i][k]*U2[k][j];
    				}
    			}
    		}
    	}
    	else{
    		cout << "unitaryCUT!=1 - cannot perform unitaryTransformed\n";
    	}
    }
}

void Graph::transformHMatrixWithU(){
    unsigned int size=basis.size();
    VecDoub H2(size*size);
    if(UCUT.size()>0 && H.size()>0){
        for(unsigned int i=0;i<size;i++){
            for(unsigned int j=0;j<size;j++){
                Doub x=0;
                for(unsigned int k=0;k<size;k++){
                    for(unsigned int l=0;l<size;l++){
                        x+=UCUT[l*size+i]*H[l*size+k]*UCUT[k*size+j];
                    }
                }
                H2[i*size+j]=x;
            }
        }
        H=H2;
    }
}

void Graph::createEmbeddedHeff(Graph const &gSub, list<vector<unsigned int> > const &embeddings, int const &embeddingFactor){
	// calculte Heff with information from subgraph gSub
	// embedding specifies which sites are embeddable
	// IMPORTANT: the specified embedding-sites must be in the same order as on the graph which should be embedded
	unsigned int basisSize=basis.size();
	if(Heff.size()==0){
		Heff.assign(basisSize*basisSize,0);
	}

	unsigned int subLLS=gSub.lls;
	unsigned int HeffSubSize=gSub.basis.size();
	unsigned int stateSize=basis[0].size();
	unsigned int embeddingSize=embeddings.size();

	// new version 1
	for(std::list<vector<unsigned int> >::const_iterator it1=embeddings.begin();it1!=embeddings.end();it1++){
		vector<unsigned int> posNot=createPosNot(stateSize,*it1);
		unsigned int posSize=it1->size();
		unsigned int posNotSize=posNot.size();
		for(unsigned int i=0;i<basisSize;i++){
			cout << "i=" << i << "\n";
			for(std::map<vector<bool>,Doub>::const_iterator stateIt_i=basis[i].stateCont.begin();stateIt_i!=basis[i].stateCont.end();stateIt_i++){
				vector<bool> kpc(posNotSize);
				for(unsigned int kk=0;kk<posNotSize;kk++){
					kpc[kk]=*(stateIt_i->first.begin()+posNot[kk]);
				}
				vector<bool> kp(posSize);
				for(unsigned int kk=0;kk<posSize;kk++){
					kp[kk]=*(stateIt_i->first.begin()+(*it1)[kk]);
				}
				for(unsigned int j=0;j<=i;j++){
					for(std::map<vector<bool>,Doub>::const_iterator stateIt_j=basis[j].stateCont.begin();stateIt_j!=basis[j].stateCont.end();stateIt_j++){
						vector<bool> lpc(posNotSize);
						for(unsigned int kk=0;kk<posNotSize;kk++){
							lpc[kk]=*(stateIt_j->first.begin()+posNot[kk]);
						}
						if(lpc==kpc){
							vector<bool> lp(posSize);
							for(unsigned int kk=0;kk<posSize;kk++){
								lp[kk]=*(stateIt_j->first.begin()+(*it1)[kk]);
							}
							Doub ampl2=0;
							for(unsigned int iSub=0;iSub<subLLS;iSub++){
								std::map<vector<bool>,Doub>::const_iterator stateIt_iSub=gSub.basis[iSub].stateCont.find(kp);
								if(stateIt_iSub!=gSub.basis[iSub].stateCont.end()){
									for(unsigned int jSub=0;jSub<subLLS;jSub++){
										std::map<vector<bool>,Doub>::const_iterator stateIt_jSub=gSub.basis[jSub].stateCont.find(lp);
										if(stateIt_jSub!=gSub.basis[jSub].stateCont.end()){
											ampl2+=(stateIt_jSub->second)*(stateIt_iSub->second)*gSub.Heff[iSub*HeffSubSize+jSub];
										}
									}
								}
							}
							Heff[i*basisSize+j]+=embeddingFactor*(stateIt_i->second)*(stateIt_j->second)*ampl2;
							Heff[j*basisSize+i]=Heff[i*basisSize+j];
						}
					}
				}
			}
		}
	}

}

VecDoub Graph::deriveEffSpinModel(){
	cout << "Derive eff Spin model:\n";
	if(n==2){
		unsigned int basisSize=basis.size();
		unsigned int rowNumber=(basisSize*(basisSize+1))/2;
		unsigned int colNumber=2;
		VecDoub J(colNumber);
		MatDoub A(rowNumber,colNumber);
		VecDoub b(rowNumber);

		vector<MatDoub> STwoSpin(1,MatDoub(basisSize,basisSize,0.0));

		unsigned int actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// constant term
				A[actualRow][0]=basis[i]*basis[j];
				// two-spin terms
				unsigned int twoSpinIndex=0;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						//A[actualRow][actualCol]=basis[i]*(0.5*(basis[j].ccDagger(2*j,2*j+1)).ccDagger(2*i+1,2*i)+0.5*(basis[j].ccDagger(2*j+1,2*j)).ccDagger(2*i,2*i+1)+0.25*((basis[j].nState(2*j)).nState(2*i)-(basis[j].nState(2*j+1)).nState(2*i)-(basis[j].nState(2*j)).nState(2*i+1)+(basis[j].nState(2*j+1)).nState(2*i+1)));
						STwoSpin[twoSpinIndex][i][j]=basis[i]*(0.5*(basis[j].ccDagger(2*site2,2*site2+1)).ccDagger(2*site1+1,2*site1)+0.5*(basis[j].ccDagger(2*site2+1,2*site2)).ccDagger(2*site1,2*site1+1)+0.25*((basis[j].nState(2*site2)).nState(2*site1)-(basis[j].nState(2*site2+1)).nState(2*site1)-(basis[j].nState(2*site2)).nState(2*site1+1)+(basis[j].nState(2*site2+1)).nState(2*site1+1)));
						STwoSpin[twoSpinIndex][j][i]=STwoSpin[twoSpinIndex][i][j];
						twoSpinIndex++;
					}
				}
				actualRow++;
			}
		}

		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// two-spin saving
				unsigned int actualCol=1;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						A[actualRow][actualCol]=STwoSpin[actualCol-1][i][j];
						actualCol++;
					}
				}
				actualRow++;
			}
		}

		// copy hamiltonmatrix
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				b[actualRow]=Heff[i*basisSize+j];
				actualRow++;
			}
		}

		// solve (overdetermined) LGS A*J=b
		SVD svdSystem(A);
		svdSystem.solve(b,J);

		// check solution
		if(testNRSolution(A, J, b)){
			return J;
		}
		else{
			return VecDoub(1,0.0);
		}
	}
	if(n==3){
		unsigned int basisSize=basis.size();
		unsigned int rowNumber=(basisSize*(basisSize+1))/2;
		unsigned int colNumber=4;
		VecDoub J(colNumber);
		MatDoub A(rowNumber,colNumber);
		VecDoub b(rowNumber);

		vector<MatDoub> STwoSpin(3,MatDoub(basisSize,basisSize,0.0));

		unsigned int actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// constant term
				A[actualRow][0]=basis[i]*basis[j];
				// two-spin terms
				unsigned int twoSpinIndex=0;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						//A[actualRow][actualCol]=basis[i]*(0.5*(basis[j].ccDagger(2*j,2*j+1)).ccDagger(2*i+1,2*i)+0.5*(basis[j].ccDagger(2*j+1,2*j)).ccDagger(2*i,2*i+1)+0.25*((basis[j].nState(2*j)).nState(2*i)-(basis[j].nState(2*j+1)).nState(2*i)-(basis[j].nState(2*j)).nState(2*i+1)+(basis[j].nState(2*j+1)).nState(2*i+1)));
						STwoSpin[twoSpinIndex][i][j]=basis[i]*(0.5*(basis[j].ccDagger(2*site2,2*site2+1)).ccDagger(2*site1+1,2*site1)+0.5*(basis[j].ccDagger(2*site2+1,2*site2)).ccDagger(2*site1,2*site1+1)+0.25*((basis[j].nState(2*site2)).nState(2*site1)-(basis[j].nState(2*site2+1)).nState(2*site1)-(basis[j].nState(2*site2)).nState(2*site1+1)+(basis[j].nState(2*site2+1)).nState(2*site1+1)));
						STwoSpin[twoSpinIndex][j][i]=STwoSpin[twoSpinIndex][i][j];
						twoSpinIndex++;
					}
				}
				actualRow++;
			}
		}

		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// two-spin saving
				unsigned int actualCol=1;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						A[actualRow][actualCol]=STwoSpin[actualCol-1][i][j];
						actualCol++;
					}
				}
				actualRow++;
			}
		}

		// copy hamiltonmatrix
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				b[actualRow]=Heff[i*basisSize+j];
				actualRow++;
			}
		}

		// solve (overdetermined) LGS A*J=b
		SVD svdSystem(A);
		svdSystem.solve(b,J);

		// check solution
		if(testNRSolution(A, J, b)){
			return J;
		}
		else{
			return VecDoub(1,0.0);
		}
	}
	if(n==4){
		unsigned int basisSize=basis.size();
		unsigned int rowNumber=(basisSize*(basisSize+1))/2;
		unsigned int colNumber=10;
		VecDoub J(colNumber);
		MatDoub A(rowNumber,colNumber);
		VecDoub b(rowNumber);

		vector<MatDoub> STwoSpin(6,MatDoub(basisSize,basisSize,0.0));

		unsigned int actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// constant term
				A[actualRow][0]=basis[i]*basis[j];
				// two-spin terms
				unsigned int twoSpinIndex=0;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						//A[actualRow][actualCol]=basis[i]*(0.5*(basis[j].ccDagger(2*j,2*j+1)).ccDagger(2*i+1,2*i)+0.5*(basis[j].ccDagger(2*j+1,2*j)).ccDagger(2*i,2*i+1)+0.25*((basis[j].nState(2*j)).nState(2*i)-(basis[j].nState(2*j+1)).nState(2*i)-(basis[j].nState(2*j)).nState(2*i+1)+(basis[j].nState(2*j+1)).nState(2*i+1)));
						STwoSpin[twoSpinIndex][i][j]=basis[i]*(0.5*(basis[j].ccDagger(2*site2,2*site2+1)).ccDagger(2*site1+1,2*site1)+0.5*(basis[j].ccDagger(2*site2+1,2*site2)).ccDagger(2*site1,2*site1+1)+0.25*((basis[j].nState(2*site2)).nState(2*site1)-(basis[j].nState(2*site2+1)).nState(2*site1)-(basis[j].nState(2*site2)).nState(2*site1+1)+(basis[j].nState(2*site2+1)).nState(2*site1+1)));
						STwoSpin[twoSpinIndex][j][i]=STwoSpin[twoSpinIndex][i][j];
						twoSpinIndex++;
					}
				}
				actualRow++;
			}
		}

		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// two-spin saving
				unsigned int actualCol=1;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						A[actualRow][actualCol]=STwoSpin[actualCol-1][i][j];
						actualCol++;
					}
				}
				actualRow++;
			}
		}

		//four-spin terms
		// S1234
		actualRow=0;
		unsigned int actualCol=7;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				A[actualRow][actualCol]=0;
				for(unsigned int k=0;k<basisSize;k++){
					A[actualRow][actualCol]+=STwoSpin[0][i][k]*STwoSpin[5][k][j];
				}
				actualRow++;
			}
		}
		actualCol++;

		// S1324
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				A[actualRow][actualCol]=0;
				for(unsigned int k=0;k<basisSize;k++){
					A[actualRow][actualCol]+=STwoSpin[1][i][k]*STwoSpin[4][k][j];
				}
				actualRow++;
			}
		}
		actualCol++;

		// S1423
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				A[actualRow][actualCol]=0;
				for(unsigned int k=0;k<basisSize;k++){
					A[actualRow][actualCol]+=STwoSpin[2][i][k]*STwoSpin[3][k][j];
				}
				actualRow++;
			}
		}
		actualCol++;

		// copy hamiltonmatrix
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				b[actualRow]=Heff[i*basisSize+j];
				actualRow++;
			}
		}

		// solve (overdetermined) LGS A*J=b
		SVD svdSystem(A);
		svdSystem.solve(b,J);

		// check solution
		if(testNRSolution(A, J, b)){
			return J;
		}
		else{
			return VecDoub(1,0.0);
		}
	}


	// n=6
	if(n==6){
		unsigned int basisSize=basis.size();
		cout << "basisSize=" << basisSize << "\n";
		unsigned int rowNumber=(basisSize*(basisSize+1))/2;
		cout << "rowNumber=" << rowNumber << "\n";
		unsigned int colNumber=15+45+1+15;
		cout << "colNumber=" << colNumber << "\n";
		VecDoub J(colNumber);
		MatDoub A(rowNumber,colNumber);
		VecDoub b(rowNumber);

		vector<MatDoub> STwoSpin(15,MatDoub(basisSize,basisSize,0.0));
		map<pair<unsigned int,unsigned int>,unsigned int> STwoSpinMapping;

		cout << " constant term and 2spin terms\n";
		unsigned int actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// constant term
				A[actualRow][0]=basis[i]*basis[j];
				// two-spin terms
				unsigned int twoSpinIndex=0;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						//A[actualRow][actualCol]=basis[i]*(0.5*(basis[j].ccDagger(2*j,2*j+1)).ccDagger(2*i+1,2*i)+0.5*(basis[j].ccDagger(2*j+1,2*j)).ccDagger(2*i,2*i+1)+0.25*((basis[j].nState(2*j)).nState(2*i)-(basis[j].nState(2*j+1)).nState(2*i)-(basis[j].nState(2*j)).nState(2*i+1)+(basis[j].nState(2*j+1)).nState(2*i+1)));
						STwoSpin[twoSpinIndex][i][j]=basis[i]*(0.5*(basis[j].ccDagger(2*site2,2*site2+1)).ccDagger(2*site1+1,2*site1)+0.5*(basis[j].ccDagger(2*site2+1,2*site2)).ccDagger(2*site1,2*site1+1)+0.25*((basis[j].nState(2*site2)).nState(2*site1)-(basis[j].nState(2*site2+1)).nState(2*site1)-(basis[j].nState(2*site2)).nState(2*site1+1)+(basis[j].nState(2*site2+1)).nState(2*site1+1)));
						STwoSpin[twoSpinIndex][j][i]=STwoSpin[twoSpinIndex][i][j];
						STwoSpinMapping.insert(pair<pair<unsigned int,unsigned int>,unsigned int>(pair<unsigned int,unsigned int>(site1,site2),twoSpinIndex));
						twoSpinIndex++;
					}
				}
				actualRow++;
			}
		}

		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				// two-spin saving
				unsigned int actualCol=1;
				for(unsigned int site1=0;site1<n-1;site1++){
					for(unsigned int site2=site1+1;site2<n;site2++){
						A[actualRow][actualCol]=STwoSpin[actualCol-1][i][j];
						actualCol++;
					}
				}
				actualRow++;
			}
		}

		cout << " 4spin terms\n";
		//four-spin terms
		unsigned int actualCol=16;
		for(unsigned int site1=0;site1<=2;site1++){
			for(unsigned int site2=site1+1;site2<=5;site2++){
				for(unsigned int site3=site1+1;site3<=5;site3++){
					if(site3!=site2){
						for(unsigned int site4=site3+1;site4<=5;site4++){
							if(site4!=site2){
								//cout << site1 << " " << site2 << " " << site3 << " " << site4 << "\n";
								actualRow=0;
								map<pair<unsigned int,unsigned int>,unsigned int>::iterator it1=STwoSpinMapping.find(pair<unsigned int,unsigned int>(site1,site2));
								map<pair<unsigned int,unsigned int>,unsigned int>::iterator it2=STwoSpinMapping.find(pair<unsigned int,unsigned int>(site3,site4));
								for(unsigned int i=0;i<basisSize;i++){
									for(unsigned int j=i;j<basisSize;j++){
										A[actualRow][actualCol]=0;

										for(unsigned int k=0;k<basisSize;k++){
											A[actualRow][actualCol]+=STwoSpin[it1->second][i][k]*STwoSpin[it2->second][k][j];
										}
										actualRow++;
									}
								}
								actualCol++;
							}
						}
					}
				}
			}
		}

		// 6spin terms
		cout << " 6spin terms\n";
		//TODO: 6spin terms
		unsigned int site1=0;
		for(unsigned int site2=1;site2<=5;site2++){
			for(unsigned int site3=1;site3<=2;site3++){
				if(site3!=site2){
					for(unsigned int site4=site3+1;site4<=5;site4++){
						if((site4!=site1) && (site4!=site2)){
							for(unsigned int site5=2;site5<=4;site5++){
								if((site5!=site2) && (site5!=site3) && (site5!=site4)){
									for(unsigned int site6=site5+1;site6<=5;site6++){
										if((site6!=site2) && (site6!=site3) && (site6!=site4)){
											actualRow=0;
											map<pair<unsigned int,unsigned int>,unsigned int>::iterator it1=STwoSpinMapping.find(pair<unsigned int,unsigned int>(site1,site2));
											map<pair<unsigned int,unsigned int>,unsigned int>::iterator it2=STwoSpinMapping.find(pair<unsigned int,unsigned int>(site3,site4));
											map<pair<unsigned int,unsigned int>,unsigned int>::iterator it3=STwoSpinMapping.find(pair<unsigned int,unsigned int>(site5,site6));

											for(unsigned int i=0;i<basisSize;i++){
												for(unsigned int j=i;j<basisSize;j++){
													A[actualRow][actualCol]=0;
													for(unsigned int k1=0;k1<basisSize;k1++){
														for(unsigned int k2=0;k2<basisSize;k2++){
															A[actualRow][actualCol]+=STwoSpin[it1->second][i][k1]*STwoSpin[it2->second][k1][k2]*STwoSpin[it3->second][k2][j];
														}
													}
													actualRow++;
												}
											}
											actualCol++;
										}
									}
								}
							}
						}
					}
				}
			}
		}

		// copy hamiltonmatrix
		actualRow=0;
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=i;j<basisSize;j++){
				b[actualRow]=Heff[i*basisSize+j];
				actualRow++;
			}
		}

		// test output
		cout << " test output of A,b:\n";
		for(unsigned int i=0;i<A.nrows();i++){
			for(unsigned int j=0;j<A.ncols();j++){
				cout << A[i][j] << " ";
			}
			cout << "\n";
		}
		cout << "\n";
		for(unsigned int i=0;i<b.size();i++){
			cout << b[i] << "\n";
		}
		cout << "\n";

		cout << " solve LGS A*J=b\n";
		// solve (overdetermined) LGS A*J=b
		SVD svdSystem(A);
		svdSystem.solve(b,J);

		cout << " test output of J:\n";
		for(unsigned int i=0;i<J.size();i++){
			cout << J[i] << "\n";
		}
		cout << "\n";

		// check solution
		if(testNRSolution(A, J, b)){
			return J;
		}
		else{
			cout << "spin model could not be derived!\n";
			return VecDoub(1,0.0);
		}
	}

	return VecDoub(1,0.0);
}

void Graph::orthogonalizeBasis(){
	// overlap matrix
	MatDoub Sover;
	unsigned int SoverSize=basis.size();
	Sover.resize(SoverSize,SoverSize);
	for(unsigned int i=0;i<SoverSize;i++){
	    Sover[i][i]=basis[i]*basis[i];
	    for(unsigned int j=0;j<i;j++){
	        Sover[i][j]=basis[i]*basis[j];
	        Sover[j][i]=Sover[i][j];
	    }
	}

	// calculate eigenvalues
	Symmeig eigSys(Sover);
	invertSorting(eigSys.d,eigSys.z);

	// calculate square root of inverse overlap matrix
	Sorth.assign(SoverSize*SoverSize,0.0);
	for(unsigned int i=0;i<SoverSize;i++){
		for(unsigned int j=0;j<SoverSize;j++){
			Doub temp=0;
	        for(unsigned int k=0;k<SoverSize;k++){
	        	temp+=eigSys.z[i][k]*eigSys.z[j][k]/sqrt(eigSys.d[k]);
	        }
	        Sorth[i*SoverSize+j]=temp;
	     }
	}

	StateVector basis2;
	for(unsigned int i=0;i<SoverSize;i++){
		StateDouble sTemp=Sorth[i*SoverSize]*basis[0];
	    for(unsigned int j=1;j<SoverSize;j++){
	    	sTemp+=Sorth[i*SoverSize+j]*basis[j];
	    }
	    basis2.push_back(sTemp);
	}

	basis.swap(basis2);
}
void Graph::transformHeffWithSorth(){
	unsigned int basisSize=basis.size();
	VecDoub Heff2(basisSize*basisSize,0.0);
	for(unsigned int i=0;i<basisSize;i++){
		for(unsigned int j=0;j<basisSize;j++){
			for(unsigned int k=0;k<basisSize;k++){
				for(unsigned int l=0;l<basisSize;l++){
					Heff2[i*basisSize+j]+=Sorth[i*basisSize+l]*Heff[l*basisSize+k]*Sorth[k*basisSize+j];
				}
			}
		}
	}
	Heff=Heff2;
}

void Graph::applyU(unsigned int blocks){
    //TODO: unitaryCUT fuer alle Funktionen implementieren und nur noch applyU aufrufen (ausserhalb dieser Funktionen!)
    unsigned int size = basis.size();
    
	unsigned int matrixSize=blocks*lls;
	if(size<matrixSize){
		matrixSize=size;
	}
    
    tempBasis.clear();
    for(unsigned int j=0;j<matrixSize;j++){
        StateDouble sTemp=UCUT[j*size]*basis[0];
        for(unsigned int k=1;k<size;k++){
            sTemp+=UCUT[j*size+k]*basis[k];
        }
        tempBasis.push_back(sTemp);
    }
    basis.swap(tempBasis);
}

StateVector Graph::getUnitaryTransformedBasis(unsigned int blocks){
    unsigned int size = basis.size();

	unsigned int matrixSize=blocks*lls;
	if(size<matrixSize){
		matrixSize=size;
	}

    tempBasis.clear();
    //TODO: andere Summationsreihenfolge
    for(unsigned int j=0;j<matrixSize;j++){
        StateDouble sTemp=UCUT[j*size]*basis[0];
        for(unsigned int k=1;k<size;k++){
            sTemp+=UCUT[j*size+k]*basis[k];
        }
        tempBasis.push_back(sTemp);
    }

    return tempBasis;
}

void Graph::unorthogonalizeHeff(){
	// this function transforms Heff -> Sorth^-1*Heff*Sorth^-1
	// important: only the lls*lls Block of Heff is transformed. The other block stays the same!
	unsigned int basisSize=basis.size();
	MatDoub NRSorth(lls,lls);
	for(unsigned int i=0;i<lls;i++){
		for(unsigned int j=0;j<lls;j++){
			NRSorth[i][j]=Sorth[i*lls+j];
		}
	}

	gaussj(NRSorth);

	MatDoub Heff2(lls,lls);

	for(unsigned int i=0;i<lls;i++){
		for(unsigned int j=0;j<lls;j++){
			Doub x=0;
			for(unsigned int k=0;k<lls;k++){
				for(unsigned int l=0;l<lls;l++){
					x+=NRSorth[i][k]*Heff[k*basisSize+l]*NRSorth[l][j];
				}
			}
			Heff2[i][j]=x;
		}
	}
	for(unsigned int i=0;i<lls;i++){
		for(unsigned int j=0;j<lls;j++){
			Heff[i*basisSize+j]=Heff2[i][j];
			cout << Heff[i*basisSize+j] << " ";
		}
		cout << "\n";
	}
}

void Graph::transformBasis(StateVector const &newBasis){
	// this function transforms the basis to newBasis and applies the transformation to the hamiltonian and the effective hamiltonian
	unsigned int basisSize=basis.size();
	if(basisSize==newBasis.size()){
		MatDoub A(basisSize,basisSize);
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=0;j<basisSize;j++){
				A[i][j]=newBasis[j]*basis[i];
			}
		}

		VecDoub H2(basisSize*basisSize);
		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=0;j<basisSize;j++){
				H2[i*basisSize+j]=0;
				for(unsigned int k=0;k<basisSize;k++){
					for(unsigned int l=0;l<basisSize;l++){
						H2[i*basisSize+j]+=A[k][i]*H[k*basisSize+l]*A[l][j];
					}
				}
			}
		}
		H=H2;

		/*
		// Heff eigenvalues
		cout << "calculate eigenvalues\n";
		unsigned int HeffMatSize=basisSize;
		MatDoub HeffMat(HeffMatSize,HeffMatSize);
		for(unsigned int i=0;i<HeffMatSize;i++){
			for(unsigned int j=0;j<HeffMatSize;j++){
				HeffMat[i][j]=this->getHeff(i,j);
			}
		}
		Symmeig eigSysHeff(HeffMat);
		invertSorting(eigSysHeff.d,eigSysHeff.z);

		for(unsigned int i=0;i<eigSysHeff.d.size();i++){
			cout << eigSysHeff.d[i] << " ";
		}
		cout << "\n";
		///////////////////////////////////////////
		 */

		for(unsigned int i=0;i<basisSize;i++){
			for(unsigned int j=0;j<basisSize;j++){
				H2[i*basisSize+j]=0;
				for(unsigned int k=0;k<basisSize;k++){
					for(unsigned int l=0;l<basisSize;l++){
						H2[i*basisSize+j]+=A[k][i]*Heff[k*basisSize+l]*A[l][j];
					}
				}
			}
		}
		Heff=H2;

		/*
		// Heff eigenvalues
		cout << "calculate eigenvalues\n";
		for(unsigned int i=0;i<HeffMatSize;i++){
			for(unsigned int j=0;j<HeffMatSize;j++){
				HeffMat[i][j]=this->getHeff(i,j);
			}
		}
		Symmeig eigSysHeff2(HeffMat);
		invertSorting(eigSysHeff2.d,eigSysHeff2.z);

		for(unsigned int i=0;i<eigSysHeff2.d.size();i++){
			cout << eigSysHeff2.d[i] << " ";
		}
		cout << "\n";
		///////////////////////////////////////////
		 */

		basis=newBasis;
	}
	else{
		cerr << "basis sizes are not the same!\n";
	}
}

unsigned int Graph::performED(VecDoub &eval, MatDoub &evec, StateVector &vs, bool zeroSz, bool halfFilling, bool eigenstates, bool StotEigenstates){
	 // performs an ED for the hubbard model and stores the eigenvalues in eval and eigenvectors in evec
		// if zeroSz (default=true) is set only for the Sz=0 sector
		// if halfFilling (default=true) is set only at halfFilling
		// return the size of eval
	    // if eigenstates (default=true) is set vs contains the hubbard-eigenstates, if not vs contains the hubbard states in the right order
	    // if StotEigenstates (default=true) is set (eigenstates must also set to true) vs contains the hubbard-eigenstates AND Stot-eigenstates
	    vector<vector<bool> > states;
	    //vector<StateDouble> vs;
	    vs.clear();
	    vector<bool> vbFalse;
	    vector<bool> vbTrue;
	    vbFalse.push_back(false);
	    vbTrue.push_back(true);
	    states.push_back(vbFalse);
	    states.push_back(vbTrue);

	    for(unsigned int j=1;j<2*n;j++){
	        unsigned int statesSize=states.size();
	        for(unsigned int i=0;i<statesSize;i++){
	            vector<bool> temp=states[i];
	            temp.push_back(true);
	            states[i].push_back(false);
	            states.push_back(temp);
	        }
	    }
	    //cout << "hilbertSpace=" << states.size() << "\n";

	    for(unsigned int i=0;i<states.size();i++){
	        unsigned int up=0;
	        unsigned int down=0;
	        for(unsigned int j=0;j<n;j++){
	            if(states[i][2*j]){
	                up++;
	            }
	            if(states[i][2*j+1]){
	                down++;
	            }
	        }
	        bool zeroSzProof=(zeroSz && (up==down)) || !zeroSz;
	        bool halfFillingProof=(halfFilling && (up+down==n)) || !halfFilling;
	        if(zeroSzProof && halfFillingProof){
	            StateDouble tempState(states[i]);
	            vs.push_back(tempState);
	        }
	    }
	    unsigned int hubbardSize=vs.size();


	    MatDoub HHubbard(hubbardSize,hubbardSize);


	    //cout << "create HHubbard:\n";
	    for(unsigned int i=0;i<hubbardSize;i++){
	    	HHubbard[i][i]=vs[i]*hubbardHamiltonian(vs[i],1);
	        for(unsigned int j=0;j<i;j++){
	            Doub x=vs[i]*hubbardHamiltonian(vs[j],1);
	            HHubbard[i][j]=x;
	            HHubbard[j][i]=x;
	            //cout << "i=" << i << "j=" << j << "\n";
	        }
	    }

	    cout << "calculate Eigenvalues/-vectors\n";


	    Symmeig eigSys(HHubbard);
	    eval=eigSys.d;
	    evec=eigSys.z;
	    invertSorting(eval,evec);

	    if(eigenstates){
	    	StateVector eigenstatesVector;
	    	createEigenstatesFromED(hubbardSize,evec,vs,eigenstatesVector);
	    	vs.swap(eigenstatesVector);

	        //calculate eigenstates of Stot in the blocks
	        if(StotEigenstates){
	            eigenstatesVector.clear();
	            unsigned int i=0;
	            while(i<vs.size()){
	            	//cout << "eval[" << i << "]=" << eval[i] << " ";
	                // find degenerate subspace
	            	StateVector blockStates;
	            	StateVector blockStatesStotOp;
	                Doub E=eval[i];
	                bool undefinedSpin=false;
	                while((myAbs(eval[i]-E)<=zeroNormQuadEPS) && i<vs.size()){
	                	//cout << "eval[i]-E=" << myAbs(eval[i]-E) << " zeroNormQuadEPS=" << zeroNormQuadEPS << "\n";
	                    StateDouble sTemp=vs[i].StotOp();
	                    Doub StotVal=vs[i].Stot(sTemp);
	                    //cout << "StotVal=" << StotVal << " ";
	                	if(StotVal<-0.5){
	                		undefinedSpin=true;
	                	}
	                    blockStates.push_back(vs[i]);
	                    blockStatesStotOp.push_back(sTemp);
	                    i++;
	                }
	                unsigned int blockStatesSize=blockStates.size();

	                if(undefinedSpin){
	                	//cout << " blockStatesSize=" << blockStatesSize << "\n";
	                    MatDoub StotMatrix(blockStatesSize,blockStatesSize);

	                    // calculate StotMatrix
	                    for(unsigned int j=0;j<blockStatesSize;j++){
	                        StotMatrix[j][j]=blockStates[j]*blockStatesStotOp[j];
	                        for(unsigned int k=0;k<j;k++){
	                            StotMatrix[j][k]=blockStates[j]*blockStatesStotOp[k];
	                            StotMatrix[k][j]=StotMatrix[j][k];
	                        }
	                    }

	                    //Symmeig eigSysStot(StotMatrix);
	                    Jacobi jac(StotMatrix);
	                    invertSorting(jac.d,jac.v);
	                    for(unsigned int j=0;j<blockStatesSize;j++){
	                        //StateDouble sTemp2=eigSysStot.z[j][0]*blockStates[0];
	                        StateDouble sTemp2=jac.v[0][j]*blockStates[0];
	                        for(unsigned int k=1;k<blockStatesSize;k++){
	                            //sTemp2+=eigSysStot.z[j][k]*blockStates[k];
	                        	sTemp2+=jac.v[k][j]*blockStates[k];
	                        }
	                        eigenstatesVector.push_back(sTemp2);
	                        //cout << " jac.d[j]=" << jac.d[j] << " Stot=" << sTemp2.Stot() << "\n";
	                    }
	                }
	                else{
	                	for(unsigned int j=0;j<blockStatesSize;j++){
	                		eigenstatesVector.push_back(blockStates[j]);
	                	}
	                }

	            }
	            vs.swap(eigenstatesVector);
	        }
	    }

	    return hubbardSize;
}

unsigned int Graph::calculateHubbardEigenvalues(VecDoub &eval){
	// performs an ED for the half-filled and Sz=0 sector of the hubbard model and stores the eigenvalues in eval
	// return the size of eval
    MatDoub evec;
    StateVector vs;
    unsigned int hubbardSize=performED(eval,evec,vs,true,true,false);
    return hubbardSize;
}

unsigned int Graph::countHubbardSinglets(){
	// performs an ED for the half-filled and Sz=0 sector of the hubbard model and counts all Eigenstates with S^2=0
	VecDoub evalHubbard;
	MatDoub evecHubbard;
	StateVector vs;
	unsigned int hubbardSize=performED(evalHubbard,evecHubbard,vs,true,true,true);
    
    unsigned int singletSize=0;
    for(unsigned int k=0;k<hubbardSize;k++){
        if(fabs(vs[k].Stot())<=zeroEPS){
            singletSize++;
        }
    }
    
    return singletSize;
}

Graph mergeGraphs(Graph const &g1, Graph const &g2){
	//TODO: implement complete merge!
	// proof if some parameters are the same (others should be also the same!)
	if((g1.n==g2.n) && (g1.order==g2.order) && (g1.graphNumber==g2.graphNumber)){
		Graph g;
		g.order=g1.order;
		g.graphNumber=g1.graphNumber;
		g.n=g1.n;
		g.t=g1.t;
		g.couplingsCont=g1.couplingsCont;
		g.withoutSqrt=g1.withoutSqrt;
		g.unitaryCUT=g1.unitaryCUT;

		unsigned int g1BasisSize=g1.basis.size();
		unsigned int g2BasisSize=g2.basis.size();

		for(unsigned int i=0;i<g1.lls;i++){
			g.basis.push_back(g1.basis[i]);
		}
		for(unsigned int i=0;i<g2.lls;i++){
			g.basis.push_back(g2.basis[i]);
		}

		/*
		if((g1.QPNumbers.size()==g1BasisSize) && (g2.QPNumbers.size()==g2BasisSize)){
			for(unsigned int i=0;i<g1.lls;i++){
				g.QPNumbers.push_back(g1.QPNumbers[i]);
			}
			for(unsigned int i=0;i<g2.lls;i++){
				g.QPNumbers.push_back(g2.QPNumbers[i]);
			}
		}
		*/

		unsigned int gBasisSize=g.basis.size();
		g.lls=gBasisSize;

		g.H.assign(gBasisSize*gBasisSize,0.0);
		g.Heff.assign(gBasisSize*gBasisSize,0.0);

		for(unsigned int i=0;i<g1.lls;i++){
			for(unsigned int j=0;j<g1.lls;j++){
				g.H[i*gBasisSize+j]=g1.H[i*g1BasisSize+j];
				g.Heff[i*gBasisSize+j]=g1.Heff[i*g1BasisSize+j];
			}
		}
		for(unsigned int i=0;i<g2.lls;i++){
			for(unsigned int j=0;j<g2.lls;j++){
				g.H[(i+g1.lls)*gBasisSize+(j+g1.lls)]=g2.H[i*g2BasisSize+j];
				g.Heff[(i+g1.lls)*gBasisSize+(j+g1.lls)]=g2.Heff[i*g2BasisSize+j];
			}
		}

		/*
		if((g1.unitaryCUT==1) && (g2.unitaryCUT==1) && (g1BasisSize*g1BasisSize==g1.UCUT.size()) && (g2BasisSize*g2BasisSize==g2.UCUT.size())){
			g.UCUT.assign(gBasisSize*gBasisSize,0.0);
			for(unsigned int i=0;i<g1.lls;i++){
				for(unsigned int j=0;j<g1.lls;j++){
					g.UCUT[i*gBasisSize+j]=g1.UCUT[i*g1BasisSize+j];
				}
			}
			for(unsigned int i=0;i<g2.lls;i++){
				for(unsigned int j=0;j<g2.lls;j++){
					g.UCUT[(i+g1.lls)*gBasisSize+(j+g1.lls)]=g2.UCUT[i*g2BasisSize+j];
				}
			}
		}
		else{
			g.UCUT.resize(0);
		}

		if((g1InitialStatesSize*g1InitialStatesSize==g1.Sorth.size()) && (g2InitialStatesSize*g2InitialStatesSize==g2.Sorth.size())){
			g.Sorth.assign(gInitialStatesSize*gInitialStatesSize,0.0);
			for(unsigned int i=0;i<g1InitialStatesSize;i++){
				for(unsigned int j=0;j<g1InitialStatesSize;j++){
					g.Sorth[i*gInitialStatesSize+j]=g1.Sorth[i*g1InitialStatesSize+j];
				}
			}
			for(unsigned int i=0;i<g2InitialStatesSize;i++){
				for(unsigned int j=0;j<g2InitialStatesSize;j++){
					g.Sorth[(i+g1InitialStatesSize)*gInitialStatesSize+(j+g1InitialStatesSize)]=g2.Sorth[i*g2InitialStatesSize+j];
				}
			}
		}
		else{
			g.Sorth.resize(0);
		}
		*/

		return g;
	}
	else{
		cerr << "some required parameters are not the same in g1 and g2!\n";
		return Graph();
	}
}

// Saving und reading of Graph class
ostream& operator<<(ostream &o, Graph const &g){
	// precision
	o << setprecision(16);
	o << "parseMode\n";
    // saving of graph variables
	o << "order ";
	o << g.order << "\n";
	o << "graphNumber ";
	o << g.graphNumber << "\n";
	o << "n ";
	o << g.n << "\n";

	o << "tSize ";
	o << g.t.size() << "\n";
	for(unsigned int i=0;i<g.t.size();i++){
		o << g.t[i] << "\n";
	}
	o << "lls ";
	o << g.lls << "\n";
    // saving of couplings
	o << "couplingsContSize ";
	o << g.couplingsCont.size() << "\n";
	for(std::list<CouplingStruct>::const_iterator it=g.couplingsCont.begin();it!=g.couplingsCont.end();it++){
		o << it->from << " " << it->to << " " << it->coupling << "\n";
	}
    // saving of basis
	unsigned int basisSize=g.basis.size();
	o << "basisSize ";
	o << basisSize << "\n";
	for(unsigned int i=0;i<basisSize;i++){
		o << g.basis[i] << "\n";
	}
	// saving of tempBasis
	unsigned int tempBasisSize=g.tempBasis.size();
	o << "tempBasisSize ";
	o << tempBasisSize << "\n";
	for(unsigned int i=0;i<tempBasisSize;i++){
		o << g.tempBasis[i] << "\n";
	}
	// saving of initialStates
	unsigned int initialStatesSize=g.initialStates.size();
	o << "initialStatesSize ";
	o << initialStatesSize << "\n";
	for(unsigned int i=0;i<initialStatesSize;i++){
		o << g.initialStates[i] << "\n";
	}
    // saving of withoutSqrt
	o << "withoutSqrt ";
    o << g.withoutSqrt << "\n";
	// saving of QPNumbers
	unsigned int QPNumbersSize=g.QPNumbers.size();
	o << "QPNumbersSize ";
	o << QPNumbersSize << "\n";
	for(unsigned int i=0;i<QPNumbersSize;i++){
		o << g.QPNumbers[i] << "\n";
	}
    // saving of Hamiltonmatrix
    unsigned int HSize=g.H.size();
    o << "HSize ";
    o << HSize << "\n";
    for(unsigned int i=0;i<HSize;i++){
    	o << g.H[i] << " ";
    }
	o << "\n";
    // saving of effective Hamiltonmatrix
    unsigned int HeffSize=g.Heff.size();
    o << "HeffSize ";
    o << HeffSize << "\n";
    for(unsigned int i=0;i<HeffSize;i++){
    	o << g.Heff[i] << " ";
    }
    o << "\n";
    // saving of Unitarymatrix
    o << "unitaryCUT ";
    o << g.unitaryCUT << "\n";
    unsigned int UCUTSize=g.UCUT.size();
    o << "UCUTSize ";
    o << UCUTSize << "\n";
    for(unsigned int i=0;i<UCUTSize;i++){
    	o << g.UCUT[i] << " ";
    }
    o<< "\n";
    // saving of orthogonalizeMatrix Sorth
    unsigned int SorthSize=g.Sorth.size();
    o << "SorthSize ";
    o << SorthSize << "\n";
    for(unsigned int i=0;i<SorthSize;i++){
    	o << g.Sorth[i] << " ";
    }

	return o;
}

istream& operator>>(istream &o, Graph &g){
	// check if file is saved in parseMode
	string tempString;
	char temp;
	unsigned int k=0;
	o>>tempString;
	if(tempString=="parseMode"){
		while(o>>tempString){
			// reading of graph variables
			if(tempString=="order"){
				if(!(o>>g.order)){
					cerr << "reading error: order\n";
				}
			}
			if(tempString=="graphNumber"){
				if(!(o>>g.graphNumber)){
					cerr << "reading error: graphNumber\n";
				}
			}
			if(tempString=="n"){
				if(!(o>>g.n)){
					cerr << "reading error: n\n";
				}
			}
			if(tempString=="tSize"){
				unsigned int k=0;
				if(!(o>>k)){
					cerr << "reading error: t.size()\n";
				}
				g.t.clear();
				for(unsigned int i=0;i<k;i++){
					Doub x=0;
					o>>x;
					g.t.push_back(x);
				}
			}
			if(tempString=="lls"){
				if(!(o>>g.lls)){
					cerr << "reading error: lls\n";
				}
			}
			// reading of couplings
			if(tempString=="couplingsContSize"){
				unsigned int k=0;
				if(!(o>>k)){
					cerr << "reading error: couplingsCont.size()\n";
				}
				g.couplingsCont.clear();
				for(unsigned int i=0;i<k;i++){
					CouplingStruct csTemp;
					o>>csTemp.from;
					o>>csTemp.to;
					o>>csTemp.coupling;
					g.couplingsCont.push_back(csTemp);
				}
			}
		    // reading of basis
			if(tempString=="basisSize"){
				if(!(o>>k)){
					cerr << "reading error: basis.size()\n";
				}
				g.basis.clear();
				for(unsigned int i=0;i<k;i++){
					StateDouble s;
					o>>s;
					g.basis.push_back(s);
				}
			}
			// reading of tempBasis
			if(tempString=="tempBasisSize"){
				if(!(o>>k)){
					cerr << "reading error: tempBasis.size()\n";
				}
				g.tempBasis.clear();
				for(unsigned int i=0;i<k;i++){
					StateDouble s;
					o>>s;
					g.tempBasis.push_back(s);
				}
			}
			// reading of initialStates
			if(tempString=="initialStatesSize"){
				if(!(o>>k)){
					cerr << "reading error: initialStates.size()\n";
				}
				g.initialStates.clear();
				for(unsigned int i=0;i<k;i++){
					StateDouble s;
					o>>s;
					g.initialStates.push_back(s);
				}
			}
			// reading of withoutSqrt
			if(tempString=="withoutSqrt"){
				if(!(o>>g.withoutSqrt)){
					cerr << "reading error: withoutSqrt\n";
				}
			}
			// reading of QPNumbers
			if(tempString=="QPNumbersSize"){
				if(!(o>>k)){
					cerr << "reading error: QPNumbers.size()\n";
				}
				g.QPNumbers.clear();
				for(unsigned int i=0;i<k;i++){
					unsigned int qq;
					o>>qq;
					g.QPNumbers.push_back(qq);
				}
			}
		    // reading of Hamiltonmatrix
		    if(tempString=="HSize"){
		    	if(!(o>>k)){
		    		cerr << "reading error: HSize\n";
		    	}
		    	o>>noskipws>>temp>>skipws;
		    	g.H.resize(k);
		    	for(unsigned int i=0;i<k;i++){
		    		o>>g.H[i];
		    	}
		    }
		    // reading of effective Hamiltonmatrix
		    if(tempString=="HeffSize"){
		    	if(!(o>>k)){
		    		cerr << "reading error: HeffSize\n";
		    	}
		    	o>>noskipws>>temp>>skipws;
		    	g.Heff.resize(k);
		    	for(unsigned int i=0;i<k;i++){
		    		o>>g.Heff[i];
		    	}
		    }
		    // reading of Unitarymatrix
		    if(tempString=="unitaryCUT"){
		    	if(!(o>>g.unitaryCUT)){
		    		cerr << "reading error: unitaryCUT\n";
		    	}
		    }
		    if(tempString=="UCUTSize"){
		    	if(!(o>>k)){
		    		cerr << "reading error: UCUTSize\n";
		    	}
		    	o>>noskipws>>temp>>skipws;
		    	g.UCUT.resize(k);
		    	for(unsigned int i=0;i<k;i++){
		    		o>>g.UCUT[i];
		    	}
		    }
		    // reading of orthonormalize Matrix Sorth
		    if(tempString=="SorthSize"){
		    	if(!(o>>k)){
		    		cerr << "reading error: SorthSize\n";
		    	}
		    	o>>noskipws>>temp>>skipws;
		    	g.Sorth.resize(k);
		    	for(unsigned int i=0;i<k;i++){
		    		o>>g.Sorth[i];
		    	}
		    }
		}
	}
	else{
		cerr << "Graph not saved in parseMode!\n";
	}
    
	return o;
}

