/*
 * SpinOperators.h
 *
 *  Created on: Feb 3, 2014
 *      Author: ixert
 */

#ifndef SPINOPERATORS_H_
#define SPINOPERATORS_H_

#include "SeriesTemplate.h"
#include "nr3.h"

using namespace std;

template<class T>
class SpinOperators {
template<class S>
friend istream& operator>>(istream &o, SpinOperators<S> &s);
template<class S>
friend ostream& operator<<(ostream &o, SpinOperators<S> &s);
template<class S>
friend bool operator==(SpinOperators<S> const& lhs, SpinOperators<S> const& rhs);
private:
	bool sameOnEveryBond;
	list<pair<T,vector<vector<short> > > > operatorSequence; // Basis is S+, S-, Sz, E; T is amplitude
public:
	SpinOperators();

	bool isSameOnEveryBond();

	unsigned int numberOfOperators();

	bool isComplex();

	SpinOperators<T> permutate(vector<unsigned int> const &positions);

	SpinOperators<T>& operator+=(SpinOperators<T> const &rhs);

	vector<SpinOperatorsOnSite<> >
};

// free function declaration //
template<class T>
bool operator != (SpinOperators<T>  const& lhs, SpinOperators<T>  const& rhs);

template<class T>
const SpinOperators<T> operator+(SpinOperators<T> const &lhs, SpinOperators<T> const rhs);

// class function implementation //
template<>
inline bool SpinOperators<Complex>::isComplex();

template<class T>
SpinOperators<T>::SpinOperators() : sameOnEveryBond(0){
}

template<class T>
bool SpinOperators<T>::isSameOnEveryBond(){
	return sameOnEveryBond;
}

template<class T>
bool SpinOperators<T>::isComplex(){
	return false;
}

template<class T>
unsigned int SpinOperators<T>::numberOfOperators(){
	if(operatorSequence.size()>0){
		return (operatorSequence.begin()->second).size();
	}
	else{
		return 0;
	}
}

template<>
inline bool SpinOperators<Complex>::isComplex(){
	list<pair<Complex,vector<vector<short> > > >::const_iterator it;
	for(it=operatorSequence.begin();it!=operatorSequence.end();it++){
		if(std::imag(it->first)!=0){
			return true;
		}
	}
	return false;
}

template<class T>
SpinOperators<T> SpinOperators<T>::permutate(vector<unsigned int> const &positions){
	// positions are the new operator positions, e.g. positions={2 3 4 1} means: op1*op2*op3*op4 -> op2*op3*op4*op1
	SpinOperators<T> soRet;
	soRet.sameOnEveryBond=sameOnEveryBond;
	typename list<pair<T,vector<vector<short> > > >::const_iterator it;
	for(it=operatorSequence.begin();it!=operatorSequence.end();it++){
		if(positions.size()!=(it->second).size()){
			cerr << "SpinOperators<T>::permutate : positions.size() does not match number of operators!\n";
			return *this;
		}
		else{
			vector<vector<short> > newVec((it->second).size());
			for(unsigned int i=0;i<positions.size();i++){
				//cout << "positions[" << i << "]=" << positions[i] << endl;
				newVec[positions[i]]=(it->second)[i];
			}
			soRet.operatorSequence.push_back(pair<T,vector<vector<short> > >(it->first,newVec));
		}
	}
	return soRet;
}

template<class T>
SpinOperators<T>& SpinOperators<T>::operator+=(SpinOperators<T> const &rhs){
	//TODO: proof if sameOnEveryBond is identical!
	typename list<pair<T,vector<vector<short> > > >::const_iterator it1;
	for(it1=rhs.operatorSequence.begin();it1!=rhs.operatorSequence.end();it1++){
		typename list<pair<T,vector<vector<short> > > >::iterator it2;
		for(it2=operatorSequence.begin();it2!=operatorSequence.end();it2++){
			if((it2->second)==(it1->second)){
				break;
			}
		}
		if(it2!=operatorSequence.end()){
			it2->first+=it1->first;
		}
		else{
			operatorSequence.push_back(*it1);
		}
	}
	return *this;
}

//// friend function implementation ////
template<class T>
istream& operator>>(istream &o, SpinOperators<T> &s){
	//TODO: test this function (on right behaviour)!

	char tempChar;
	tempChar=o.peek();
	if(tempChar=='\n'){
		o>>noskipws>>tempChar>>skipws;
	}

	// get complete line of stream
	string tempstring;
	getline(o,tempstring);
	istringstream iss(tempstring);

	s.operatorSequence.clear();

	iss>>s.sameOnEveryBond;
	iss>>noskipws>>tempChar;

	T tempAmpl;
	int vz=1;
	if(iss.peek()=='+'){
		iss>>tempChar;
		vz=1;
	}
	else{
		if(iss.peek()=='-'){
			iss>>tempChar;
			vz=-1;
		}
	}
	while(iss>>tempAmpl){
		vector<vector<short> > productOperator;

		T ampl=T(vz)*tempAmpl;
		while(iss.peek()=='*'){
			iss>>tempChar;
			iss>>tempChar;
			vector<short> amplVec(4,0);
			//identify operator
			if(tempChar=='E'){
				amplVec[3]=1;
			}
			else{
				if(tempChar=='S'){
					iss>>tempChar;
					switch(tempChar){
						case 'p':{
							amplVec[0]=1;
							break;
						}
						case 'm':{
							amplVec[1]=1;
							break;
						}
						case 'z':{
							amplVec[2]=1;
							break;
						}
						case 'x':{
							amplVec[0]=1;
							amplVec[1]=1;
							break;
						}
						case 'y':{
							amplVec[0]=1;
							amplVec[1]=-1;
							break;
						}
					}
				}
			}
			productOperator.push_back(amplVec);
		}

		s.operatorSequence.push_back(pair<T,vector<vector<short> > >(ampl,productOperator));

		if(iss.peek()=='+'){
			iss>>tempChar;
			vz=1;
		}
		else{
			if(iss.peek()=='-'){
				iss>>tempChar;
				vz=-1;
			}
		}
	}

	if(s.operatorSequence.size()==0){
		o.setstate(ios_base::failbit);
		//cout << "Leaving with failbit ..." << endl;
	}

	return o;
}

template<class T>
ostream& operator<<(ostream &o, SpinOperators<T> &s){
	//TODO: implement this properly!
	typename list<pair<T,vector<vector<short> > > >::const_iterator it;
	o << s.sameOnEveryBond << " ";
	for(it=s.operatorSequence.begin();it!=s.operatorSequence.end();it++){
		o << "+" << it->first;
		for(unsigned int i=0;i<(it->second).size();i++){
			if((it->second)[i][0]!=0){
				if((it->second)[i][0]==(it->second)[i][1]){
					o << "*Sx";
				}
				else{
					if((it->second)[i][0]==-1*(it->second)[i][1]){
						o << "*Sy";
					}
					else{
						o << "*Sp";
					}
				}
			}
			else{
				if((it->second)[i][1]!=0){
					o << "*Sm";
				}
				else{
					if((it->second)[i][2]!=0){
						o << "*Sz";
					}
					else{
						if((it->second)[i][3]!=0){
							o << "*E";
						}
					}
				}
			}
		}
	}
	return o;
}

template<class T>
bool operator==(SpinOperators<T> const& lhs, SpinOperators<T> const& rhs){
	if((lhs.sameOnEveryBond==rhs.sameOnEveryBond) && (lhs.operatorSequence.size()==rhs.operatorSequence.size())){
		//cout << "global attributes are the same" << endl;
		// create copy of rhs
		list<pair<T,vector<vector<short> > > > operatorSequenceCopy=rhs.operatorSequence;
		typename list<pair<T,vector<vector<short> > > >::const_iterator it1;
		// iterate through the first operator
		for(it1=lhs.operatorSequence.begin();it1!=lhs.operatorSequence.end();it1++){
			// now search the second operator (the copied sequence) for the operator
			typename list<pair<T,vector<vector<short> > > >::iterator it2;
			bool foundElement=false;
			for(it2=operatorSequenceCopy.begin();it2!=operatorSequenceCopy.end();it2++){
				if((*it2)==(*it1)){
					foundElement=true;
					break;
				}
			}
			if(foundElement){
				//cout << "found element" << endl;
				// if element is found delete this one from the copied operatorSequence
				operatorSequenceCopy.erase(it2);
			}
			else{
				// if element is not found the operators are not equal!
				return false;
			}
		}
		return true;
	}
	else{
		return false;
	}
}

// free function implementation //
template<class T>
bool operator != (SpinOperators<T>  const& lhs, SpinOperators<T>  const& rhs){
  return ! (lhs==rhs);
}


template<class T>
const SpinOperators<T> operator+(SpinOperators<T> const &lhs, SpinOperators<T> const rhs){
	SpinOperators<T> ret=lhs;
	lhs+=rhs;
	return ret;
}

#endif /* SPINOPERATORS_H_ */
