/*
 * SpinOperators.h
 *
 *  Created on: Feb 3, 2014
 *      Author: ixert
 */

#ifndef SPINOPERATORS_H_
#define SPINOPERATORS_H_

#include "nr3.h"
#include<vector>
#include<complex>

using namespace std;

template<class T>
class SpinOperatorsOnSites {
template<class S>
friend istream& operator>>(istream &o, SpinOperatorsOnSites<S> &s);
template<class S>
friend ostream& operator<<(ostream &o, SpinOperatorsOnSites<S> &s);
//template<class S> TODO: output!
//friend ostream& operator<<(ostream &o, SpinOperatorsOnSites<S> &s);
template<class S>
friend bool operator==(SpinOperatorsOnSites<S> const& lhs, SpinOperatorsOnSites<S> const& rhs);
private:
	T prefactor; // this is not the Coupling, but can be used to identify the difference between the pauli-matrices and the spin-operator
	vector<pair<unsigned short, vector<short> > > operatorSequence; // Basis is Sx, Sy, Sz, E (where Sx=sigma_x, Sy=i*sigma_y, Sz=sigma_z
public:
	SpinOperatorsOnSites();
	unsigned int numberOfOperators();
	bool isComplex();
	vector<unsigned short> getSites();
	bool areSameSites(vector<unsigned short> const &sites);
	bool containsSite(unsigned short const &site);
	SpinOperatorsOnSites<T> permutate(vector<unsigned short> const &positions);
	void assignNewSites();
	void assignNewSites(vector<unsigned short> const &newSites);
	void assignNewSites(vector<unsigned short> const &oldSites, vector<unsigned short> const &newSites);
	void orderAgainstSites();

	bool isUnity();
};

// free function declaration //
template<class T>
bool operator != (SpinOperatorsOnSites<T>  const& lhs, SpinOperatorsOnSites<T>  const& rhs);

// class function implementation //
template<>
inline bool SpinOperatorsOnSites<Complex>::isComplex();

template<class T>
SpinOperatorsOnSites<T>::SpinOperatorsOnSites(){
}

template<class T>
bool SpinOperatorsOnSites<T>::isComplex(){
	return false;
}

template<class T>
unsigned int SpinOperatorsOnSites<T>::numberOfOperators(){
	return operatorSequence.size();
}

template<>
inline bool SpinOperatorsOnSites<Complex>::isComplex(){
	return (std::imag(prefactor)!=0);
}

template<class T>
vector<unsigned short> SpinOperatorsOnSites<T>::getSites(){
	vector<unsigned short> tempVec(operatorSequence.size());
	for(unsigned int i=0;i<operatorSequence.size();i++){
		tempVec[i]=operatorSequence[i].first;
	}
	return tempVec;
}

template<class T>
bool SpinOperatorsOnSites<T>::areSameSites(vector<unsigned short> const &sites){
	if(sites.size()!=operatorSequence.size()){
		return false;
	}
	for(unsigned int i=0;i<operatorSequence.size();i++){
		if(sites[i]!=operatorSequence[i].first){
			return false;
		}
	}
	return true;
}
template<class T>
bool SpinOperatorsOnSites<T>::containsSite(unsigned short const &site){
	for(unsigned int i=0;i<operatorSequence.size();i++){
		if(operatorSequence[i].first==site){
			return true;
		}
	}
	return false;
}

template<class T>
SpinOperatorsOnSites<T> SpinOperatorsOnSites<T>::permutate(vector<unsigned short> const &positions){
	// positions are the new operator positions, e.g. positions={2 3 4 1} means: op1*op2*op3*op4 -> op2*op3*op4*op1
	SpinOperatorsOnSites<T> soRet;
	soRet.prefactor=prefactor;
	if(positions.size()!=operatorSequence.size()){
		cerr << "SpinOperators<T>::permutate : positions.size() does not match number of operators!\n";
		return *this;
	}
	for(unsigned int i=0;i<positions.size();i++){
		soRet.operatorSequence.push_back(operatorSequence[positions[i]]);
	}
	return soRet;
}

template<class T>
void SpinOperatorsOnSites<T>::assignNewSites(){
	//this function assigns the new sites 0,1,2,..
	for(unsigned int i=0;i<operatorSequence.size();i++){
		operatorSequence[i].first=i;
	}
}
template<class T>
void SpinOperatorsOnSites<T>::assignNewSites(vector<unsigned short> const &newSites){
	//this function sets the sites to the newSites, BUT does not order them!
	if(newSites.size()!=operatorSequence.size()){
		cerr << "SpinOperators<T>::assignNewSites : newSites.size() does not match number of operators!\n";
	}
	else{
		for(unsigned int i=0;i<newSites.size();i++){
			operatorSequence[i].first=newSites[i];
		}
	}
}
template<class T>
void SpinOperatorsOnSites<T>::assignNewSites(vector<unsigned short> const &oldSites, vector<unsigned short> const &newSites){
	//this function sets the sites in oldSites to the sites in newSites, but does not order them.
	for(unsigned int i=0;i<operatorSequence.size();i++){
		for(unsigned int j=0;j<oldSites.size();j++){
			if(operatorSequence[i].first==oldSites[j]){
				operatorSequence[i].first=newSites[j];
				break;
			}
		}
	}
}

template<class T>
void SpinOperatorsOnSites<T>::orderAgainstSites(){
	// this function orders the operatorSequence in ascending order with respect to the siteNumbers
	SpinOperatorsOnSites<T> soRet;
	soRet.prefactor=prefactor;
	for(unsigned int i=0;i<operatorSequence.size();i++){
		unsigned short minSite=operatorSequence[i].first;
		unsigned int minSiteIndex=i;
		for(unsigned int j=i+1;j<operatorSequence.size();j++){
			if(operatorSequence[j].first<minSite){
				minSite=operatorSequence[j].first;
				minSiteIndex=j;
			}
		}
		if(minSiteIndex!=i){
			pair<unsigned short,vector<short> > tempPair=operatorSequence[minSiteIndex];
			operatorSequence[minSiteIndex]=operatorSequence[i];
			operatorSequence[i]=tempPair;
		}
	}
}

template<class T>
bool SpinOperatorsOnSites<T>::isUnity(){
	// returns true if the Operator is the Unity-Operator on ONE SINGLE SITE!
	if((operatorSequence.size()==1) && (prefactor==T(1))){
		if((operatorSequence[0].second[0]==0) && (operatorSequence[0].second[1]==0) && (operatorSequence[0].second[2]==0) && (operatorSequence[0].second[3]==1)){
			return true;
		}
	}
	return false;
}

//// friend function implementation ////
template<class T>
istream& operator>>(istream &o, SpinOperatorsOnSites<T> &s){
	// Input
	// This function reads an Single Operator in the Form (0.25,0)*Sx*Sx (almost the same as in in SpinOperators.h, see kaneMeleHubbardToSpinDerivator)
	// IMPORTANT: If no sites are specified the Sites are just set to 0,1,2,3,... and must be set to the correct sites using the assignSites-Method!
	// 28.08.2014: now this function also works with specified sites (e.g. Sx0*Sx2)
	char tempChar;

	s.operatorSequence.clear();

	T tempAmpl;
	int vz=1;
	if(o.peek()=='+'){
		o>>tempChar;
		vz=1;
	}
	else{
		if(o.peek()=='-'){
			o>>tempChar;
			vz=-1;
		}
	}
	o>>tempAmpl;
	s.prefactor=T(vz)*tempAmpl;

	unsigned short siteIndex=0;
	while(o.peek()=='*'){
		o>>tempChar;
		o>>tempChar;
		vector<short> amplVec(4,0);
		//identify operator
		if(tempChar=='E'){
			amplVec[3]=1;
		}
		else{
			if(tempChar=='S'){
				o>>tempChar;
				switch(tempChar){
					case 'x':{
						amplVec[0]=1;
						break;
					}
					case 'y':{
						amplVec[1]=1;
						break;
					}
					case 'z':{
						amplVec[2]=1;
						break;
					}
				}
			}
		}
		// check if a site is already defined
		unsigned short x;
		if(o>>x){
			s.operatorSequence.push_back(pair<unsigned short, vector<short> >(x,amplVec));
		}
		else{
			o.clear();
			s.operatorSequence.push_back(pair<unsigned short, vector<short> >(siteIndex,amplVec));
			siteIndex++;
		}
	}


	if(s.operatorSequence.size()==0){
		o.setstate(ios_base::failbit);
		//cout << "Leaving with failbit ..." << endl;
	}
	else{
		o.clear();
	}

	return o;
}

template<class T>
ostream& operator<<(ostream &o, SpinOperatorsOnSites<T> &s){
	o << s.prefactor;
	for(unsigned int i=0;i<s.operatorSequence.size();i++){
		o << "*";
		if(s.operatorSequence[i].second[3]){
			o << "E";
		}
		else{
			if(s.operatorSequence[i].second[0]){
				o << "Sx";
			}
			else{
				if(s.operatorSequence[i].second[1]){
					o << "Sy";
				}
				else{
					if(s.operatorSequence[i].second[2]){
						o << "Sz";
					}
				}
			}
		}
		o << s.operatorSequence[i].first;
	}
	return o;
}

template<class T>
bool operator==(SpinOperatorsOnSites<T> const& lhs, SpinOperatorsOnSites<T> const& rhs){
	return ((lhs.prefactor==rhs.prefactor) && (lhs.operatorSequence==rhs.operatorSequence));
}

// free function implementation //
template<class T>
bool operator != (SpinOperatorsOnSites<T>  const& lhs, SpinOperatorsOnSites<T>  const& rhs){
  return ! (lhs==rhs);
}

#endif /* SPINOPERATORS_H_ */
