/*
 * Solver.cpp
 *
 *  Created on: 26.10.2010
 *      Author: klagges
 *
 */

#include <Framework/Solver.h>
#include <fstream>

// To load global configuration and the module configuration.
Solver::Solver() {

	// Parse the command arguments to set the configuration files folder.

	std::ifstream ifs("/proc/self/cmdline");

	if (ifs.good()) {

		std::string temp;

		ifs >> temp;
		ifs.close();

		std::string::const_iterator endIt = temp.begin();
		std::string::const_iterator startIt = temp.begin();

		while (startIt != temp.end()) {
			if (*startIt == '\0') {
				startIt++;
				break;
			}
			startIt++;
		}

		endIt = startIt;

		while (endIt != temp.end()) {
			if (*endIt == '\0')
				break;
			endIt++;
		}

		if (startIt == endIt) {

			// the default configuration directory is 'Config/'

			configurationDirectory = std::string("Config");

		} else {

			configurationDirectory = std::string(startIt, endIt);

		}

	} else {

		// the default configuration directory is 'Config/'
		configurationDirectory = std::string("Config");

	}

	// Set the mpf_class default precision.
	// You can specify it in the Interface/Interface.h file.
	mpf_set_default_prec(DEFAULT_MPF_CLASS_PRECISION);

	// Set the starting time. Used to calculate the processing time of the solver.
	startingTime = time(NULL);

	// First read the global configurations
	globalConfiguration.readFromConfigFile("GlobalConfiguration.cfg");

	// Then read the module configuration
	modules.readFromConfigFile("Modules.cfg");

	// and create the instances of the modules.
	modules.createInstances();

	if (globalConfiguration.verbosityLevel > 0)
		std::cerr << "\n";

}

// Call the modules in the appropriate order.
int Solver::run() {

	// First check if a ModelProvider is loaded.
	if (!modules.pointerToModelProvider) {

		if (globalConfiguration.verbosityLevel > 0)
			std::cerr
					<< "ModelProvider disabled. Skipping execution of LatticeProvider and DataProcessor.\n";

	} else {

		while (true) {

			// This is the main loop of the LatticeProvider and the DataProcessor.

			if (modules.pointerToLatticeProvider) {
				if (globalConfiguration.verbosityLevel > 0)
					std::cerr << "Executing LatticeProvider.\n";

				if (modules.pointerToLatticeProvider->execute()) {

					// The lattice provider returned true, so we have to update
					// the globalProperties and call the DataProcessor.

					solver.globalProperties.bitsPerBaseState =
							solver.modelProperties.bitsPerLatticeSite
									* solver.latticeProperties.numberOfLatticeSites
									+ solver.latticeProperties.numberOfExtraBits;

					solver.globalProperties.eigenEnergy =
							modules.pointerToModelProvider->getH_0Eigenvalue(
									solver.inputStates.right.begin()->first);

					// Check if the BaseState class supplies enough bits to represent the lattice.

					if (solver.globalProperties.bitsPerBaseState
							> sizeof(BaseState) * 8) {
						std::cerr
								<< "Error! The number of bits to represent a base state of the lattice is "
								<< solver.globalProperties.bitsPerBaseState
								<< ", but the BaseState class only supports "
								<< (sizeof(BaseState) * 8)
								<< " bits.\n"
								<< "Please change the BaseState typedef in the Interface/Interface.h file and recompile.\n\n";
						exit(-1);

					} else {
						if (solver.globalConfiguration.verbosityLevel > 0
								&& (solver.globalProperties.bitsPerBaseState
										+ 32) < sizeof(BaseState) * 8) {
							std::cerr
									<< "Warning!  The number of bits to represent a base state of the lattice is "
									<< solver.globalProperties.bitsPerBaseState
									<< ", but the BaseState class supports "
									<< (sizeof(BaseState) * 8)
									<< " bits.\n"
									<< "To save memory you might want to change the BaseState typedef in the Interface/Interface.h file and recompile.\n\n";

						}
					}

					// Now call the DataProcessor.

					solver.outputAmplitude = stateAmplitude(amplitude(0));
					solver.outputState.clear();

					if (modules.pointerToDataProcessor) {
						if (globalConfiguration.verbosityLevel > 0)
							std::cerr << "Executing DataProcessor.\n";
						modules.pointerToDataProcessor->execute();
					}

					else if (globalConfiguration.verbosityLevel > 0)
						std::cerr << "DataProcessor disabled.\n";

				} else
					// The LatticeProvider returned false. We are finished.
					break;

			} else if (globalConfiguration.verbosityLevel > 0) {
				std::cerr
						<< "LatticeProvider disabled. Skipping execution of DataProcessor.\n";
				break;
			}

		}

	}

	if (globalConfiguration.verbosityLevel > 0)
		std::cerr << "\n";

	// Call the DataWriter to output the processed data.

	if (modules.pointerToDataWriter) {
		if (globalConfiguration.verbosityLevel > 0)
			std::cerr << "Executing DataWriter.\n";
		modules.pointerToDataWriter->execute();
	}

	else if (globalConfiguration.verbosityLevel > 0)
		std::cerr << "DataWriter disabled.\n";

	// The template module type.

	//	if (modules.pointerToTemplateModuleType) {
	//		std::cerr << "Executing TemplateModuleType.\n";
	//		modules.pointerToTemplateModuleType->execute();
	//	}
	//
	//	else
	//		std::cerr << "TemplateModuleType disabled.\n";

	return EXIT_SUCCESS;

}
