/*
 * BaseStateTemplate.h
 *
 *  Created on: 02.05.2011
 *      Author: klagges
 */

#ifndef BASESTATE_H_
#define BASESTATE_H_

#include <list>
#include <unordered_map>
#include <iostream>

#define BITS_PER_FIELD (sizeof(T) * 8)
#define FIELD_NUMBER(pos) ((pos) / BITS_PER_FIELD)
#define POS_IN_FIELD(pos) ((pos) % BITS_PER_FIELD)
#define FIELD_MASK(pos) ((T)1 << POS_IN_FIELD(pos))
#define POS_MASK(pos) ((T)1 << pos)

/** Lookup table that holds the number of bits set to one of an unsigned char number.*/
const unsigned int countTable[] = {

0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3,
		4, 3, 4, 4, 5, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 2, 3, 3,
		4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3,
		4, 3, 4, 4, 5, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 2, 3, 3,
		4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5,
		6, 5, 6, 6, 7, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 2, 3, 3,
		4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4,
		5, 4, 5, 5, 6, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 2, 3, 3,
		4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5,
		6, 5, 6, 6, 7, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 4, 5, 5,
		6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 };

/** Lookup table that holds the number of the first bit set to one of an unsigend char number.*/
const unsigned int countTable2[] = {

0, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2,
		1, 3, 1, 2, 1, 6, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2,
		1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 7, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2,
		1, 3, 1, 2, 1, 5, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 6, 1, 2,
		1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2,
		1, 3, 1, 2, 1, 8, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2,
		1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 6, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2,
		1, 3, 1, 2, 1, 5, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 7, 1, 2,
		1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2,
		1, 3, 1, 2, 1, 6, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 5, 1, 2,
		1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1 };

/** A class to represent a base state.
 *
 * Base states are represented by a number of bits. They are implemented as an array of some
 * basic data type. The data type and the size of the array can be specified as a template parameter.*/
template<typename T, unsigned int n>
class BaseStateTemplate {
public:

	/** The array to represent the bits of the base state.*/
	T data[n];

	/** Default constructor. */
	inline BaseStateTemplate() {
		memset(data, 0, sizeof(T[n]));
	}

	/** Create a base state representing the base type value i.*/
	inline BaseStateTemplate(T i) {
		memset(data, 0, sizeof(T[n]));
		data[0] = i;
	}

	/** Create a base state from a list of boolean values. Each boolean value
	 * is interpreted as a bit of the base state. */
	inline BaseStateTemplate(const std::list<bool>& list) {
		memset(data, 0, sizeof(T[n]));
		unsigned int index = 0;
		for (std::list<bool>::const_iterator it = list.begin();
				it != list.end(); it++) {
			if (*it)
				set(index);
			index++;
		}

	}

	/** Returns the bit value at position pos.*/
	inline bool get(const unsigned int pos) const {

		return (data[FIELD_NUMBER(pos)]) & FIELD_MASK(pos);

	}

	/** For lists with a SINGLE 1: returns the position of the first 1.*/
	inline unsigned int find1() const {

		unsigned int result = 0;

		const unsigned char* pointer = (const unsigned char*) data;

		for (unsigned int i = 0; i < sizeof(T[n]); i++) {

			if (countTable2[*pointer])
				result += i * BITS_PER_FIELD + countTable2[*pointer];

			pointer++;

		}

		return result;
	}

	/** Sets all bit values to zero (false).*/
	inline void clear() {
		memset(data, 0, sizeof(T[n]));
	}

	/** Set the value at position pos to one (true).*/
	inline void set(unsigned int pos) {

		data[FIELD_NUMBER(pos)] |= FIELD_MASK(pos);

	}

	/** Set the value at position pos to zero (false)*/
	inline void unSet(unsigned int pos) {

		data[FIELD_NUMBER(pos)] &= ~FIELD_MASK(pos);

	}

	/** Set the value at position pos to value.*/
	inline void setTo(const unsigned int pos, const bool value) {
		if (value)
			set(pos);
		else
			unSet(pos);
	}

	/** Interpret the base state as an value of the base type and return this (truncated) value.*/
	inline T toInteger() const {
		return data[0];
	}

	/** Interpret the bits of the base state from position minPos (inclusively) to maxPos (exclusively)
	 * as an unsigned integer and return it.*/
	inline unsigned int toInteger(const unsigned int minPos,
			const unsigned int maxPos) const {

		unsigned int result = 0;
		unsigned int mask = 1;

		for (unsigned int pos = minPos; pos < maxPos; pos++) {
			if (get(pos))
				result |= mask;
			mask <<= 1;

		}

		return result;
	}

	/** Canonical comparison of two base state objects.*/
	inline bool operator<(const BaseStateTemplate<T, n>& r) const {

		for (int i = n - 1; i > 0; i--) {
			if (data[i] < r.data[i])
				return true;
			if (data[i] > r.data[i])
				return false;
		}
		if (data[0] < r.data[0])
			return true;
		return false;

	}

	/** Test for equality of all bits.*/
	inline bool operator==(const BaseStateTemplate<T, n>& r) const {

		for (int i = n - 1; i >= 0; i--) {
			if (data[i] != r.data[i])
				return false;
		}

		return true;
	}

	/** Count and return the number of bits set to one in this base state.*/
	inline unsigned int count() const {

		unsigned int result = 0;

		const unsigned char* pointer = (const unsigned char*) data;

		for (unsigned int i = 0; i < sizeof(T[n]); i++) {

			result += countTable[*pointer];
			pointer++;

		}

		return result;
	}

	/** Count and return the number of bits set to one in the bytes from up to to.*/
	inline unsigned int count(const unsigned int from,
			const unsigned int to) const {

		unsigned int result = 0;

		const unsigned char* pointer = (const unsigned char*) &data;
		pointer += from;

		for (unsigned int i = from; i <= to; i++) {

			result += countTable[*pointer];
			pointer++;

		}

		return result;
	}

	/** Bitwise and operator.*/
	inline BaseStateTemplate<T, n> operator&(
			const BaseStateTemplate<T, n>& b) const {

		BaseStateTemplate<T, n> result(*this);

		for (int i = n - 1; i >= 0; i--) {
			result.data[i] &= b.data[i];
		}

		return result;

	}

	/** Bitwise and operator.*/
	inline BaseStateTemplate<T, n>& operator&=(
			const BaseStateTemplate<T, n>& b) {

		for (int i = n - 1; i >= 0; i--) {
			this->data[i] &= b.data[i];
		}

		return *this;

	}

	/** Bitwise or operator.*/
	inline BaseStateTemplate<T, n> operator|(
			const BaseStateTemplate<T, n>& b) const {

		BaseStateTemplate<T, n> result(*this);

		for (int i = n - 1; i >= 0; i--) {
			result.data[i] |= b.data[i];
		}

		return result;

	}

	/** Bitwise or operator.*/
	inline BaseStateTemplate<T, n>& operator|=(
			const BaseStateTemplate<T, n>& b) {

		for (int i = n - 1; i >= 0; i--) {
			this->data[i] |= b.data[i];
		}

		return *this;

	}

	/** Bitwise xor operator.*/
	inline BaseStateTemplate<T, n> operator^(
			const BaseStateTemplate<T, n>& b) const {

		BaseStateTemplate<T, n> result(*this);

		for (int i = n - 1; i >= 0; i--) {
			result.data[i] ^= b.data[i];
		}

		return result;

	}

	/** Bitwise xor operator.*/
	inline BaseStateTemplate<T, n>& operator^=(
			const BaseStateTemplate<T, n>& b) {

		for (int i = n - 1; i >= 0; i--) {
			this->data[i] ^= b.data[i];
		}

		return *this;

	}

	/** Bitwise not operator.
	 * Be aware, that all bits are inverted (and not only
	 * the first solver.globalProperties.numberOfBits bits).*/
	inline BaseStateTemplate<T, n> operator~() const {

		BaseStateTemplate<T, n> result;

		for (int i = n - 1; i >= 0; i--) {
			result.data[i] = ~this->data[i];
		}

		return result;

	}

};

/** A template specialization for base states represented by exactly one value of a basic data type.*/
template<typename T>
class BaseStateTemplate<T, 1> {
public:

	/** The basic data type to represent the bits of the base state.*/
	T data;

	/** Default constructor. */
	inline BaseStateTemplate() {
		data = 0;
	}

	/** Create a base state representing the base type value i.*/
	inline BaseStateTemplate(T i) {
		data = i;
	}

	/** Create a base state from a list of boolean values. Each boolean value
	 * is interpreted as a bit of the base state. */
	inline BaseStateTemplate(const std::list<bool>& list) {
		data = 0;
		unsigned int index = 0;
		for (std::list<bool>::const_iterator it = list.begin();
				it != list.end(); it++) {
			if (*it)
				set(index);
			index++;
		}

	}

	/** Returns the bit value at position pos.*/
	inline bool get(const unsigned int pos) const {

		return data & POS_MASK(pos);

	}

	/** Sets all bit values to zero.*/
	inline void clear() {
		data = 0;
	}

	/** Set the value at position pos to one (true).*/
	inline void set(unsigned int pos) {

		data |= POS_MASK(pos);

	}

	/** Set the value at position pos to zero (false)*/
	inline void unSet(unsigned int pos) {

		data &= ~POS_MASK(pos);

	}

	/** Set the value at position pos to value.*/
	inline void setTo(const unsigned int pos, const bool value) {
		if (value)
			set(pos);
		else
			unSet(pos);
	}

	/** Interprete the base state as an value of the base type and return this value.*/
	inline T toInteger() const {
		return data;
	}

	/** Interpret the bits of the base state from position minPos (inclusively) to maxPos (exclusively)
	 * as an unsigned integer and return it.*/
	inline unsigned int toInteger(const unsigned int minPos,
			const unsigned int maxPos) const {

		unsigned int result = 0;
		unsigned int mask = 1;

		for (unsigned int pos = minPos; pos < maxPos; pos++) {
			if (get(pos))
				result |= mask;
			mask <<= 1;

		}

		return result;
	}

	/** Canonical comparison of two base state objects.*/
	inline bool operator<(const BaseStateTemplate<T, 1>& r) const {

		if (data < r.data)
			return true;
		return false;

	}

	/** Test for equality of all bits.*/
	inline bool operator==(const BaseStateTemplate<T, 1>& r) const {

		if (data == r.data)
			return true;

		return false;
	}

	/** Count and return the number of bits set to one in this base state.*/
	inline unsigned int count() const {

		unsigned int result = 0;

		const unsigned char* pointer = (const unsigned char*) &data;

		for (unsigned int i = 0; i < sizeof(T); i++) {

			result += countTable[*pointer];
			pointer++;

		}

		return result;
	}

	/** Count and return the number of bits set to one in the bytes from up to to.*/
	inline unsigned int count(const unsigned int from,
			const unsigned int to) const {

		unsigned int result = 0;

		const unsigned char* pointer = (const unsigned char*) &data;
		pointer += from;

		for (unsigned int i = from; i <= to; i++) {

			result += countTable[*pointer];
			pointer++;

		}

		return result;
	}

	/** Bitwise and operator.*/
	inline BaseStateTemplate<T, 1> operator&(
			const BaseStateTemplate<T, 1>& b) const {

		BaseStateTemplate<T, 1> result(*this);

		result.data &= b.data;

		return result;

	}

	/** Bitwise and operator.*/
	inline BaseStateTemplate<T, 1>& operator&=(
			const BaseStateTemplate<T, 1>& b) {

		this->data &= b.data;

		return *this;

	}

	/** Bitwise or operator.*/
	inline BaseStateTemplate<T, 1> operator|(
			const BaseStateTemplate<T, 1>& b) const {

		BaseStateTemplate<T, 1> result(*this);

		result.data |= b.data;

		return result;

	}

	/** Bitwise or operator.*/
	inline BaseStateTemplate<T, 1>& operator|=(
			const BaseStateTemplate<T, 1>& b) {

		this->data |= b.data;

		return *this;

	}

	/** Bitwise xor operator.*/
	inline BaseStateTemplate<T, 1> operator^(
			const BaseStateTemplate<T, 1>& b) const {

		BaseStateTemplate<T, 1> result(*this);

		result.data ^= b.data;

		return result;

	}

	/** Bitwise xor operator.*/
	inline BaseStateTemplate<T, 1>& operator^=(
			const BaseStateTemplate<T, 1>& b) {

		this->data ^= b.data;

		return *this;

	}

	/** Bitwise not operator.
	 * Be aware, that all bits are inverted (and not only
	 * the first solver.globalProperties.numberOfBits bits).*/
	inline BaseStateTemplate<T, 1> operator~() const {

		return BaseStateTemplate<T, 1>(~this->data);

	}

};

namespace std {

/** Hash method for the BaseStateTemplate class.
 * Implements the Fowler-Noll-Vo algorithm. */
template<typename T>
class hash<BaseStateTemplate<T, 1> > {
public:

	/** Hash method for the BaseStateTemplate class.
	 * Implements the Fowler-Noll-Vo algorithm. */
	inline std::size_t operator()(const BaseStateTemplate<T, 1> &key) const {

		std::size_t hash = 14695981039346656037ull;
		// Use this for 32-bit computers:
		//std::size_t hash = 2166136261ul;
		const unsigned char* pointer = (const unsigned char*) &key.data;
		for (unsigned int i = 0; i < sizeof(T); i++) {
			hash ^= *pointer++;

			hash += (hash << 1) + (hash << 4) + (hash << 5) + (hash << 7)
					+ (hash << 8) + (hash << 40);

			// Use this for 32-bit computers:
			// hash += (hash << 1) + (hash << 4) + (hash << 7) + (hash << 8)
			//	   	+ (hash << 24);

		}

		return hash;

	}

};

/** Hash method for the BaseStateTemplate class.
 * Implements the Fowler-Noll-Vo algorithm. */
template<typename T, unsigned int n>
class hash<BaseStateTemplate<T, n> > {
public:

	/** Hash method for the BaseStateTemplate class.
	 * Implements the Fowler-Noll-Vo algorithm. */
	inline std::size_t operator()(const BaseStateTemplate<T, n> &key) const {

		std::size_t hash = 14695981039346656037ull;
		// Use this for 32-bit computers:
		// std::size_t hash = 2166136261ul;
		const unsigned char* pointer = (const unsigned char*) key.data;
		for (unsigned int i = 0; i < sizeof(T[n]); i++) {
			hash ^= *pointer++;

			hash += (hash << 1) + (hash << 4) + (hash << 5) + (hash << 7)
					+ (hash << 8) + (hash << 40);

			// Use this for 32-bit computers:
			// hash += (hash << 1) + (hash << 4) + (hash << 7) + (hash << 8)
			//         + (hash << 24);

		}

		return hash;

	}

};

// You might squeeze out a little bit more performance by implementing a custom hash function for your data type.
/*template<>
 class hash<BaseStateTemplate<unsigned long long, 1> > {
 public:
 inline std::size_t operator()(const BaseStateTemplate<unsigned long long, 1> &key) const {

 return key.data;

 }

 };*/

// You might squeeze out a little bit more performance by implementing a custom hash function for your data type.
/*template<>
 class hash<BaseStateTemplate<unsigned long long, 2> > {
 public:
 inline std::size_t operator()(const BaseStateTemplate<unsigned long long, 2> &key) const {

 return key.data[0] ^ key.data[1];

 }

 };*/

}

#endif /* BASESTATE_H_ */
