/*
 * ExpandedAmplitudeTemplate.h
 *
 *  Created on: 07.01.2011
 *      Author: klagges
 */

#ifndef EXPANDEDAMPLITUDE_H_
#define EXPANDEDAMPLITUDE_H_

#include <utility>
#include <Utilities/Utilities.h>
#include <Utilities/EqualsZero.h>

/** Representation of an expanded amplitude of for example a ::BaseState or an energy correction.
 * The expansion goes over the expansion parameters of the model. For example:
 * \f[
 * c_0\cdot h_1^3 + c_1\cdot h_1^2 h_2^2 + c_2\cdot h_1 h_2^2 + c_3\cdot h_2^4
 * \f]
 * In this example there are two expansion parameters. The sequence is stored symbolically by storing all non zero
 * (::amplitude) c amplitudes in a associative container with the according ::Exponents object as the key.
 * You can specify the type of the used container, the ::Exponents class and the type of the ::amplitude
 * as the template parameters in the Interface/Interface.h file.
 */
template<typename cont, typename exp, typename amp>
class ExpandedAmplitudeTemplate: public cont {

public:

	/** Create an expanded amplitude representing the value '0'.
	 * This means, that no coefficients are present. */
	inline ExpandedAmplitudeTemplate() :
			cont() {

	}

	/** Create an expanded amplitude representing the ::amplitude value x.
	 * This means, that only the Exponents term with all exponents being zero with coefficient x exists. */
	inline ExpandedAmplitudeTemplate(const amp& x) :
			cont() {
		if (!Utilities::equalsZero(x))
			this->insert(typename cont::value_type(exp(), x));

	}

	/** Copy constructor.*/
	inline ExpandedAmplitudeTemplate(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) :
			cont(e) {

	}

	/** Assignment operator.*/
	inline ExpandedAmplitudeTemplate& operator=(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) {

		cont::operator=(e);

		return *this;

	}

	/** Assignment operator.*/
	inline ExpandedAmplitudeTemplate& operator=(const amp& e) {

		cont::operator=(ExpandedAmplitudeTemplate<cont, exp, amp>(e));

		return *this;

	}

	/** The sum of two expanded coefficients.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp> operator+(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) const {

		ExpandedAmplitudeTemplate<cont, exp, amp> result(e);

		for (typename cont::const_iterator it = this->begin();
				it != this->end(); it++) {

			std::pair<typename cont::iterator, bool> insertResult =
					result.insert(*it);

			if (!insertResult.second) {

				if (Utilities::equalsZero(
						insertResult.first->second += it->second))
					result.erase(insertResult.first);

			}

		}

		return result;

	}

	/** The sum of two expanded coefficients.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp>& operator+=(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) {

		for (typename cont::const_iterator it = e.begin(); it != e.end();
				it++) {

			std::pair<typename cont::iterator, bool> insertResult = insert(*it);

			if (!insertResult.second) {

				if (Utilities::equalsZero(
						insertResult.first->second += it->second))
					erase(insertResult.first);

			}

		}

		return *this;
	}

	/** Multiplying two expanded coefficients.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp> operator*(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) const {

		ExpandedAmplitudeTemplate<cont, exp, amp> result;

		for (typename cont::const_iterator it = this->begin();
				it != this->end(); it++) {
			for (typename cont::const_iterator it2 = e.begin(); it2 != e.end();
					it2++) {

				typename cont::value_type resultingEntry(it->first * it2->first,
						it->second * it2->second);

				std::pair<typename cont::iterator, bool> insertResult =
						result.insert(resultingEntry);

				if (!insertResult.second) {

					if (Utilities::equalsZero(insertResult.first->second +=
							resultingEntry.second))
						result.erase(insertResult.first);

				}
			}
		}

		return result;

	}

	/** Multiplying two expanded coefficients.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp>& operator*=(
			const ExpandedAmplitudeTemplate<cont, exp, amp>& e) {

		ExpandedAmplitudeTemplate<cont, exp, amp> result;

		for (typename cont::const_iterator it = this->begin();
				it != this->end(); it++) {
			for (typename cont::const_iterator it2 = e.begin(); it2 != e.end();
					it2++) {

				typename cont::value_type resultingEntry(it->first * it2->first,
						it->second * it2->second);

				std::pair<typename cont::iterator, bool> insertResult =
						result.insert(resultingEntry);

				if (!insertResult.second) {

					if (Utilities::equalsZero(insertResult.first->second +=
							resultingEntry.second))
						result.erase(insertResult.first);

				}

			}

		}

		this->swap(result);

		return *this;

	}

	/** Multiply an ::amplitude number to the expanded coefficient.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp> operator*(
			const amp& r) const {

		if (Utilities::equalsZero(r))
			return ExpandedAmplitudeTemplate<cont, exp, amp>();

		ExpandedAmplitudeTemplate<cont, exp, amp> result(*this);

		for (typename cont::iterator it = result.begin(); it != result.end();
				it++) {
			it->second *= r;
		}

		return result;

	}

	/** Multiply an ::amplitude number to the expanded coefficient.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp>& operator*=(const amp& r) {

		if (Utilities::equalsZero(r)) {
			this->clear();
			return *this;
		}

		for (typename cont::iterator it = this->begin(); it != this->end();
				it++) {
			it->second *= r;
		}

		return *this;

	}

	/** Divide the expanded coefficient by an ::amplitude number.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp> operator/(
			const amp& r) const {

		ExpandedAmplitudeTemplate<cont, exp, amp> result(*this);

		for (typename cont::iterator it = result.begin(); it != result.end();
				it++) {
			it->second /= r;
		}

		return result;

	}

	/** Divide the expanded coefficient by an ::amplitude number.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp>& operator/=(const amp& r) {

		for (typename cont::iterator it = this->begin(); it != this->end();
				it++) {
			it->second /= r;
		}

		return *this;

	}

	/** Comparison operator with ::amplitude.*/
	inline bool operator==(const amp& r) const {

		if (Utilities::equalsZero(r)) {
			if (this->size() == 0)
				return true;
			return false;
		}
		if (this->size() == 1 && this->begin()->first.equalsOne()
				&& this->begin()->second == r)
			return true;
		return false;

	}

	/** Negate the expanded amplitude.*/
	inline ExpandedAmplitudeTemplate<cont, exp, amp> operator-() const {

		ExpandedAmplitudeTemplate<cont, exp, amp> result(*this);

		for (typename cont::iterator it = result.begin(); it != result.end();
				it++) {
			it->second = -it->second;
		}

		return result;

	}

	/** Add the term r*e to the ExpandedAmplitudeTemplate.*/
	inline void add(const amp& r, const exp& e) {

		typename cont::value_type insertValue(e, r);

		std::pair<typename cont::iterator, bool> insertResult = this->insert(
				insertValue);

		if (!insertResult.second) {

			if (Utilities::equalsZero(insertResult.first->second += r))
				erase(insertResult.first);

		}

	}

};

namespace Utilities {

template<typename cont, typename exp, typename amp>
inline bool equalsZero(
		const ExpandedAmplitudeTemplate<cont, exp, amp> & value) {
	if (value.size() == 0)
		return true;
	return false;
}

}
#endif /* EXPANDEDAMPLITUDE_H_ */
