/*
 * ExponentsTemplate.h
 *
 *  Created on: 10.01.2011
 *      Author: klagges
 */

#ifndef EXPONENTS_H_
#define EXPONENTS_H_

#include <unordered_map>

/**
 * A class to represent a sequence of powers of expansion parameters.
 * For example:
 * \f[
 * \lambda_1^{e_1}\cdot \lambda_2^{e_2}\cdot \lambda_3^{e_3}
 * \f]
 * In this example there are three expansion parameters.
 * The sequence is stored symbolically by storing the exponents in an array of unsigned chars.
 * You can specify the number of expansion parameters as the template parameter in the
 * Interface/Interface.h file.
 */
template<unsigned int n>
class ExponentsTemplate {

public:

	/** The array to store the exponents of this object.*/
	char e[n];

	/** Create a ExponentsTemplate object representing the constant value '1'.
	 * That means, all exponents are zero.
	 */
	inline ExponentsTemplate() {
		memset(e, 0, n);
	}

	/** Copy constructor.*/
	inline ExponentsTemplate(const ExponentsTemplate<n>& ex) {
		memcpy(e, ex.e, n);

	}

	/** Assignment operator.*/
	inline ExponentsTemplate& operator=(const ExponentsTemplate<n>& ex) {

		memcpy(e, ex.e, n);

		return *this;
	}

	/** Check if the object represents the constant '1' value (all exponents are zero). */
	inline bool equalsOne() const {

		for (unsigned int i = 0; i < n; i++) {

			if (e[i] != 0)
				return false;

		}

		return true;

	}

	/** Multiplying two sequences of expansion parameters means adding their exponents.*/
	inline ExponentsTemplate operator*(const ExponentsTemplate<n>& ex) const {

		ExponentsTemplate<n> result(*this);

		for (unsigned int index = 0; index < n; index++) {

			result.e[index] += ex.e[index];

		}

		return result;

	}

	/** Multiplying two sequences of expansion parameters means adding their exponents.*/
	inline ExponentsTemplate& operator*=(const ExponentsTemplate<n>& ex) {

		for (unsigned int index = 0; index < n; index++) {

			e[index] += ex.e[index];

		}

		return *this;

	}

	/** Return the order (sum of exponents) of this ExponentsTemplate object.*/
	inline unsigned int getOrder() const {

		unsigned int order = 0;

		for (unsigned int i = 0; i < n; i++) {

			order += e[i];

		}

		return order;

	}

	/** Canonical comparison of two ExponentsTemplate objects.*/
	inline bool operator<(const ExponentsTemplate<n>& r) const {

		for (unsigned int i = n - 1; i > 0; i--) {
			if (e[i] < r.e[i])
				return true;
			if (e[i] > r.e[i])
				return false;
		}
		if (e[0] < r.e[0])
			return true;
		return false;

	}

	/** Test for equality.*/
	inline bool operator==(const ExponentsTemplate<n>& r) const {

		for (unsigned int i = 0; i < n; i++) {
			if (e[i] != r.e[i])
				return false;

		}
		return true;

	}

};

namespace std {

/** Hash method for ExponentsTemplate objects.
 * Implements the Fowler-Noll-Vo algorithm. */
template<unsigned int n>
class hash<ExponentsTemplate<n> > {
public:
	inline std::size_t operator()(const ExponentsTemplate<n> &key) const {

		std::size_t hash = 14695981039346656037ull;
		// Use this for 32-bit computers:
		// std::size_t hash = 2166136261ul;
		const unsigned char* pointer = key.e;
		for (unsigned int i = 0; i < n; i++) {
			hash ^= *pointer++;

			hash += (hash << 1) + (hash << 4) + (hash << 5) + (hash << 7)
					+ (hash << 8) + (hash << 40);

			// Use this for 32-bit computers:
			// hash += (hash << 1) + (hash << 4) + (hash << 7) + (hash << 8)
			//		+ (hash << 24);

		}

		return hash;

	}

};

}

#endif /* EXPONENTS_H_ */
