/*
 * Interface.h
 *
 *  Created on: 05.05.2011
 *      Author: klagges
 */

#ifndef INTERFACE_H_
#define INTERFACE_H_

#include <Interface/ExpandedAmplitude.h>
#include <Interface/State.h>
#include <Interface/Exponents.h>
#include <Interface/BaseState.h>
#include <gmpxx.h>
#include <Utilities/AssociativeList.h>
#include <complex.h>
#include <Utilities/PoolAllocator.h>

typedef mpq_class rational;
#define DEFAULT_MPF_CLASS_PRECISION 512

/** You may choose the data type to represent the scalar amplitudes in the solver.
 * You may choose between rational numbers (gmp), various float types and complex numbers.
 * The optimal choice depends on the properties of the model.*/

//typedef rational amplitude;
typedef mpf_class amplitude;
//typedef double amplitude;
//typedef float amplitude;
//typedef std::complex<rational> amplitude;
//typedef std::complex<mpf_class> amplitude;
//typedef std::complex<double> amplitude;
//typedef std::complex<float> amplitude;

/** You may choose how to represent base states.
 * Base states are represented by an array of some basic data type.
 * You may choose the base data type and the size of the array of this data type.
 * It must be chosen large enough to contain all bits necessary for your model/lattice.
 * If it is possible to fit all bits into a single basic data type this should be preferred. */

//typedef BaseStateTemplate<unsigned long long, 2> BaseState;
typedef BaseStateTemplate<unsigned int, 1> BaseState;
//typedef BaseStateTemplate<unsigned short, 1> BaseState;
//typedef BaseStateTemplate<unsigned char, 6> BaseState;

//typedef BaseStateTemplate<unsigned long long, 2> BaseState;
//typedef BaseStateTemplate<unsigned int, 4> BaseState;
//typedef BaseStateTemplate<unsigned short, 8> BaseState;
//typedef BaseStateTemplate<unsigned char, 16> BaseState;


/** You may choose the number of exponents stored in the ::Exponents class.
 * It should match the number of expansion parameters of your model. */

typedef ExponentsTemplate<1> Exponents;
//typedef ExponentsTemplate<2> Exponents;
//typedef ExponentsTemplate<3> Exponents;
//typedef ExponentsTemplate<4> Exponents;
//typedef ExponentsTemplate<5> Exponents;
//typedef ExponentsTemplate<6> Exponents;
//typedef ExponentsTemplate<8> Exponents;

/** You may choose which associative container should be used by the ::ExpandedAmplitude class
* to store the linear combination of exponents.
* You may choose between the Utilities::associative_list, the std::map and the std::unordered_map.
* The Utilities::associative_list has the best memory performance and for most cases also the best
* speed performance. For some special models with a very high number of expansion parameters one of
* the other containers might be a bit faster. */

typedef Utilities::associative_list<Exponents, amplitude>
		expandedAmplitudeContainer;
//typedef std::map<Exponents, amplitude> expandedAmplitudeContainer;
//typedef std::unordered_map<Exponents, amplitude> expandedAmplitudeContainer;

/** Typedef for the ::ExpandedAmplitude class. */
typedef ExpandedAmplitudeTemplate<expandedAmplitudeContainer, Exponents,
		amplitude> ExpandedAmplitude;

/** You may choose the type of the amplitude of a ::BaseState used in a ::State object.
 * You may choose between the ::ExpandedAmplitude class and the ::amplitude type.
 * Choosing the ::ExpandedAmplitude means, that the solver keeps track of the expansion
 * parameters. Choosing the ::amplitude type effectively sets all expansion parameters
 * to one.
 * You must use the ::ExpandedAmplitude class if you want to keep track of multiple
 * expansion parameters. If you only want a single expansion parameter you can use ::amplitude
 * (saving a lot of memory and time) but must calculate every order separately. Only the Loewdin
 * solvers are capable to calculate all orders simultaneously in this case. */

typedef ExpandedAmplitude stateAmplitude;
//typedef amplitude stateAmplitude;

/** You may choose which associative container should be used by the ::State class
 * to represent the sparse vectors of states.
 * You may choose for example between the std::map and the std::unordered map
 * either with the default allocator or with a pool allocator.
 * The std::unordered_map is more memory efficient then the std::map. The speed
 * efficiency strongly depends on the model and the used method.
 * The std::unordered_map seems be more sensitive to memory fragmentation.
 * So in cases where a lot of different states need to be tracked (for example in
 * case of a state splitting by the TwoSideSolvers) the std::map might be preferred.
 *
 * In both cases the PoolAllocators seems to perform better, but it still is
 * a bit experimental. One should note, that the SimplePoolAllocator only
 * deallocates memory back to the operating system, if the state is completely cleared.
 * If this is not acceptable you should use the TrackedPoolAllocator
 * or the default allocator. */

//typedef std::unordered_map<BaseState, stateAmplitude, std::hash<BaseState>,
//		std::equal_to<BaseState>, Utilities::SimplePoolAllocator<std::pair<
//				const BaseState, stateAmplitude> > > stateContainer;

//typedef std::map<BaseState, stateAmplitude, std::less<BaseState>,
//		Utilities::SimplePoolAllocator<std::pair<const BaseState, stateAmplitude> > >
//		stateContainer;

//typedef std::unordered_map<BaseState, stateAmplitude, std::hash<BaseState>,
//		std::equal_to<BaseState>, Utilities::TrackedPoolAllocator<std::pair<
//				const BaseState, stateAmplitude> > > stateContainer;

//typedef std::map<BaseState, stateAmplitude, std::less<BaseState>,
//		Utilities::TrackedPoolAllocator<std::pair<const BaseState, stateAmplitude> > >
//		stateContainer;

typedef std::unordered_map<BaseState, stateAmplitude, std::hash<BaseState>,
		std::equal_to<BaseState> > stateContainer;
//typedef std::map<BaseState, stateAmplitude> stateContainer;

/** Typedef for the ::State class. */
typedef StateTemplate<stateContainer, BaseState, stateAmplitude> State;

#endif /* INTERFACE_H_ */
