/*
 * State.h
 *
 *  Created on: 23.11.2010
 *      Author: klagges
 */

#ifndef STATE_H_
#define STATE_H_

#include <map>
#include <vector>
#include <Utilities/EqualsZero.h>

/** The data structure to represent a state as a sparse vector of ::BaseState objects.
 You can specify the type of the used container class, the ::BaseState class and the type of the ::stateAmplitude
 as the template parameters in the Interface/Interface.h file.
 */
template<typename cont, typename base, typename amp>
class StateTemplate: public cont {

public:

	/** Enumeration to describe if the state is a bra or ket vector.*/
	enum Type {
		bra, ket
	};

	/** The type of the state (bra or ket vector).*/
	Type type;

	/** Create an empty StateTemplate.*/
	StateTemplate() :
		cont(), type(ket) {

	}

	/** Copy constructor.*/
	StateTemplate(const StateTemplate& s) :
		cont(), type(s.type) {
		insert(s.begin(), s.end());

	}

	/** Assignment operator.*/
	StateTemplate& operator=(const StateTemplate& s) {

		this->clear();
		insert(s.begin(), s.end());
		type = s.type;

		return *this;

	}

	/** Component vice sum.*/
	StateTemplate& operator+=(const StateTemplate& s) {

		for (typename cont::const_iterator it = s.begin(); it != s.end(); it++) {

			std::pair<typename cont::iterator, bool> insertResult = insert(*it);

			if (!insertResult.second) {

				if (Utilities::equalsZero(insertResult.first->second
						+= it->second))
					erase(insertResult.first);

			}

		}

		return *this;

	}

	/** Component vice sum.*/
	StateTemplate operator+(const StateTemplate& s) const {

		StateTemplate result(*this);

		for (typename cont::const_iterator it = s.begin(); it != s.end(); it++) {

			std::pair<typename cont::iterator, bool> insertResult =
					result.insert(*it);

			if (!insertResult.second) {

				if (Utilities::equalsZero(insertResult.first->second
						+= it->second))
					result.erase(insertResult.first);

			}

		}

		return result;

	}

	/** Multiplication with a scalar.*/
	StateTemplate& operator*=(const amp& m) {

		for (typename cont::iterator it = this->begin(); it != this->end(); it++) {
			it->second *= m;
		}

		return *this;

	}

	/** Multiplication with a scalar.*/
	StateTemplate operator*(const amp& m) const {

		StateTemplate result(*this);

		for (typename cont::iterator it = result.begin(); it != result.end(); it++) {
			it->second *= m;

		}
		return result;

	}

	/** this = this + m * s;*/
	void addMultiple(const StateTemplate<cont, base, amp>& s, const amp& m) {

		for (typename cont::const_iterator it = s.begin(); it != s.end(); it++) {

			typename cont::value_type resultingEntry(it->first, it->second * m);

			std::pair<typename cont::iterator, bool> insertResult = insert(
					resultingEntry);

			if (!insertResult.second) {

				if (Utilities::equalsZero(insertResult.first->second
						+= resultingEntry.second))
					erase(insertResult.first);

			}

		}

	}

	/** this = this + m * bs;*/
	void add(const base& bs, const amp& m) {

		typename cont::value_type newEntry(bs, m);

		std::pair<typename cont::iterator, bool> insertResult =
				insert(newEntry);

		if (!insertResult.second) {

			if (Utilities::equalsZero(insertResult.first->second += m))
				erase(insertResult.first);

		}

	}

	/** Calculate the norm <i|i>.*/
	amp norm() const;

	/** Calculate the value <i|S|i>.*/
	amp normWithS() const;

	/** Split the state into numberOfSplits states of roughly equal length. All
	 * split states have no common base states. Base states are distributed among
	 * the split states in a way, so that base states with equal or similar eigen energy
	 * should be in the same split state.*/
	std::vector<StateTemplate<cont, base, amp> > splitState(
			unsigned int numberOfSplits);

	/** Count the base states of the state, that have common energy levels.
	 * In the resulting map, every double energy level is associated to the number
	 * of occurring base states with this energy level. */
	std::map<double, unsigned int> countStatesByEnergy() const;

};

#endif /* STATE_H_ */
