/*
 * StateIO.h
 *
 *  Created on: 23.11.2010
 *      Author: klagges
 */

#include <Utilities/Utilities.h>
#include <Interface/BaseStateIO.h>
#include <Interface/ExpandedAmplitudeIO.h>
#include <Utilities/TypeCast.h>
#include <Utilities/IsPositive.h>
#include <Utilities/ComplexConjugate.h>

/** Stream out as a string like 1|0000>-3*h0|0001>+2*h0^2*h1|0101> ;*/
template<typename cont, typename base, typename amp>
std::ostream& operator<<(std::ostream& o,
		const StateTemplate<cont, base, amp>& d) {

	typename cont::const_iterator it = d.begin();

	if (it != d.end()) {

		if (Utilities::isPositive(it->second))
			o << "+";

		o << it->second;

		o << it->first << " ";

		it++;

	}

	while (it != d.end()) {

		o << "\n";
		o << it->second;

		o << (BaseState&) it->first << " ";

		it++;

	}

	o << ";";

	return o;

}

/** Stream in from a string like 1|0000> -3*h0|0001> +2*h0^2 h1|0101> ;*/
template<typename cont, typename base, typename amp>
std::istream& operator>>(std::istream& i, StateTemplate<cont, base, amp>& d) {

	std::string s;

	while (i.good()) {

		char peek = i.get();

		if (peek == ';' || peek == EOF)
			break;

		s.push_back(peek);

	}

	std::stringstream ss(s);

	while (ss.good()) {

		ExpandedAmplitude ec;

		ss >> ec;

		if (!ss.good())
			break;

		BaseState b;
		ss >> b;

		stateAmplitude amplitude;
		Utilities::typeCast(ec, amplitude);

		d.add(b, amplitude);

	}

	return i;
}

template<typename cont, typename base, typename amp>
amp operator*(const StateTemplate<cont, base, amp>& left
		, const StateTemplate<cont, base, amp>& right) {

	amp result(amplitude(0));

	typename cont::const_iterator stateIt;

	if (left.size() < right.size()) {

		for (typename cont::const_iterator it = left.begin(); it != left.end();
				it++) {

			stateIt = right.find(it->first);

			if (stateIt != right.end()) {

				amp conjugate(it->second);
				Utilities::ComplexConjugate(conjugate);
				result += conjugate * stateIt->second;

			}

		}

	} else {

		for (typename cont::const_iterator it = right.begin();
				it != right.end(); it++) {

			stateIt = left.find(it->first);

			if (stateIt != left.end()) {

				amp conjugate(stateIt->second);
				Utilities::ComplexConjugate(conjugate);
				result += conjugate * it->second;

			}

		}

	}

	return result;

}

template<typename cont, typename base, typename amp>
amp StateTemplate<cont, base, amp>::norm() const {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "norm\n";

	amp result(amplitude(0));

	for (typename cont::const_iterator it = this->begin(); it != this->end();
			it++) {

		amp conjugate(it->second);
		Utilities::ComplexConjugate(conjugate);

		result += conjugate * it->second;

	}

	return result;

}

template<typename cont, typename base, typename amp>
amp StateTemplate<cont, base, amp>::normWithS() const {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "normWithS\n";

	amp result(amplitude(0));

	for (typename cont::const_iterator it = this->begin(); it != this->end();
			it++) {

		double correctionFactor = solver.globalProperties.eigenEnergy
				- solver.modules.pointerToModelProvider->getH_0Eigenvalue(
						it->first);

		if (correctionFactor == 0)
			continue;

		amplitude castedCorrectionFactor;
		Utilities::typeCast(correctionFactor, castedCorrectionFactor);

		amp conjugate(it->second);
		Utilities::ComplexConjugate(conjugate);

		result += (conjugate * it->second) / castedCorrectionFactor;

	}

	return result;

}

template<typename cont, typename base, typename amp>
std::map<double, unsigned int> StateTemplate<cont, base, amp>::countStatesByEnergy() const {

	std::map<double, unsigned int> result;

	for (typename cont::const_iterator it = this->begin(); it != this->end();
			it++) {

		result[solver.modules.pointerToModelProvider->getH_0Eigenvalue(
				it->first)] += 1;

	}

	return result;

}

template<typename cont, typename base, typename amp>
std::vector<StateTemplate<cont, base, amp> > StateTemplate<cont, base, amp>::splitState(
		unsigned int numberOfSplits) {

	std::vector<StateTemplate<cont, base, amp> > result(numberOfSplits);

	unsigned int statesPerSplit = this->size() / numberOfSplits;

	std::map<double, unsigned int> energyCounter = countStatesByEnergy();

	std::map<double, std::list<std::pair<unsigned int, unsigned int> > > associatedBaseStates;

	unsigned int currentSplitNumber = 0;
	unsigned int currentBaseStateCounter = statesPerSplit;

	for (std::map<double, unsigned int>::iterator it = energyCounter.begin();
			it != energyCounter.end(); it++) {

		while (it->second > 0) {

			if (currentBaseStateCounter > it->second
					|| currentSplitNumber == numberOfSplits - 1) {

				associatedBaseStates[it->first].push_back(
						std::pair<unsigned int, unsigned int>(
								currentSplitNumber, it->second));
				currentBaseStateCounter -= it->second;
				it->second = 0;

			} else {
				associatedBaseStates[it->first].push_back(
						std::pair<unsigned int, unsigned int>(
								currentSplitNumber, currentBaseStateCounter));
				it->second -= currentBaseStateCounter;
				currentBaseStateCounter = statesPerSplit;
				currentSplitNumber++;

			}

		}

	}

	typename cont::iterator it = this->begin();

	typename cont::iterator tempIt;

	while (it != this->end()) {

		std::map<double, std::list<std::pair<unsigned int, unsigned int> > >::iterator it2 =
				associatedBaseStates.find(
						solver.modules.pointerToModelProvider->getH_0Eigenvalue(
								it->first));
		result[it2->second.begin()->first].insert(*it);
		it2->second.begin()->second--;
		if (it2->second.begin()->second == 0)
			it2->second.pop_front();

		tempIt = it;

		it++;

		erase(tempIt);

	}

	return result;
}
