/*
 * Term.h
 *
 *  Created on: 22.11.2010
 *      Author: klagges
 */

#ifndef LOEWDINTERM_H_
#define LOEWDINTERM_H_

#include <iostream>
#include <list>
#include <vector>
#include <Interface/Interface.h>

/** Name space to capsule all classes related to the Loewdin degenerated perturbation solvers.
 *
 * In case of the operator formalism of degenerated perturbation theory one separates the
 * solutions of the general Schrödingergleichung into a part parallel to the \f$ P \f$ eigen space
 * and into a part parallel to \f$ Q = 1 - P \f$.
 *
 * \f{eqnarray*}{
 * |\psi_{i}\rangle&=&P|\psi_{i}\rangle+Q|\psi_{i}\rangle\\
 * &=:&|\psi_{i}\rangle_{L}+|\psi_{i}\rangle_{\bar{L}}.
 * \f}
 *
 * By substitution into the Schrödingergleichung, multiplying with \f$ P \f$ and \f$ Q \f$ and
 * seperating the parts, you get an effective eigenvalue equation only acting in the \f$ P \f$
 * eigen space, with the same eingen values as the initial Schrödingergleichung.
 *
 * \f[
 * \Theta_{i}|\psi_{i}\rangle_{L}=(E_{i}-E_{L}^{(0)})|\psi_{i}\rangle_{L}
 * \f]
 * with:
 * \f[
 * \Theta_{i}:=(PVP)+(PVQ)(E_{i}-H_{0}-QVQ)^{-1}(QVP)
 * \f]
 *
 * This equation can be solved perturbatively by expanding \f$ \Theta \f$ into powers of \f$ V \f$.
 *
 * State corrections can be achieved by resubstitution.
 *
 * The class Loewdin::Term is used to represent the coefficients of this effective Hamiltonian.
 *
 *
 *
 * The projector operator formalism was introduced by Loewdin in:
 *
 * Per-Olov Löwdin.
 * Studies in perturbation theory.
 * Chapter IV: Solution of eigenvalue problem by projection operator formalism.
 * Journal of Mathematical Physics, 3(5):969-982, 1962.
 *
 * A more recent review can be found in:
 *
 * Demin Yao and Jicong Shi.
 * Projection operator approach to time-independent perturbation theory in quantum mechanics.
 * American Journal of Physics, 68(3):278-281, 2000
 *
 * and an application with a brief review in:
 *
 * Tom Griffin and Stephen D. Bartlett.
 * Spin lattices with two-body hamiltonians for which the ground state encodes a cluster state.
 * Phys. Rev. A, 78:062306, May 2008.
 *
 * */
namespace Loewdin {

/**  Represents a coefficient of the Loewdin degenerated perturbation method.
 * For example a term like : \f$   -1 E_1 PVSSVP   \f$.*/
/*                                 /    \       \
                                  /      \       \
                               factor  E[0]=1   Operators as a string.
 */

class Term {

public:

	/** Create an empty term.*/
	Term();

	/** The exponents of the E factors. For example E[1] = 3 means, that \f$E_2\f$ appears with
	 the exponent three in the term.*/
	std::vector<unsigned int> E;

	/** The pre factor of the term. */
	amplitude factor;

	/** Operator part as a string.
	 - P :   Projector to the eigenspace of \f$ H_0 \f$ in question
	 - V :   Perturbation part of H.
	 - S :   The operator : \f$ \frac{1-P}{E_P - H_0} \f$  with \f$ E_P \f$ being the \f$ H_0 \f$ eigen energy of the eigen space in question.
	 */
	std::string operators;

	/** The order of the term. It is three in the example.*/
	unsigned int order;

	/** The order of the operator part of the term. It is two in the example.*/
	unsigned int operatorOrder;

};

/** Stream out as a string like 3 -1 1 PVSSVP*/
std::ostream& operator<<(std::ostream& o, const Term& d);

/** Stream in from a string like 3 -1 1 PVSSVP*/
std::istream& operator>>(std::istream& i, Term& d);

}

#endif /* LOEWDINTERM_H_ */
