/*
 * DirectSolver.cpp
 *
 *  Created on: 01.12.2010
 *      Author: klagges
 */

#include <Modules/DataProcessor/Loewdin/DirectSolver/DirectSolver.h>
#include <Utilities/Utilities.h>
#include <Interface/ExpandedAmplitudeIO.h>
#include <Utilities/TypeCast.h>

Loewdin::DirectSolver::DirectSolver() {

	// Read the configuration file.

	configuration.readFromConfigFile("LoewdinConfiguration.cfg");

	// And load the coefficients.

	coefficients.readFromCoefficientsFile(configuration.nameOfCoefficientsFile);

}

void Loewdin::DirectSolver::execute() {

	// Used to store the calculated energy corrections of every order.
	std::vector<stateAmplitude> E;
	E.resize(solver.globalConfiguration.order);

	for (std::map<std::string, ListedOperator>::reverse_iterator it =
			coefficients.operatorResults.rbegin();
			it != coefficients.operatorResults.rend(); it++) {
		// First we calculate the eigen value of all occurring operator parts.
		// We go through the map in reverse order, because so the "bad term"
		// (i.e. PVSVSVSVSVSVSVSVP) comes first.

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Processing operator :" << it->first << "\n";

		// We calculate the eigen value of the operator part, by applying it
		// directly to the input state...
		State operatorOutputState(solver.inputStates.right);

		// The initial P operator.

		solver.modules.pointerToModelProvider->applyP(operatorOutputState);

		// Apply all operators directly

		for (std::list<Utilities::OperatorDescription>::iterator operatorIt =
				it->second.listedOperators.begin();
				operatorIt != it->second.listedOperators.end(); operatorIt++) {

			operatorIt->apply(operatorOutputState);

		}

		// and compare the output state to the input state.

		if (operatorOutputState.begin() != operatorOutputState.end()) {
			// As only eigenstates are considered it is sufficient to compare the amplitudes
			// of an arbitrary base state before and after the transformation to determine
			// the eigenvalue of the terms. If no base states are present, the eigenvalue is zero.

			amplitude factor;

			Utilities::typeCast(
					solver.inputStates.right.find(
							operatorOutputState.begin()->first)->second,
					factor);

			it->second.eigenValue = operatorOutputState.begin()->second
					/ factor;

		} else
			it->second.eigenValue = stateAmplitude();

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	// We now calculate the energy corrections.
	for (unsigned int order = 0; order < coefficients.c.size(); order++) {

		for (std::list<Term>::iterator termIt = coefficients.c[order].begin();
				termIt != coefficients.c[order].end(); termIt++) {

			// We iterate over every term (of every order)

			// Calculate the contribution of the E-part of the term
			// By multiplying the (already calculated) energy corrections
			// of lower order.
			stateAmplitude ETermFactor(amplitude(1));

			for (unsigned int j = 0; j < termIt->E.size(); j++) {

				for (unsigned int k = 0; k < termIt->E[j]; k++) {
					ETermFactor *= E[j];

				}

			}

			// The term energy correction is the product of the term factor, the E-part factor and
			// the operator effect.
			E[order] +=
					(ETermFactor * termIt->factor)
							* coefficients.operatorResults[termIt->operators].eigenValue;

		}

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Calculated energy corrections of order "
					<< (order + 1) << ": " << E[order] << ".\n";

		if ((order + 1) >= solver.globalConfiguration.minOrder) {

			if (configuration.extraOutput) {

				if (Utilities::isPositive(E[order]))
					std::cout << "+";
				if (order == 0)
					std::cout << E[order] << "*" << "h";
				else
					std::cout << E[order] << "*" << "h^" << (order + 1);

			}

			solver.outputAmplitude += E[order];

		}

	}

	if (solver.globalConfiguration.verbosityLevel > 1)
		std::cerr << "\n";

}
