/*
 * TwoSideCommonSolver.cpp
 *
 *  Created on: 19.02.2011
 *      Author: klagges
 */

#include <Modules/DataProcessor/Loewdin/TwoSideCommonSolver/TwoSideCommonSolver.h>
#include <Interface/StateIO.h>

Loewdin::TwoSideCommonSolver::TwoSideCommonSolver() {

	// Read the configuration file.

	configuration.readFromConfigFile("LoewdinConfiguration.cfg");

	// And load the coefficients.

	coefficients.readFromCoefficientsFile(configuration.nameOfCoefficientsFile);

	// States with more than maximumNumberOfStates base states should
	// be split up before processing.

	maximumNumberOfStates.readFromConfigFile("NumberOfStates.cfg");

}

void Loewdin::TwoSideCommonSolver::execute() {

	// Used to store the calculated energy corrections of every order.
	std::vector<stateAmplitude> E;
	E.resize(solver.globalConfiguration.order);

	// Every input states is handled separately and equally.

	// Calculate <i|i>.
	amplitude normFactor;
	Utilities::typeCast(solver.inputStates.right.norm(), normFactor);

	for (std::map<std::string, ListedOperator>::reverse_iterator it =
			coefficients.operatorResults.rbegin();
			it != coefficients.operatorResults.rend(); it++) {
		// First we calculate the eigenvalue of all occurring operator parts.
		// We go through the map in reverse order, because so the "bad term"
		// (i.e. PVSVSVSVSVSVSVSVP) comes first.

		State outputState(solver.inputStates.right);

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Processing Operator :" << it->first << "\n";

		// The initial P operator.

		solver.modules.pointerToModelProvider->applyP(outputState);

		// Calculate <i|O|i> recursively.
		recursiveCommon(outputState, it->second.eigenValue,
				it->second.rightSideOperators.begin(),
				it->second.leftSideOperators.begin(), it->second.order);

		it->second.eigenValue /= normFactor;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	// We now calculate the energy corrections.
	for (unsigned int order = 0; order < coefficients.c.size(); order++) {

		for (std::list<Term>::iterator termIt = coefficients.c[order].begin();
				termIt != coefficients.c[order].end(); termIt++) {

			// We iterate over every term (of every order)

			// Calculate the contribution of the E-part of the term
			// By multiplying the (already calculated) energy corrections
			// of lower order.
			stateAmplitude ETermFactor(amplitude(1));

			for (unsigned int j = 0; j < termIt->E.size(); j++) {

				for (unsigned int k = 0; k < termIt->E[j]; k++) {
					ETermFactor *= E[j];

				}

			}

			// The term output state is the product of the term factor, the E-part factor and
			// the operator effect.

			E[order] +=
					ETermFactor * termIt->factor
							* coefficients.operatorResults[termIt->operators].eigenValue;

		}

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Calculated energy corrections of order : "
					<< (order + 1) << ": " << E[order] << ".\n";

		if ((order + 1) >= solver.globalConfiguration.minOrder) {

			if (configuration.extraOutput) {

				if (Utilities::isPositive(E[order]))
					std::cout << "+";
				if (order == 0)
					std::cout << E[order] << "*" << "h";
				else
					std::cout << E[order] << "*" << "h^" << (order + 1);

			}

			solver.outputAmplitude += E[order];

		}

	}

	if (solver.globalConfiguration.verbosityLevel > 1)
		std::cerr << "\n";

}

void Loewdin::TwoSideCommonSolver::recursiveCommon(State& source,
		stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
		std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// The operator is symmetric up to an application of a final V operator.
		// i.e:
		// PVSVSVP
		//
		// We apply the final V operator directly.

		solver.modules.pointerToModelProvider->applyV(source, source, result,
				rightOperator->remainingV);

		return;

	}

	if (remainingOrder == 2 && source.size() <= maximumNumberOfStates.value) {

		// The operator is symmetric up to the possible application of a S operator
		// i.e.:
		// PVSVP
		// or
		// PVSSVP
		//
		// the current state is also small enough to be processed directly.

		Utilities::OperatorDescription tempDescriptor(*rightOperator);
		tempDescriptor.numberOfS /= 2;

		// We apply the final V operator, but with halve of the S operators
		// (VS in the example PVSSVP)
		tempDescriptor.apply(source);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << source.size() << "\n";

		// And calculate the value <i|O|i> with or without a final S operator.
		if (rightOperator->numberOfS % 2 == 0) {

			result = source.norm();

		} else {

			result = source.normWithS();
		}

		return;

	}

	if (source.size() > maximumNumberOfStates.value) {

		// The current state is to big to be processed directly.
		// We need to split it up to continue the calculation.

		State leftState(source);

		recursiveSplit(source, leftState, result, rightOperator, leftOperator,
				remainingOrder);

		return;

	}

	if (!(*leftOperator == *rightOperator)) {

		// The next operator is no longer symmetric at both sides.

		if (rightOperator->numberOfS < leftOperator->numberOfS) {

			// We still can apply one V operator for both sides,
			rightOperator->apply(source);

			if (solver.globalConfiguration.verbosityLevel > 2)
				std::cerr << "Number of States: " << source.size() << "\n";

			State leftState(source);

			// but must apply the differencing S operators to the other side.

			solver.modules.pointerToModelProvider->applyS(leftState,
					leftOperator->remainingV,
					leftOperator->numberOfS - rightOperator->numberOfS);

			// Continue not symmetric.
			recursive(source, leftState, result, ++rightOperator,
					++leftOperator, remainingOrder - 2);

		} else {

			// We still can apply one V operator for both sides,
			leftOperator->apply(source);

			if (solver.globalConfiguration.verbosityLevel > 2)
				std::cerr << "Number of States: " << source.size() << "\n";

			State rightState(source);

			// but must apply the differencing S operators to the other side.

			solver.modules.pointerToModelProvider->applyS(rightState,
					rightOperator->remainingV,
					rightOperator->numberOfS - leftOperator->numberOfS);

			// Continue not symmetric.
			recursive(rightState, source, result, ++rightOperator,
					++leftOperator, remainingOrder - 2);

		}

		return;

	}

	// The general case, when the operators from both sides are still symmetric, and
	// The state is still small enough to be processed without splitting.

	// Apply the next operator..
	rightOperator->apply(source);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << source.size() << "\n";

	// and continue.
	recursiveCommon(source, result, ++rightOperator, ++leftOperator,
			remainingOrder - 2);

}

void Loewdin::TwoSideCommonSolver::recursive(State& rightSource,
		State& leftSource, stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
		std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyV(rightSource,
					leftSource, result, rightOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(leftSource,
					rightSource, result, leftOperator->remainingV);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value
			&& leftSource.size() > maximumNumberOfStates.value) {

		// Both states are to big to be processed further.
		// We start to split the right state.
		recursiveSplit(rightSource, leftSource, result, rightOperator,
				leftOperator, remainingOrder);

		return;

	}

	// There is still a state small enough to be processed without splitting.
	// Apply the next operator and continue recursively.
	if (rightSource.size() <= leftSource.size()) {

		rightOperator->apply(rightSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << rightSource.size() << "\n";

		recursive(rightSource, leftSource, result, ++rightOperator,
				leftOperator, remainingOrder - 1);

	} else {

		leftOperator->apply(leftSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << leftSource.size() << "\n";

		recursive(rightSource, leftSource, result, rightOperator,
				++leftOperator, remainingOrder - 1);

	}

}

void Loewdin::TwoSideCommonSolver::recursiveSplit(State& rightSource,
		const State& leftSource, stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
		std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyV(rightSource,
					leftSource, result, rightOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(leftSource,
					rightSource, result, leftOperator->remainingV);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value) {

		// The right state is to big to be processed directly.

		unsigned int numberOfSplits = rightSource.size()
				/ maximumNumberOfStates.value + 1;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Splitting right state into " << numberOfSplits
					<< " sub states.\n";

		// So we must split it up.
		std::vector<State> splitRightState = rightSource.splitState(
				numberOfSplits);

		for (std::vector<State>::iterator it = splitRightState.begin();
				it != splitRightState.end(); it++) {

			// And process every split state separately.
			stateAmplitude intermediateResult;

			recursiveSplit(*it, leftSource, intermediateResult, rightOperator,
					leftOperator, remainingOrder);

			it->clear();

			result += intermediateResult;

		}

		return;

	}

	// No split up necessary. Apply the operator and continue.
	rightOperator->apply(rightSource);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << rightSource.size() << "\n";

	recursiveSplit(rightSource, leftSource, result, ++rightOperator,
			leftOperator, remainingOrder - 1);

}

