/*
 * TwoSideFirstApplicationSolver.cpp
 *
 *  Created on: 19.02.2011
 *      Author: klagges
 */

#include <Modules/DataProcessor/Loewdin/TwoSideFirstApplicationSolver/TwoSideFirstApplicationSolver.h>
#include <Utilities/Utilities.h>
#include <Interface/ExpandedAmplitudeIO.h>
#include <Interface/StateIO.h>

Loewdin::TwoSideFirstApplicationSolver::TwoSideFirstApplicationSolver() {

	// Read the configuration file.

	configuration.readFromConfigFile("LoewdinConfiguration.cfg");

	// And load the coefficients.

	coefficients.readFromCoefficientsFile(configuration.nameOfCoefficientsFile);

	// Load the correction parameter.

	correction.readFromConfigFile(
			"LoewdinTwoSideFirstApplicationCorrection.cfg");

	// States with more than maximumNumberOfStates base states should
	// be split up before processing.

	maximumNumberOfStates.readFromConfigFile("NumberOfStates.cfg");

}

void Loewdin::TwoSideFirstApplicationSolver::execute() {

	// Used to store the calculated energy corrections of every order.
	std::vector<stateAmplitude> E;
	E.resize(solver.globalConfiguration.order);

	// Calculate <i|i>.
	amplitude normFactor;
	Utilities::typeCast(solver.inputStates.right.norm(), normFactor);

	for (std::map<std::string, ListedOperator>::reverse_iterator it =
			coefficients.operatorResults.rbegin();
			it != coefficients.operatorResults.rend(); it++) {
		// First we calculate the eigenvalue of all occurring operator parts.
		// We go through the map in reverse order, because so the "bad term"
		// (i.e. PVSVSVSVSVSVSVSVP) comes first.

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Processing Operator :" << it->first << "\n";

		State rightOutputState(solver.inputStates.right);
		State leftOutputState(solver.inputStates.left);

		// The initial P operator.
		solver.modules.pointerToModelProvider->applyP(rightOutputState);

		if (it->second.order == 1) {
			// This is the order 1 PVP Term. Just calculate the scalar product of the input states.
			it->second.eigenValue = (leftOutputState * rightOutputState)
					/ normFactor;

		} else {

			// And the initial S operators.

			std::list<Utilities::OperatorDescription>::const_iterator leftIt =
					it->second.leftSideOperators.begin();

			leftIt->apply(leftOutputState);
			leftIt++;

			recursive(rightOutputState, leftOutputState, it->second.eigenValue,
					it->second.rightSideOperators.begin(), leftIt,
					it->second.order - 1);

			it->second.eigenValue /= normFactor;

		}

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	std::cout.precision(16);

	// We now calculate the energy corrections.
	for (unsigned int order = 0; order < coefficients.c.size(); order++) {

		for (std::list<Term>::iterator termIt = coefficients.c[order].begin();
				termIt != coefficients.c[order].end(); termIt++) {

			// We iterate over every term (of every order)

			// Calculate the contribution of the E-part of the term
			// By multiplying the (already calculated) energy corrections
			// of lower order.
			stateAmplitude ETermFactor(amplitude(1));

			for (unsigned int j = 0; j < termIt->E.size(); j++) {

				for (unsigned int k = 0; k < termIt->E[j]; k++) {
					ETermFactor *= E[j] * correction.value;

				}

			}

			// The term output state is the product of the term factor, the E-part factor and
			// the operator effect.

			E[order] +=
					ETermFactor * termIt->factor
							* coefficients.operatorResults[termIt->operators].eigenValue;

		}

		std::cerr.precision(16);

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Calculated energy corrections of order : "
					<< (order + 1) << ": " << E[order] << ".\n";

		if ((order + 1) >= solver.globalConfiguration.minOrder) {

			if (configuration.extraOutput) {

				if (Utilities::isPositive(E[order]))
					std::cout << "+";
				if (order == 0)
					std::cout << E[order] << "*" << "h";
				else
					std::cout << E[order] << "*" << "h^" << (order + 1);

			}

			solver.outputAmplitude += E[order];
		}

	}

	if (solver.globalConfiguration.verbosityLevel > 1)
		std::cerr << "\n";

}

void Loewdin::TwoSideFirstApplicationSolver::recursive(State& rightSource,
		State& leftSource, stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
		std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyV(rightSource,
					leftSource, result, rightOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(leftSource,
					rightSource, result, leftOperator->remainingV);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value
			&& leftSource.size() > maximumNumberOfStates.value) {

		// Both states are to big to be processed further.
		// We start to split one of the states.

		if (rightSource.size() <= leftSource.size()) {

			recursiveSplit(rightSource, leftSource, result, rightOperator,
					leftOperator, remainingOrder);

		} else {

			recursiveSplit(leftSource, rightSource, result, leftOperator,
					rightOperator, remainingOrder);

		}

		return;

	}

	// There is still a state small enough to be processed without splitting.
	// Apply the next operator and continue recursively.
	if (rightSource.size() <= leftSource.size()) {

		rightOperator->apply(rightSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << rightSource.size() << "\n";

		recursive(rightSource, leftSource, result, ++rightOperator,
				leftOperator, remainingOrder - 1);

	} else {

		leftOperator->apply(leftSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << leftSource.size() << "\n";

		recursive(rightSource, leftSource, result, rightOperator,
				++leftOperator, remainingOrder - 1);

	}

}

void Loewdin::TwoSideFirstApplicationSolver::recursiveSplit(State& splitSource,
		const State& otherSource, stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator splitOperator,
		std::list<Utilities::OperatorDescription>::const_iterator otherOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.

		if (splitSource.size() < otherSource.size()) {

			solver.modules.pointerToModelProvider->applyV(splitSource,
					otherSource, result, splitOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(otherSource,
					splitSource, result, otherOperator->remainingV);

		}

		return;

	}

	if (splitSource.size() > maximumNumberOfStates.value) {

		// The split state is to big to be processed directly.

		unsigned int numberOfSplits = splitSource.size()
				/ maximumNumberOfStates.value + 1;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Splitting state into " << numberOfSplits
					<< " sub states.\n";

		// So we must split it up.
		std::vector<State> splits = splitSource.splitState(numberOfSplits);

		for (std::vector<State>::iterator it = splits.begin();
				it != splits.end(); it++) {

			// And process every split state separately.
			stateAmplitude intermediateResult;

			recursiveSplit(*it, otherSource, intermediateResult, splitOperator,
					otherOperator, remainingOrder);

			it->clear();

			result += intermediateResult;

		}

		return;

	}

	// No split up necessary. Apply the operator and continue.
	splitOperator->apply(splitSource);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << splitSource.size() << "\n";

	recursiveSplit(splitSource, otherSource, result, ++splitOperator,
			otherOperator, remainingOrder - 1);

}
