/*
 * TwoSideFirstApplicationSolver.h
 *
 *  Created on: 19.02.2011
 *      Author: klagges
 */

#ifndef LOEWDINTWOSIDEFIRSTAPPLICATIONSOLVER_H_
#define LOEWDINTWOSIDEFIRSTAPPLICATIONSOLVER_H_

#include <Modules/DataProcessor/DataProcessor.h>
#include <Interface/ExpandedAmplitude.h>
#include <Modules/DataProcessor/Loewdin/Common/Term.h>
#include <Modules/DataProcessor/Loewdin/Common/Configuration.h>
#include <Utilities/OperatorDescription.h>
#include <Utilities/StreamableBaseType.h>
#include <Utilities/StreamableAmplitude.h>

namespace Loewdin {

/** This DataProcessor does the same as the Loewdin::TwoSideCommonSolver, but skips the application
 * of the left most V operator.
 *
 * Instead, the Solver::inputStates.left state is used.
 *
 * It can be used for example by appropriately choosing the inputStates.left to simulate
 * a calculation where the first application of the V operator differs from the others.
 *
 */
class TwoSideFirstApplicationSolver: public DataProcessor {
public:

	/** Load and process the coefficients file.*/
	TwoSideFirstApplicationSolver();

	void execute();

private:

	/** Used to store a operator string (like PVSVSSVSSVSVSVSVP) in a listed form.
	 *
	 * Additionally its order and its calculated eigen value is stored.
	 * */
	class ListedOperator {
	public:

		/** The eigenvalue of the operator string represented by this object.*/
		stateAmplitude eigenValue;

		/** The order of the operator. */
		unsigned int order;

		/** Represents the operator string listed from the right.
		 */
		std::list<Utilities::OperatorDescription> rightSideOperators;

		/** Represents the operator string listed from the left.
		 */
		std::list<Utilities::OperatorDescription> leftSideOperators;

	};
	friend std::ostream& operator<<(std::ostream& o,
			TwoSideFirstApplicationSolver::ListedOperator& d);

	/**
	 * Class to store and stream in Loewdin coefficients.
	 * Stores every term separately. Additionally all occurring
	 * operators are stored in a map of ListedOperator objects.
	 */
	class Coefficients: public Utilities::ParameterSet {

	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The coefficients as c[order][term of order]*/
		std::vector<std::list<Term> > c;

		/** To store the operator parts in a listed form. */
		std::map<std::string, ListedOperator> operatorResults;

	};

	/** Configuration to chose from different coefficient files*/
	Configuration configuration;

	/** The coefficients.*/
	Coefficients coefficients;

	/** The parameter to correct the energy calculations.
	 * It is multiplied to all calculated energy corrections. This is for example necessary
	 * to calculate eigen energies of lattice states, where the first application of the V operator
	 * only acts on a single bond.*/
	Utilities::StreamableAmplitude<unsigned int> correction;

	/** The maximum number of base states that should be processed without splitting the state.*/
	Utilities::StreamableBaseType<unsigned int> maximumNumberOfStates;

	/** Recursive method to calculate the value \f$\langle i|H_{eff}|i\rangle \f$.
	 * This method is intended for the part of the calculation, where the left and right part
	 * of the effective Hamilton need to be calculated separately, but it is not necessary
	 * to split the states up for memory reasons. */
	void
			recursive(
					State& rightSource,
					State& leftSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
					std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
					unsigned int remainingOrder);

	/** Recursive method to calculate the value \f$\langle i|H_{eff}|i\rangle \f$.
	 * This method is intended for the part of the calculation, where it is necessary to split up
	 * one state to save memory. */
	void
			recursiveSplit(
					State& splitSource,
					const State& otherSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator splitOperator,
					std::list<Utilities::OperatorDescription>::const_iterator otherOperator,
					unsigned int remainingOrder);

};

/** Stream out as a string.*/
std::ostream& operator<<(std::ostream& o,
		TwoSideFirstApplicationSolver::ListedOperator& d);

}

#endif /* LOEWDINTWOSIDEFIRSTAPPLICATIONSOLVER_H_ */
