/*
 * TwoSideSolver.h
 *
 *  Created on: 19.02.2011
 *      Author: klagges
 */

#ifndef LOEWDINTWOSIDESOLVER_H_
#define LOEWDINTWOSIDESOLVER_H_

#include <Modules/DataProcessor/DataProcessor.h>
#include <Interface/ExpandedAmplitude.h>
#include <Modules/DataProcessor/Loewdin/Common/Term.h>
#include <Modules/DataProcessor/Loewdin/Common/Configuration.h>
#include <Utilities/OperatorDescription.h>
#include <Utilities/StreamableBaseType.h>

namespace Loewdin {

/** Calculates the energy corrections according to the degenerated perturbation
 method in the projector operator formalism by Loewdin.

 The input states must be eigen states of the terms of the effective Hamiltonian.

 This module reads the Solver::inputStates.left and Solver::inputStates.right members
 and writes the eigen energies into the Solver::outputAmplitude
 member of the solver object.

 In this implementation the eigen values of the operator parts of the coefficients
 are calculated by splitting each operator string into two parts of (roughly) equal
 length

 \f[ h = h_l h_r \f]

 and calculates the states

 \f[ h_r|i\rangle \f]

 and

 \f[ h_l^\dagger|i\rangle \f]

 separately. The eigen value is then calculated as

 \f[ \frac {\langle i|h_l\,h_r|i\rangle}{\langle i|i\rangle}. \f]

 Although the left the left and right input states must be identical, they can differ
 in extra information stored in the BaseState objects. For this reason in this module
 the left and right part are calculated separately even if the operators are the same.

 The energy corrections of every order is calculated successively
 from this eigen values and the energy corrections of lower orders.
 */
class TwoSideSolver: public DataProcessor {
public:

	/** Load and process the coefficients file.*/
	TwoSideSolver();

	void execute();

private:

	/** Used to store a operator string (like PVSVSSVSSVSVSVSVP) in a listed form.
	 *
	 * Additionally its calculated eigen value is stored.
	 * */
	class ListedOperator {
	public:

		/** The eigenvalue of the operator string represented by this object.*/
		stateAmplitude eigenValue;

		/** The order of the operator. */
		unsigned int order;

		/** Represents the operator string listed from the right.
		 */
		std::list<Utilities::OperatorDescription> rightSideOperators;

		/** Represents the operator string listed from the left.
		 */
		std::list<Utilities::OperatorDescription> leftSideOperators;

	};
	friend std::ostream& operator<<(std::ostream& o,
			TwoSideSolver::ListedOperator& d);

	/**
	 * Class to store and stream in Loewdin coefficients.
	 * Stores every term separately. Additionally all occurring
	 * operators are stored in a map of ListedOperator objects.
	 */
	class Coefficients: public Utilities::ParameterSet {

	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The coefficients as c[order][term of order]*/
		std::vector<std::list<Term> > c;

		/** To store the operator parts in a listed form. */
		std::map<std::string, ListedOperator> operatorResults;

	};

	/** Configuration to chose from different coefficient files*/
	Configuration configuration;

	/** The coefficients.*/
	Coefficients coefficients;

	/** The maximum number of base states that should be processed without splitting the state.*/
	Utilities::StreamableBaseType<unsigned int> maximumNumberOfStates;

	/** Recursive method to calculate the value \f$\langle i|H_{eff}|i\rangle \f$.
	 * This method is intended for the part of the calculation, where the left and right part
	 * of the effective Hamilton need to be calculated separately, but it is not necessary
	 * to split the states up for memory reasons. */
	void
			recursive(
					State& rightSource,
					State& leftSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
					std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
					unsigned int remainingOrder);

	/** Recursive method to calculate the value \f$\langle i|H_{eff}|i\rangle \f$.
	 * This method is intended for the part of the calculation, where it is necessary to split up
	 * one state to save memory. */
	void
			recursiveSplit(
					State& splitSource,
					const State& otherSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator splitOperator,
					std::list<Utilities::OperatorDescription>::const_iterator otherOperator,
					unsigned int remainingOrder);

};

/** Stream out as a string.*/
std::ostream& operator<<(std::ostream& o, TwoSideSolver::ListedOperator& d);

}

#endif /* LOEWDINTWOSIDESOLVER_H_ */
