/*
 * TreeSolver.h
 *
 *  Created on: 15.04.2010
 *      Author: kamfor, klagges
 *
 */

#ifndef PCUTTREESOLVER_H_
#define PCUTTREESOLVER_H_

#include <Modules/DataProcessor/DataProcessor.h>
#include <Modules/DataProcessor/PCUT/Common/Term.h>
#include <Utilities/OperatorDescription.h>
#include <Utilities/StreamableBaseType.h>
#include <set>

namespace PCUT {

/** Applies an effective Hamiltonian according to the PCUT method.

 This module reads the Solver::inputStates.right member and writes the states after
 the application of the effective Hamiltonian to the Solver::outputState member
 of the Solver object.

 In this implementation the operator parts of the terms are brought into a tree
 structure. For example the tree for some terms up to order three:

 - T0
 - T1 T-1
 - T1 T0 T-1
 - T0 T1 T-1
 - T1 T-1 T0

 looks like:
 \code
            I
          /   \
        T0     T1
       /      /  \
	  T1    T-1   T0
     /       |     \
	T-1		T_0    T_-1
 \endcode

 Each term is associated to the corresponding node of the tree.

 The output states of each term is calculated by a depth first search (DFS)
 through the tree.

 This method is (linearly in the order) more memory intensive then the PCUT::DirectSolver,
 but requires (linearly in the order) less time.
 */
class TreeSolver: public DataProcessor {

public:

	TreeSolver();

	void execute();

private:

	class TreeNode {
	public:

		/** Default constructor.*/
		TreeNode();

		/** Construct a TreeNode associated with the operator T_c.*/
		TreeNode(const char c, const unsigned int currentOrder);

		/** Recursively update the tree with this node as the root so
		 * that it contains term.*/
		void insertTerm(Term& term);

		/** Recursively stream out all terms contained in the tree with
		 * this node as the root.*/
		void streamOutRecursively(Term t, std::ostream& o);

		/** Stream out this node and all sub nodes.*/
		void streamOutRecursivelyAsTree(std::list<char> l, std::ostream& o);

		/** Sort the lists of children of each node.*/
		void sortRecursively();

		/** Do a depth first search (DFS) through the tree to calculate
		 * the output states of all terms.*/
		void calculateOutputStateRecursively(State & inputState,
				State & outputState, Term t);

		/** The T operator associated to this node.
		 * The value -128 means, that no operator is associated with this node.*/
		char op;

		/** The children of this node in the tree.*/
		std::list<TreeNode> children;

		/** The intermediate state calculated during the processing of the tree.*/
		State intermediateState;

		/** If there is a term associated to this node, factor is the pre factor of this term.
		 * Otherwise it is zero.*/
		amplitude factor;

		/** The order of the current calculation.
		 * Equals to the number of T operators in the parent nodes of this node plus one.*/
		unsigned int currentOrder;

	};
	friend bool operator<(const TreeSolver::TreeNode& l,
			const TreeSolver::TreeNode& r);

	/** Class to store and stream in PCUT coefficients.
	 * Every term is inserted into a tree.*/
	class Coefficients: public Utilities::ParameterSet {
	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The coefficients as a tree.*/
		TreeNode root;

	};
	friend std::ostream& operator<<(std::ostream& o,
			TreeSolver::Coefficients& c);

	/** Configuration to choose from different configuration files.*/
	Utilities::StreamableBaseType<std::string> nameOfCoefficientsFile;

	/** The coefficients.*/
	Coefficients coefficients;

};
/** Stream out the coefficients directly and in tree form.*/
std::ostream& operator<<(std::ostream& o, TreeSolver::Coefficients& c);

/** Operator to compare TreeNode objects. Used to sort the children lists of TreeNodes.*/
bool operator<(const TreeSolver::TreeNode& l, const TreeSolver::TreeNode& r);

}
#endif /* PCUTTREESOLVER_H_ */
