/*
 * TwoSideSolver.cpp
 *
 *  Created on: 04.05.2011
 *      Author: klagges
 */

#include <Modules/DataProcessor/PCUT/TwoSideSolver/TwoSideSolver.h>
#include <Framework/Solver.h>

PCUT::TwoSideSolver::TwoSideSolver() {


	pcutConfiguration.readFromConfigFile("PCUTConfiguration.cfg");

	coefficients.numberOfQuasiParticles
			= pcutConfiguration.numberOfQuasiParticles;

	coefficients.readFromCoefficientsFile(
			pcutConfiguration.nameOfCoefficientsFile);

//	numberOfBitsToDrop.readFromConfigFile("BitsToDrop.cfg");
//
//
//
//
//	nameOfCoefficientsFile.readFromConfigFile("PCUTConfiguration.cfg");
//
//	coefficients.readFromCoefficientsFile(nameOfCoefficientsFile.value);

	maximumNumberOfStates.readFromConfigFile("NumberOfStates.cfg");

}

void PCUT::TwoSideSolver::execute() {

	for (std::list<Term>::const_iterator termIt = coefficients.c.begin(); termIt
			!= coefficients.c.end(); termIt++) {

		// We iterate over every term

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Processing the Term: " << *termIt << "\n";

		State rightState = solver.inputStates.right;

		State leftState = solver.inputStates.left;

		stateAmplitude result;

		recursive(rightState, leftState, result, termIt->operators.begin(),
				termIt->operators.rbegin(), termIt->operators.size());

		// And sum up all results.
		solver.outputAmplitude += result * termIt->factor;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	if (solver.globalConfiguration.verbosityLevel > 1)
		std::cerr << "\n";
}

void PCUT::TwoSideSolver::recursive(State & rightSource, State & leftSource,
		stateAmplitude & result,
		std::list<char>::const_iterator rightOperator,
		std::list<char>::const_reverse_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one T operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyT(rightSource,
					leftSource, result, *rightOperator);

		} else {

			solver.modules.pointerToModelProvider->applyT(leftSource,
					rightSource, result, -*leftOperator);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value
			&& leftSource.size() > maximumNumberOfStates.value) {

		// Both states are to big to be processed further.
		// We start to split one of the states.

		if (rightSource.size() <= leftSource.size()) {

			recursiveSplitRight(rightSource, leftSource, result, rightOperator,
					leftOperator, remainingOrder);

		} else {

			recursiveSplitLeft(rightSource, leftSource, result, rightOperator,
					leftOperator, remainingOrder);

		}

		return;

	}

	// There is still a state small enough to be processed without splitting.
	// Take the smaller one, apply the next operator and continue recursively.
	if (rightSource.size() <= leftSource.size()) {

		{
			State intermediateState;
			solver.modules.pointerToModelProvider->applyT(rightSource,
					intermediateState, *rightOperator);
			rightSource.swap(intermediateState);
		}

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << rightSource.size() << "\n";

		recursive(rightSource, leftSource, result, ++rightOperator,
				leftOperator, remainingOrder - 1);

	} else {

		{
			State intermediateState;
			solver.modules.pointerToModelProvider->applyT(leftSource,
					intermediateState, -*leftOperator);
			leftSource.swap(intermediateState);
		}

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << leftSource.size() << "\n";

		recursive(rightSource, leftSource, result, rightOperator,
				++leftOperator, remainingOrder - 1);

	}

}

void PCUT::TwoSideSolver::recursiveSplitRight(State & rightSource,
		const State & leftSource, stateAmplitude & result,
		std::list<char>::const_iterator rightOperator,
		std::list<char>::const_reverse_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one T operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyT(rightSource,
					leftSource, result, *rightOperator);

		} else {

			solver.modules.pointerToModelProvider->applyT(leftSource,
					rightSource, result, -*leftOperator);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value) {

		// The split state is to big to be processed directly.

		unsigned int numberOfSplits = rightSource.size()
				/ maximumNumberOfStates.value + 1;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Splitting right state into " << numberOfSplits
					<< " sub states.\n";

		// So we must split it up.
		std::vector<State> splits = rightSource.splitState(numberOfSplits);

		for (std::vector<State>::iterator it = splits.begin(); it
				!= splits.end(); it++) {

			// And process every split state separately.
			stateAmplitude intermediateResult;

			recursiveSplitRight(*it, leftSource, intermediateResult,
					rightOperator, leftOperator, remainingOrder);

			it->clear();

			result += intermediateResult;

		}

		return;

	}

	// No split up necessary. Apply the operator and continue.
	{
		State intermediateState;
		solver.modules.pointerToModelProvider->applyT(rightSource,
				intermediateState, *rightOperator);
		rightSource.swap(intermediateState);
	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << rightSource.size() << "\n";

	recursiveSplitRight(rightSource, leftSource, result, ++rightOperator,
			leftOperator, remainingOrder - 1);

}

void PCUT::TwoSideSolver::recursiveSplitLeft(const State & rightSource,
		State & leftSource, stateAmplitude & result,
		std::list<char>::const_iterator rightOperator,
		std::list<char>::const_reverse_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one T operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyT(rightSource,
					leftSource, result, *rightOperator);

		} else {

			solver.modules.pointerToModelProvider->applyT(leftSource,
					rightSource, result, -*leftOperator);

		}

		return;

	}

	if (leftSource.size() > maximumNumberOfStates.value) {

		// The split state is to big to be processed directly.

		unsigned int numberOfSplits = leftSource.size()
				/ maximumNumberOfStates.value + 1;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Splitting left state into " << numberOfSplits
					<< " sub states.\n";

		// So we must split it up.
		std::vector<State> splits = leftSource.splitState(numberOfSplits);

		for (std::vector<State>::iterator it = splits.begin(); it
				!= splits.end(); it++) {

			// And process every split state separately.
			stateAmplitude intermediateResult;

			recursiveSplitLeft(rightSource, *it, intermediateResult,
					rightOperator, leftOperator, remainingOrder);

			it->clear();

			result += intermediateResult;

		}

		return;

	}

	// No split up necessary. Apply the operator and continue.
	{
		State intermediateState;
		solver.modules.pointerToModelProvider->applyT(leftSource,
				intermediateState, -*leftOperator);
		leftSource.swap(intermediateState);
	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << rightSource.size() << "\n";

	recursiveSplitLeft(rightSource, leftSource, result, rightOperator,
			++leftOperator, remainingOrder - 1);
}

