/*
 * TwoSideSolver.h
 *
 *  Created on: 04.05.2011
 *      Author: klagges
 */

#ifndef PCUTTWOSIDESOLVER_H_
#define PCUTTWOSIDESOLVER_H_

#include <Modules/DataProcessor/PCUT/Common/Configuration.h>
#include <Modules/DataProcessor/DataProcessor.h>
#include <Utilities/StreamableBaseType.h>
#include <Modules/DataProcessor/PCUT/Common/Term.h>

namespace PCUT {

/** Calculates matrix elements of the effective Hamiltonian according to the
 PCUT method.

 This module reads the Solver::inputStates.right and the Solver::inputStates.left members
 and writes the matrix element \f$<l|H_{eff}|r>\f$ into the Solver::outputAmplitude
 member of the solver object.

 In this implementation the matrix elements of the operator parts of the coefficients
 are calculated by splitting each operator string into two parts of (roughly) equal
 length

 \f[ h = h_l h_r \f]

 and calculate the states

 \f[ h_r|i\rangle \f]

 and

 \f[ h_l^\dagger|li\rangle \f]

 separately. The matrix element is then calculated as

 \f[ \langle li|h_l\,h_r|i\rangle. \f]

 The matrix element of the whole effective Hamiltonian is calculated by summing over all terms.
 */
class TwoSideSolver: public DataProcessor {
public:

	TwoSideSolver();

	void execute();

private:

	/** Class to store and stream in PCUT coefficients.
	 * Stores every term in a set.*/
	class Coefficients: public Utilities::ParameterSet {
	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The coefficients.*/
		std::list<Term> c;

		short numberOfQuasiParticles;
	};
	friend std::ostream& operator<<(std::ostream& o,
			TwoSideSolver::Coefficients& c);

	/** Configuration to choose from different configuration files.*/
	Utilities::StreamableBaseType<std::string> nameOfCoefficientsFile;

	/** PCUT initial configuration: coefficients-file, qp-number. */
	Configuration pcutConfiguration;

	/** The coefficients.*/
	Coefficients coefficients;

	/** The maximum number of base states that should be processed without splitting the state.*/
	Utilities::StreamableBaseType<unsigned int> maximumNumberOfStates;

	/** Recursive method to calculate the value \f$\langle l|H_{eff}|l\rangle \f$.
	 * This method is intended for the part of the calculation, where the left and right part
	 * of the effective Hamilton need to be calculated separately, but it is not necessary
	 * to split the states up for memory reasons. */
	void
	recursive(State& rightSource, State& leftSource, stateAmplitude& result,
			std::list<char>::const_iterator rightOperator,
			std::list<char>::const_reverse_iterator leftOperator,
			unsigned int remainingOrder);

	/** Recursive method to calculate the value \f$\langle l|H_{eff}|l\rangle \f$.
	 * This method is intended for the part of the calculation, where it is necessary to split up
	 * the right state to save memory. */
	void
	recursiveSplitRight(State& rightSource, const State& leftSource,
			stateAmplitude& result,
			std::list<char>::const_iterator rightOperator,
			std::list<char>::const_reverse_iterator leftOperator,
			unsigned int remainingOrder);

	/** Recursive method to calculate the value \f$\langle l|H_{eff}|l\rangle \f$.
	 * This method is intended for the part of the calculation, where it is necessary to split up
	 * the left state to save memory. */
	void
	recursiveSplitLeft(const State& rightSource, State& leftSource,
			stateAmplitude& result,
			std::list<char>::const_iterator rightOperator,
			std::list<char>::const_reverse_iterator leftOperator,
			unsigned int remainingOrder);
};

/** Stream out all coefficients. */
std::ostream& operator<<(std::ostream& o, TwoSideSolver::Coefficients& c);

}

#endif /* PCUTTWOSIDESOLVER_H_ */
