/*
 * Term.h
 *
 *  Created on: 06.02.2011
 *      Author: klagges
 */

#ifndef TAKAHASHITERM_H_
#define TAKAHASHITERM_H_

#include <Interface/Interface.h>

/** Name space to capsule all classes related to the Takahashi degenerated perturbation solvers.
 *
 * In the degenerated perturbation approach by Takahashi, the Hamiltonian is transfered by the
 * base transformation:
 *
 * \f[
 * \Gamma = \bar{P}P(P\bar{P}P)^{-\frac{1}{2}}
 * \f]
 *
 * \f[
 * H_{eff} = \Gamma^\dagger H \Gamma
 * \f]
 *
 * where \f$ \bar{P} \f$ is the projector into the space of the perturbed eigen vectors of \f$ H \f$
 * corresponding to the unperturbed \f$ P \f$ space.
 *
 * It can be shown, that
 *
 * \f[
 * \Gamma^\dagger \Gamma = P.
 * \f]
 *
 * So the effective Hamiltonian only acts in the \f$ P \f$ eigen space, while preserving the spectral
 * properties of the initial Hamiltonian.
 *
 * By expanding the effective Hamiltonian into powers of \f$ V \f$ you get perturbation coefficients
 * as they are represented by the class Takahashi::Term.
 *
 *
 * Based on the works of Kato:
 *
 * Tosio Kato,
 * On the Convergence of the Perturbation Method.
 * Prog. Theor. Phys. Vol. 4 No. 4 (1949) pp. 514-523
 *
 * this degenerated perturbation method was introduced by Takahashi:
 *
 * Minoru Takahashi,
 * Half-filed Hubbard model at low temperature.
 * J. Phys. C: Solid State Phys., Vol. 10. 1977.
 *
 * An example for an application can be found here:
 *
 * E. Müller-Hartmann and A. Reischl,
 * Derivation of effective spin models from a three band model for CuO2-planes.
 * Eur. Phys. J. B 28, 173-183 (2002)
 *
 *
 * */
namespace Takahashi {

/** Represents a coefficient of the Takahashi degenerated perturbation method.
 * For example:
 * \f[ -\frac{1}{2}PVSSVPVP \f]
 * which is one of the 3-rd order coefficients.*/
class Term {
public:

	/** The order of the term. */
	unsigned int order;

	/** The prefactor of the term. */
	amplitude factor;

	/** Operator part as a string.
	 - P :   Projector to the eigenspace of \f$ H_0 \f$ in question
	 - V :   Perturbation part of H.
	 - S :   The operator : \f$ \frac{1-P}{E_P - H_0} \f$  with \f$ E_P \f$ being the \f$ H_0 \f$ eigen energy of the eigen space in question. */
	std::string operators;

};

/** Stream out as a string like 3 -1/2 PVPVSSVP*/
std::ostream& operator<<(std::ostream& o, Term& d);

/** Stream in from a string like 3 -1/2 PVPVSSVP*/
std::istream& operator>>(std::istream& i, Term& d);

}

#endif /* TAKAHASHITERM_H_ */
