/*
 * TreeNode.cpp
 *
 *  Created on: 23.02.2011
 *      Author: klagges
 */

#include <Modules/DataProcessor/Takahashi/TreeSolver/TreeSolver.h>
#include <Framework/Solver.h>
#include <sstream>

Takahashi::TreeSolver::TreeNode::TreeNode() :
	factor(0), currentOrder(1) {

}

Takahashi::TreeSolver::TreeNode::TreeNode(
		const Utilities::OperatorDescription& o, const unsigned int currentOr) :
	op(o), factor(0), currentOrder(currentOr) {

}

void Takahashi::TreeSolver::TreeNode::insertTerm(Takahashi::Term & term) {

	if (term.operators.size() == 0) {

		// No operators left, so associate this term with this node.
		factor = term.factor;
		return;
	}

	Utilities::OperatorDescription descriptor(*term.operators.rbegin());
	term.operators.resize(term.operators.size() - 1);

	std::list<Takahashi::TreeSolver::TreeNode>::iterator it = children.begin();

	// Search through the list of children to see if we already have a children
	// node with this operator.
	while (it != children.end() && !(it->op == descriptor))
		it++;

	if (it != children.end())
		// If so, insert into the existing child.
		it->insertTerm(term);
	else {
		// If not, create a new one.
		if (op.applyV)
			children.push_back(TreeNode(descriptor, currentOrder + 1));
		else
			children.push_back(TreeNode(descriptor, currentOrder));

		children.rbegin()->insertTerm(term);
	}

}

unsigned int Takahashi::TreeSolver::TreeNode::calculateRemainingVRecursively() {

	if (children.size() == 0)
		return op.remainingV;

	unsigned int tempV = 0;

	// Calculate the remainingV values of all children and store the maximum.
	for (std::list<TreeNode>::iterator it = children.begin(); it
			!= children.end(); it++) {

		unsigned int childrensRemainingV = it->calculateRemainingVRecursively();

		if (childrensRemainingV > tempV)
			tempV = childrensRemainingV;

	}

	// If this node is a P operator its remainingV value is zero.
	if (op.applyV == false && op.numberOfS == 0 && op.remainingV == 0)
		return 0;

	// Otherwise it is the maximum of all children.
	op.remainingV = tempV;

	// If the operator of this node is a V operator we have to increment
	// the remainingV value for the parent.
	if (op.applyV == true)
		return tempV + 1;
	else
		return tempV;

}

void Takahashi::TreeSolver::TreeNode::shrinkNodes() {

	// First shrink the node of all child trees.
	for (std::list<TreeNode>::iterator it = children.begin(); it
			!= children.end(); it++) {

		it->shrinkNodes();

	}

	if (children.size() == 1 && Utilities::equalsZero(factor)) {

		// We only can shrink nodes, that are not associated to a term
		// and that have exactly one successor.
		if (op.applyV && op.remainingV == 0) {

			// This is the case of a V operator, that is followed
			// only by a P operator. We can skip the P, as it is done
			// implicitly by the applySSV method.
			factor = children.begin()->factor;

			std::list<TreeNode> temp;

			temp.swap(children);

			children.swap(temp.begin()->children);

		}

		else {
			if (children.begin()->op.applyV == false) {

				// The child node is just a number of S operators.
				// We associate them to this node.

				op.numberOfS += children.begin()->op.numberOfS;

				factor = children.begin()->factor;

				std::list<TreeNode> temp;

				temp.swap(children);

				children.swap(temp.begin()->children);

			}

		}

	}

}

void Takahashi::TreeSolver::TreeNode::sortRecursively() {

	children.sort();

	for (std::list<TreeNode>::iterator it = children.begin(); it
			!= children.end(); it++) {

		it->sortRecursively();

	}

}

void Takahashi::TreeSolver::TreeNode::calculateOutputStateRecursively(
		State & inputState, State & outputState, std::string s) {

	solver.globalConfiguration.currentOrder = currentOrder;

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << s << " " << op << "\n";

	if (solver.globalConfiguration.verbosityLevel > 1) {
		std::stringstream ss;
		ss << op;
		s += ss.str();
	}

	if (op.applyV == true && op.numberOfS == 0 && op.remainingV > 0) {

		// We need to apply the single V operator by hand, because the
		// the implementation of the OperatorDescription would apply
		// the (1-P)V operator instead.

		solver.modules.pointerToModelProvider->applyV(inputState,
				intermediateState, op.remainingV);
	} else {

		// Apply the operator of this node.
		op.apply(inputState, intermediateState);
	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

	if (!Utilities::equalsZero(factor)) {
		// This node is associated with a Takahashi term. So add it to the result.

		outputState.addMultiple(intermediateState, factor);
		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Applying term: " << factor << " " << s << "\n";
		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	// Repeat the method with all children (DFS).
	for (std::list<TreeNode>::iterator it = children.begin(); it
			!= children.end(); it++) {

		it->calculateOutputStateRecursively(intermediateState, outputState, s);

	}

	// This node and all children are processed, so we can delete
	// the intermediate state to save memory.
	intermediateState.clear();

}

std::ostream& Takahashi::operator<<(std::ostream& o, TreeSolver::TreeNode& lt) {

	lt.streamOutRecursively(std::string(), o);

	o << "\n";

	lt.streamOutRecursivelyAsTree(std::string(), o);

	return o;
}

bool Takahashi::operator<(const TreeSolver::TreeNode& l,
		const TreeSolver::TreeNode& r) {

	// Take the "most dificult nodes" first.
	if (l.op.remainingV > r.op.remainingV)
		return true;

	// If two nodes have the same dificulty, prefere 'V' over 'S' and 'S' over 'P'.
	else if (l.op.remainingV == r.op.remainingV) {
		if (l.op.applyV == true && l.op.applyV == false)
			return true;
		else if (l.op.applyV == r.op.applyV) {
			if (l.op.numberOfS < r.op.numberOfS)
				return true;

		}

	}

	return false;

}

void Takahashi::TreeSolver::TreeNode::streamOutRecursively(std::string s,
		std::ostream& o) {

	std::stringstream ss;
	ss << op;
	s += ss.str();

	if (!Utilities::equalsZero(factor))
		o << factor << " " << s << "\n";

	for (std::list<TreeNode>::iterator termIt = children.begin(); termIt
			!= children.end(); termIt++) {

		termIt->streamOutRecursively(s, o);

	}

}

void Takahashi::TreeSolver::TreeNode::streamOutRecursivelyAsTree(std::string s,
		std::ostream& o) {

	o << s << " Node: " << op << " Remaining V: " << op.remainingV
			<< " currentOrder:" << currentOrder;
	if (!Utilities::equalsZero(factor))
		o << " Factor: " << factor;
	o << "\n";

	std::stringstream ss;
	ss << op;
	s += ss.str();

	for (std::list<TreeNode>::iterator termIt = children.begin(); termIt
			!= children.end(); termIt++) {

		termIt->streamOutRecursivelyAsTree(s, o);

	}

}

