/*
 * TreeSolver.h
 *
 *  Created on: 17.11.2010
 *      Author: klagges
 *
 */

#ifndef TAKAHASHITREESOLVER_H_
#define TAKAHASHITREESOLVER_H_

#include <Modules/DataProcessor/DataProcessor.h>
#include <Modules/DataProcessor/Takahashi/Common/Term.h>
#include <Utilities/OperatorDescription.h>
#include <Utilities/StreamableBaseType.h>
#include <set>

namespace Takahashi {

/** Applies an effective Hamiltonian according to the degenerated perturbation
 method by Takahashi.

 This module reads the Solver::inputStates.right member and writes the states after
 the application of the effective Hamiltonian to the Solver::outputState member
 of the solver object.

 In this implementation the operator parts of the terms are brought into a tree
 structure. For example the tree for the terms up to order three:

 - PVP
 - PVSVP
 - PVPVSSVP
 - PVSVSVP
 - PVSSVPVP

 looks like:
 \code
            P
            |
            V
         /     \
        S       P
       / \       \
      V   S      SSV
     / \   \       \
    S   P  PV      PV
   /       /
  PV      PV
 \endcode

 Each term is associated to the corresponding node of the tree.

 The output states of each term is calculated by a depth first search (DFS)
 through the tree.

 This method is (linearly in the order) more memory intensive then the Takahashi::DirectSolver,
 but requires (linearly in the order) less time.
 */
class TreeSolver: public DataProcessor {

public:

	TreeSolver();

	void execute();

private:

	/**
	 Represents a node in the tree of operators.
	 */
	class TreeNode {
	public:

		/** Default constructor.*/
		TreeNode();

		/** Create node with the operator o.*/
		TreeNode(const Utilities::OperatorDescription& o,
				const unsigned int currentOrder);

		/** Recursively update the tree with this node as the root so
		 * that it contains term.*/
		void insertTerm(Term& term);

		/** Recursively stream out all terms contained in the tree with
		 * this node as the root.*/
		void streamOutRecursively(std::string s, std::ostream& o);

		/** Stream out this node and all sub nodes.*/
		void streamOutRecursivelyAsTree(std::string s, std::ostream& o);

		/** Go through the tree recursively and calculate the
		 according OperatorDescription::remainingV values.*/
		unsigned int calculateRemainingVRecursively();

		/** Convert nodes with only one successor and there successors
		 to single notes if possible.*/
		void shrinkNodes();

		/** Sort the lists of children of each node.*/
		void sortRecursively();

		/** Do a depth first search (DFS) through the tree to calculate
		 * the output states of all terms.*/
		void calculateOutputStateRecursively(State & inputState,
				State & outputState, std::string s);

		/** The operator associated to this node.*/
		Utilities::OperatorDescription op;

		/** The children of this node in the tree.*/
		std::list<TreeNode> children;

		/** The intermediate state calculated during the processing of the tree.*/
		State intermediateState;

		/** If there is a term associated to this node, factor is the pre factor of this term.
		 * Otherwise it is zero.*/
		amplitude factor;

		/** The order of the current calculation. Equals the number of V operators in the parent nodes plus one.*/
		unsigned int currentOrder;

	};
	friend bool operator<(const TreeSolver::TreeNode& l,
			const TreeSolver::TreeNode& r);
	friend std::ostream& operator<<(std::ostream& o, TreeSolver::TreeNode& d);

	/** Class to store and stream in Takahashi coefficients.
	 The terms are stored in a tree structure.*/
	class Coefficients: public Utilities::ParameterSet {
	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The root of the tree storing the coefficients.*/
		TreeNode root;

	};

	/** Configuration to choose from different configuration files.*/
	Utilities::StreamableBaseType<std::string> nameOfCoefficientsFile;

	/** The coefficients.*/
	Coefficients coefficients;

};

/** Used by TreeNode::sortRecursively() to sort the children lists of each node.
 * They are sorted by there OperatorDescription::remainingV property (most difficult first)
 * and nodes with equal remainingV property are sorted by:
 * Prefer 'V' over 'S' and 'S' over 'P'.*/
bool operator<(const TreeSolver::TreeNode& l, const TreeSolver::TreeNode& r);

/** Stream out the tree with the TreeNode as the root.*/
std::ostream& operator<<(std::ostream& o, TreeSolver::TreeNode& tn);

}
#endif /* TAKAHASHITREESOLVER_H_ */
