/*
 * TwoSideSolver.cpp
 *
 *  Created on: 22.02.2011
 *      Author: klagges
 */

#include <Modules/DataProcessor/Takahashi/TwoSideSolver/TwoSideSolver.h>
#include <Framework/Solver.h>

Takahashi::TwoSideSolver::TwoSideSolver() {
	// Read the configuration file.

	nameOfCoefficientsFile.readFromConfigFile("TakahashiConfiguration.cfg");

	// And load the coefficients.

	coefficients.readFromCoefficientsFile(nameOfCoefficientsFile.value);

	// States with more than maximumNumberOfStates base states should
	// be split up before processing.

	maximumNumberOfStates.readFromConfigFile("NumberOfStates.cfg");

}

void Takahashi::TwoSideSolver::execute() {

	for (std::set<ListedTerm>::reverse_iterator termIt =
			coefficients.c.rbegin(); termIt != coefficients.c.rend(); termIt++) {

		// We iterate over every term

		if (solver.globalConfiguration.verbosityLevel > 1)
			std::cerr << "Processing the Term: " << (Term&) termIt->term
					<< "\n";

		State rightState = solver.inputStates.right;

		solver.modules.pointerToModelProvider->applyP(rightState);

		State leftState = solver.inputStates.left;

		solver.modules.pointerToModelProvider->applyP(leftState);

		stateAmplitude result;

		recursive(rightState, leftState, result,
				termIt->rightSideOperators.begin(),
				termIt->leftSideOperators.begin(), termIt->term.order);

		solver.outputAmplitude += result * termIt->term.factor;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "\n";

	}

	if (solver.globalConfiguration.verbosityLevel > 1)
		std::cerr << "\n";

}

void Takahashi::TwoSideSolver::recursive(
		State& rightSource,
		State& leftSource,
		stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
		std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.

		if (rightSource.size() < leftSource.size()) {

			solver.modules.pointerToModelProvider->applyV(rightSource,
					leftSource, result, rightOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(leftSource,
					rightSource, result, leftOperator->remainingV);

		}

		return;

	}

	if (rightSource.size() > maximumNumberOfStates.value
			&& leftSource.size() > maximumNumberOfStates.value) {

		// Both states are to big to be processed further.
		// We start to split one of the states.

		if (rightSource.size() <= leftSource.size()) {

			recursiveSplit(rightSource, leftSource, result, rightOperator,
					leftOperator, remainingOrder);

		} else {

			recursiveSplit(leftSource, rightSource, result, leftOperator,
					rightOperator, remainingOrder);

		}

		return;

	}

	// There is still a state small enough to be processed without splitting.
	// Apply the next operator and continue recursively.
	if (rightSource.size() <= leftSource.size()) {

		rightOperator->apply(rightSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << rightSource.size() << "\n";

		recursive(rightSource, leftSource, result, ++rightOperator,
				leftOperator, remainingOrder - 1);

	} else {

		leftOperator->apply(leftSource);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Number of States: " << leftSource.size() << "\n";

		recursive(rightSource, leftSource, result, rightOperator,
				++leftOperator, remainingOrder - 1);

	}

}

void Takahashi::TwoSideSolver::recursiveSplit(
		State& splitSource,
		const State& otherSource,
		stateAmplitude& result,
		std::list<Utilities::OperatorDescription>::const_iterator splitOperator,
		std::list<Utilities::OperatorDescription>::const_iterator otherOperator,
		unsigned int remainingOrder) {

	if (remainingOrder == 1) {

		// Only one V operator is left.
		// We apply it directly.


		if (splitSource.size() < otherSource.size()) {

			solver.modules.pointerToModelProvider->applyV(splitSource,
					otherSource, result, splitOperator->remainingV);

		} else {

			solver.modules.pointerToModelProvider->applyV(otherSource,
					splitSource, result, otherOperator->remainingV);

		}

		return;

	}

	if (splitSource.size() > maximumNumberOfStates.value) {

		// The split state is to big to be processed directly.

		unsigned int numberOfSplits = splitSource.size()
				/ maximumNumberOfStates.value + 1;

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "Splitting state into " << numberOfSplits
					<< " sub states.\n";

		// So we must split it up.
		std::vector<State> splits = splitSource.splitState(numberOfSplits);

		for (std::vector<State>::iterator it = splits.begin(); it
				!= splits.end(); it++) {

			// And process every split state separately.
			stateAmplitude intermediateResult;

			recursiveSplit(*it, otherSource, intermediateResult, splitOperator,
					otherOperator, remainingOrder);

			it->clear();

			result += intermediateResult;

		}

		return;

	}

	// No split up necessary. Apply the operator and continue.
	splitOperator->apply(splitSource);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "Number of States: " << splitSource.size() << "\n";

	recursiveSplit(splitSource, otherSource, result, ++splitOperator,
			otherOperator, remainingOrder - 1);

}

