/*
 * TwoSideSolver.h
 *
 *  Created on: 22.02.2011
 *      Author: klagges
 */

#ifndef TAKAHASHITWOSIDESOLVER_H_
#define TAKAHASHITWOSIDESOLVER_H_

#include <Modules/DataProcessor/DataProcessor.h>
#include <Modules/DataProcessor/Takahashi/Common/Term.h>
#include <Utilities/OperatorDescription.h>
#include <Utilities/StreamableBaseType.h>
#include <set>

namespace Takahashi {

/** Calculates matrix elements of the effective Hamiltonian according to the
 degenerated perturbation method by Takahashi.

 This module reads the Solver::inputStates.right and the Solver::inputStates.left members
 and writes the matrix element \f$<l|H_{eff}|r>\f$ into the Solver::outputAmplitude
 member of the solver object.

 In this implementation the matrix elements of the operator parts of the coefficients
 are calculated by splitting each operator string into two parts of (roughly) equal
 length

 \f[ h = h_l h_r \f]

 and calculate the states

 \f[ h_r|i\rangle \f]

 and

 \f[ h_l^\dagger|li\rangle \f]

 separately. The matrix element is then calculated as

 \f[ \langle li|h_l\,h_r|i\rangle. \f]

 The matrix element of the whole effective Hamiltonian is calculated by summing over all terms.
 */
class TwoSideSolver: public DataProcessor {
public:

	TwoSideSolver();

	void execute();

private:

	/** Used to store a Takahashi coefficient term (like for example 1/2 PVSSVPVSVSVP) in a listed form.
	 */
	class ListedTerm {
	public:

		/** The term represented by this object */
		Term term;

		/** The operators of the right side of the splitting as a list of OperatorDescription objects. */
		std::list<Utilities::OperatorDescription> rightSideOperators;

		/** The operators of the left side of the splitting as a list of OperatorDescription objects. */
		std::list<Utilities::OperatorDescription> leftSideOperators;

	};
	friend std::ostream& operator<<(std::ostream& o,
			TwoSideSolver::ListedTerm& d);
	friend bool operator<(const TwoSideSolver::ListedTerm& l,
			const TwoSideSolver::ListedTerm& r);

	/** Class to store and stream in Takahashi coefficients.
	 * Stores every term in a set, with its operator part in a split form.*/
	class Coefficients: public Utilities::ParameterSet {

	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The coefficients. */
		std::set<ListedTerm> c;

	};

	/** Configuration to chose from different coefficient files*/
	Utilities::StreamableBaseType<std::string> nameOfCoefficientsFile;

	/** The coefficients.*/
	Coefficients coefficients;

	/** The maximum number of base states that should be processed without splitting the state.*/
	Utilities::StreamableBaseType<unsigned int> maximumNumberOfStates;

	/** Recursive method to calculate the value \f$\langle l|H_{eff}|l\rangle \f$.
	 * This method is intended for the part of the calculation, where the left and right part
	 * of the effective Hamilton need to be calculated separately, but it is not necessary
	 * to split the states up for memory reasons. */
	void
			recursive(
					State& rightSource,
					State& leftSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator rightOperator,
					std::list<Utilities::OperatorDescription>::const_iterator leftOperator,
					unsigned int remainingOrder);

	/** Recursive method to calculate the value \f$\langle l|H_{eff}|l\rangle \f$.
	 * This method is intended for the part of the calculation, where it is necessary to split up
	 * a state to save memory. */
	void
			recursiveSplit(
					State& splitSource,
					const State& otherSource,
					stateAmplitude& result,
					std::list<Utilities::OperatorDescription>::const_iterator splitOperator,
					std::list<Utilities::OperatorDescription>::const_iterator otherOperator,
					unsigned int remainingOrder);

};

/** Stream out as a string.*/
std::ostream& operator<<(std::ostream& o, TwoSideSolver::ListedTerm& d);

/** Used to sort the coefficients from higher to lower order.
 */
bool operator<(const TwoSideSolver::ListedTerm& l,
		const TwoSideSolver::ListedTerm& r);

}

#endif /* TAKAHASHITWOSIDESOLVER_H_ */
