/*
 * ClusterInFieldDispersion.cpp
 *
 *  Created on: 26.10.2010
 *      Author: kalis
 *
 */

#include <Modules/LatticeProvider/ClusterInFieldDispersion/ClusterInFieldDispersion.h>
#include <Framework/Solver.h>
#include <Interface/StateIO.h>
#include <math.h>

// The constructor should initialize all data needed for the implementation.
ClusterInFieldDispersion::ClusterInFieldDispersion() {
	/** set precision for all output */
	std::cout.precision(15);

	/** read from config file */
	configuration.readFromConfigFile("ClusterInField.cfg");

	if (configuration.fullHopping) {
		initializeCompleteHoppingList();
	}

	/** variables to control output (can be read from config file)*/
	configuration.executed = false;
	configuration.outputControlQP = configuration.startWith1QP;

	/** copy list for output */
	configuration.dist_copy = configuration.dist;
}

// In the module implementations this method is no longer virtual.
bool ClusterInFieldDispersion::execute() {
	/* if first calculation finished, store results in list */
	if (configuration.executed) {
		results.push_back(solver.outputAmplitude);
	} else {
		/** every command here will only be performed by first call of module */
		std::cout << "dx \t dy \t oS \t pR \t pD" << std::endl;
		configuration.executed = true;
	}

	/** iterators to navigate through vector and list; it_erase is always at the first element */
	std::list<std::vector<unsigned int>>::iterator it_erase =
			configuration.dist.begin();

	/** if no hopping elements left stop processing and create output */
	if (configuration.dist.empty()) {
		if ((configuration.file_out) && (configuration.console_out)) {
			results_copy = results;
		}
		output();
		return false;
	} else {
		createOptimzedBondfile(it_erase);
		it_erase = configuration.dist.erase(it_erase);
		return true;
	}
}

/*
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */
/** all functions delcared here */

std::list<Bond> ClusterInFieldDispersion::renameBondsAndStates(
		unsigned int bonds[300][300], unsigned int sx, unsigned int sy,
		unsigned int fx, unsigned int fy, BaseState& initial, BaseState& final) {
	unsigned int count = 0;

	/** list of Bonds returned at the end of procedure */
	std::list<Bond> RETURN;

	/** some helpful iterators */
	std::list<unsigned int>::iterator it;
	std::vector<std::list<unsigned int>>::iterator itt;
	std::map<unsigned int, unsigned int>::iterator ittt;

	/** Connect new Bond Numbers with old Bond Numbers */
	std::map<unsigned int, unsigned int> SiteMap;

	/** Bond & Site numbers of current Coordinate and temp list to make a copy */
	std::list<unsigned int> curCoor;
	std::list<unsigned int> temp;

	/** vector to store lists of Sites touched by every Bond */
	std::vector<std::list<unsigned int>> allBonds;

	/** List with all touched Sites and Bonds */
	std::list<unsigned int> allCoor;

	for (unsigned int y = 0; y < 300; y++) {
		for (unsigned int x = 0; x < 300; x++) {
			if (bonds[y][x] == 1) {
				curCoor = findSitesForBond(x, y);
				temp = curCoor;
				allCoor.merge(temp); // merge makes an empty list out of temp
				allBonds.push_back(curCoor);
				curCoor.clear();
				temp.clear();
			}
		}
	}
	/** save all touched Sites */
	allCoor.sort();
	allCoor.unique();
	unsigned int SIZE = ((unsigned int) allCoor.size());

	Bond newBond, newhzBond;
	newBond.index = 0;
	newBond.operatorNumber = 1;
	newhzBond.index = 0;
	newhzBond.operatorNumber = 0;

	for (unsigned int i = 0; i < SIZE; i++) {
		newhzBond.sites.push_back(count);
		RETURN.push_back(newhzBond);
		newhzBond.sites.clear();
		SiteMap.insert(
				std::pair<unsigned int, unsigned int>(allCoor.front(), count));
		count++;
		allCoor.pop_front();
	}

	solver.latticeProperties.numberOfLatticeSites = count;

	/** create initial and final State with new site numbers */
	unsigned int START = (sy * 300) + sx;
	unsigned int STOP = (fy * 300) + fx;
	unsigned int NEWSTART = SiteMap.find(START)->second;
	unsigned int NEWSTOP = SiteMap.find(STOP)->second;
	initial.set(NEWSTART);
	final.set(NEWSTOP);

	/*	  //debug cout for old site number <-> new site numbers
	 for ( ittt=SiteMap.begin() ; ittt != SiteMap.end(); ittt++ ){
	 std::cout << (*ittt).first << "=>" << (*ittt).second << std::endl;
	 }
	 */

	for (itt = allBonds.begin(); itt < allBonds.end(); itt++) {
		for (it = (*itt).begin(); it != (*itt).end(); it++) {
			newBond.sites.push_back(SiteMap.find((*it))->second);
		}
		RETURN.push_back(newBond);
		newBond.sites.clear();
	}
	return RETURN;
}

void ClusterInFieldDispersion::consoleOutput() {
	unsigned int control = 0;
	unsigned int x, y;
	std::list<std::vector<unsigned int>>::iterator it;
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		if (control == 0) {
			std::cout << solver.configurationDirectory << "zeroQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else if (control == 1) {
			std::cout << solver.configurationDirectory << "oneQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else {
			x = *(it->begin());
			y = *((it->begin()) + 1);
			std::cout << solver.configurationDirectory << "t[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
		}
	}
	return;
}

void ClusterInFieldDispersion::consoleOutputWithDispersion() {
	results_copy = results;
	unsigned int control = 0;
	unsigned int size;
	unsigned int x, y;
	std::list<std::vector<unsigned int>>::iterator it;
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		if (control == 0) {
			std::cout << solver.configurationDirectory << "zeroQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else if (control == 1) {
			std::cout << solver.configurationDirectory << "oneQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else {
			x = *(it->begin());
			y = *((it->begin()) + 1);
			std::cout << solver.configurationDirectory << "t[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				std::cout << results.front() << " ;" << std::endl;
			} else {
				std::cout << 0 << " ;" << std::endl;
			}
			results.pop_front();
		}
	}
	control = 0;
	results = results_copy;
	std::cout << std::endl;
	std::cout << solver.configurationDirectory
			<< "Dispersion := simplify(sort(1+";
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		size = (unsigned int) results.size();
		if (size > 1) {
			if (control == 0) {
				std::cout << "(-1*(" << solver.configurationDirectory
						<< "zeroQPt[" << x << y << "] ) + ";
				results.pop_front();
				control++;
			} else if (control == 1) {
				std::cout << "(" << solver.configurationDirectory << "oneQPt["
						<< x << y << "])) + ";
				results.pop_front();
				control++;
			} else {
				if (!Utilities::equalsZero(results.front())) {
					if (x == y) {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx + " << y << "*ky) + cos(" << x
								<< "*kx - " << y << "*ky)) + ";
					} else if (y == 0) {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx) + cos(" << x << "*ky)) + ";
					} else {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x << "*kx+"
								<< y << "*ky) + cos(" << x << "*kx-" << y
								<< "*ky) + cos(" << y << "*kx+" << x
								<< "*ky) + cos(-" << y << "*kx+" << x
								<< "*ky)) + ";
					}
				} else {
					std::cout << " 0 + ";
				}
				results.pop_front();
			}
		} else {
			if (control == 0) {
				std::cout << "(" << results.front() << " - ";
				results.pop_front();
				control++;
			} else if (control == 1) {
				std::cout << results.front() << ") + ";
				results.pop_front();
				control++;
			} else {
				if (!Utilities::equalsZero(results.front())) {
					x = *(it->begin());
					y = *((it->begin()) + 1);
					if (x == y) {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx + " << y << "*ky) + cos(" << x
								<< "*kx - " << y << "*ky))";
					} else if (y == 0) {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx) + cos(" << x << "*ky))";
					} else {
						std::cout << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x << "*kx+"
								<< y << "*ky) + cos(" << x << "*kx-" << y
								<< "*ky) + cos(" << y << "*kx+" << x
								<< "*ky) + cos(-" << y << "*kx+" << x
								<< "*ky))";
					}
				} else {
					std::cout << "0";
				}
				results.pop_front();
			}
		}
		std::cout << std::endl;
	}
	std::cout << ",[h0,h1],ascending));" << std::endl;
	return;
}

void ClusterInFieldDispersion::fileOutput() {
	unsigned int control = 0;
	unsigned int x, y;
	std::ofstream result;
	result.precision(15);
	std::string LP = solver.modules.nameOfLatticeProvider;
	std::string order = convertInt(solver.globalConfiguration.order);
	std::string file = LP + "_O" + order + "_" + solver.configurationDirectory
			+ ".res";
	result.open(file.c_str());
	std::list<std::vector<unsigned int>>::iterator it;
	std::vector<unsigned int>::iterator itv;
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		if (control == 0) {
			result << solver.configurationDirectory << "zeroQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else if (control == 1) {
			result << solver.configurationDirectory << "oneQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else {
			x = *(it->begin());
			y = *((it->begin()) + 1);
			result << solver.configurationDirectory << "t[" << x << y << "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
		}
	}
	result.close();
	return;
}

void ClusterInFieldDispersion::fileOutputWithDispersion() {
	results_copy = results;
	unsigned int control = 0;
	unsigned int size;
	unsigned int x, y;
	std::ofstream result;
	result.precision(15);
	std::string LP = solver.modules.nameOfLatticeProvider;
	std::string order = convertInt(solver.globalConfiguration.order);
	std::string file = LP + "_O" + order + "_" + solver.configurationDirectory
			+ "_Dispersion.res";
	result.open(file.c_str());
	std::list<std::vector<unsigned int>>::iterator it;
	std::vector<unsigned int>::iterator itv;
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		if (control == 0) {
			result << solver.configurationDirectory << "zeroQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else if (control == 1) {
			result << solver.configurationDirectory << "oneQPt[" << x << y
					<< "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
			control++;
		} else {
			x = *(it->begin());
			y = *((it->begin()) + 1);
			result << solver.configurationDirectory << "t[" << x << y << "]:= ";
			if (!Utilities::equalsZero(results.front())) {
				result << results.front() << " ;" << std::endl;
			} else {
				result << 0 << " ;" << std::endl;
			}
			results.pop_front();
		}
	}
	control = 0;
	results = results_copy;
	result << std::endl;
	result << solver.configurationDirectory << "Dispersion := simplify(sort(1+";
	for (it = configuration.dist_copy.begin(); it
			!= configuration.dist_copy.end(); it++) {
		x = *(it->begin());
		y = *((it->begin()) + 1);
		size = (unsigned int) results.size();
		if (size > 1) {
			if (control == 0) {
				result << "(-1*(" << solver.configurationDirectory
						<< "zeroQPt[" << x << y << "] ) + ";
				results.pop_front();
				control++;
			} else if (control == 1) {
				result << "(" << solver.configurationDirectory << "oneQPt["
						<< x << y << "])) + ";
				results.pop_front();
				control++;
			} else {
				if (!Utilities::equalsZero(results.front())) {
					if (x == y) {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx + " << y << "*ky) + cos(" << x
								<< "*kx - " << y << "*ky)) + ";
					} else if (y == 0) {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx) + cos(" << x << "*ky)) + ";
					} else {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x << "*kx+"
								<< y << "*ky) + cos(" << x << "*kx-" << y
								<< "*ky) + cos(" << y << "*kx+" << x
								<< "*ky) + cos(-" << y << "*kx+" << x
								<< "*ky)) + ";
					}
				} else {
					result << " 0 + ";
				}
				results.pop_front();
			}
		} else {
			if (control == 0) {
				result << "(" << results.front() << " - ";
				results.pop_front();
				control++;
			} else if (control == 1) {
				result << results.front() << ") + ";
				results.pop_front();
				control++;
			} else {
				if (!Utilities::equalsZero(results.front())) {
					x = *(it->begin());
					y = *((it->begin()) + 1);
					if (x == y) {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx + " << y << "*ky) + cos(" << x
								<< "*kx - " << y << "*ky))";
					} else if (y == 0) {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x
								<< "*kx) + cos(" << x << "*ky))";
					} else {
						result << "2*(" << solver.configurationDirectory
								<< "t[" << x << y << "] )*(cos(" << x << "*kx+"
								<< y << "*ky) + cos(" << x << "*kx-" << y
								<< "*ky) + cos(" << y << "*kx+" << x
								<< "*ky) + cos(-" << y << "*kx+" << x
								<< "*ky))";
					}
				} else {
					result << "0";
				}
				results.pop_front();
			}
		}
		result << std::endl;
	}
	result << ",[h0,h1],ascending));" << std::endl;
	result.close();
	return;
}

std::list<Bond> ClusterInFieldDispersion::createPeriodicSquareCluster(
		unsigned int size, unsigned int Dx, unsigned int Dy,
		BaseState& initial, BaseState& final) {
	std::list<Bond> RETURN;

	unsigned int NEWSTART = ((size) * (size)) / 2;
	unsigned int NEWSTOP;
	if ((Dx > (size / 2)) && (Dy > (size / 2))) {
		NEWSTOP = ((size) * (size)) / 2 - (size - Dx) - ((size - Dy) * size);
	} else if ((Dx > (size / 2)) && (Dy < (size / 2))) {
		NEWSTOP = ((size) * (size)) / 2 - (size - Dx) + (Dy * size);
	} else if ((Dx < (size / 2)) && (Dy > (size / 2))) {
		NEWSTOP = ((size) * (size)) / 2 + Dx - ((size - Dy) * size);
	} else if ((Dx <= (size / 2)) && (Dy <= (size / 2))) {
		NEWSTOP = ((size) * (size)) / 2 + Dx + (Dy * size);
	} else {
		NEWSTOP = NEWSTART;
		std::cout << "Something unexpected happend" << std::endl;
	}
	//std::cout << "start: " << NEWSTART << " stop: " << NEWSTOP << std::endl;

	initial.set(NEWSTART);
	final.set(NEWSTOP);

	Bond hzBond, hxBond;
	hzBond.index = 0;
	hxBond.index = 0;
	hxBond.operatorNumber = 1;
	hzBond.operatorNumber = 0;

	for (unsigned int i = 0; i < (size * size); i++) {
		hzBond.sites.push_back(i);
		if (i < size) {
			hxBond.sites.push_back((size * size) + i - size);
		} else {
			hxBond.sites.push_back(i - (size));
		}
		if (i % size == 0) {
			hxBond.sites.push_back(i - 1 + size);
		} else {
			hxBond.sites.push_back(i - 1);
		}
		hxBond.sites.push_back(i);
		if (i % size == (size - 1)) {
			hxBond.sites.push_back(i - (size - 1));
		} else {
			hxBond.sites.push_back(i + 1);
		}
		if (i >= ((size * size) - size)) {
			hxBond.sites.push_back(i % size);
		} else {
			hxBond.sites.push_back(i + size);
		}

		RETURN.push_back(hzBond);
		RETURN.push_back(hxBond);
		hzBond.sites.clear();
		hxBond.sites.clear();
	}

	return RETURN;
}

void ClusterInFieldDispersion::comparePeriodicRectangleAndOpenBoundary(
		unsigned int nx, unsigned int ny, unsigned int x, unsigned int y,
		unsigned int periodicRectangleSize, BaseState& initial,
		BaseState& final) {
	/** make sure to compare cluster-sizes in follwing order: newSize = compare(periodicPectangle,OpenBoundary) -> size = compare(newSize,periodicDiamond) */
	std::cout << solver.latticeProperties.numberOfLatticeSites << "\t"
			<< periodicRectangleSize << "\t";
	//if (solver.latticeProperties.numberOfLatticeSites > periodicRectangleSize) {
	/** always take rectangle for taka */
	if (true) {
		solver.latticeProperties.bonds.clear();
		initial.clear();
		final.clear();
		solver.latticeProperties.numberOfLatticeSites = periodicRectangleSize;
		solver.latticeProperties.bonds = createPeriodicRectangleCluster(nx, ny,
				x, y, periodicRectangleSize, initial, final);
	}
	return;
}

void ClusterInFieldDispersion::comparePeriodicDiamondAndOthers(unsigned int nx,
		unsigned int ny, unsigned int x, unsigned int y,
		unsigned int periodicDiamondSize, BaseState& initial, BaseState& final) {
	/** make sure to compare cluster-sizes in follwing order: newSize = compare(periodicPectangle,OpenBoundary) -> size = compare(newSize,periodicDiamond) */
	std::cout << periodicDiamondSize << std::endl;
	if (solver.latticeProperties.numberOfLatticeSites > periodicDiamondSize) {
		solver.latticeProperties.bonds.clear();
		initial.clear();
		final.clear();
		solver.latticeProperties.numberOfLatticeSites = periodicDiamondSize;
		solver.latticeProperties.bonds = createPeriodicDiamondCluster(x, y,
				periodicDiamondSize, initial, final);
	}
	return;
}

std::list<Bond> ClusterInFieldDispersion::createPeriodicDiamondCluster(
		unsigned int Dx, unsigned int Dy, unsigned int periodicDiamondSize,
		BaseState& initial, BaseState& final) {
	std::list<Bond> RETURN;
	unsigned int asymetricConstant = abs(Dx - Dy);
	unsigned int bmax, hmax;
	bmax = 2 * solver.globalConfiguration.order + 1 + asymetricConstant;
	hmax = 2 * solver.globalConfiguration.order + Dy + 1;
	unsigned int NEWSTART = 0;
	unsigned int NEWSTOP = 0;
	for (unsigned int i = 0; i < solver.globalConfiguration.order; i++) {
		NEWSTART += 2 * i + asymetricConstant + 1;
	}
	NEWSTART += solver.globalConfiguration.order + Dy * bmax;
	NEWSTOP = NEWSTART + Dx - (Dy * bmax) - Dy;

	//std::cout << "start: " << NEWSTART << " stop: " << NEWSTOP << std::endl;


	initial.set(NEWSTART);
	final.set(NEWSTOP);

	Bond hzBond, hxBond;
	hzBond.index = 0;
	hxBond.index = 0;
	hxBond.operatorNumber = 1;
	hzBond.operatorNumber = 0;

	unsigned int count = 0;
	for (unsigned int y = 0; y < hmax; y++) {
		unsigned int x, X;
		x = 0;
		if ((y <= solver.globalConfiguration.order)) {
			X = 2 * y + asymetricConstant + 1;
		} else if ((y > solver.globalConfiguration.order) && (y
				< solver.globalConfiguration.order + Dy + 1)) {
			X = bmax;
		} else if (y >= solver.globalConfiguration.order + Dy + 1) {
			X = 2 * (hmax - y - 1) + asymetricConstant + 1;
		} else {
			std::cout << "" << std::endl;
		}
		while (x < X) {
			//at top of the cluster
			if (y == 0) {
				if (asymetricConstant != 0) {
					if (x == 0) {
						// top
						unsigned int t = NEWSTOP
								+ solver.globalConfiguration.order;
						hxBond.sites.push_back(t);
						// left
						if ((y + 1) < (Dy + 1)) {
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order
									+ (y + 1) * bmax;
							hxBond.sites.push_back(l);
						} else {
							unsigned int ynew = (y + 1) - (Dy + 1);
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								l += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(l);
						}
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					} else if (x == (X - 1)) {
						// top
						unsigned int t = periodicDiamondSize - 2;
						hxBond.sites.push_back(t);
						// left
						hxBond.sites.push_back(count - 1);
						// on site
						hxBond.sites.push_back(count);
						// right
						unsigned int r = (NEWSTART
								- solver.globalConfiguration.order);
						hxBond.sites.push_back(r);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					} else {
						//top
						unsigned int t = periodicDiamondSize - 1
								- (asymetricConstant + 1) + count;
						hxBond.sites.push_back(t);
						// left
						hxBond.sites.push_back(count - 1);
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					}
				} else {
					// top
					unsigned int t = NEWSTOP + solver.globalConfiguration.order;
					hxBond.sites.push_back(t);
					// left
					if ((y + 1) < (Dy + 1)) {
						unsigned int l = NEWSTOP
								+ solver.globalConfiguration.order + (y + 1)
								* bmax;
						hxBond.sites.push_back(l);
					} else {
						unsigned int ynew = (y + 1) - (Dy + 1);
						unsigned int l = NEWSTOP
								+ solver.globalConfiguration.order + Dy * bmax;
						for (unsigned int i = 0; i <= ynew; i++) {
							l += (bmax - 2 * (i + 1));
						}
						hxBond.sites.push_back(l);
					}
					// on site
					hxBond.sites.push_back(count);
					// right
					unsigned int r = (NEWSTART
							- solver.globalConfiguration.order);
					hxBond.sites.push_back(r);
					// bottom
					hxBond.sites.push_back(count + X + 1);
				}
			}
			// not at top or bottom of the cluster but at the ...
			else if ((y != 0) && (y != (hmax - 1))) {
				// ... left edge of the cluster
				if (x == 0) {
					if ((y < solver.globalConfiguration.order)) {
						//TODO: check if this case is the same as the next ... always!
						// top
						if (y < (Dy + 1)) {
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + y
									* bmax;
							hxBond.sites.push_back(t);
						} else {
							unsigned int ynew = y - (Dy + 1);
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								t += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(t);
						}
						// left
						if ((y + 1) < (Dy + 1)) {
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order
									+ (y + 1) * bmax;
							hxBond.sites.push_back(l);
						} else {
							unsigned int ynew = (y + 1) - (Dy + 1);
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								l += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(l);
						}
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					} else if ((y >= solver.globalConfiguration.order) && (y
							< solver.globalConfiguration.order + Dy)) {
						// top
						if (y < (Dy + 1)) {
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + y
									* bmax;
							hxBond.sites.push_back(t);
						} else {
							unsigned int ynew = y - (Dy + 1);
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								t += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(t);
						}

						// left
						if ((y + 1) < (Dy + 1)) {
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order
									+ (y + 1) * bmax;
							hxBond.sites.push_back(l);
						} else {
							unsigned int ynew = (y + 1) - (Dy + 1);
							unsigned int l = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								l += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(l);
						}

						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					} else if (y == solver.globalConfiguration.order + Dy) {
						// top
						if (y < (Dy + 1)) {
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + y
									* bmax;
							hxBond.sites.push_back(t);
						} else {
							unsigned int ynew = y - (Dy + 1);
							unsigned int t = NEWSTOP
									+ solver.globalConfiguration.order + Dy
									* bmax;
							for (unsigned int i = 0; i <= ynew; i++) {
								t += (bmax - 2 * (i + 1));
							}
							hxBond.sites.push_back(t);
						}
						// left
						hxBond.sites.push_back(asymetricConstant);
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						hxBond.sites.push_back(2 * (asymetricConstant + 1) + 1);
					} else if ((y > solver.globalConfiguration.order + Dy)) {
						// top
						hxBond.sites.push_back(count - X - 1);
						// left
						unsigned int ynew = y
								- (solver.globalConfiguration.order + Dy);
						unsigned int l = 2 * (asymetricConstant + 1) + 1;
						for (unsigned int i = 2; i <= ynew; i++) {
							l += (asymetricConstant + 1) + 2 * i;
						}
						hxBond.sites.push_back(l);
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						unsigned int b = 2 * (asymetricConstant + 1) + 1;
						for (unsigned int i = 2; i <= (ynew + 1); i++) {
							b += (asymetricConstant + 1) + 2 * i;
						}
						hxBond.sites.push_back(b);
					} else {
						std::cout << "hier stimmt was nicht1" << std::endl;
					}
				}
				// ...right edge of the cluster
				else if (x == (X - 1)) {
					// devide into three parts
					if ((y < solver.globalConfiguration.order)) {
						// top
						unsigned int t = (NEWSTART
								- solver.globalConfiguration.order);
						for (unsigned int i = 1; i < y; i++) {
							t += (bmax - 2 * (i - 1));
						}
						hxBond.sites.push_back(t);
						// left
						hxBond.sites.push_back(count - 1);
						// on site
						hxBond.sites.push_back(count);
						// right
						unsigned int r = (NEWSTART
								- solver.globalConfiguration.order);
						for (unsigned int i = 1; i < (y + 1); i++) {
							r += (bmax - 2 * (i - 1));
						}
						hxBond.sites.push_back(r);
						// bottom
						hxBond.sites.push_back(count + X + 1);
					} else if (y == solver.globalConfiguration.order) {
						// top
						unsigned int t = (NEWSTART
								- solver.globalConfiguration.order);
						for (unsigned int i = 1; i < y; i++) {
							t += (bmax - 2 * (i - 1));
						}
						hxBond.sites.push_back(t);
						// left
						hxBond.sites.push_back(count - 1);
						// on site
						hxBond.sites.push_back(count);
						// right
						unsigned int r = (NEWSTART
								- solver.globalConfiguration.order);
						for (unsigned int i = 1; i < (y + 1); i++) {
							r += (bmax - 2 * (i - 1));
						}
						hxBond.sites.push_back(r);
						//bottom
						hxBond.sites.push_back(0);
					} else if ((y > solver.globalConfiguration.order) && (y
							<= solver.globalConfiguration.order + Dy)) {
						//TODO: check if this case is the same as the next ... always!
						//top
						hxBond.sites.push_back(count - X - 1);
						//left
						hxBond.sites.push_back(count - 1);
						//on site
						hxBond.sites.push_back(count);
						//right
						unsigned int ynew = y
								- solver.globalConfiguration.order;
						unsigned int r = 0;
						for (unsigned int i = 1; i < ynew; i++) {
							r += (asymetricConstant - 1) + 2 * i;
						}
						hxBond.sites.push_back(r);
						//bottom
						unsigned int ynew2 = (y + 1)
								- solver.globalConfiguration.order;
						unsigned int b = 0;
						for (unsigned int i = 1; i < ynew2; i++) {
							b += (asymetricConstant - 1) + 2 * i;
						}
						hxBond.sites.push_back(b);
					} else if (y > solver.globalConfiguration.order + Dy) {
						//top
						hxBond.sites.push_back(count - X - 1);
						//left
						hxBond.sites.push_back(count - 1);
						//on site
						hxBond.sites.push_back(count);
						//right
						unsigned int ynew = y
								- solver.globalConfiguration.order;
						unsigned int r = 0;
						for (unsigned int i = 1; i < ynew; i++) {
							r += (asymetricConstant - 1) + 2 * i;
						}
						hxBond.sites.push_back(r);
						//bottom
						unsigned int ynew2 = (y + 1)
								- solver.globalConfiguration.order;
						unsigned int b = 0;
						for (unsigned int i = 1; i < ynew2; i++) {
							b += (asymetricConstant - 1) + 2 * i;
						}
						hxBond.sites.push_back(b);
					} else {
						std::cout << "hier stimmt was nicht2" << std::endl;
					}
				}
				// ... everything NOT at the edge of the cluster
				else {
					if (Dy != 0) {
						// devide into three parts
						if ((y < solver.globalConfiguration.order)) {
							// top
							hxBond.sites.push_back(count - X + 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X + 1);
						} else if (y == solver.globalConfiguration.order) {
							// top
							hxBond.sites.push_back(count - X + 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X + 1);
						} else if ((y > solver.globalConfiguration.order) && (y
								< solver.globalConfiguration.order + Dy)) {
							// top
							hxBond.sites.push_back(count - X - 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X + 1);
						} else if (y >= solver.globalConfiguration.order + Dy) {
							// top
							hxBond.sites.push_back(count - X - 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X - 1);
						}
					} else {
						// devide into three parts
						if ((y < solver.globalConfiguration.order)) {
							// top
							hxBond.sites.push_back(count - X + 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X + 1);
						} else if (y == solver.globalConfiguration.order) {
							// top
							hxBond.sites.push_back(count - X + 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X - 1);
						} else if (y > solver.globalConfiguration.order) {
							// top
							hxBond.sites.push_back(count - X - 1);
							// left
							hxBond.sites.push_back(count - 1);
							// on site
							hxBond.sites.push_back(count);
							// right
							hxBond.sites.push_back(count + 1);
							// bottom
							hxBond.sites.push_back(count + X - 1);
						}
					}

				}
			}
			//at bottom of the cluster
			else if (y == (hmax - 1)) {
				if (asymetricConstant != 0) {
					if (x == 0) {
						// top
						hxBond.sites.push_back(count - X - 1);
						// left
						unsigned int ynew = y
								- (solver.globalConfiguration.order + Dy);
						unsigned int l = 2 * (asymetricConstant + 1) + 1;
						for (unsigned int i = 2; i <= ynew; i++) {
							l += (asymetricConstant + 1) + 2 * i;
						}
						hxBond.sites.push_back(l);
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						unsigned int b = (asymetricConstant + 1)
								- (periodicDiamondSize - 1 - count);
						hxBond.sites.push_back(b);
					} else if (x == (X - 1)) {
						//top
						hxBond.sites.push_back(count - X - 1);
						//left
						hxBond.sites.push_back(count - 1);
						//on site
						hxBond.sites.push_back(count);
						//right
						unsigned int ynew = y
								- solver.globalConfiguration.order;
						unsigned int r = 0;
						for (unsigned int i = 1; i < ynew; i++) {
							r += (asymetricConstant - 1) + 2 * i;
						}
						hxBond.sites.push_back(r);
						//bottom
						unsigned int b = NEWSTART
								- solver.globalConfiguration.order;
						hxBond.sites.push_back(b);
					} else {
						// top
						hxBond.sites.push_back(count - X - 1);
						// left
						hxBond.sites.push_back(count - 1);
						// on site
						hxBond.sites.push_back(count);
						// right
						hxBond.sites.push_back(count + 1);
						// bottom
						unsigned int b = (asymetricConstant + 1)
								- (periodicDiamondSize - 1 - count);
						hxBond.sites.push_back(b);
					}
				} else {
					// top
					hxBond.sites.push_back(count - X - 1);
					// left
					unsigned int ynew = y - (solver.globalConfiguration.order
							+ Dy);
					unsigned int l = 2 * (asymetricConstant + 1) + 1;
					for (unsigned int i = 2; i <= ynew; i++) {
						l += (asymetricConstant + 1) + 2 * i;
					}
					hxBond.sites.push_back(l);
					// on site
					hxBond.sites.push_back(count);
					// right
					ynew = y - solver.globalConfiguration.order;
					unsigned int r = 0;
					for (unsigned int i = 1; i < ynew; i++) {
						r += (asymetricConstant - 1) + 2 * i;
					}
					hxBond.sites.push_back(r);
					// bottom
					unsigned int b = NEWSTART
							- solver.globalConfiguration.order;
					hxBond.sites.push_back(b);
				}

			}
			//nowhere in the cluster?
			else {
				std::cout << "was?" << std::endl;
			}
			RETURN.push_back(hxBond);
			hzBond.sites.push_back(count);
			RETURN.push_back(hzBond);
			hxBond.sites.clear();
			hzBond.sites.clear();
			x++;
			count++;
		}
	}
	return RETURN;
}

std::list<Bond> ClusterInFieldDispersion::createPeriodicRectangleCluster(
		unsigned int nx, unsigned int ny, unsigned int Dx, unsigned int Dy,
		unsigned int periodicRectangleSize, BaseState& initial,
		BaseState& final) {
	std::list<Bond> RETURN;
	/** TODO: catch case dy or dx = 0 will results in Bond entry 6 5 6 7 6 */
	unsigned int Ny = ((Dy + 1) + ny);
	unsigned int Nx = ((Dx + 1) + nx);
	unsigned int NEWSTART = (Ny - 1) * Nx;
	unsigned int NEWSTOP = NEWSTART + Dx - (Dy * Nx);

	//std::cout << "start: " << NEWSTART << " stop: " << NEWSTOP << std::endl;


	initial.set(NEWSTART);
	final.set(NEWSTOP);

	Bond hzBond, hxBond;
	hzBond.index = 0;
	hxBond.index = 0;
	hxBond.operatorNumber = 1;
	hzBond.operatorNumber = 0;

	for (unsigned int i = 0; i < (periodicRectangleSize); i++) {
		if ((i < Nx)) {
			hxBond.sites.push_back(periodicRectangleSize - Nx + i);
		} else {
			hxBond.sites.push_back(i - Nx);
		}
		if (i % Nx == 0) {
			hxBond.sites.push_back(i + (Nx - 1));
		} else {
			hxBond.sites.push_back(i - 1);
		}
		hxBond.sites.push_back(i);
		hzBond.sites.push_back(i);
		if (i % Nx == (Nx - 1)) {
			hxBond.sites.push_back(i - (Nx - 1));
		} else {
			hxBond.sites.push_back(i + 1);
		}
		if (i >= (periodicRectangleSize - Nx)) {
			hxBond.sites.push_back(i % Nx);
		} else {
			hxBond.sites.push_back(i + Nx);
		}
		RETURN.push_back(hxBond);
		RETURN.push_back(hzBond);
		hxBond.sites.clear();
		hzBond.sites.clear();
	}
	return RETURN;
}

void ClusterInFieldDispersion::output() {
	if (!configuration.generate_dispersion) {
		if (configuration.file_out && configuration.console_out) {
			fileOutput();
			results = results_copy;
			consoleOutput();
		} else if (configuration.file_out) {
			fileOutput();
		} else if (configuration.console_out) {
			consoleOutput();
		}
	} else if ((configuration.generate_dispersion)
			&& (configuration.outputControlQP)) {
		std::cerr
				<< "Please check configuration! Cannot compute dispersion without any 0QP expression! Swichting to normal output!"
				<< std::endl;
		if (configuration.file_out && configuration.console_out) {
			fileOutput();
			results = results_copy;
			consoleOutput();
			;
		} else if (configuration.file_out) {
			fileOutput();
		} else if (configuration.console_out) {
			consoleOutput();
		}
	} else {
		if (configuration.file_out && configuration.console_out) {
			fileOutputWithDispersion();
			results = results_copy;
			consoleOutputWithDispersion();
		} else if (configuration.file_out) {
			fileOutputWithDispersion();
		} else if (configuration.console_out) {
			consoleOutputWithDispersion();
		}
	}
	return;
}

void ClusterInFieldDispersion::createOptimzedBondfile(
		std::list<std::vector<unsigned int>>::iterator it_erase) {
	/* set initial and final BaseStates */
	BaseState initial;
	BaseState final;

	/** start,final site; number of extra rows in x an y direction */
	unsigned int x, y, sx, sy, fx, fy, nx, ny;
	unsigned int maxDist = 2 * solver.globalConfiguration.order;

	/** set starting point in the middle of the big bond array */
	sx = 150;
	sy = 150;

	/** create big bond storage arrray */
	unsigned int bonds[300][300];
	resetBondArray(bonds);

	/** use auxiluary variables to store distance */
	x = *(it_erase->begin());
	y = *((it_erase->begin()) + 1);

	unsigned int asymetricConstant = abs(x - y);
	/** number of extra rows for rectangles with periodic boundary */
	nx = solver.globalConfiguration.order - 1;
	ny = solver.globalConfiguration.order - 1;
	if ((y == 0) && ((solver.globalConfiguration.order % 2) == 0)) {
		nx += 1;
	}
	if ((y == 0) && (x == 0)) {
		nx += 1;
		ny += 1;
	}
	std::cout << x << "\t" << y << "\t";

	/** calculate size of periodic rectangle cluster */
	unsigned int periodicRectangleSize = ((y + 1) + ny) * ((x + 1) + nx);

	/** characteristics of periodic diamond */
	unsigned int bmax, hmax;
	bmax = 2 * solver.globalConfiguration.order + 1 + asymetricConstant;
	hmax = 2 * solver.globalConfiguration.order + y + 1;
	unsigned int periodicDiamondSize = bmax * (y + 1);
	for (unsigned int i = 1; i <= solver.globalConfiguration.order; i++) {
		periodicDiamondSize += 2 * (bmax - 2 * i);
	}

	/** set final sites for closed boundary */
	fx = 150 + x;
	fy = 150 + y;

	/** distance fixed, first compare periodic and closed boundary then create bonds */
	if ((x + y) == 0) {
		markAllBondsWithinDistance(bonds, sx, sy, maxDist);
		//printBigBondArray(bonds,sx-maxDist,sy-maxDist,sx+maxDist,sy+maxDist);
		solver.latticeProperties.bonds = renameBondsAndStates(bonds, sx, sy,
				fx, fy, initial, final);

		/** compare open and periodic bonds and pick smaller one */
		comparePeriodicRectangleAndOpenBoundary(nx, ny, x, y,
				periodicRectangleSize, initial, final);
		comparePeriodicDiamondAndOthers(nx, ny, x, y, periodicDiamondSize,
				initial, final);
		createInputStates(initial, final);
	} else if ((x + y) == maxDist) {
		/** mark all type 1 bonds in big bond array */
		markAllTypeOneBonds(bonds, sx, sy, fx, fy);
		solver.latticeProperties.bonds = renameBondsAndStates(bonds, sx, sy,
				fx, fy, initial, final);
		//std::cout << solver.latticeProperties;
		//printBigBondArray(bonds,sx-maxDist,sy-maxDist,sx+maxDist,sy+maxDist);
		comparePeriodicRectangleAndOpenBoundary(nx, ny, x, y,
				periodicRectangleSize, initial, final);
		comparePeriodicDiamondAndOthers(nx, ny, x, y, periodicDiamondSize,
				initial, final);
		createInputStates(initial, final);
		/*
		 std::cout << std::endl;
		 std::cout << "x: " << x << " y: " << y << std::endl;
		 std::cout << std::endl;
		 */
	} else if ((x + y) == (maxDist - 1)) {
		/** mark all type 2 bonds in big bond array */
		markAllTypeTwoBonds(bonds, sx, sy, fx, fy);
		//printBigBondArray(bonds,sx-maxDist,sy-maxDist,sx+maxDist,sy+maxDist);
		solver.latticeProperties.bonds = renameBondsAndStates(bonds, sx, sy,
				fx, fy, initial, final);
		comparePeriodicRectangleAndOpenBoundary(nx, ny, x, y,
				periodicRectangleSize, initial, final);
		comparePeriodicDiamondAndOthers(nx, ny, x, y, periodicDiamondSize,
				initial, final);
		createInputStates(initial, final);
	} else if (((x + y) < (maxDist - 1)) && ((x + y) % 2 == 0)) {
		/** mark all type 2 bonds in big bond array */
		std::list<std::vector<unsigned int>> markedBonds =
				markAllTypeOneBondsReturnMarkedBonds(bonds, sx, sy, fx, fy);
		while (!markedBonds.empty()) {
			std::vector<unsigned int> marked = markedBonds.front();
			unsigned int X, Y, dist;
			X = marked.front();
			Y = marked.back();
			dist = maxDist - (x + y);
			markAllBondsWithinDistance(bonds, X, Y, dist);
			markedBonds.pop_front();
		}
		solver.latticeProperties.bonds = renameBondsAndStates(bonds, sx, sy,
				fx, fy, initial, final);
		//printBigBondArray(bonds,sx-maxDist,sy-maxDist,sx+maxDist,sy+maxDist);
		comparePeriodicRectangleAndOpenBoundary(nx, ny, x, y,
				periodicRectangleSize, initial, final);
		comparePeriodicDiamondAndOthers(nx, ny, x, y, periodicDiamondSize,
				initial, final);
		//std::cout << solver.latticeProperties << std::endl;
		createInputStates(initial, final);

	} else if (((x + y) < (maxDist - 1)) && ((x + y) % 2 == 1)) {
		/** mark all type 2 bonds in big bond array */
		std::list<std::vector<unsigned int>> markedBonds =
				markAllTypeTwoBondsReturnMarkedBonds(bonds, sx, sy, fx, fy);
		while (!markedBonds.empty()) {
			std::vector<unsigned int> marked = markedBonds.front();
			unsigned int X, Y, dist;
			X = marked.front();
			Y = marked.back();
			dist = 2 * (solver.globalConfiguration.order - (((x + y) / 2) + 1));
			markAllBondsWithinDistance(bonds, X, Y, dist);
			markedBonds.pop_front();
		}

		solver.latticeProperties.bonds = renameBondsAndStates(bonds, sx, sy,
				fx, fy, initial, final);

		comparePeriodicRectangleAndOpenBoundary(nx, ny, x, y,
				periodicRectangleSize, initial, final);
		comparePeriodicDiamondAndOthers(nx, ny, x, y, periodicDiamondSize,
				initial, final);
		//std::cout << solver.latticeProperties << std::endl;
		/*
		 printBigBondArray(bonds,sx-maxDist,sy-maxDist,sx+maxDist,sy+maxDist);
		 std::cout << std::endl;
		 std::cout << solver.latticeProperties;
		 std::cout << std::endl;
		 */
		createInputStates(initial, final);
		/*
		 std::cout << std::endl;
		 std::cout << solver.inputStates.right;
		 std::cout << std::endl;
		 std::cout << solver.inputStates.left;
		 */
	} else {
		std::cout << std::endl;
		std::cout << "Nicht alle Faelle bedacht!" << std::endl;
		std::cout << std::endl;
		return;
	}
	return;
}

void ClusterInFieldDispersion::initializeCompleteHoppingList() {

	std::vector<unsigned int> currDist;
	for (unsigned int y = 0; y <= solver.globalConfiguration.order - 2; y++) {
		for (unsigned int x = y; x < solver.globalConfiguration.order; x++) {
			if ((y + x) <= 2 * solver.globalConfiguration.order) {
				if (((x == 0) && (y == 0)) && (!configuration.startWith1QP)) {
					/** two [0,0] elements needed for 0qp groundstate fluctuations and 1 qp local hopping */
					currDist.push_back(x);
					currDist.push_back(y);
					configuration.dist.push_back(currDist);
					currDist.clear();

				}
				currDist.push_back(x);
				currDist.push_back(y);
				configuration.dist.push_back(currDist);
				currDist.clear();
			}
		}
	}

	/*
	 std::list<std::vector<unsigned int>>::iterator it;
	 std::vector<unsigned int>::iterator itv;

	 for (it=dist.begin(); it!=dist.end(); it++){
	 for ( itv=it->begin() ; itv < it->end(); itv++ ){
	 std::cout << " " << *itv;
	 }
	 std::cout << std::endl;
	 }
	 */

}
