/*
 * RotatedClusterDispersion.h
 *
 *  Created on: 26.10.2010
 *      Author: kalis
 *
 *
 */

#ifndef ROTATEDCLUSTERDISPERSION_H_
#define ROTATEDCLUSTERDISPERSION_H_

#include <Modules/LatticeProvider/LatticeProvider.h>
#include <Utilities/ParameterSet.h>
#include <Interface/Interface.h>
#include <Interface/Bond.h>
#include <list>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>

/** Lattice Provider for Rotated Cluster in Field
 *  */

class RotatedClusterDispersion: public LatticeProvider {

public:

	/** The constructor should initialize all data needed for the implementation.*/
	RotatedClusterDispersion();

	/** In the module implementations this method needs to be overloaded.*/
	bool execute();

	/** functions for manipulating hopping list e.g. for calculating just one hopping */
	void initializeCompleteHoppingList();
	void clearHoppingList();
	void addElementToHoppinglist(unsigned int dx, unsigned int dy);
	void removeElementFromHoppinglist(unsigned int dx, unsigned int dy);

	void createOptimzedBondfile(std::list<std::vector<unsigned int>>::iterator it_erase);

	void markBond(unsigned int bonds[300][300], unsigned int posx, unsigned int posy);
	void unMarkBond(unsigned int bonds[300][300], unsigned int posx, unsigned int posy);

	/** Type One means all distance = uneven (from start and final bond)*/
	/*
	 * **************************************
	 * x	   	x			x			x	*
	 * 	   x		  x			  x			*
	 * x	   	x			x			x	*
	 * 	   x		  x			  x			*
	 * x	   	x			x			x	*
	 *****************************************/
	void markAllTypeOneBonds(unsigned int bonds[300][300], unsigned int sposx, unsigned int sposy, unsigned int fposx, unsigned int fposy);

	/** Type Two means all Bonds in rectangle spanned by start and end Site*/
	/*
	 * **************************************
	 * x   x	x     x		x	  x		x	*
	 * x   x	x     x		x	  x		x	*
	 * x   x	x     x		x	  x		x	*
	 * x   x	x     x		x	  x		x	*
	 * x   x	x     x		x	  x		x	*
	 *****************************************/
	void markAllTypeTwoBonds(unsigned int bonds[300][300], unsigned int sposx, unsigned int sposy, unsigned int fposx, unsigned int fposy);

	/** Type One means all distance = uneven (from start and final bond), returns number of marked Bonds*/
	std::list<std::vector<unsigned int>> markAllTypeOneBondsReturnMarkedBonds(unsigned int bonds[300][300], unsigned int sposx, unsigned int sposy, unsigned int fposx, unsigned int fposy);

	/** Type Two means all Bonds in rectangle spanned by start and end Site, returns number of marked Bonds */
	std::list<std::vector<unsigned int>> markAllTypeTwoBondsReturnMarkedBonds(unsigned int bonds[300][300], unsigned int sposx, unsigned int sposy, unsigned int fposx, unsigned int fposy);
	void markAllBondsWithinDistance(unsigned int bonds[300][300], unsigned int xCoor, unsigned int yCoor, unsigned int dist);

	/** Print Array from Site sposx,sposy to fposx,fposy */
	void printBigBondArray(unsigned int bonds[300][300], unsigned int sposx, unsigned int sposy, unsigned int fposx, unsigned int fposy);

	/** set all bits to 0 */
	void resetBondArray(unsigned int bonds[300][300]);

	/** rename or recount marked bonds */
	std::list<Bond> renameBondsAndStates(unsigned int bonds[300][300], unsigned int sx, unsigned int sy, unsigned int fx, unsigned int fy, BaseState& initial, BaseState& final);

	std::list<Bond> createPeriodicSquareCluster(unsigned int size, unsigned int Dx, unsigned int Dy, BaseState& initial, BaseState& final);
	std::list<Bond> createPeriodicRectangleCluster(unsigned int nx, unsigned int ny, unsigned int Dx, unsigned int Dy, unsigned int periodicRectangleSize, BaseState& initial, BaseState& final);
	std::list<Bond> createPeriodicDiamondCluster(unsigned int Dx, unsigned int Dy, unsigned int periodicDiamondSize, BaseState& initial, BaseState& final);

	/** creating input States for already generated input BaseStates */
	void createInputStates(BaseState& initial, BaseState& final);


	/** function compares bond sizes and insert smaller bond-setting into solver.latticeProperties.bonds */
	void comparePeriodicRectangleAndOpenBoundary(unsigned int nx, unsigned int ny, unsigned int x, unsigned int y, unsigned int periodicRectangleSize, BaseState& initial, BaseState& final);
	void comparePeriodicDiamondAndOthers(unsigned int nx, unsigned int ny, unsigned int x, unsigned int y, unsigned int periodicDiamondSize, BaseState& initial, BaseState& final);

	/** funtction to find neighbours (in absolute numbers eg.: B(150,150) will return 45150, 45149, 45151, 44850, 45450) of Bond B(x,y) in big bond array */
	/** Hint: function will act nasty when you reach boundary */
	std::list<unsigned int> findSitesForBond(unsigned int Bx, unsigned int By);

	std::string convertInt(unsigned int number);

	/** output functions */
	void consoleOutput();
	void fileOutput();
	void consoleOutputWithDispersion();
	void fileOutputWithDispersion();
	void output();


private:

	class Configuration: public Utilities::ParameterSet {

		public:

			void streamOutExample(std::ostream& stream);
			void streamIn(std::istream& stream);

			std::list< std::vector <unsigned int> > dist;
			std::list< std::vector <unsigned int> > dist_copy;

			/** parameters to control output */
			bool startWith1QP;
			bool outputControlQP;
			bool generate_dispersion;
			bool file_out;
			bool console_out;
			bool executed;
			bool fullHopping;

	};

	/** The configuration */
	Configuration configuration;

	std::list<stateAmplitude> results;
	std::list<stateAmplitude> results_copy;




};

#endif /* ROTATEDCLUSTERDISPERSION_H_ */
