/*
 * FermionicHubbardIsing.cpp
 */

#include <Modules/ModelProvider/FermionicHubbardIsing/FermionicHubbardIsing.h>
#include <Interface/BaseStateIO.h>
#include <Interface/ExpandedAmplitudeIO.h>
#include <Framework/Solver.h>

void FermionicHubbardIsing::updateOutputAmplitude(const BaseState & outputState,
		const BaseState & inputState, stateAmplitude & outputAmplitude,
		const stateAmplitude & inputAmplitude, const Bond & bond) {

	//std::cout << "\n\nOutputState: " << outputState << "\n";
	//std::cout << "OutputAmplitude: " << outputAmplitude << "\n";
	//std::cout << "InputState: " << inputState << "\n";
	//std::cout << "InputAmplitude: " << inputAmplitude << "\n";

	unsigned int destroyedBit=0;
	unsigned int createdBit=0;
	amplitude vz(1);

	for(unsigned int i=0;i<solver.modelProperties.bitsPerLatticeSite*solver.latticeProperties.numberOfLatticeSites;i++){
		if(!outputState.get(i) && inputState.get(i)){
			destroyedBit=i;
		}
		else{
			if(outputState.get(i) && !inputState.get(i)){
				createdBit=i;
			}
		}
	}
	//std::cout << "destroyedBit=" << destroyedBit << " createdBit=" << createdBit << "\n";
	for(unsigned int i=0;i<destroyedBit;i++){
		if(inputState.get(i)){
			vz*=-1;
		}
	}
	for(unsigned int i=0;i<createdBit;i++){
		if(outputState.get(i)){
			vz*=-1;
		}
	}
	//std::cout << "vz=" << vz << "\n\n";
	outputAmplitude *= vz;

}

double FermionicHubbardIsing::getH_0Eigenvalue(const BaseState & bs) {

	double eigenValue = 0;

	// Calculate the eigen value of the global base state by summing over the eigen values
	// of the states of all lattice sites.

	for (unsigned int latticeSiteCounter = 0;
			latticeSiteCounter < solver.latticeProperties.numberOfLatticeSites;
			latticeSiteCounter++) {

		unsigned int localBaseState = bs.toInteger(
				latticeSiteCounter
						* latticeModelConfiguration.bitsPerLatticeSite,
				(latticeSiteCounter + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);

		eigenValue += h_0[localBaseState];

	}


	// now calculate the ising-interaction terms in the H0-operator! they should correspond

	double eigenValue2;

	for (std::list<Bond>::iterator bondIt = solver.latticeProperties.bonds.begin();
		     bondIt != solver.latticeProperties.bonds.end(); bondIt++) {

		  if (bondIt->operatorNumber==2){

		    unsigned int sitesStart = *(bondIt->sites.begin());
		    unsigned int sitesEnd = *(bondIt->sites.rbegin());

		    unsigned int localBaseState1 = bs.toInteger(
					sitesStart
							* latticeModelConfiguration.bitsPerLatticeSite,
					(sitesStart + 1)
							* latticeModelConfiguration.bitsPerLatticeSite);

		    unsigned int localBaseState2 = bs.toInteger(
					sitesEnd
							* latticeModelConfiguration.bitsPerLatticeSite,
					(sitesEnd + 1)
							* latticeModelConfiguration.bitsPerLatticeSite);


		    //Dies entspricht gerade |10 10> !!!
		    if ((localBaseState1==1)&&(localBaseState2==1)) {

		      eigenValue2 += 1.0;

		    }

		  }
		}

	// This is a numerical trick.
	// This command sets the 16 least significant bits of the calculated eigenvalue to zero,
	// which effectively reduces the precision of the double value.
	// This is necessary, as it is possible, that base states of the same energy level
	// calculates to slightly different energy values for numerical reasons.
	// The truncation of the precision ensures, that the energies still can be compared with
	// the == operator for example by the P operator methods.

	unsigned char* pointer = (unsigned char*) &eigenValue;
	pointer[0] = 0;
	//pointer[1] = 0;


	eigenValue+=eigenValue2;

	return eigenValue;

}


double Shastry::getH_0Eigenvalue(const BaseState & bs) {

	double eigenValue = 0;

	// Calculate the eigen value of the global base state by summing over the eigen values
	// of the states of all lattice sites.

	for (unsigned int latticeSiteCounter = 0;
			latticeSiteCounter < solver.latticeProperties.numberOfLatticeSites;
			latticeSiteCounter++) {

		unsigned int localBaseState = bs.toInteger(
				latticeSiteCounter
						* latticeModelConfiguration.bitsPerLatticeSite,
				(latticeSiteCounter + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);

		eigenValue += h_0[localBaseState];

		//std::cout<<localBaseState<<" "<<h_0[localBaseState]<<" "<<latticeSiteCounter<<std::endl;

	}
	//std::cout<<std::endl;

	double eigenValue2 = 0;

	// Calculate first the additional part from n_i*n_j
	// Start with Loop over bonds

	for (std::list<Bond>::iterator bondIt = solver.latticeProperties.bonds.begin();
	     bondIt != solver.latticeProperties.bonds.end(); bondIt++) {

	  if (bondIt->operatorNumber==2){

	    unsigned int sitesStart = *(bondIt->sites.begin());
	    unsigned int sitesEnd = *(bondIt->sites.rbegin());

	    unsigned int localBaseState1 = bs.toInteger(
				sitesStart
						* latticeModelConfiguration.bitsPerLatticeSite,
				(sitesStart + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);

	    unsigned int localBaseState2 = bs.toInteger(
				sitesEnd
						* latticeModelConfiguration.bitsPerLatticeSite,
				(sitesEnd + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);


	    //Dies entspricht gerade |10 10> !!!
	    if ((localBaseState1==1)&&(localBaseState2==1)) {

	      eigenValue2 += 1.0;

	    }

	  }
	}

	// This is a numerical trick.
	// This command sets the 128 least significant bits of the calculated eigenvalue to zero,
	// which effectively reduces the precision of the double value.
	// This is necessary, as it is possible, that base states of the same energy level
	// calculates to slightly different energy values for numerical reasons.
	// The truncation of the precision ensures, that the energies still can be compared with
	// the == operator for example by the P operator methods.

	unsigned char* pointer = (unsigned char*) &eigenValue;
	pointer[0] = 0;
	//pointer[1] = 0;

	eigenValue+=eigenValue2;

	return eigenValue;

}
