/*
 * Shastry.cpp
 *
 *  Created on: 31.01.2011
 *      Author: klagges
 */

#include <Modules/ModelProvider/HubbardIsing/HubbardIsing.h>
#include <Framework/Solver.h>
#include <Utilities/Utilities.h>
#include <Utilities/ComplexConjugate.h>
#include <sstream>

HubbardIsing::HubbardIsing() {

	// Read in the name of the sub folder to use
	latticeModelFolder.readFromConfigFile("LatticeModel.cfg");

	// and load the configuration from this sub folder.
	latticeModelConfiguration.readFromConfigFile(
			latticeModelFolder.value + "/LatticeModelConfiguration.cfg");

	// Set some of the Solver::modelProperties.
	solver.modelProperties.Nmax = latticeModelConfiguration.Nmax;
	solver.modelProperties.numberOfExpansionParameters =
			latticeModelConfiguration.numberOfExpansionParameters;
	solver.modelProperties.bitsPerLatticeSite =
			latticeModelConfiguration.bitsPerLatticeSite;

	// And the local operators
	h_0.readFromConfigFile(
			latticeModelFolder.value + "/"
					+ latticeModelConfiguration.operatorsSubFolderName
					+ "h_0.op");

	// Calculate the maximal distance between two neighboring h_0 energy levels.
       
	solver.modelProperties.energyDelta = h_0.getMaximalDelta();

	t.resize(latticeModelConfiguration.numberOfPerturbationOperators);
	v.resize(latticeModelConfiguration.numberOfPerturbationOperators);

	for (unsigned int j = 0;
			j < latticeModelConfiguration.numberOfPerturbationOperators; j++) {

		t[j].resize(latticeModelConfiguration.Nmax * 2 + 1);

		for (int i = -latticeModelConfiguration.Nmax;
				i <= latticeModelConfiguration.Nmax; i++) {

			std::stringstream filename;
			filename << latticeModelFolder.value << "/"
					<< latticeModelConfiguration.operatorsSubFolderName
					<< "/t^(" << j << ")_" << i << ".op";

			t[j][i].readFromConfigFile(filename.str());

			v[j] += t[j][i];

		}

	}

	pthread_mutex_init(&bondFileMutex, NULL);
	pthread_mutex_init(&outputMutex, NULL);

}

// This routine has to be changed!
// One has to add the "QPs" coming from the n_i n_j 

double HubbardIsing::getH_0Eigenvalue(const BaseState & bs) {

	double eigenValue = 0;

	// Calculate the eigen value of the global base state by summing over the eigen values
	// of the states of all lattice sites.

	for (unsigned int latticeSiteCounter = 0;
			latticeSiteCounter < solver.latticeProperties.numberOfLatticeSites;
			latticeSiteCounter++) {

		unsigned int localBaseState = bs.toInteger(
				latticeSiteCounter
						* latticeModelConfiguration.bitsPerLatticeSite,
				(latticeSiteCounter + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);

		eigenValue += h_0[localBaseState];

		//std::cout<<localBaseState<<" "<<h_0[localBaseState]<<" "<<latticeSiteCounter<<std::endl;

	}
	//std::cout<<std::endl;

	double eigenValue2 = 0;

	// Calculate first the additional part from n_i*n_j
	// Start with Loop over bonds

	for (std::list<Bond>::iterator bondIt = solver.latticeProperties.bonds.begin();
	     bondIt != solver.latticeProperties.bonds.end(); bondIt++) {

	  if (bondIt->operatorNumber==2){

	    unsigned int sitesStart = *(bondIt->sites.begin());
	    unsigned int sitesEnd = *(bondIt->sites.rbegin());

	    unsigned int localBaseState1 = bs.toInteger(
				sitesStart
						* latticeModelConfiguration.bitsPerLatticeSite,
				(sitesStart + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);

	    unsigned int localBaseState2 = bs.toInteger(
				sitesEnd
						* latticeModelConfiguration.bitsPerLatticeSite,
				(sitesEnd + 1)
						* latticeModelConfiguration.bitsPerLatticeSite);


	    //Dies entspricht gerade |10 10> !!!
	    if ((localBaseState1==1)&&(localBaseState2==1)) {
	     
	      eigenValue2 += 1.0;
	  
	    }

	  }
	}

	// This is a numerical trick.
	// This command sets the 128 least significant bits of the calculated eigenvalue to zero,
	// which effectively reduces the precision of the double value.
	// This is necessary, as it is possible, that base states of the same energy level
	// calculates to slightly different energy values for numerical reasons.
	// The truncation of the precision ensures, that the energies still can be compared with
	// the == operator for example by the P operator methods.

	unsigned char* pointer = (unsigned char*) &eigenValue;
	pointer[0] = 0;
	//pointer[1] = 0;

	eigenValue+=eigenValue2;

	return eigenValue;

}

void HubbardIsing::applyT(const State& source, State& result, int c) {

	// Prepare the calculation, by copying all data needed by the threads into
	// class member variables.

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyT: Bonds: ";

	bondFileIterator = solver.latticeProperties.bonds.begin();

	sourceState = &source;
	resultingState = &result;

	this->c = c;

	std::vector<pthread_t> threads(solver.globalConfiguration.numberOfThreads);

	// Create the requested number of threads.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_create(&(*threadIt), NULL, &threadStarterApplyT, this);

	}

	// Wait for all threads to complete.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_join(*threadIt, NULL);

	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

}

void HubbardIsing::applyT(const State& source, const State& target,
		stateAmplitude& result, int c) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyT(direct): Bonds: ";

	// Prepare the calculation, by copying all data needed by the threads into
	// class member variables.

	bondFileIterator = solver.latticeProperties.bonds.begin();

	sourceState = &source;
	targetState = &target;
	resultingAmplitude = &result;

	*resultingAmplitude = stateAmplitude();

	this->c = c;

	std::vector<pthread_t> threads(solver.globalConfiguration.numberOfThreads);

	// Create the requested number of threads.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_create(&(*threadIt), NULL, &threadStarterApplyTDirect, this);

	}

	// Wait for all threads to complete.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_join(*threadIt, NULL);

	}

	// We calculated the scalar product assuming that the source state
	// is a bra vector. If it is not, we have to conjugate the result.
	if (sourceState->type == State::ket)
		Utilities::ComplexConjugate(result);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

}

void HubbardIsing::applyV(const State& source, State& result,
		unsigned int remainingV) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyV: Bonds:";

	// Prepare the calculation, by copying all data needed by the threads into
	// class member variables.

	/* Daniels version
	min = solver.globalProperties.eigenEnergy
			- (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax * 1.00001);
	max = solver.globalProperties.eigenEnergy
			+ (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax * 1.00001);
	*/

	//In our case: energyDelta=1, solver.modelProperties.Nmax has to be 4!
	min = solver.globalProperties.eigenEnergy- (remainingV * 28 * 1.00001);
	max = solver.globalProperties.eigenEnergy+ (remainingV * 28 * 1.00001);


	//min=-4000.00001;
	//max=4000.00001;

	bondFileIterator = solver.latticeProperties.bonds.begin();

	sourceState = &source;
	resultingState = &result;

	std::vector<pthread_t> threads(solver.globalConfiguration.numberOfThreads);

	// Create the requested number of threads.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_create(&(*threadIt), NULL, &threadStarterApplyV, this);

	}

	// Wait for all threads to complete.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_join(*threadIt, NULL);

	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

}

void HubbardIsing::applyV(const State& source, const State& target,
		stateAmplitude& result, unsigned int remainingV) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyV(direct): Bonds: ";

	// Prepare the calculation, by copying all data needed by the threads into
	// class member variables.

	bondFileIterator = solver.latticeProperties.bonds.begin();

	sourceState = &source;
	targetState = &target;
	resultingAmplitude = &result;

	*resultingAmplitude = stateAmplitude();

	std::vector<pthread_t> threads(solver.globalConfiguration.numberOfThreads);

	// Create the requested number of threads.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_create(&(*threadIt), NULL, &threadStarterApplyVDirect, this);

	}

	// Wait for all threads to complete.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_join(*threadIt, NULL);

	}

	// We calculated the scalar product assuming that the source state
	// is a bra vector. If it is not, we have to conjugate the result.
	if (sourceState->type == State::ket)
		Utilities::ComplexConjugate(result);

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

}

void HubbardIsing::applySSV(const State& source, State& result,
		unsigned int remainingV, unsigned int numberOfS) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applySSV: Bonds: ";

	// Prepare the calculation, by copying all data needed by the threads into
	// class member variables.

	/* Daniels version!

	min = solver.globalProperties.eigenEnergy
	  - (remainingV * solver.modelProperties.energyDelta
	     * solver.modelProperties.Nmax * 1.00001);
	
	max = solver.globalProperties.eigenEnergy
	  + (remainingV * solver.modelProperties.energyDelta
	     * solver.modelProperties.Nmax * 1.00001);
        */
	

	// In our case: energyDelta=1, solver.modelProperties.Nmax has to be 4!
	//min = solver.globalProperties.eigenEnergy
	//- (remainingV * 4 * 1.00001);
			//max = solver.globalProperties.eigenEnergy
	//		+ (remainingV * 4 * 1.00001);

	//std::cout<<"Here "<<min<<" "<<max<<std::endl;

	min = solver.globalProperties.eigenEnergy- (remainingV * 28 * 1.00001);
	max = solver.globalProperties.eigenEnergy+ (remainingV * 28 * 1.00001);

	//std::cout<<"Here "<<min<<" "<<max<<std::endl;


	//min = -4000.00001;
	//max = 4000.00001;


	this->numberOfS = numberOfS;

	bondFileIterator = solver.latticeProperties.bonds.begin();

	sourceState = &source;
	resultingState = &result;

	std::vector<pthread_t> threads(solver.globalConfiguration.numberOfThreads);

	// Create the requested number of threads.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_create(&(*threadIt), NULL, &threadStarterApplySSV, this);

	}

	// Wait for all threads to complete.
	for (std::vector<pthread_t>::iterator threadIt = threads.begin();
			threadIt != threads.end(); threadIt++) {

		pthread_join(*threadIt, NULL);

	}

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "\n";

}

inline unsigned int HubbardIsing::copyGlobalToLocal(
		const BaseState & global, const std::list<unsigned int> & sites) {

	unsigned int local = 0;
	unsigned int mask = 1;
	for (std::list<unsigned int>::const_iterator latticeSiteIterator =
			sites.begin(); latticeSiteIterator != sites.end();
			latticeSiteIterator++) {
		for (unsigned int bitCounter = 0;
				bitCounter < latticeModelConfiguration.bitsPerLatticeSite;
				bitCounter++) {

			if (global.get(
					*latticeSiteIterator
							* latticeModelConfiguration.bitsPerLatticeSite
							+ bitCounter))
				local |= mask;
			mask <<= 1;

		}

	}

	return local;

}

inline void HubbardIsing::copyLocalToGlobal(const std::list<bool>& local,
		BaseState & global, const std::list<unsigned int> & sites) {

	std::list<bool>::const_iterator localIterator = local.begin();
	for (std::list<unsigned int>::const_iterator latticeSiteIterator =
			sites.begin(); latticeSiteIterator != sites.end();
			latticeSiteIterator++) {
		for (unsigned int bitCounter = 0;
				bitCounter < latticeModelConfiguration.bitsPerLatticeSite;
				bitCounter++) {

			global.setTo(
					*latticeSiteIterator
							* latticeModelConfiguration.bitsPerLatticeSite
							+ bitCounter, *localIterator);
			localIterator++;

		}
	}
}

void *HubbardIsing::threadStarterApplyT(void *obj) {

	return ((HubbardIsing*) obj)->threadMethodApplyT();

}

void *HubbardIsing::threadMethodApplyT() {

	while (true) {

		// Request a new bond to process.
		pthread_mutex_lock(&bondFileMutex);

		if (bondFileIterator == solver.latticeProperties.bonds.end()) {
			// No bonds left. Quit.
			pthread_mutex_unlock(&bondFileMutex);
			break;

		}

		std::list<Bond>::const_iterator bond = bondFileIterator;

		bondFileIterator++;

		pthread_mutex_unlock(&bondFileMutex);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "O";

		// The requested t part of the perturbation operator of this bond
		// Does not have a contribution at all and we can continue with the
		// next bond.
		if (t[bond->operatorNumber][c].size() == 0)
			continue;

		for (State::const_iterator sourceStateIterator = sourceState->begin();
				sourceStateIterator != sourceState->end();
				sourceStateIterator++) {
			// We iterate over all input base states

			// Check, if this input base state should be processed with this bond.
			if (acceptInputState(sourceStateIterator->first,
					sourceStateIterator->second, *bond)) {

				unsigned int localBaseState = copyGlobalToLocal(
						sourceStateIterator->first, bond->sites);

				// Now we apply the local t^(i) operator (with the requested i according to
				// the bond file entry) to the local input state.
				const std::list<std::pair<std::list<bool>, stateAmplitude> > & localResult =
						t[bond->operatorNumber][c][localBaseState];

				if (!localResult.empty()) {

					// Will be used to store the global base states resulting from the local operation.
					BaseState globalBaseState(sourceStateIterator->first);

					for (std::list<std::pair<std::list<bool>, stateAmplitude> >::const_iterator localResultBaseStateIt =
							localResult.begin();
							localResultBaseStateIt != localResult.end();
							localResultBaseStateIt++) {

						// We iterate over all resulting local base states

						copyLocalToGlobal(localResultBaseStateIt->first,
								globalBaseState, bond->sites);

						// Update the extra information and check if this state should be stored.
						if (updateExtraInformation(globalBaseState,
								sourceStateIterator->first,
								sourceStateIterator->second, *bond)) {

							stateAmplitude calculatedCoefficient =
									localResultBaseStateIt->second
											* sourceStateIterator->second;

							updateOutputAmplitude(globalBaseState,
									sourceStateIterator->first,
									calculatedCoefficient,
									sourceStateIterator->second, *bond);

							// and add the resulting global base state with its according amplitude to the result.

							pthread_mutex_lock(&outputMutex);

							resultingState->add(globalBaseState,
									calculatedCoefficient);

							pthread_mutex_unlock(&outputMutex);

						}

					}

				}

			}

		}

	}

	return 0;

}

void *HubbardIsing::threadStarterApplyTDirect(void *obj) {

	return ((HubbardIsing*) obj)->threadMethodApplyTDirect();

}

void *HubbardIsing::threadMethodApplyTDirect() {

	while (true) {

		// Request a new bond to process.
		pthread_mutex_lock(&bondFileMutex);

		if (bondFileIterator == solver.latticeProperties.bonds.end()) {
			// No bonds left. Quit.
			pthread_mutex_unlock(&bondFileMutex);
			break;

		}

		std::list<Bond>::const_iterator bond = bondFileIterator;

		bondFileIterator++;

		pthread_mutex_unlock(&bondFileMutex);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "O";

		// The requested t part of the perturbation operator of this bond
		// Does not have a contribution at all and we can continue with the
		// next bond.
		if (t[bond->operatorNumber][c].size() == 0)
			continue;

		for (State::const_iterator sourceStateIterator = sourceState->begin();
				sourceStateIterator != sourceState->end();
				sourceStateIterator++) {
			// We iterate over all input base states

			// Check, if this input base state should be processed with this bond.
			if (acceptInputState(sourceStateIterator->first,
					sourceStateIterator->second, *bond)) {

				unsigned int localBaseState = copyGlobalToLocal(
						sourceStateIterator->first, bond->sites);

				// Now we apply the local t^(i) operator (with the requested i according to
				// the bond file entry) to the local input state.
				const std::list<std::pair<std::list<bool>, stateAmplitude> > & localResult =
						t[bond->operatorNumber][c][localBaseState];

				if (!localResult.empty()) {

					// Will be used to store the global base states resulting from the local operation.
					BaseState globalBaseState(sourceStateIterator->first);

					for (std::list<std::pair<std::list<bool>, stateAmplitude> >::const_iterator localResultBaseStateIt =
							localResult.begin();
							localResultBaseStateIt != localResult.end();
							localResultBaseStateIt++) {

						// We iterate over all resulting local base states.

						copyLocalToGlobal(localResultBaseStateIt->first,
								globalBaseState, bond->sites);

						// Update the extra information and check if this state should not be skipped.
						if (updateExtraInformation(globalBaseState,
								sourceStateIterator->first,
								sourceStateIterator->second, *bond)) {

							// Check, if the resulting base state is part of the target state.
							State::const_iterator targetStateIt =
									targetState->find(globalBaseState);

							if (targetStateIt != targetState->end()) {

								// If so, update the output amplitude

								stateAmplitude calculatedCoefficient =
										(localResultBaseStateIt->second
												* sourceStateIterator->second);

								updateOutputAmplitude(globalBaseState,
										sourceStateIterator->first,
										calculatedCoefficient,
										sourceStateIterator->second, *bond);

								// Conjugate the calculated coefficient to properly
								// multiply it with the target state.
								// (Assuming, that the source state is a bra vector.)

								Utilities::ComplexConjugate(
										calculatedCoefficient);

								// Calculate the product

								calculatedCoefficient *= targetStateIt->second;

								// and add it to the result.

								pthread_mutex_lock(&outputMutex);

								*resultingAmplitude += calculatedCoefficient;

								pthread_mutex_unlock(&outputMutex);

							}
						}

					}
				}

			}

		}

	}

	return 0;
}

void *HubbardIsing::threadStarterApplyV(void *obj) {

	return ((HubbardIsing*) obj)->threadMethodApplyV();

}

void *HubbardIsing::threadMethodApplyV() {

	while (true) {

		// Request a new bond to process.
		pthread_mutex_lock(&bondFileMutex);

		if (bondFileIterator == solver.latticeProperties.bonds.end()) {
			// No bonds left. Quit.
			pthread_mutex_unlock(&bondFileMutex);
			break;

		}

		std::list<Bond>::const_iterator bond = bondFileIterator;

		bondFileIterator++;

		pthread_mutex_unlock(&bondFileMutex);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "O";

		for (State::const_iterator sourceStateIterator = sourceState->begin();
				sourceStateIterator != sourceState->end();
				sourceStateIterator++) {
			// We iterate over all input base states

			// Check, if this input base state should be processed with this bond.
			if (acceptInputState(sourceStateIterator->first,
					sourceStateIterator->second, *bond)) {

				unsigned int localBaseState = copyGlobalToLocal(
						sourceStateIterator->first, bond->sites);

				// Now we apply the local v^(i) operator (with the requested i according to
				// the bond file entry) to the local input state.
				const std::list<std::pair<std::list<bool>, stateAmplitude> > & localResult =
						v[bond->operatorNumber][localBaseState];

				if (!localResult.empty()) {

					// Will be used to store the global base states resulting from the local operation.
					BaseState globalBaseState(sourceStateIterator->first);

					for (std::list<std::pair<std::list<bool>, stateAmplitude> >::const_iterator localResultBaseStateIt =
							localResult.begin();
							localResultBaseStateIt != localResult.end();
							localResultBaseStateIt++) {

						// We iterate over all resulting local base states.

						copyLocalToGlobal(localResultBaseStateIt->first,
								globalBaseState, bond->sites);

						// Update the extra information and check if this state should be stored.
						if (updateExtraInformation(globalBaseState,
								sourceStateIterator->first,
								sourceStateIterator->second, *bond)) {

							double eigenEnergy = getH_0Eigenvalue(
									globalBaseState);


							if (eigenEnergy <= max && eigenEnergy >= min) {

								stateAmplitude calculatedCoefficient =
										localResultBaseStateIt->second
												* sourceStateIterator->second;

								updateOutputAmplitude(globalBaseState,
										sourceStateIterator->first,
										calculatedCoefficient,
										sourceStateIterator->second, *bond);

								pthread_mutex_lock(&outputMutex);

								resultingState->add(globalBaseState,
										calculatedCoefficient);

								pthread_mutex_unlock(&outputMutex);

							}
						}
					}
				}

			}

		}

	}

	return 0;
}

void *HubbardIsing::threadStarterApplyVDirect(void *obj) {

	return ((HubbardIsing*) obj)->threadMethodApplyVDirect();

}

void *HubbardIsing::threadMethodApplyVDirect() {

	while (true) {

		// Request a new bond to process.
		pthread_mutex_lock(&bondFileMutex);

		if (bondFileIterator == solver.latticeProperties.bonds.end()) {
			// No bonds left. Quit.
			pthread_mutex_unlock(&bondFileMutex);
			break;

		}

		std::list<Bond>::const_iterator bond = bondFileIterator;

		bondFileIterator++;

		pthread_mutex_unlock(&bondFileMutex);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "O";

		for (State::const_iterator sourceStateIterator = sourceState->begin();
				sourceStateIterator != sourceState->end();
				sourceStateIterator++) {
			// We iterate over all input base states

			// Check, if this input base state should be processed with this bond.
			if (acceptInputState(sourceStateIterator->first,
					sourceStateIterator->second, *bond)) {

				unsigned int localBaseState = copyGlobalToLocal(
						sourceStateIterator->first, bond->sites);

				// Now we apply the local v^(i) operator (with the requested i according to
				// the bond file entry) to the local input state.
				const std::list<std::pair<std::list<bool>, stateAmplitude> > & localResult =
						v[bond->operatorNumber][localBaseState];

				if (!localResult.empty()) {

					// Will be used to store the global base states resulting from the local operation.
					BaseState globalBaseState(sourceStateIterator->first);

					for (std::list<std::pair<std::list<bool>, stateAmplitude> >::const_iterator localResultBaseStateIt =
							localResult.begin();
							localResultBaseStateIt != localResult.end();
							localResultBaseStateIt++) {

						// We iterate over all resulting local base states.

						copyLocalToGlobal(localResultBaseStateIt->first,
								globalBaseState, bond->sites);

						// Update the extra information and check if this state should not be skipped.
						if (updateExtraInformation(globalBaseState,
								sourceStateIterator->first,
								sourceStateIterator->second, *bond)) {

							// Check, if the resulting base state is part of the target state.
							State::const_iterator targetStateIt =
									targetState->find(globalBaseState);

							if (targetStateIt != targetState->end()) {

								// If so, update the output amplitude

								stateAmplitude calculatedCoefficient =
										(localResultBaseStateIt->second
												* sourceStateIterator->second);

								updateOutputAmplitude(globalBaseState,
										sourceStateIterator->first,
										calculatedCoefficient,
										sourceStateIterator->second, *bond);

								// Conjugate the calculated coefficient to properly
								// multiply it with the target state.
								// (Assuming, that the source state is a bra vector.)

								Utilities::ComplexConjugate(
										calculatedCoefficient);

								// Calculate the product

								calculatedCoefficient *= targetStateIt->second;

								// and add it to the result.

								pthread_mutex_lock(&outputMutex);

								*resultingAmplitude += calculatedCoefficient;

								pthread_mutex_unlock(&outputMutex);

							}
						}

					}
				}

			}

		}

	}

	return 0;
}

void *HubbardIsing::threadStarterApplySSV(void *obj) {

	return ((HubbardIsing*) obj)->threadMethodApplySSV();
}

void *HubbardIsing::threadMethodApplySSV() {

	bool applyP;
	if (min == max)
		applyP = true;
	else
		applyP = false;

	while (true) {

		// Request a new bond to process.
		pthread_mutex_lock(&bondFileMutex);

		if (bondFileIterator == solver.latticeProperties.bonds.end()) {
			// No bonds left. Quit.
			pthread_mutex_unlock(&bondFileMutex);
			break;

		}

		std::list<Bond>::const_iterator bond = bondFileIterator;

		bondFileIterator++;

		pthread_mutex_unlock(&bondFileMutex);

		if (solver.globalConfiguration.verbosityLevel > 2)
			std::cerr << "O";

		for (State::const_iterator sourceStateIterator = sourceState->begin();
				sourceStateIterator != sourceState->end();
				sourceStateIterator++) {
			// We iterate over all input base states

			// Check, if this input base state should be processed with this bond.
			if (acceptInputState(sourceStateIterator->first,
					sourceStateIterator->second, *bond)) {

				unsigned int localBaseState = copyGlobalToLocal(
						sourceStateIterator->first, bond->sites);

				// Now we apply the local v^(i) operator (with the requested i according to
				// the bond file entry) to the local input state.
				const std::list<std::pair<std::list<bool>, stateAmplitude> > & localResult =
						v[bond->operatorNumber][localBaseState];

				if (!localResult.empty()) {

					// Will be used to store the global base states resulting from the local operation.
					BaseState globalBaseState = sourceStateIterator->first;

					for (std::list<std::pair<std::list<bool>, stateAmplitude> >::const_iterator localResultBaseStateIt =
							localResult.begin();
							localResultBaseStateIt != localResult.end();
							localResultBaseStateIt++) {

						// We iterate over all resulting local base states

						copyLocalToGlobal(localResultBaseStateIt->first,
								globalBaseState, bond->sites);

						// Update the extra information and check if this state should be stored.
						if (updateExtraInformation(globalBaseState,
								sourceStateIterator->first,
								sourceStateIterator->second, *bond)) {

							double eigenEnergy = getH_0Eigenvalue(
									globalBaseState);

							if (applyP
									&& eigenEnergy
											== solver.globalProperties.eigenEnergy) {

								// If we should apply a P operator and the resulting state is in
								// the eigen space in question, we add it to the result.

								stateAmplitude calculatedCoefficient =
										localResultBaseStateIt->second
												* sourceStateIterator->second;

								updateOutputAmplitude(globalBaseState,
										sourceStateIterator->first,
										calculatedCoefficient,
										sourceStateIterator->second, *bond);

								pthread_mutex_lock(&outputMutex);

								resultingState->add(globalBaseState,
										calculatedCoefficient);

								pthread_mutex_unlock(&outputMutex);

							} else {
							  
								if (eigenEnergy
										!= solver.globalProperties.eigenEnergy
										&& eigenEnergy <= max
										&& eigenEnergy >= min) {

									// One or more S operators should be applied.

									stateAmplitude calculatedCoefficient =
											localResultBaseStateIt->second
													* sourceStateIterator->second;

									updateOutputAmplitude(globalBaseState,
											sourceStateIterator->first,
											calculatedCoefficient,
											sourceStateIterator->second, *bond);

									// We calculate the S operator contribution

									amplitude correctionFactor = amplitude(1);
									correctionFactor /=
											solver.globalProperties.eigenEnergy
													- eigenEnergy;

									amplitude correction = Utilities::pow(
											correctionFactor, numberOfS);

									calculatedCoefficient *= correction;

									// and add it to the result.

									pthread_mutex_lock(&outputMutex);

									resultingState->add(globalBaseState,
											calculatedCoefficient);

									pthread_mutex_unlock(&outputMutex);

								}

							}

						}

					}

				}

			}

		}

	}

	return 0;
}


bool HubbardIsing::acceptInputState(const BaseState& inputState,
		const stateAmplitude& inputAmplitude, const Bond& bond) {

	// The default implementation of this method is to accept all input states.
	return true;

}




bool HubbardIsing::updateExtraInformation(BaseState& outputState,
		const BaseState& inputState, const stateAmplitude& inputAmplitude,
		const Bond& bond) {

	// The default implementation of this method is to ignore the extra information
	// and to skip no output states.
	return true;
}





void HubbardIsing::updateOutputAmplitude(const BaseState& outputState,
		const BaseState& inputState, stateAmplitude& outputAmplitude,
		const stateAmplitude& inputAmplitude, const Bond& bond) {

	// The default implementation of this method is to keep
	// the outputAmplitude unchanged.
}

