/*
 * LWLowField.h
 *
 *  Created on: 31.01.2011
 *      Author: klagges
 *
 *
 * How to create a configuration for a lattice model:
 *
 * - Make sure the LWLowField is selected in the /Config/Modules.cfg.
 *
 * - Create a new sub folder for your configuration for example /Config/LatticeModels/MyModel
 *
 * - Enter this folder into the /Config/LatticeModel.cfg file.
 *
 * - Execute the solver. You will find an example LatticeModelConfiguration.cfg file in
 *   your sub folder.
 *
 * - Enter the properties of your model into the LatticeModelConfiguration.cfg file.
 *
 * - Execute the solver again. You will find example files for all local operators
 *   and the bond file.
 *
 * - Enter all necessary informations into the bond file.
 *
 * - Execute the solver again. You will find example files for your input states in your folder.
 *
 * - For an example how to fill the configuration and operator files, see
 *   /Config/LatticeModels/ExampleLatticeModel
 *
 * - There is a maple document to automatically generate the t operator files from maple matrices
 *   in Trunk/Utilities/TOperatorGenerator.mw.
 *
 */

#ifndef LWLOWFIELD_H_
#define LWLOWFIELD_H_

#include <Modules/ModelProvider/ModelProvider.h>
#include <Interface/LatticeProperties.h>
#include <Utilities/DiagonalOperator.h>
#include <Utilities/OperatorVector.h>
#include <Utilities/StreamableBaseType.h>
#include <pthread.h>

/** A LWLowField is a ModelProvider representing a model living on a lattice \f$ L \f$.
 *
 * Living on a lattice means, that the Hilbert space of the system can be represented
 * as the direct product of a number of equal subspaces \f$ l \f$ (the lattice sites).
 *
 * On this lattice, the unperturbed Hamiltonian \f$ H_0 \f$ can be expressed as the sum
 * of operators each acting equally and independently on every lattice site.
 *
 * \f[ H_0 = \sum_{l\in L} h_l\f]
 *
 * The base states to represent a lattice site must be the eigen states of the local
 * Hamiltonians (\f$ h \f$ is diagonal in this basis). It is given explicitly as a
 * Utilities::DiagonalOperator.
 *
 * The perturbation operators \f$ V_i \f$ can be expressed locally, as a sum
 * of local perturbation operators each acting equally and independently on finite
 * sets of lattice sites (on bonds).
 *
 * \f[ V_i = \sum_{k\subset L,k=bond} v_{ik}\f]
 *
 * For every local \f$ v_i \f$ operator, let the \f$ t_c^{(i)} \f$ operators be
 * such, that:
 *
 * \f[ v_i = t_0^{(i)} + t_1^{(i)} + t_{-1}^{(i)} + ... + t_{N_{max}}^{(i)} + t_{-N_{max}}^{(i)} \f]
 *
 * and \f$ t_c^{(i)} \f$ transfers a state in the i-th eigen state of \f$ H_0 \f$ to a state in the (i+c)-th
 * eigen space of \f$ H_0 \f$.
 *
 * The \f$ t_c^{(i)} \f$ are specified explicitly as Utilities::Operator objects.
 *
 * The lattice sites on which the perturbation operators act
 * are given in a 'bond file'.
 * */
class LWLowField: public ModelProvider {
public:

	/** Load all configurations.*/
	LWLowField();

	void applyT(const State& source, State& result, int c);

	void applyT(const State& source, const State& target,
			stateAmplitude& result, int c);

	double getH_0Eigenvalue(const BaseState& bs);

	void applyV(const State& source, State& result, unsigned int remainingV);

	void applyV(const State& source, const State& target,
			stateAmplitude& result, unsigned int remainingV);

	void applySSV(const State& source, State& result, unsigned int remainingV,
			unsigned int numberOfS);

protected:

	/** Class to store and stream in properties of a lattice. It is streamed in from the
	 * LatticeModelConfiguration.cfg file in the sub folder you specified in the
	 * /Config/LatticeModel.cfg file.*/
	class Configuration: public Utilities::ParameterSet {

	public:

		void streamOutExample(std::ostream& stream);

		void streamIn(std::istream& stream);

		/** The number of bits used to represent the hilbert space of a single lattice site.*/
		unsigned int bitsPerLatticeSite;

		/** The number of expansion parameters. */
		unsigned int numberOfExpansionParameters;

		/** The number of perturbation operators.
		 Can be bigger then the number of expansion parameters if multiple operators are
		 associated with the same expansion parameter.
		 */
		unsigned int numberOfPerturbationOperators;

		/** The smallest Number N, so that all perturbation Operators can be written in the T operator form.

		 That is:
		 \f[
		 V = T_0 + T_1 + T_{-1} + ... + T_N + T_{-N}
		 \f]
		 Where \f$ T_i \f$ is an operator transferring a state in the n-th \f$ H_0 \f$ eigenspace into a
		 state in the (n+i)-th \f$ H_0 \f$ eigenspace.
		 */
		int Nmax;

		/** The sub folder where to find the definitions of the local h_0 and t_i operators. */
		std::string operatorsSubFolderName;

	};

	/** Decide whether the bond should be applied to a given input state or not.
	 * The input state itself, its input amplitude or all extra information
	 * stored in the input state can be used. */
	virtual bool acceptInputState(const BaseState& inputState,
			const stateAmplitude& inputAmplitude, const Bond& bond);

	/** Should update the extra information stored in outputState.
	 * The outputState is the state that results if you apply bond to inputState.
	 * If the outputState was updated the method should return true.
	 * If the outputState should be discarded the method should return false.*/
	virtual bool updateExtraInformation(BaseState& outputState,
			const BaseState& inputState, const stateAmplitude& inputAmplitude,
			const Bond& bond);

	/** Should update the output state amplitude.
	 * The other parameter are the same as with the updateExtraInformation method.*/
	virtual void updateOutputAmplitude(const BaseState& outputState,
			const BaseState& inputState, stateAmplitude& outputAmplitude,
			const stateAmplitude& inputAmplitude, const Bond& bond);

	/** Creates a base state of the Hilbert space of a local perturbation operator from a base
	 * state of the global Hilbert space.
	 *
	 * The base state is created, by copying the states of the listed lattice sites to the
	 * local base state.*/
	inline unsigned int copyGlobalToLocal(const BaseState& global,
			const std::list<unsigned int>& sites);

	/** Creates a base state of the global Hilbert space from a base
	 * state of the Hilbert space of a local perturbation operator.
	 *
	 * The base state is created, by copying the local base state to the listed lattice sites
	 * of the global base state (leaving all other lattice sites untouched).*/
	inline void copyLocalToGlobal(const std::list<bool>& local,
			BaseState& global, const std::list<unsigned int>& sites);

	/** The subfolder to use.*/
	Utilities::StreamableBaseType<std::string> latticeModelFolder;

	/** The lattice model configuration.*/
	Configuration latticeModelConfiguration;

	/** The local \f$ t_c^{(i)} \f$ operators.*/
	std::vector<Utilities::OperatorVector> t;

	/** it is \f$ v_i = t_0^{(i)} + t_1^{(i)} + t_{-1}^{(i)} + ... + t_{N_{max}}^{(i)} + t_{-N_{max}}^{(i)} \f$.*/
	std::vector<Utilities::Operator> v;

	/** The local unperturbed Hamiltonian.*/
	Utilities::DiagonalOperator h_0;

	// This are the method and members needed for the communication between the threads.

	/** Method to start a thread.*/
	static void* threadStarterApplyV(void* obj);
	/** Method called by a thread starter.*/
	void* threadMethodApplyV();

	/** Method to start a thread.*/
	static void* threadStarterApplyVDirect(void* obj);
	/** Method called by a thread starter.*/
	void* threadMethodApplyVDirect();

	/** Method to start a thread.*/
	static void* threadStarterApplyTDirect(void* obj);
	/** Method called by a thread starter.*/
	void* threadMethodApplyTDirect();

	/** Method to start a thread.*/
	static void* threadStarterApplySSV(void* obj);
	/** Method called by a thread starter.*/
	void* threadMethodApplySSV();

	/** Method to start a thread.*/
	static void* threadStarterApplyT(void* obj);
	/** Method called by a thread starter.*/
	void* threadMethodApplyT();

	/** Mutex to coordinate the access of the threads to the bond file.*/
	pthread_mutex_t bondFileMutex;

	/** Mutex to coordinate the access to the output data structure of the threads. */
	pthread_mutex_t outputMutex;

	/** The bond file iterator used by all threads.*/
	std::list<Bond>::const_iterator bondFileIterator;

	/** A pointer to the source state to be used by all threads.*/
	const State* sourceState;

	/** A pointer to the target state to be used by all threads.*/
	const State* targetState;

	/** A pointer to the output state to be used by all threads.*/
	State* resultingState;

	/** A pointer to the output coefficient to be used by all threads.*/
	stateAmplitude* resultingAmplitude;

	/** The minimum energy to truncate the Hilber space.*/
	double min;

	/** The maximum energy to truncate the Hilber space.*/
	double max;

	/** The number of S operators to be applied by all threads.*/
	unsigned int numberOfS;

	/** The T_c operator to be applied by all threads.*/
	int c;

};

#endif /* LWLOWFIELD_H_ */
