/*
 * ModelProvider.cpp
 *
 *  Created on: 17.11.2010
 *      Author: klagges
 *
 */

#include <Framework/Solver.h>
#include <Utilities/Utilities.h>
#include <Interface/StateIO.h>

void ModelProvider::applyT(const State& source, const State& target,
		stateAmplitude& result, int c) {

	State intermediateResult;

	applyT(source, intermediateResult, c);

	// The state multiplication is not commutative for complex States.
	if (source.type == State::ket)
		result = target * intermediateResult;
	else
		result = intermediateResult * target;

}

void ModelProvider::applyV(const State& source, State& result,
		unsigned int remainingV) {

	// The default implementation to apply V is to apply the sum of all T operators.

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyV\n";

	for (int i = -solver.modelProperties.Nmax; i <= solver.modelProperties.Nmax;
			i++) {

		applyT(source, result, i);

	}

	double min = solver.globalProperties.eigenEnergy
			- (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax);
	double max = solver.globalProperties.eigenEnergy
			+ (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax);

	State::iterator it = result.begin();

	State::iterator oldit;

	while (it != result.end()) {

		double eigenEnergy = getH_0Eigenvalue(it->first);

		if (eigenEnergy > max || eigenEnergy < min) {
			oldit = it;
			it++;
			result.erase(oldit);
		} else
			it++;

	}

}

void ModelProvider::applyV(const State& source, const State& target,
		stateAmplitude& result, unsigned int remainingV) {

	State intermediateResult;

	applyV(source, intermediateResult, remainingV);

	// The state multiplication is not commutative for complex States.
	if (source.type == State::ket)
		result = target * intermediateResult;
	else
		result = intermediateResult * target;

}

void ModelProvider::applyS(State & source, unsigned int remainingV,
		unsigned int numberOfS) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyS\n";

	double min = solver.globalProperties.eigenEnergy
			- (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax * 1.00001);
	double max = solver.globalProperties.eigenEnergy
			+ (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax * 1.00001);

	State::iterator it = source.begin();

	while (it != source.end()) {

		double eigenEnergy = getH_0Eigenvalue(it->first);

		if (eigenEnergy == solver.globalProperties.eigenEnergy
				|| eigenEnergy > max || eigenEnergy < min) {
			State::iterator oldit = it;
			it++;
			source.erase(oldit);
		} else {

			amplitude correctionFactor = amplitude(1);
			correctionFactor /= solver.globalProperties.eigenEnergy
					- eigenEnergy;

			amplitude correction = Utilities::pow(correctionFactor, numberOfS);

			it->second *= correction;

			it++;

		}

	}

}

void ModelProvider::applyS(const State & source, State & result,
		unsigned int remainingV, unsigned int numberOfS) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyS\n";

	double min = solver.globalProperties.eigenEnergy
			- (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax);
	double max = solver.globalProperties.eigenEnergy
			+ (remainingV * solver.modelProperties.energyDelta
					* solver.modelProperties.Nmax);

	for (State::const_iterator it = source.begin(); it != source.end(); it++) {

		double eigenEnergy = getH_0Eigenvalue(it->first);

		if (eigenEnergy != solver.globalProperties.eigenEnergy
				&& eigenEnergy <= max && eigenEnergy >= min) {

			amplitude correctionFactor = amplitude(1);
			correctionFactor /= solver.globalProperties.eigenEnergy
					- eigenEnergy;

			amplitude correction = Utilities::pow(correctionFactor, numberOfS);

			result.insert(
					State::value_type(it->first, it->second * correction));

		}

	}

}

void ModelProvider::applyP(State& source) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyP\n";

	State::iterator it = source.begin();

	while (it != source.end()) {

		if (getH_0Eigenvalue(it->first)
				!= solver.globalProperties.eigenEnergy) {
			State::iterator oldit = it;
			it++;
			source.erase(oldit);
		} else
			it++;

	}

}

void ModelProvider::applyP(const State& source, State& result) {

	if (solver.globalConfiguration.verbosityLevel > 2)
		std::cerr << "applyP\n";

	for (State::const_iterator it = source.begin(); it != source.end(); it++) {

		if (getH_0Eigenvalue(it->first) == solver.globalProperties.eigenEnergy)
			result.insert(*it);

	}

}

void ModelProvider::applySSV(const State& source, State& result,
		unsigned int remainingV, unsigned int numberOfS) {

	applyV(source, result, remainingV);

	if (remainingV == 0) {

		applyP(result);

	} else {

		applyS(result, remainingV, numberOfS);

	}

}
