/*
 * ModelProvider.h
 *
 *  Created on: 17.11.2010
 *      Author: klagges
 *
 */

#ifndef MODELPROVIDER_H_
#define MODELPROVIDER_H_

#include <Interface/Interface.h>
#include <Interface/ModelProperties.h>

/** A ModelProvider should fill the Solver::modelProperties fields of the solver
 and provide model specific methods to the DataProcessor.

 To correctly apply the methods it can use the solver.latticeProperties and solver.globalProperties fields
 of the ::solver object.

 - The ModelProvider::applyT() and ModelProvider::getH_0Eigenvalue() methods must be provided by all model providers.
 For the other methods there are default implementations.
 You may of course override them for performance reasons.

 - The ModelProvider::getH_0Eigenvalue() method should return the \f$ H_0 \f$ eigenvalue of the base state bs.

 - The ModelProvider::applyX() methods should apply the operator:
 - P :   Projector to the eigen space of \f$ H_0 \f$ in question
 - V :   Perturbation part of H.
 - S :   The operator : \f$ \frac{1-P}{E_P - H_0} \f$  with \f$ E_P \f$ being the \f$ H_0 \f$ eigen energy of the eigen space in question.
 - \f$ T_c \f$ : Operator, that transfers a state in the i-th eigen state of \f$ H_0 \f$ to a state in the (i+c)-th
 eigen space of \f$ H_0 \f$ in a way, so that \f$V = T_0 + T_1 + T_{-1} + ... + T_{Nmax} + T_{-Nmax} \f$
 (also see ModelProperties::Nmax)
 - SSV : A combination of the above operators.

 You find an instruction how to create new a new ModelProvider in
 Modules/ModelProvider/TemplateModelProvider/TemplateModelProvider.h
 */
class ModelProvider {

public:

	// These methods must be overridden by all ModelProvider classes.

	/** Return the eigen value of the base state bs according to the unperturbed
	 * Hamiltonian \f$ H_0 \f$. */
	virtual double getH_0Eigenvalue(const BaseState& bs) = 0;

	/** Apply the \f$ T_c \f$ operator to source and write the resulting state to result. */
	virtual void applyT(const State& source, State& result, int c) = 0;

	// There are default implementations for these methods.
	// You may of course override them for performance reasons.

	/** Calculate the value \f$ <target|T_c|source> \f$ and write it to result. */
	virtual void applyT(const State& source, const State& target,
			stateAmplitude& result, int c);

	/** Apply the perturbation operator V to source and write the resulting state to result.
	 * The parameter remainingV should give the number of V operators that are expected to
	 * be applied after the application of this operator, before the next P operator is
	 * applied. It is used to truncate the Hilbert space of the resulting state from
	 * not contributing states. */
	virtual void applyV(const State& source, State& result,
			unsigned int remainingV);

	/** Calculate the value <target| V |source> and write is to result. The parameter remainingV
	 * should be the number of V operators used to  calculate target. */
	virtual void applyV(const State& source, const State& target,
			stateAmplitude& result, unsigned int remainingV);

	/** Apply the resolvent operator S to source in place. The parameter remainingV should
	 * give the number of V operators that are expected to  be applied after the application
	 * of this operator, before the next P operator is applied. It is used to truncate
	 * the Hilbert space of the resulting state from not contributing states. */
	virtual void applyS(State& source, unsigned int remainingV,
			unsigned int numberOfS);

	/** Apply the resolvent operator S to source and write the resulting state to result.
	 * The parameter remainingV should give the number of V operators that are expected to
	 * be applied after the application of this operator, before the next P operator is
	 * applied. It is used to truncate the Hilbert space of the resulting state from
	 * not contributing states. */
	virtual void applyS(const State& source, State& result,
			unsigned int remainingV, unsigned int numberOfS);

	/** Apply the projector P to source in place.*/
	virtual void applyP(State& source);

	/** Apply the projector P to source and write the result to the state result. */
	virtual void applyP(const State& source, State& result);

	/** Apply a sequence of operators to source and write the result to the state result.
	 * First a V operator should be applied. Then, if remainingV is zero, a P operator should
	 * be applied, otherwise the S operator should be applied numberOfS times (truncating
	 * the Hilber space as in the applyS() methods). */
	virtual void applySSV(const State& source, State& result,
			unsigned int remainingV, unsigned int numberOfS);

};

#endif /* MODELPROVIDER_H_ */
