/*
 * AssociativeList.h
 *
 *  Created on: 28.06.2011
 *      Author: klagges
 */

#ifndef ASSOCIATIVELIST_H_
#define ASSOCIATIVELIST_H_

#include <utility>

namespace Utilities {

/** Associative container that stores the entries in a single linked list.
 * More memory efficient than the std::map or std::unordered_map.*/
template<typename Key, typename Tp>
class associative_list {
public:

	typedef Key key_type;
	typedef Tp mapped_type;
	typedef std::pair<const Key, Tp> value_type;

	/** Default constructor. */
	inline associative_list() :
			first(0) {

	}

	/** Default destructor. */
	inline ~associative_list() {

		ListNode* pointer = first;

		while (pointer != 0) {

			ListNode* next = pointer->next;
			delete (pointer);
			pointer = next;

		}

	}

	/** Copy constructor. */
	associative_list(const associative_list& s) {

		if (s.first == 0) {
			first = 0;
			return;
		}

		first = new ListNode(s.first->value, 0);

		ListNode* sourcePointer = s.first->next;
		ListNode* targetPointer = first;

		while (sourcePointer != 0) {

			targetPointer->next = new ListNode(sourcePointer->value, 0);
			targetPointer = targetPointer->next;
			sourcePointer = sourcePointer->next;

		}

	}

	/** Assignment operator.*/
	associative_list& operator=(const associative_list& s) {

		clear();

		if (s.first == 0) {
			first = 0;
			return *this;
		}

		first = new ListNode(s.first->value, 0);

		ListNode* sourcePointer = s.first->next;
		ListNode* targetPointer = first;

		while (sourcePointer != 0) {

			targetPointer->next = new ListNode(sourcePointer->value, 0);
			targetPointer = targetPointer->next;
			sourcePointer = sourcePointer->next;

		}

		return *this;

	}

private:

	/** Consists of the data to store and a pointer to the next node.*/
	class ListNode {
	public:

		inline ListNode(const value_type& v, ListNode* n) :
				value(v), next(n) {

		}
		inline ListNode() :
				next(0) {

		}
		/** Copy constructor.*/
		ListNode(const ListNode& s) :
				value(s.value), next(s.next) {

		}

		/** Assignment operator.*/
		ListNode& operator=(const ListNode& s) {
			value = s.value;
			next = s.next;

			return *this;

		}

		/** The data to store.*/
		value_type value;

		/** The next node in the list.*/
		ListNode* next;

	};

	/** Pointer to the first node of the list.*/
	ListNode* first;

public:

	class iterator {
	public:

		inline iterator(ListNode* p) {
			pointer = p;

		}

		inline void operator++() {
			pointer = pointer->next;

		}

		inline void operator++(int) {
			pointer = pointer->next;

		}

		inline value_type& operator*() const {

			return pointer->value;

		}

		inline value_type* operator->() const {

			return &pointer->value;

		}

		inline bool operator!=(const iterator& r) const {

			if (r.pointer != pointer)
				return true;
			return false;

		}

		inline bool operator==(const iterator& r) const {

			if (r.pointer == pointer)
				return true;
			return false;

		}

		ListNode* pointer;

	};

	class const_iterator {
	public:

		inline const_iterator(ListNode* p) {
			pointer = p;

		}

		inline const_iterator(const iterator p) {
			pointer = p.pointer;

		}

		inline void operator++() {
			pointer = pointer->next;

		}

		inline void operator++(int) {
			pointer = pointer->next;

		}

		inline const value_type& operator*() const {

			return pointer->value;

		}

		inline const value_type* operator->() const {

			return &pointer->value;

		}

		inline bool operator!=(const const_iterator& r) const {

			if (r.pointer != pointer)
				return true;
			return false;

		}

		inline bool operator==(const const_iterator& r) const {

			if (r.pointer == pointer)
				return true;
			return false;

		}

	private:
		ListNode* pointer;

	};

public:

	inline iterator begin() const {

		return iterator(first);

	}

	inline iterator end() const {

		return iterator(0);
	}

	inline std::pair<iterator, bool> insert(const value_type& x) {

		if (first == 0) {
			first = new ListNode(x, 0);
			return std::pair<iterator, bool>(iterator(first), true);
		}

		if (first->value.first == x.first)
			return std::pair<iterator, bool>(iterator(first), false);

		if (x.first < first->value.first) {

			first = new ListNode(x, first);
			return std::pair<iterator, bool>(iterator(first), true);

		}

		ListNode* pointer = first;

		while (pointer->next != 0) {

			if (pointer->next->value.first == x.first)
				return std::pair<iterator, bool>(iterator(pointer->next), false);

			if (x.first < pointer->next->value.first) {

				pointer->next = new ListNode(x, pointer->next);
				return std::pair<iterator, bool>(iterator(pointer->next), true);

			}

			pointer = pointer->next;

		}

		pointer->next = new ListNode(x, 0);
		return std::pair<iterator, bool>(iterator(pointer->next), true);

	}

	inline iterator find(const key_type& key) const {

		ListNode* pointer = first;

		while (pointer != 0 && pointer->value.first != key) {

			if (key < pointer->value.first)
				return iterator(0);

			pointer = pointer->next;

		}

		return iterator(pointer);

	}

	inline size_t size() const {

		size_t s = 0;
		ListNode* pointer = first;

		while (pointer != 0) {

			s++;
			pointer = pointer->next;

		}

		return s;

	}

	inline void clear() {

		ListNode* pointer = first;

		while (pointer != 0) {

			ListNode* next = pointer->next;
			delete (pointer);
			pointer = next;

		}
		first = 0;

	}

	inline void erase(iterator& it) {

		if (it == end())
			return;
		if (begin() == end())
			return;

		iterator search = begin();

		if (search == it) {

			ListNode* pointer = search.pointer->next;
			delete (search.pointer);
			first = pointer;
			return;

		}

		iterator next = search;
		next++;

		while (next != it) {
			search = next;
			next++;
		}

		ListNode* pointer = next.pointer->next;
		delete (next.pointer);
		search.pointer->next = pointer;

	}

	inline void swap(associative_list<Key, Tp>& s) {
		ListNode* temp = first;
		first = s.first;
		s.first = temp;
	}

};

}

#endif /* ASSOCIATIVELIST_H_ */
