/*
 * PoolAllocator.h
 *
 *  Created on: 23.11.2011
 *      Author: klagges
 */

#ifndef POOLALLOCATOR_H_
#define POOLALLOCATOR_H_

#include <limits>
#include <cstddef>

namespace Utilities {

#define MAXIMAL_BLOCK_SIZE_SPA 8388608
//1048576
//8388608

#define OBJECTS_PER_BLOCK_SPA ((MAXIMAL_BLOCK_SIZE_SPA-8)/sizeof(T))

/** Memory allocator for constant size objects.
 * To reduce overhead, memory is allocated in blocks with a size
 * large against the size of the object size (the pool). Memory
 * for the objects is allocated by pointer increment. The blocks
 * are ordered in a linear list. When the number of stored objects
 * is reduced to zero, all blocks are deallocated. */
template<typename T>
class SimplePoolAllocator {
public:
	//    typedefs

	typedef T value_type;
	typedef value_type* pointer;
	typedef const value_type* const_pointer;
	typedef value_type& reference;
	typedef const value_type& const_reference;
	typedef std::size_t size_type;
	typedef std::ptrdiff_t difference_type;

public:
	//    convert an allocator<T> to allocator<U>

	template<typename U>
	struct rebind {
		typedef SimplePoolAllocator<U> other;
	};

public:

	/** A class to store a block of memory. Together with the data, the
	 * pointer to the next block in the list is stored.*/
	class PoolAllocatorBlock {

	public:

		/** The block of objects.*/
		T data[OBJECTS_PER_BLOCK_SPA];
		/** Pointer to the next block in the list.*/
		PoolAllocatorBlock* nextBlock;

	};

	/** Pointer to the next free slice of memory in the current block. */
	T* pointerToFreeData;
	/** The number of free memory slices in the current block. */
	unsigned int spaceRemaining;
	/** The number of allocated objects in all blocks.*/
	unsigned int objectCounter;

	/** The pointer to the first block of the list. */
	PoolAllocatorBlock* pointerToFirstBlock;
	/** The pointer to the last (current) block of the list. */
	PoolAllocatorBlock* pointerToLastBlock;

	inline explicit SimplePoolAllocator() :
			pointerToFreeData(0), spaceRemaining(0), objectCounter(0), pointerToFirstBlock(
					0), pointerToLastBlock(0) {

	}

	inline ~SimplePoolAllocator() {

	}

	template<typename U>
	inline explicit SimplePoolAllocator(SimplePoolAllocator<U> const& p) :
			pointerToFreeData(0), spaceRemaining(0), objectCounter(0), pointerToFirstBlock(
					0), pointerToLastBlock(0) {
		// We do not copy allocators to make sure all allocators have a unique memory pool.

	}

	//    address

	inline pointer address(reference r) {
		return &r;
	}
	inline const_pointer address(const_reference r) {
		return &r;
	}

	//    memory allocation

	inline pointer allocate(size_type cnt,
			typename std::allocator<void>::const_pointer = 0) {

		// We only allocate memory for single block. All other allocations
		// are forwarded to the new operator.
		if (cnt > 1)
			return reinterpret_cast<pointer>(::operator new(cnt * sizeof(T)));

		// The current block has remaining space.
		// We increment the pointer and return the free space.
		if (spaceRemaining > 0) {
			pointer p = pointerToFreeData;
			pointerToFreeData++;
			spaceRemaining--;
			objectCounter++;
			return p;
		}

		if (pointerToFirstBlock == 0) {

			// No objects allocated so far. We create the first block.

			pointerToFirstBlock =
					reinterpret_cast<PoolAllocatorBlock*>(::operator new(
							sizeof(PoolAllocatorBlock)));
			pointerToFirstBlock->nextBlock = 0;
			pointerToLastBlock = pointerToFirstBlock;

		} else {

			// The current block is full, so we have to allocate a new block.
			pointerToLastBlock->nextBlock =
					reinterpret_cast<PoolAllocatorBlock*>(::operator new(
							sizeof(PoolAllocatorBlock)));
			pointerToLastBlock = pointerToLastBlock->nextBlock;
			pointerToLastBlock->nextBlock = 0;

		}

		spaceRemaining = OBJECTS_PER_BLOCK_SPA - 1;
		pointerToFreeData = pointerToLastBlock->data + 1;
		objectCounter++;

		return reinterpret_cast<pointer>(pointerToLastBlock->data);

	}

	inline void deallocate(pointer p, size_type cnt) {

		// We only allocate memory for single block. All other allocations
		// are forwarded to the delete operator.
		if (cnt > 1)
			::operator delete(p);
		else {
			objectCounter--;
			if (objectCounter == 0) {
				// This was the last object to deallocate.
				while (pointerToFirstBlock != 0) {

					// We go through the list of blocks and deallocate all of them.
					PoolAllocatorBlock* pnt = pointerToFirstBlock;
					pointerToFirstBlock = pointerToFirstBlock->nextBlock;
					::operator delete(pnt);

				}

				pointerToFreeData = 0;
				spaceRemaining = 0;
				pointerToLastBlock = 0;

			}

		}

	}

	//    size

	inline size_type max_size() const {
		return std::numeric_limits<size_type>::max() / sizeof(T);
	}

	//    construction/destruction

	inline void construct(pointer p, const T& t) {
		new (p) T(t);
	}
	inline void destroy(pointer p) {
		p->~T();
	}

	inline bool operator==(SimplePoolAllocator const&) {
		return false;
	}
	inline bool operator!=(SimplePoolAllocator const& a) {
		return !operator==(a);
	}

};

#define MAXIMAL_BLOCK_SIZE_TPA 8388608
//1048576
//8388608

#define OBJECTS_PER_BLOCK_TPA ((MAXIMAL_BLOCK_SIZE_TPA-16)/sizeof(T))

/** Memory allocator for constant size objects.
 * To reduce overhead, memory is allocated in blocks with a size
 * large against the size of the object size (the pool). Memory
 * for the objects is allocated by pointer increment. The blocks
 * are ordered in a linear list. To every block the number of stored
 * objects is tracked. If the number of objects of a block reaches zero,
 * The whole block is deallocated. */
template<typename T>
class TrackedPoolAllocator {
public:
	//    typedefs

	typedef T value_type;
	typedef value_type* pointer;
	typedef const value_type* const_pointer;
	typedef value_type& reference;
	typedef const value_type& const_reference;
	typedef std::size_t size_type;
	typedef std::ptrdiff_t difference_type;

public:
	//    convert an allocator<T> to allocator<U>

	template<typename U>
	struct rebind {
		typedef TrackedPoolAllocator<U> other;
	};

public:

	/** A class to store a block of memory. Together with the data, the
	 * pointer to the next block in the list is stored and the number of
	 * allocated slices of this block*/
	class PoolAllocatorBlock {

	public:

		/** The block of objects.*/
		T data[OBJECTS_PER_BLOCK_TPA];

		/** Pointer to the next block in the list.*/
		PoolAllocatorBlock* nextBlock;

		/** the number of allocated objects in this block. */
		unsigned long long int objectCounter;

	};

	/** Pointer to the next free slice of memory in the current block. */
	T* pointerToFreeData;

	/** The number of free memory slices in the current block. */
	unsigned long long int spaceRemaining;

	/** The pointer to the first block of the list. */
	PoolAllocatorBlock* pointerToFirstBlock;

	/** The pointer to the last (current) block of the list. */
	PoolAllocatorBlock* pointerToLastBlock;

	inline explicit TrackedPoolAllocator() :
			pointerToFreeData(0), spaceRemaining(0), pointerToFirstBlock(0), pointerToLastBlock(
					0) {

	}

	inline ~TrackedPoolAllocator() {

	}

	template<typename U>
	inline explicit TrackedPoolAllocator(TrackedPoolAllocator<U> const& p) :
			pointerToFreeData(0), spaceRemaining(0), pointerToFirstBlock(0), pointerToLastBlock(
					0) {
		// We do not copy allocators to make sure all allocators have a unique memory pool.

	}

	//    address

	inline pointer address(reference r) {
		return &r;
	}
	inline const_pointer address(const_reference r) {
		return &r;
	}

	//    memory allocation

	inline pointer allocate(size_type cnt,
			typename std::allocator<void>::const_pointer = 0) {

		// We only allocate memory for single block. All other allocations
		// are forwarded to the new operator.
		if (cnt > 1)
			return reinterpret_cast<pointer>(::operator new(cnt * sizeof(T)));

		if (spaceRemaining > 0) {

			// The current block has remaining space.
			// We increment the pointer and return the free space.

			pointer p = pointerToFreeData;
			pointerToFreeData++;
			spaceRemaining--;
			pointerToLastBlock->objectCounter++;
			return p;
		}

		if (pointerToFirstBlock == 0) {

			// No objects allocated so far. We create the first block.

			pointerToFirstBlock =
					reinterpret_cast<PoolAllocatorBlock*>(::operator new(
							sizeof(PoolAllocatorBlock)));
			pointerToLastBlock = pointerToFirstBlock;

		} else {

			// The current block is full, so we have to allocate a new block.

			pointerToLastBlock->nextBlock =
					reinterpret_cast<PoolAllocatorBlock*>(::operator new(
							sizeof(PoolAllocatorBlock)));
			pointerToLastBlock = pointerToLastBlock->nextBlock;

		}

		pointerToLastBlock->nextBlock = 0;
		pointerToLastBlock->objectCounter = 1;
		spaceRemaining = OBJECTS_PER_BLOCK_TPA - 1;
		pointerToFreeData = pointerToLastBlock->data + 1;

		return reinterpret_cast<pointer>(pointerToLastBlock->data);

	}

	inline void deallocate(pointer p, size_type cnt) {

		// We only allocate memory for single block. All other allocations
		// are forwarded to the delete operator.
		if (cnt > 1)
			::operator delete(p);
		else {

			// We try to find the block associated with the pointer p.
			// We do so by calculating the distance of the pointer
			// to the block pointers.

			ptrdiff_t delta = (p - pointerToFirstBlock->data);

			if (delta >= 0 && delta < (ptrdiff_t) OBJECTS_PER_BLOCK_TPA) {

				// The pointer p is part of the first block.

				if (pointerToFirstBlock->objectCounter == 1) {

					// It was the last object of the first block.
					// We deallocate the first block.

					if (pointerToLastBlock == pointerToFirstBlock) {

						// The first block was the last allocated block
						// The allocator pool is empty again.
						::operator delete(pointerToFirstBlock);
						pointerToFirstBlock = 0;
						pointerToLastBlock = 0;
						spaceRemaining = 0;
						pointerToFreeData = 0;

						return;
					}

					PoolAllocatorBlock* temp = pointerToFirstBlock;
					pointerToFirstBlock = pointerToFirstBlock->nextBlock;
					::operator delete(temp);
					return;

				}

				// There are other objects left in the first block so we just decrement the counter.
				pointerToFirstBlock->objectCounter--;

				return;

			}

			PoolAllocatorBlock* beforeBlock = pointerToFirstBlock;
			PoolAllocatorBlock* currentBlock = pointerToFirstBlock->nextBlock;

			while (true) {

				delta = (p - currentBlock->data);

				if (delta >= 0 && delta < (ptrdiff_t) OBJECTS_PER_BLOCK_TPA) {

					// We found the block associated with the pointer p.

					if (currentBlock->objectCounter == 1) {

						// It was the last object of this block, so we deallocate the block.

						if (pointerToLastBlock == currentBlock) {

							// This block was the last (current) block of the allocator.
							// We reset it.

							pointerToLastBlock = beforeBlock;
							pointerToLastBlock->nextBlock = 0;
							spaceRemaining = 0;
							pointerToFreeData = 0;

							return;
						}

						beforeBlock->nextBlock = currentBlock->nextBlock;
						::operator delete(currentBlock);
						return;

					}

					// There are other objects left in this block, so we only decrement the counter.

					currentBlock->objectCounter--;

					return;

				}

				beforeBlock = currentBlock;
				currentBlock = beforeBlock->nextBlock;

			}

		}

	}

	//    size

	inline size_type max_size() const {
		return std::numeric_limits<size_type>::max() / sizeof(T);
	}

	//    construction/destruction

	inline void construct(pointer p, const T& t) {
		new (p) T(t);
	}
	inline void destroy(pointer p) {
		p->~T();
	}

	inline bool operator==(TrackedPoolAllocator const&) {
		return false;
	}
	inline bool operator!=(TrackedPoolAllocator const& a) {
		return !operator==(a);
	}

};

}

#endif /* POOLALLOCATOR_H_ */
