/*
 * LatticeSymmetryClass.h
 *
 *  Created on: Sep 11, 2014
 *      Author: ixert
 */

#ifndef LATTICESYMMETRYCLASS_H_
#define LATTICESYMMETRYCLASS_H_

#include<vector>
#include<utility>
#include<math.h>
#include<iostream>
#include<map>
#include<tuple>
#include<vector>
#include<set>
#include "globalVariables.h"

using namespace std;

class LatticeSymmetryClass {
private:
	static unsigned int xMax,yMax;

	static vector<tuple<int,int,int> > invSiteMap; // specifies the x,y Component of each lattice site, where x,y are the prefactors of the lattice vectors!
	static map<tuple<int,int,int>, unsigned int> siteMap;

	static vector<vector<vector<int> > > shiftedSites; // shiftedSite=shiftedSites[site][jx][jy]
public:
	static void initializeSiteMap(unsigned int const &xMaxValue, unsigned int const &yMaxValue, vector<pair<int,int> > const &invSiteMapValue, vector<pair<int,int> > const &interUnitCellVectorsValue, vector<pair<int,int> > const &intraUnitCellVectorsValue);

	static void initializeShiftedSitesForReflectionSymmetry(vector<pair<int,int> > const &reflectionMap);

	static unsigned int getxMax();
	static unsigned int getyMax();

	// reflections
	template<class S,class T>
	static void getNormForReflections(S const &bitState, int const &refx, int const &refy, T &normQuad);

	template<class S>
	static S getRepresentativeAndReflectionNumbers(S const &bitState, unsigned int &rxRep, unsigned int &ryRep);

	// k-symm
	template<class S>
	static S applyTranslationOperator(S const & bitState, unsigned int const &jx, unsigned int const &jy, bool spinFlip);

	template<class S>
	static S applySublatticeTranslationOperator(S const & bitState, unsigned int const &jx, unsigned int const &jy, bool spinFlip, unsigned int const &ucShift);

	template<class S,class T>
	static bool getRepresentativeAndPhase(S &bitState, int const &kx, int const &ky, int const &spinFlip, T &phase);

	template<class S,class T>
	static T getRepresentativeAndPhaseAndNorm(S &bitState, int const &kx, int const &ky, int const &spinFlip, T &phase);

	template<class S,class T>
	static void getNorm(S const &bitState, int const &kx, int const &ky, int const &spinFlip, T &normQuad);

	template<class S>
	static bool connectedThroughTranslation(S const &state1, S const &state2, int const &spinFlip);

	template<class S>
	static S getRepresentativeAndTranslationNumbers(S const &bitState, unsigned int &jxRep, unsigned int &jyRep, bool &spinFlip, unsigned int const &ucShift, bool const &sublatticeVersion=true);

	template<class S>
	static void getOrbitsOfState(S const &bitState, set<tuple<unsigned int,unsigned int,bool> > &orbits, int const &spinFlip, unsigned int const &ucShift);

	template<class T>
	static void calculatePhase(unsigned int const &jx, unsigned int const &jy, int const &kx, int const &ky, T &phase);

	template<class T>
	static void calculatePhaseForSites(unsigned int const &site1, unsigned int const &site2, int const &kx, int const &ky, T &phase);

	template<class S,class T>
	static T findPhase(S &bitState, map<S,T> const &sum);
};

// reflection
template<class S,class T>
void LatticeSymmetryClass::getNormForReflections(S const &bitState, int const &refx, int const &refy, T &normQuad){
	map<S,int> sum;
	// apply Pk (without 1/n factor in front, since this is not necessary for the definition of the norm) on bitState
	for(unsigned int rx=0;rx<2;rx++){
		for(unsigned int ry=0;ry<2;ry++){
			S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,rx,ry,false);
			int phaseTemp=(1+(refx-1)*rx)*(1+(refy-1)*ry);
			pair<typename map<S,int>::iterator,bool> ret=sum.insert(pair<S,int>(translatedState,phaseTemp));
			if(!ret.second){
				ret.first->second+=phaseTemp;
			}
			if(ret.first->second==0){
				sum.erase(ret.first);
			}
		}
	}

	// since each element should have a prefactor with norm 1 (or 0, if the state is non-existent), the complete (square of) norm is just the number of elements in the map
	normQuad=T(sum.size());
}

template<class S>
S LatticeSymmetryClass::getRepresentativeAndReflectionNumbers(S const &bitState, unsigned int &rxRep, unsigned int &ryRep){
	// this function returns the representative for a given bitState and the reflection numbers needed to obtain this state through reflection of the representative
	S representative=bitState;
	rxRep=0;
	ryRep=0;;
	for(unsigned int rx=0;rx<2;rx++){
		for(unsigned int ry=0;ry<2;ry++){
			if((rx!=0) || (ry!=0)){
				S translatedState;
				translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,rx,ry,false);
				if(translatedState<representative){
					representative=translatedState;
				}
			}
		}
	}

	// now reflect the representative since you get the state
	if(representative!=bitState){
		for(unsigned int rx=0;rx<2;rx++){
			for(unsigned int ry=0;ry<2;ry++){
				if((rx!=0) || (ry!=0)){
					S translatedState;
					translatedState=LatticeSymmetryClass::applyTranslationOperator(representative,rx,ry,false);
					if(translatedState==bitState){
						rxRep=rx;
						ryRep=ry;
						return representative;
					}
				}
			}
		}
	}
	return representative;
}


// k-symm
template<class S>
S LatticeSymmetryClass::applyTranslationOperator(S const & bitState, unsigned int const &jx,unsigned int const &jy, bool spinFlip){
	// translates all sites by jx and jy (jx and jy is number of latticeVectors in x,y direction)
	// translate all sites => find newPositions of the sites

	S retState=S(0);
	S fullLocalState=(S(1)<<localHilbertSpace)-1;

	for(int i=0;i<shiftedSites.size();i++){
		int j=shiftedSites[i][jx][jy];

		S shifted3=fullLocalState<<(localHilbertSpace*i); // shift a fullLocalState to position i

		S extractedBit=(bitState&shifted3);
		if(j>i){
			extractedBit<<=(localHilbertSpace*(j-i));
		}
		else{
			extractedBit>>=(localHilbertSpace*(i-j));
		}
		retState|=extractedBit; // set bit j in retState to the extractedBit
	}

	S flipState=(S(1)<<(localHilbertSpace*numberOfSites))-1;

	if(spinFlip){
		retState^=flipState;
	}

	return retState;
}

template<class S>
S LatticeSymmetryClass::applySublatticeTranslationOperator(S const & bitState, unsigned int const &jx,unsigned int const &jy, bool spinFlip, unsigned int const &ucShift){
	// translates all sites by jx and jy (jx and jy is number of latticeVectors in x,y direction)
	// translate all sites => find newPositions of the sites
	// IMPORTANT:  ucShift=1 for two site uc and ucShift=2 for one site uc

	S retState=S(0);
	S fullLocalState=(S(1)<<localHilbertSpace)-1;

	for(int i=0;i<invSiteMap.size();i++){
		int j=shiftedSites[i][ucShift*jx][jy];
		// here it is assumed that the unit cells are oriented along jx and that by adding a jx you get to the other sublattice. Therefore a single translation in x-Direction on a sublattice is performed by 2 x-translations on the full lattice

		S shifted3=fullLocalState<<(localHilbertSpace*i); // shift a 3 to position i (actual site)
		S extractedBit=(bitState&shifted3);
		if(j>i){
			extractedBit<<=(localHilbertSpace*(j-i));
		}
		else{
			extractedBit>>=(localHilbertSpace*(i-j));
		}
		retState|=extractedBit; // set bit j in retState to the extractedBit
	}

	S flipState=(S(1)<<(localHilbertSpace*numberOfSites/2))-1;

	if(spinFlip){
		retState^=flipState;
	}

	return retState;
}

template<class S,class T>
bool LatticeSymmetryClass::getRepresentativeAndPhase(S &bitState, int const &kx, int const &ky, int const &spinFlip, T &phase){
	// return the phase (prefactor, without Normalization) of the bitState, after applying the TranslationProjection(kx,ky)
	// the bitState is then changed to the corresponding representative!

	// spinFlipParity: 0,1,-1

	S repState=bitState;
	T repPhase(1);
	phase=T(1); //maybe i can assume that this is already the case when the function gets called?

	T tempPhase(0);
	//TODO: ueberlegen wielange eigentlich der translationsop angewendet werden muss.
	// fuer jx=jy=0 ist das ebenfalls einfach und kann sein gelassen werden => ist derselbe Zustand, phase=1;
	unsigned int jxMax=xMax;
	unsigned int jyMax=yMax;

	for(unsigned int jx=0;jx<jxMax;jx++){
		for(unsigned int jy=0;jy<jyMax;jy++){
			if(jx!=0 || jy!=0){ // for jx=jy=0 the phase is simply 1 and the state remains itself!
				S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,false);

				if(translatedState<repState){
					repState=translatedState;
					LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,repPhase);
				}
				else{
					if(translatedState==repState){
						LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,tempPhase);
						repPhase+=tempPhase;
					}
					if(translatedState==bitState){
						LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,tempPhase);
						phase+=tempPhase;
					}
				}
				if(spinFlip){
					S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,true);

					if(translatedState<repState){
						repState=translatedState;
						LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,repPhase);
						repPhase*=T(spinFlip);
					}
					else{
						if(translatedState==repState){
							LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,tempPhase);
							repPhase+=T(spinFlip)*tempPhase;
						}
						if(translatedState==bitState){
							LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,tempPhase);
							phase+=T(spinFlip)*tempPhase;
						}
					}
				}
			}
		}
	}

	if(myAbs(repPhase)>zeroEPS){
		phase/=repPhase;
	}
	bitState=repState;

	bool returnValue=false;
	if(myAbs(phase)>zeroEPS){
		returnValue=true;
	}

	return returnValue;
}

template<class S,class T>
T LatticeSymmetryClass::getRepresentativeAndPhaseAndNorm(S &bitState, int const &kx, int const &ky, int const &spinFlip, T &phase){
	map<S,T> sum;
	// apply Pk on bitState
	for(unsigned int jx=0;jx<xMax;jx++){
		for(unsigned int jy=0;jy<yMax;jy++){
			S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy, false);
			T phaseTemp(0);
			LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,phaseTemp);
			pair<typename map<S,T>::iterator,bool> ret=sum.insert(pair<S,T>(translatedState,phaseTemp));
			if(!ret.second){
				ret.first->second+=phaseTemp;
			}
			if(spinFlip){
				S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy, true);
				T phaseTemp(0);
				LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,phaseTemp);
				phaseTemp*=T(spinFlip);
				pair<typename map<S,T>::iterator,bool> ret=sum.insert(pair<S,T>(translatedState,phaseTemp));
				if(!ret.second){
					ret.first->second+=phaseTemp;
				}
			}
		}
	}


	T normQuad(0);
	T firstPhase=sum.begin()->second; // as the sum is a map, the representative is the first element
	if(myAbs(firstPhase)>zeroEPS){
		typename map<S,T>::iterator itNorm;
		for(itNorm=sum.begin();itNorm!=sum.end();itNorm++){
			//divide all phases by the first phase, so that the phase in front of the representative is 1
			itNorm->second/=firstPhase;
			normQuad+=conj(itNorm->second)*itNorm->second;
		}

		phase=LatticeSymmetryClass::findPhase(bitState,sum);
	}

	return normQuad;
}

template<class S,class T>
void LatticeSymmetryClass::getNorm(S const &bitState, int const &kx, int const &ky, int const &spinFlip, T &normQuad){
	map<S,T> sum;
	// apply Pk (without 1/n factor in front, since this is not necessary for the definition of the norm) on bitState
	for(unsigned int jx=0;jx<xMax;jx++){
		for(unsigned int jy=0;jy<yMax;jy++){
			S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,false);
			T phaseTemp(0);
			LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,phaseTemp);
			pair<typename map<S,T>::iterator,bool> ret=sum.insert(pair<S,T>(translatedState,phaseTemp));
			if(!ret.second){
				ret.first->second+=phaseTemp;
			}
			if(myAbs(ret.first->second)<=zeroEPS){
				sum.erase(ret.first);
			}
			if(spinFlip){
				S translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,true);
				T phaseTemp(0);
				LatticeSymmetryClass::calculatePhase(jx,jy,kx,ky,phaseTemp);
				phaseTemp*=T(spinFlip);
				pair<typename map<S,T>::iterator,bool> ret=sum.insert(pair<S,T>(translatedState,phaseTemp));
				if(!ret.second){
					ret.first->second+=phaseTemp;
				}
				if(myAbs(ret.first->second)<=zeroEPS){
					sum.erase(ret.first);
				}
			}
		}
	}

	// since each element should have a prefactor with norm 1 (or 0, if the state is non-existent), the complete (square of) norm is just the number of elements in the map
	normQuad=T(sum.size());
}

template<class S>
bool LatticeSymmetryClass::connectedThroughTranslation(S const &state1, S const &state2, int const &spinFlip){
	for(unsigned int jx=0;jx<xMax;jx++){
		for(unsigned int jy=0;jy<yMax;jy++){
			S translatedState=LatticeSymmetryClass::applyTranslationOperator(state1,jx,jy,false);
			if(translatedState==state2){
				return true;
			}
			if(spinFlip){
				S translatedState=LatticeSymmetryClass::applyTranslationOperator(state1,jx,jy,true);
				if(translatedState==state2){
					return true;
				}
			}
		}
	}
	return false;
}

template<class S>
S LatticeSymmetryClass::getRepresentativeAndTranslationNumbers(S const &bitState, unsigned int &jxRep, unsigned int &jyRep, bool &spinFlip, unsigned int const &ucShift, bool const &sublatticeVersion){
	// this function returns the representative for a given bitState and the translation numbers needed to obtain this state through translation of the representative
	bool spinFlipBool=spinFlip;

	S representative=bitState;
	jxRep=0;
	jyRep=0;
	spinFlip=false;
	for(unsigned int jx=0;jx<xMax/ucShift;jx++){
		for(unsigned int jy=0;jy<yMax;jy++){
			if((jx!=0) || (jy!=0)){
				S translatedState;
				if(sublatticeVersion){
					translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(bitState,jx,jy,false,ucShift);
				}
				else{
					translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,false);
				}
				if(translatedState<representative){
					representative=translatedState;
				}
			}
			if(spinFlipBool){
				S translatedState;
				if(sublatticeVersion){
					translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(bitState,jx,jy,true,ucShift);
				}
				else{
					translatedState=LatticeSymmetryClass::applyTranslationOperator(bitState,jx,jy,true);
				}
				if(translatedState<representative){
					representative=translatedState;
				}
			}
		}
	}

	// now translate the representative since you get the state //TODO: this could be optimized as it needs only to be done once per cycle: start with representative and translate until you have all states within the cycle
	if(representative!=bitState){
		for(unsigned int jx=0;jx<xMax/ucShift;jx++){
			for(unsigned int jy=0;jy<yMax;jy++){
				if((jx!=0) || (jy!=0)){
					S translatedState;
					if(sublatticeVersion){
						translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(representative,jx,jy,false,ucShift);
					}
					else{
						translatedState=LatticeSymmetryClass::applyTranslationOperator(representative,jx,jy,false);
					}
					if(translatedState==bitState){
						jxRep=jx;
						jyRep=jy;
						spinFlip=false;
						return representative;
					}
				}
				if(spinFlipBool){
					S translatedState;
					if(sublatticeVersion){
						translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(representative,jx,jy,true,ucShift);
					}
					else{
						translatedState=LatticeSymmetryClass::applyTranslationOperator(representative,jx,jy,true);
					}
					if(translatedState==bitState){
						jxRep=jx;
						jyRep=jy;
						spinFlip=true;
						return representative;
					}
				}
			}
		}
	}
	return representative;
}

template<class S>
void LatticeSymmetryClass::getOrbitsOfState(S const &bitState, set<tuple<unsigned int,unsigned int,bool> > &orbits, int const &spinFlip, unsigned int const &ucShift){
	//cout << "getOrbitsOfState:\n";
	//cout << "bitState=" << bitState << "\n";
	for(unsigned int jx=0;jx<xMax/ucShift;jx++){
		for(unsigned int jy=0;jy<yMax;jy++){
			//cout << "jx=" << jx << " jy=" << jy << "\n";
			if((jx!=0) || (jy!=0)){
				S translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(bitState,jx,jy,false,ucShift);
				if(translatedState==bitState){
					// check if this translation is a direct multiple of an existing translation within the orbit
					set<tuple<unsigned int,unsigned int,bool> >::iterator itSet;
					bool isMultiple=false;
					for(itSet=orbits.begin();itSet!=orbits.end();++itSet){
						if(!get<2>(*itSet)){
							if(jx==0){
								if(get<0>(*itSet)==0){
									isMultiple=true;
								}
							}
							else{
								if(get<0>(*itSet)!=0){
									if(jx%get<0>(*itSet)==0){
										isMultiple=true;
									}
								}
							}
							if(isMultiple){
								isMultiple=false;
								if(jy==0){
									if(get<1>(*itSet)==0){
										isMultiple=true;
									}
								}
								else{
									if(get<1>(*itSet)!=0){
										if(jx%get<1>(*itSet)==0){
											isMultiple=true;
										}
									}
								}
							}
							if(isMultiple){
								break;
							}
						}
					}
					if(!isMultiple){
						orbits.insert(tuple<unsigned int,unsigned int,bool>(jx,jy,false));
					}
				}
			}
			if(spinFlip){
				S translatedState=LatticeSymmetryClass::applySublatticeTranslationOperator(bitState,jx,jy,true,ucShift);
				if(translatedState==bitState){
					// check if this translation is a direct multiple of an existing translation within the orbit
					set<tuple<unsigned int,unsigned int,bool> >::iterator itSet;
					bool isMultiple=false;
					for(itSet=orbits.begin();itSet!=orbits.end();++itSet){
						if(get<2>(*itSet)){
							if(jx==0){
								if(get<0>(*itSet)==0){
									isMultiple=true;
								}
							}
							else{
								if(get<0>(*itSet)!=0){
									if(jx%get<0>(*itSet)==0){
										isMultiple=true;
									}
								}
							}
							if(isMultiple){
								isMultiple=false;
								if(jy==0){
									if(get<1>(*itSet)==0){
										isMultiple=true;
									}
								}
								else{
									if(get<1>(*itSet)!=0){
										if(jx%get<1>(*itSet)==0){
											isMultiple=true;
										}
									}
								}
							}
							if(isMultiple){
								break;
							}
						}
					}
					if(!isMultiple){
						orbits.insert(tuple<unsigned int,unsigned int,bool>(jx,jy,true));
					}
				}
			}
		}
	}
}

template<class T>
inline void LatticeSymmetryClass::calculatePhase(unsigned int const &jx, unsigned int const &jy, int const &kx, int const &ky, T &phase){
	phase=T(cos((2*M_PI*jx*kx)/xMax+(2*M_PI*jy*ky)/yMax),sin((2*M_PI*jx*kx)/xMax+(2*M_PI*jy*ky)/yMax));
}

template<class T>
void LatticeSymmetryClass::calculatePhaseForSites(unsigned int const &site1, unsigned int const &site2, int const &kx, int const &ky, T &phase){
	int ix=get<0>(invSiteMap[site1]);
	int iy=get<1>(invSiteMap[site1]);
	int jx=get<0>(invSiteMap[site2]);
	int jy=get<1>(invSiteMap[site2]);

	phase=T(cos((2*M_PI*(ix-jx)*kx)/xMax+(2*M_PI*(iy-jy)*ky)/yMax),sin((2*M_PI*(ix-jx)*kx)/xMax+(2*M_PI*(iy-jy)*ky)/yMax));
}

template<class S,class T>
T LatticeSymmetryClass::findPhase(S &bitState, map<S,T> const &sum){
	typename map<S,T>::const_iterator it=sum.find(bitState);
	bitState=sum.begin()->first; // set bitState to representative
	return it->second;
}

#endif /* LATTICESYMMETRYCLASS_H_ */
