/*
 * SeriesTemplate.h
 *
 *  Created on: Jun 11, 2013
 *      Author: ixert
 */

#ifndef SERIESTEMPLATE_H_
#define SERIESTEMPLATE_H_

#include <iostream>
#include <list>
#include <vector>
#include "globalVariables.h"
#include "ExponentsTemplate.h"
#include "mathFunctions.h"
#include "nr3.h"

using namespace std;

template<class T,class S>
class SeriesTemplate {
template<class R,class Q>
friend ostream& operator<<(ostream &o, SeriesTemplate<R,Q> const &s);
friend inline ostream& operator<<(ostream &o, SeriesTemplate<Doub,ExponentsTemplate<int> > const &s);
template<class R, class Q>
friend istream& operator>>(istream &o, SeriesTemplate<R,Q> &s);
template<class R,class Q>
friend bool operator==(SeriesTemplate<R,Q> const& lhs, SeriesTemplate<R,Q> const& rhs);
private:
	list<pair<T,S> > series; // T is amplitude, S is ExponentsTemplate
public:
	SeriesTemplate();
	SeriesTemplate(T const &x);
	T evaluate(vector<T> const &h) const;

	T getMaxAmplitude() const;

	void clear();

	template<class R>
	void truncateUpToOrder(R const &order);

	SeriesTemplate<T,S>& insert(T const &x,S const &exp);

	SeriesTemplate<T,S>& operator+=(SeriesTemplate<T,S> const& rhs);
	SeriesTemplate<T,S>& operator-=(SeriesTemplate<T,S> const& rhs);
	SeriesTemplate<T,S>& operator*=(T const& rhs);
	SeriesTemplate<T,S>& operator/=(T const& rhs);
	SeriesTemplate<T,S>& operator*=(SeriesTemplate<T,S> const &rhs);

	SeriesTemplate<Doub,S> real() const;
	SeriesTemplate<Doub,S> imag() const;
	SeriesTemplate<T,S> conj() const;
};

// free functions declaration //
template<class T,class S>
const SeriesTemplate<T,S> operator+(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> operator-(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> operator*(SeriesTemplate<T,S> const& lhs, T const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> operator*(T const& lhs, SeriesTemplate<T,S> const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> operator/(SeriesTemplate<T,S> const& lhs, T const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> operator*(SeriesTemplate<T,S> const &s1,SeriesTemplate<T,S> const &s2);

template<class R,class Q>
ostream& operator<<(ostream &o, SeriesTemplate<R,Q> const &s);
inline ostream& operator<<(ostream &o, SeriesTemplate<Doub,ExponentsTemplate<int> > const &s);
template<class R, class Q>
istream& operator>>(istream &o, SeriesTemplate<R,Q> &s);

template<class T,class S>
bool operator==(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs);

template<class T,class S>
bool operator != (SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs);

template<class T,class S>
const SeriesTemplate<T,S> conj(SeriesTemplate<T,S> const &s);

//template<class T,class S>
//bool operator==(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs);


// member functions implementation //
template<class T,class S>
SeriesTemplate<T,S>::SeriesTemplate(){
}

template<class T,class S>
SeriesTemplate<T,S>::SeriesTemplate(T const &x){
	if(x!=T(0)){
		series.push_back(pair<T,S>(x,S(0)));
	}
}

template<class T,class S>
T SeriesTemplate<T,S>::evaluate(vector<T> const &h) const{
	T result=T(0);
	typename list<pair<T,S> >::const_iterator it;
	for(it=series.begin();it!=series.end();it++){
		result+=(it->first)*(it->second.evaluate(h));
	}
	return result;
}

template<class T,class S>
T SeriesTemplate<T,S>::getMaxAmplitude() const{
	T maxAmplitude=0;
	typename list<pair<T,S> >::const_iterator it;
	for(it=series.begin();it!=series.end();it++){
		if(myAbs(it->first)>maxAmplitude){
			maxAmplitude=myAbs(it->first);
		}
	}
	return maxAmplitude;
}

template<class T,class S>
void SeriesTemplate<T,S>::clear(){
	series.clear();
}

template<class T,class S>
template<class R>
void SeriesTemplate<T,S>::truncateUpToOrder(R const &order){
	// kicks out all Elements which are greater than order
	typename list<pair<T,S> >::iterator it=series.begin();
	while(it!=series.end()){
		//cout << "it->second.order()=" << it->second.order() << " order=" << order << "\n";
		if(it->second.order()>order){
			it=series.erase(it);
		}
		else{
			it++;
		}
	}
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::insert(T const &x,S const &exp){
	typename list<pair<T,S> >::iterator it2;
	bool inserted=false;
	for(it2=series.begin();it2!=series.end();it2++){
		if((it2->second)==(exp)){
			it2->first+=x;
			if(it2->first==T(0)){
				it2=series.erase(it2);
			}
			inserted=true;
			break;
		}
	}
	if(!inserted){
		series.push_back(pair<T,S>(x,exp));
	}

	return *this;
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::operator+=(SeriesTemplate<T,S> const& rhs){
	typename list<pair<T,S> >::const_iterator it1;
	for(it1=rhs.series.begin();it1!=rhs.series.end();it1++){
		typename list<pair<T,S> >::iterator it2;
		bool inserted=false;
		for(it2=series.begin();it2!=series.end();it2++){
			if((it2->second)==(it1->second)){
				it2->first+=it1->first;
				if(it2->first==T(0)){
					it2=series.erase(it2);
				}
				inserted=true;
				break;
			}
		}
		if(!inserted){
			series.push_back(pair<T,S>(it1->first,it1->second));
		}
	}

	return *this;
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::operator-=(SeriesTemplate<T,S> const& rhs){
	(*this)+=T(-1)*rhs;
	return *this;
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::operator*=(T const& rhs){
	if(rhs!=T(0)){
		typename list<pair<T,S> >::iterator it1;
		for(it1=series.begin();it1!=series.end();it1++){
			it1->first*=rhs;
		}
	}
	else{
		series.clear();
	}

	return *this;
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::operator/=(T const& rhs){
	if(rhs!=0){
		typename list<pair<T,S> >::iterator it1;
		for(it1=series.begin();it1!=series.end();it1++){
			it1->first/=rhs;
		}
	}
	else{
		cerr << "ERROR! Division by zero!\n";
	}

	return *this;
}

template<class T,class S>
SeriesTemplate<T,S>& SeriesTemplate<T,S>::operator*=(SeriesTemplate<T,S> const &rhs){
	typename list<pair<T,S> >::const_iterator it1,it2;
	typename list<pair<T,S> >::iterator it3;
	list<pair<T,S> > newList;
	for(it1=series.begin();it1!=series.end();it1++){
		for(it2=rhs.series.begin();it2!=rhs.series.end();it2++){
			T newAmpl=it1->first*it2->first;
			S newExp=it1->second*it2->second;
			bool foundElement=false;
			for(it3=newList.begin();it3!=newList.end();it3++){
				if(it3->second==newExp){
					foundElement=true;
					break;
				}
			}
			if(foundElement){
				it3->first+=newAmpl;
			}
			else{
				newList.push_back(pair<T,S>(newAmpl,newExp));
			}
		}
	}
	series.swap(newList);

	return *this;
}

template<class T,class S>
SeriesTemplate<Doub,S> SeriesTemplate<T,S>::real() const{
	SeriesTemplate<Doub,S> s2;
	typename list<pair<T,S> >::const_iterator it;
	for(it=series.begin();it!=series.end();it++){
		s2.insert(it->first.real(),it->second);
	}
	return s2;
}

template<class T,class S>
SeriesTemplate<Doub,S> SeriesTemplate<T,S>::imag() const{
	SeriesTemplate<Doub,S> s2;
	typename list<pair<T,S> >::const_iterator it;
	for(it=series.begin();it!=series.end();it++){
		s2.insert(it->first.imag(),it->second);
	}
	return s2;
}

template<class T, class S>
SeriesTemplate<T,S> SeriesTemplate<T,S>::conj() const{
	SeriesTemplate<T,S> s2;
	typename list<pair<T,S> >::const_iterator it;
	for(it=series.begin();it!=series.end();it++){
		s2.series.push_back(pair<T,S>(std::conj(it->first),it->second));
	}
	return s2;
}

// friend functions implementation //
template<class T,class S>
ostream& operator<<(ostream &o, SeriesTemplate<T,S> const &s){
	bool showposset=false;
	ios::fmtflags f;
	f = cout.flags();
	if(f & ios::showpos){
		showposset=true;
	}

	typename list<pair<T,S> >::const_iterator it;
	if(s.series.size()==0){
		// if the series is empty get out 0.
		o << T(0);
	}

	for(it=s.series.begin();it!=s.series.end();it++){
		o << showpos;
		o<<it->first << "*" << noshowpos << it->second;
	}

	if(showposset){
		o << showpos;
	}
	else{
		o << noshowpos;
	}

	return o;
}

inline ostream& operator<<(ostream &o, SeriesTemplate<Doub,ExponentsTemplate<int> > const &s){
	bool showposset=false;
	ios::fmtflags f;
	f = cout.flags();
	if(f & ios::showpos){
		showposset=true;
	}

	list<pair<Doub,ExponentsTemplate<int> > >::const_iterator it;
	if(s.series.size()==0){
		// if the series is empty get out 0.
		o << 0;
	}

	for(it=s.series.begin();it!=s.series.end();it++){
		o << showpos;
		if((myAbs(it->first)>zeroEPS)){
			o<<it->first << "*" << noshowpos << it->second;
		}
	}

	if(showposset){
		o << showpos;
	}
	else{
		o << noshowpos;
	}

	return o;
}

template<class T,class S>
istream& operator>>(istream &o, SeriesTemplate<T,S> &s){
	char tempChar;

	T x;
	s.series.clear();
	T vz=T(1);
	if(o.peek()=='+'){
		o>>tempChar;
	}
	else{
		if(o.peek()=='-'){
			o>>tempChar;
			vz=T(-1);
		}
	}
	while(o>>x){
		tempChar=o.peek();
		if(tempChar=='*'){
			o>>tempChar;
		}
		S exp;
		if(!(o>>exp)){
			cout << "reading error in SeriesTemplate!\n";
		}
		//cout << exp << "\n";
		s.series.push_back(pair<T,S>(vz*x,exp));
		vz=T(1);
		if(o.peek()=='+'){
			o>>tempChar;
		}
		else{
			if(o.peek()=='-'){
				o>>tempChar;
				vz=T(-1);
			}
		}
	}

	//if(s.series.size()>0){
		o.clear(); // this is not so good... but what should happen if no series is given? yeah... just saying that if he cannot read a series he assumes this means the series is empty!
	//}

	//cout << "s.series.size()=" << s.series.size() << "\n";

	//cout << "streamIsGood=" << o.good() << "\n";

	return o;
}


// free functions implementation //
template<class T,class S>
const SeriesTemplate<T,S> operator+(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs){
	SeriesTemplate<T,S> ret=lhs;
	ret+=rhs;
	return ret;
}

template<class T,class S>
const SeriesTemplate<T,S> operator-(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs){
	SeriesTemplate<T,S> ret=lhs;
	ret-=rhs;
	return ret;
}

template<class T,class S>
const SeriesTemplate<T,S> operator*(SeriesTemplate<T,S> const& lhs, T const& rhs){
	SeriesTemplate<T,S> ret=lhs;
	ret*=rhs;
	return ret;
}

template<class T,class S>
const SeriesTemplate<T,S> operator*(T const& lhs, SeriesTemplate<T,S> const& rhs){
	SeriesTemplate<T,S> ret=rhs;
	ret*=lhs;
	return ret;
}

template<class T,class S>
const SeriesTemplate<T,S> operator/(SeriesTemplate<T,S> const& lhs, T const& rhs){
	SeriesTemplate<T,S> ret=lhs;
	ret/=rhs;
	return ret;
}

template<class T,class S>
const SeriesTemplate<T,S> operator*(SeriesTemplate<T,S> const &s1,SeriesTemplate<T,S> const &s2){
	SeriesTemplate<T,S> s3=s1;
	s3*=s2;
	return s3;
}

template<class T,class S>
bool operator==(SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs){
	if(lhs.series==rhs.series){
		return true;
	}
	else{
		return false;
	}
}

template<class T,class S>
bool operator != (SeriesTemplate<T,S> const& lhs, SeriesTemplate<T,S> const& rhs){
	return !(lhs==rhs);
}

template<class T,class S>
const SeriesTemplate<T,S> conj(SeriesTemplate<T,S> const &s){
	return s.conj();
}


typedef SeriesTemplate<Doub,ExponentsTemplate<int> > SeriesDouble;
typedef SeriesTemplate<Complex,ExponentsTemplate<int> > SeriesComplex;

#endif /* SERIESTEMPLATE_H_ */
