/*
 * SpinModelClassTemplate.h
 *
 *  Created on: 02.09.2014
 *      Author: dominik-ixert
 */

#ifndef SPINMODELCLASSTEMPLATE_H_
#define SPINMODELCLASSTEMPLATE_H_

#include "nr3.h"
#include<list>
#include<utility>
#include "SeriesTemplate.h"
#include "SpinOperatorsOnSitesXYBasis.h"
#include "SpinOperatorsOnSitesPMBasis.h"
#include<iostream>
#include "pairFunctions.h"

using namespace std;

template<class T>
class SpinModelClassTemplate {
template<class S>
friend istream& operator>>(istream &o, SpinModelClassTemplate<S> &s);
template<class S>
friend ostream& operator<<(ostream &o, SpinModelClassTemplate<S> const &s);
protected:
	list<pair<T,SpinOperatorsOnSitesPMBasis> > spinModel;

public:
	SpinModelClassTemplate();
	SpinModelClassTemplate(T const &value, SpinOperatorsOnSitesPMBasis const &spinOperator);

	int numberOfOperators() const;

	void assignNewSites();
	void assignNewSites(vector<unsigned short> const &newSites);
	void assignNewSites(vector<unsigned short> const &oldSites, vector<unsigned short> const &newSites);
	void orderAgainstSites();

	void divideGroundStateEnergyBy(Doub const &n);
	void setGroundStateSiteTo(unsigned short const &initialSite);

	list<vector<unsigned short> > getAllSiteCombinations();
	SpinModelClassTemplate<T> getOperatorsOnSpecifiedSites(vector<unsigned short> const &newSites);
	SpinModelClassTemplate<T> getOperatorsWhichContainSite(unsigned short const &site);
	void deleteOperatorsWithoutSite(unsigned short const &site);

	template<class S>
	void applyOperatorSequenceOnBitState(map<S,T> &sum, S const &state, T const &prefactor) const;

	SpinModelClassTemplate<Complex> evaluateSpinModel(vector<Doub> const &values);

	SpinModelClassTemplate<T>& operator+=(SpinModelClassTemplate<T> const& rhs);
	SpinModelClassTemplate<T>& operator-=(SpinModelClassTemplate<T> const& rhs);
	SpinModelClassTemplate<T>& operator*=(T const& rhs);
	SpinModelClassTemplate<T>& operator/=(Doub const& rhs);
};

// free functions declaration //
template<class T>
const SpinModelClassTemplate<T> operator+(SpinModelClassTemplate<T> const& lhs, SpinModelClassTemplate<T> const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator-(SpinModelClassTemplate<T> const& lhs, SpinModelClassTemplate<T> const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator*(SpinModelClassTemplate<T> const& lhs, Doub const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator*(Doub const& lhs, SpinModelClassTemplate<T> const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator*(SpinModelClassTemplate<T> const& lhs, T const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator*(T const& lhs, SpinModelClassTemplate<T> const& rhs);
template<class T>
const SpinModelClassTemplate<T> operator/(SpinModelClassTemplate<T> const& lhs, Doub const& rhs);

// class function implementation //

template<class T>
SpinModelClassTemplate<T>::SpinModelClassTemplate(){
}

template<class T>
SpinModelClassTemplate<T>::SpinModelClassTemplate(T const &value, SpinOperatorsOnSitesPMBasis const &spinOperator){
	spinModel.push_back(pair<T,SpinOperatorsOnSitesPMBasis>(value,spinOperator));
}

template<class T>
int SpinModelClassTemplate<T>::numberOfOperators() const{
	// returns the number of operators if the number of operators in each summand is the same, -1 otherwise
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::const_iterator it=spinModel.begin();
	//cout << "size=" << spinModel.size() << "\n";
	//cout << *this << "\n";
	int number=it->second.numberOfOperators();
	//cout << "number=" << number << "\n";
	it++;
	while(it!=spinModel.end()){
		//cout << " " << it->second.numberOfOperators() << "\n";
		if(it->second.numberOfOperators()!=number){
			return -1;
		}
		it++;
	}
	return number;
}

template<class T>
void SpinModelClassTemplate<T>::assignNewSites(){
	//assigns the new sites 0,1,2,...
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		it->second.assignNewSites();
	}
}
template<class T>
void SpinModelClassTemplate<T>::assignNewSites(vector<unsigned short> const &newSites){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		it->second.assignNewSites(newSites);
	}
}
template<class T>
void SpinModelClassTemplate<T>::assignNewSites(vector<unsigned short> const &oldSites, vector<unsigned short> const &newSites){
	// in contrast to assignNewSites(vector<unsigned short> const &newSites) this version sets the oldSites specified in oldSites to the newSites specified in newSites!
	// this method also works if the SpinModel consists of for example, two AND four spin operators
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		it->second.assignNewSites(oldSites,newSites);
	}
}
template<class T>
void SpinModelClassTemplate<T>::orderAgainstSites(){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		it->second.orderAgainstSites();
	}
}

template<class T>
void SpinModelClassTemplate<T>::divideGroundStateEnergyBy(Doub const &n){
	if(n!=0){
		typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
		for(it=spinModel.begin();it!=spinModel.end();it++){
			if(it->second.isUnity()){ // check if this operator is the unity, i.e. the gsEnergy Operator.
				it->first.imag()/=n;
				it->first.real()/=n;
			}
		}
	}
	else{
		cout << "Divison by zero!\n";
	}
}

template<class T>
void SpinModelClassTemplate<T>::setGroundStateSiteTo(unsigned short const &initialSite){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		if(it->second.isUnity()){ // check if this operator is the unity, i.e. the gsEnergy Operator.
			vector<unsigned short> vus(1,initialSite);
			it->second.assignNewSites(vus);
		}
	}
}

template<class T>
list<vector<unsigned short> > SpinModelClassTemplate<T>::getAllSiteCombinations(){
	list<vector<unsigned short> > siteCombinations;
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	list<vector<unsigned short> >::iterator itSiteComb;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		bool foundComb=false;
		vector<unsigned short> tempVec=it->second.getSites();
		for(itSiteComb=siteCombinations.begin();itSiteComb!=siteCombinations.end();itSiteComb++){
			if((*itSiteComb)==tempVec){
				foundComb=true;
				break;
			}
		}
		if(!foundComb){
			siteCombinations.push_back(tempVec);
		}
	}

	return siteCombinations;
}
template<class T>
SpinModelClassTemplate<T> SpinModelClassTemplate<T>::getOperatorsOnSpecifiedSites(vector<unsigned short> const &newSites){
	SpinModelClassTemplate<T> newSpinModel;
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		if(it->second.areSameSites(newSites)){
			newSpinModel.spinModel.push_back(*it);
		}
	}
	return newSpinModel;
}
template<class T>
SpinModelClassTemplate<T> SpinModelClassTemplate<T>::getOperatorsWhichContainSite(unsigned short const &site){
	SpinModelClassTemplate<T> newSpinModel;
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		if(it->second.containsSite(site)){
			newSpinModel.spinModel.push_back(*it);
		}
	}
	return newSpinModel;
}
template<class T>
void SpinModelClassTemplate<T>::deleteOperatorsWithoutSite(unsigned short const &site){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it=spinModel.begin();
	while(it!=spinModel.end()){
		if(!(it->second.containsSite(site))){
			it=spinModel.erase(it);
		}
		else{
			it++;
		}
	}
}

template<class T>
template<class S>
void SpinModelClassTemplate<T>::applyOperatorSequenceOnBitState(map<S,T> &sum, S const &state, T const &prefactor) const{
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::const_iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		S tempState=state;
		// apply operator product onto tempState to obtain ONE new State with a prefactor!
		//cout << "tempState=" << tempState << "\n";
		//cout << "operator=" << it->second << "\n";
		double vz=it->second.applyOperatorSequenceOnBitState(tempState);
		//cout << "vz=" << vz << " tempState=" << tempState << "\n";
		if(vz!=0){
			T tempPrefactor=prefactor*it->first*T(vz);
			pair<typename map<S,T>::iterator,bool> ret=sum.insert(pair<S,T>(tempState,tempPrefactor));
			if(!ret.second){
				ret.first->second+=tempPrefactor;
			}
		}
	}
}

template<class T>
SpinModelClassTemplate<Complex> SpinModelClassTemplate<T>::evaluateSpinModel(vector<Doub> const &values){
	SpinModelClassTemplate<Complex> evaluatedSpinModel;
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::const_iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		evaluatedSpinModel.spinModel.push_back(pair<Complex,SpinOperatorsOnSitesPMBasis>(Complex(it->first),it->second));
	}
	return evaluatedSpinModel;
}

template<>
inline SpinModelClassTemplate<Complex> SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> >::evaluateSpinModel(vector<Doub> const &values){
	SpinModelClassTemplate<Complex> evaluatedSpinModel;
	list<pair<pair<SeriesDouble,SeriesDouble>,SpinOperatorsOnSitesPMBasis> >::const_iterator it;
	for(it=spinModel.begin();it!=spinModel.end();it++){
		Complex c(it->first.first.evaluate(values),it->first.second.evaluate(values));
		SpinModelClassTemplate<Complex> tempSpinModel(c,it->second);
		evaluatedSpinModel+=tempSpinModel;
	}
	return evaluatedSpinModel;
}


template<class T>
SpinModelClassTemplate<T>& SpinModelClassTemplate<T>::operator+=(SpinModelClassTemplate<T> const& rhs){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::const_iterator it1;
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it2;

	for(it1=rhs.spinModel.begin();it1!=rhs.spinModel.end();it1++){
		bool foundOperators=false;;
		for(it2=spinModel.begin();it2!=spinModel.end();it2++){
			if(it1->second==it2->second){
				it2->first+=it1->first;
				if(it2->first==T(0)){
					spinModel.erase(it2);
				}
				foundOperators=true;
				break;
			}
		}
		if(!foundOperators){
			spinModel.push_back(*it1);
		}
	}

	return *this;
}
template<>
inline SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> >& SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> >::operator+=(SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> > const& rhs){
	typename list<pair<pair<SeriesDouble,SeriesDouble>,SpinOperatorsOnSitesPMBasis> >::const_iterator it1;
	typename list<pair<pair<SeriesDouble,SeriesDouble>,SpinOperatorsOnSitesPMBasis> >::iterator it2;

	for(it1=rhs.spinModel.begin();it1!=rhs.spinModel.end();it1++){
		bool foundOperators=false;;
		for(it2=spinModel.begin();it2!=spinModel.end();it2++){
			if(it1->second==it2->second){
				it2->first+=it1->first;
				if(it2->first==pair<SeriesDouble,SeriesDouble>(0,0)){
					spinModel.erase(it2);
				}
				foundOperators=true;
				break;
			}
		}
		if(!foundOperators){
			spinModel.push_back(*it1);
		}
	}

	return *this;
}
template<class T>
SpinModelClassTemplate<T>& SpinModelClassTemplate<T>::operator-=(SpinModelClassTemplate<T> const& rhs){
	*this+=Doub(-1)*rhs;
	return *this;
}
template<class T>
SpinModelClassTemplate<T>& SpinModelClassTemplate<T>::operator*=(T const& rhs){
	if(rhs==T(0)){
		spinModel.clear();
	}
	else{
		typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::iterator it1;
		for(it1=spinModel.begin();it1!=spinModel.end();it1++){
			it1->first*=rhs;
		}
	}
	return *this;
}

template<class T>
SpinModelClassTemplate<T>& SpinModelClassTemplate<T>::operator/=(Doub const& rhs){
	if(rhs==0){
		cout << "Error: Division by zero is not allowed!\n";
	}
	else{
		*this*=1/rhs;
	}
	return *this;
}

template<class T>
const SpinModelClassTemplate<T> operator+(SpinModelClassTemplate<T> const& lhs, SpinModelClassTemplate<T> const& rhs){
	SpinModelClassTemplate<T> smc=lhs;
	smc+=rhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator-(SpinModelClassTemplate<T> const& lhs, SpinModelClassTemplate<T> const& rhs){
	SpinModelClassTemplate<T> smc=lhs;
	smc-=rhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator*(SpinModelClassTemplate<T> const& lhs, Doub const& rhs){
	SpinModelClassTemplate<T> smc=lhs;
	smc*=rhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator*(Doub const& lhs, SpinModelClassTemplate<T> const& rhs){
	SpinModelClassTemplate<T> smc=rhs;
	smc*=lhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator*(SpinModelClassTemplate<T> const& lhs, T const& rhs){
	SpinModelClassTemplate<T> smc=lhs;
	smc*=rhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator*(T const& lhs, SpinModelClassTemplate<T> const& rhs){
	SpinModelClassTemplate<T> smc=rhs;
	smc*=lhs;
	return smc;
}
template<class T>
const SpinModelClassTemplate<T> operator/(SpinModelClassTemplate<T> const& lhs, Doub const& rhs){
	SpinModelClassTemplate<T> smc=lhs;
	smc/=rhs;
	return smc;
}

template<class T>
istream& operator>>(istream &o, SpinModelClassTemplate<T> &s){
	char tempChar;
	tempChar=o.peek();
	if(tempChar=='\n'){
		o>>noskipws>>tempChar>>skipws;
	}

	// get complete line of stream
	string tempstring;
	getline(o,tempstring);
	istringstream iss(tempstring);

	//cout << "tempstring=" << tempstring << "\n";

	s.spinModel.clear();

	int vz=1;
	if(iss.peek()=='+'){
		iss>>tempChar;
	}
	else{
		if(iss.peek()=='-'){
			iss>>tempChar;
			vz=-1;
		}
	}
	T prefactor1=T(0);
	T prefactor2=T(0);

	iss>>prefactor1;

	while(prefactor1!=T(0)){
		//cout << "iss.peek()=" << char(iss.peek()) << "\n";
		//cout << "prefactor1=" << prefactor1 << "\n";
		//int xxx;
		//cin>>xxx;
		if(iss.peek()=='*'){
			iss>>tempChar;
			iss>>prefactor2;
			//cout << "prefactor2=" << prefactor2 << "\n";
			prefactor1*=prefactor2;
		}
		prefactor1*=vz;
		//cout << "prefactor1=" << prefactor1 << "\n";
		//read OperatorSequence
		SpinOperatorsOnSitesXYBasis tempSpinOperator;
		//cout << "read tempSpinOperator...\n";
		iss>>tempSpinOperator;
		//cout << "XYBasis=" << tempSpinOperator << "\n";
		//convert XYBasis to PMBasis
		list<pair<SpinOperatorsOnSitesPMBasis,int> > pmList=tempSpinOperator.convertToPMBasis();
		list<pair<SpinOperatorsOnSitesPMBasis,int> >::iterator it;
		for(it=pmList.begin();it!=pmList.end();it++){
			SpinModelClassTemplate<T> tempSpinModel(T(it->second)*prefactor1,it->first);
			s+=tempSpinModel;
		}

		if(iss>>prefactor1){
			//cout << "prefactor readed!\n";
		}
		else{
			//cout << "not readed...\n";
			prefactor1=T(0);
		}
		//cout << "prefactor1=" << prefactor1 << "\n";
	}

	if(s.spinModel.size()==0){
		o.setstate(ios_base::failbit);
		//cout << "Leaving with failbit ..." << endl;
	}
	else{
		o.clear();
	}

	//cout << "streamIsGood=" << o.good() << "\n";

	return o;
}

template<>
inline istream& operator>>(istream &o, SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> > &s){
	char tempChar;
	tempChar=o.peek();
	if(tempChar=='\n'){
		o>>noskipws>>tempChar>>skipws;
	}

	// get complete line of stream
	string tempstring;
	getline(o,tempstring);
	istringstream iss(tempstring);

	//cout << "tempstring=" << tempstring << "\n";

	s.spinModel.clear();

	int vz=1;
	if(iss.peek()=='+'){
		iss>>tempChar;
	}
	else{
		if(iss.peek()=='-'){
			iss>>tempChar;
			vz=-1;
		}
	}
	pair<SeriesDouble,SeriesDouble> prefactor1=pair<SeriesDouble,SeriesDouble>(0,0);
	pair<SeriesDouble,SeriesDouble> prefactor2=pair<SeriesDouble,SeriesDouble>(0,0);

	iss>>prefactor1;

	while(prefactor1!=pair<SeriesDouble,SeriesDouble>(0,0)){
		cout << "iss.peek()=" << char(iss.peek()) << "\n";
		cout << "prefactor1=" << prefactor1 << "\n";
		//int xxx;
		//cin>>xxx;
		if(iss.peek()=='*'){
			iss>>tempChar;
			iss>>prefactor2;
			cout << "prefactor2=" << prefactor2 << "\n";
			prefactor1*=prefactor2;
		}
		prefactor1*=vz;
		cout << "prefactor1=" << prefactor1 << "\n";
		//read OperatorSequence
		SpinOperatorsOnSitesXYBasis tempSpinOperator;
		//cout << "read tempSpinOperator...\n";
		iss>>tempSpinOperator;
		//cout << "XYBasis=" << tempSpinOperator << "\n";
		//convert XYBasis to PMBasis
		list<pair<SpinOperatorsOnSitesPMBasis,int> > pmList=tempSpinOperator.convertToPMBasis();
		list<pair<SpinOperatorsOnSitesPMBasis,int> >::iterator it;
		for(it=pmList.begin();it!=pmList.end();it++){
			SpinModelClassTemplate<pair<SeriesDouble,SeriesDouble> > tempSpinModel(pair<SeriesDouble,SeriesDouble>(it->second,0)*prefactor1,it->first);
			s+=tempSpinModel;
		}

		if(iss>>prefactor1){
			cout << "prefactor readed!\n";
		}
		else{
			cout << "not readed...\n";
			prefactor1=pair<SeriesDouble,SeriesDouble>(0,0);
		}
		cout << "prefactor1=" << prefactor1 << "\n";
	}

	if(s.spinModel.size()==0){
		o.setstate(ios_base::failbit);
		//cout << "Leaving with failbit ..." << endl;
	}
	else{
		o.clear();
	}

	//cout << "streamIsGood=" << o.good() << "\n";

	return o;
}

template<class T>
ostream& operator<<(ostream &o, SpinModelClassTemplate<T> const &s){
	typename list<pair<T,SpinOperatorsOnSitesPMBasis> >::const_iterator it1;
	for(it1=s.spinModel.begin();it1!=s.spinModel.end();it1++){
		o << "+" << it1->first << "*" << it1->second << "\n";
	}
	return o;
}


#endif /* SPINMODELCLASSTEMPLATE_H_ */
