/*
 * SpinStateTemplate.h
 *
 *  Created on: Sep 1, 2014
 *      Author: ixert
 */

#ifndef SPINSTATETEMPLATE_H_
#define SPINSTATETEMPLATE_H_

#include<iostream>
#include<sstream>
#include<utility>
#include<gmpxx.h>
#include<gmp.h>
#include<vector>
#include<tuple>
#include<omp.h>
#include<set>
#include<list>
#include "SpinModelClassTemplate.h"
#include "LocalSpinModelClassTemplate.h"
#include "LatticeSymmetryClass.h"
#include "functions.h"
#include "globalVariables.h"
#include "nr3.h"


typedef std::tuple<unsigned int,unsigned int,bool,unsigned int,unsigned int,bool> tuple6bool;

template<class S,class T>
class LocalSpinModelClassTemplate;

using namespace std;

template<class S,class T>
class SpinStateTemplate {
template<class Q, class R>
friend ostream& operator<<(ostream &o, SpinStateTemplate<Q,R> const &s);
template<class Q, class R>
friend R operator*(SpinStateTemplate<Q,R> const &s1, SpinStateTemplate<Q,R> const &s2);
template<class R,class Q>
friend bool operator==(SpinStateTemplate<R,Q> const& lhs, SpinStateTemplate<R,Q> const& rhs);

public:
	static unsigned int orbitIndex;
	static unsigned int xMax;
	static unsigned int yMax;
	static subStateInt maxNumber;

	static unsigned int orbitxyxy;
	static unsigned int xyxy;
	static unsigned int yxy;
	static unsigned int xy;

	// there are 4 ways to find the representative:
	// 1) sublattice coding with one site UC: orbitTranslationVec (lt,eq,gt)
	// 2) sublattice coding with two site UC: orbitTranslationVec_eq
	// 3) no symmetry: state corresponds to index
	// 4) createRepresentativesOnlySz (stateToRepresentativeIndex holds the index for the state)
	// 5) createRepresentativesWithoutSublatticeCoding (stateToRepresentativeIndex holds index of rep for the state, stateToTranslationNumbers hold the corresponding translationNumbers)
	// 6) createRepresentativesReflectionSymmetry (stateToRepresentativeIndex holds index of rep for the state, stateToTranslationNumbers holds the corresponding reflectionNumbers)
	// 7) createRepresentativesNoSymmPartialBits (if not all bits of the local hilbert space are used, this method is necessary, since the no symmetry method will create invalid states)
	// 8) createRepresentativesOnlySzLinTables (stateToTranslationNumber_x(y) codes the Lin Tables Ja(Jb))
	static unsigned int repMethod; // this variable stores the used method for easy calling

	// combined state
	static vector<S> indexToRepresentative;

	static vector<unsigned int> sublatticeRepsToIndexVec; // (rA,rB,jx,jy,spinFlip) -> index
	static unsigned int allRepsSize;

	// sublattice states
	static vector<unsigned int> stateToRepresentativeIndex;
	static vector<unsigned int> stateToTranslationNumber_x;
	static vector<unsigned int> stateToTranslationNumber_y;
	static vector<bool> stateToTranslationNumber_spinFlip;
	static vector<unsigned int> stateToOrbitNumber;

	static vector<tuple6bool> orbitTranslationVec_lt; // (ga,gb,jax,jay,jbx,jby,spinFlip_a,spinFlip_b) -> (ix,iy,spinFlip_i,jx,jy,spinFlip_j)
	static vector<tuple6bool> orbitTranslationVec_eq; // this one is also used for the two site UC version!
	static vector<tuple6bool> orbitTranslationVec_gt;

	static vector<Doub> normalizationFactors;
	vector<T> stateCont;

	static void createRepresentativesForSublatticeCodingOneSiteUC(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm);
	static void createRepresentativesForSublatticeCodingTwoSiteUC(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm);
	static void createRepresentativesWithoutSublatticeCoding(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm);
	static void createRepresentativesReflectionSymmetry(mpq_class const &Sz, int const &refx, int const &refy, bool const &useSzSymm);
	static void createRepresentativesOnlySz(mpq_class const &Sz);
	static void createRepresentativesOnlySzLinTables(mpq_class const &Sz);
	static void createRepresentativesNoSymm();
	static void createRepresentativesNoSymmPartialBits();


public:
	SpinStateTemplate();
	SpinStateTemplate(SpinStateTemplate<S,T> const &s);
	SpinStateTemplate(unsigned int const &index);
	SpinStateTemplate(unsigned int const &index, T const &value);

	static Doub getNormalizationFactor(unsigned int const &index);

	SpinStateTemplate<S,T>& operator=(SpinStateTemplate<S,T> const &s);

	T normQuad() const;

	static unsigned int getRepresentativeIndex(subStateInt const &stateA, subStateInt const &stateB, unsigned int &ix, unsigned int &iy, bool &spinFlip_i);
	static unsigned int getStateToRepresentativeIndex(subStateInt const &state);
	static unsigned int getStateToTranslationNumber_x(subStateInt const &state);
	static unsigned int getStateToTranslationNumber_y(subStateInt const &state);
	static unsigned int getStateToTranslationNumber_spinFlip(subStateInt const &state);

	void createRandomState();

	void swap(SpinStateTemplate<S,T> &s2);
	void clear();

	unsigned int size() const;

	static void createRepresentatives(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const& refy, bool const &useSzSymm, bool const &periodicBC, unsigned int const &useSublatticeCoding, bool const &useReflectionSymm);

	void applyLocalSpinHamiltonian(SpinStateTemplate<S,T> &outputState, vector<LocalSpinModelClassTemplate<S,T> > const &localSpinModel, list<pair<vector<unsigned short>,unsigned int> > const &siteCombinationsToReferenceSpinModel, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const &refy);

	void applyLocalOperator(SpinStateTemplate<S,T> &outputState, LocalSpinModelClassTemplate<S,T> const &localOperator, vector<vector<unsigned short> > const &siteCombinationsToOperatorModel, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const &refy);

    SpinStateTemplate<S,T>& operator+=(SpinStateTemplate<S,T> const &s);
    SpinStateTemplate<S,T>& operator-=(SpinStateTemplate<S,T> const &s);
    SpinStateTemplate<S,T>& operator*=(T const &x);
    SpinStateTemplate<S,T>& operator/=(T const &x);
};

template<class S,class T>
unsigned int SpinStateTemplate<S,T>::orbitIndex=0;
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::xMax=0;
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::yMax=0;
template<class S,class T>
subStateInt SpinStateTemplate<S,T>::maxNumber=0;

template<class S,class T>
unsigned int SpinStateTemplate<S,T>::orbitxyxy=0;
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::xyxy=0;
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::yxy=0;
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::xy=0;

template<class S,class T>
unsigned int SpinStateTemplate<S,T>::repMethod=0;

template<class S,class T>
vector<S> SpinStateTemplate<S,T>::indexToRepresentative=vector<S>();

template<class S,class T>
vector<unsigned int> SpinStateTemplate<S,T>::sublatticeRepsToIndexVec=vector<unsigned int>();
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::allRepsSize=0;

template<class S,class T>
vector<unsigned int> SpinStateTemplate<S,T>::stateToRepresentativeIndex=vector<unsigned int>();
template<class S,class T>
vector<unsigned int> SpinStateTemplate<S,T>::stateToTranslationNumber_x=vector<unsigned int>();
template<class S,class T>
vector<unsigned int> SpinStateTemplate<S,T>::stateToTranslationNumber_y=vector<unsigned int>();
template<class S,class T>
vector<bool> SpinStateTemplate<S,T>::stateToTranslationNumber_spinFlip=vector<bool>();
template<class S,class T>
vector<unsigned int> SpinStateTemplate<S,T>::stateToOrbitNumber=vector<unsigned int>();

template<class S,class T>
vector<tuple6bool> SpinStateTemplate<S,T>::orbitTranslationVec_lt=vector<tuple6bool>();
template<class S,class T>
vector<tuple6bool> SpinStateTemplate<S,T>::orbitTranslationVec_eq=vector<tuple6bool>();
template<class S,class T>
vector<tuple6bool> SpinStateTemplate<S,T>::orbitTranslationVec_gt=vector<tuple6bool>();

template<class S,class T>
vector<Doub> SpinStateTemplate<S,T>::normalizationFactors=vector<Doub>();

//// friend functions declaration ////
template<class S,class T>
ostream& operator<<(ostream &o, SpinStateTemplate<S,T> const &s);
template<class S,class T>
T operator*(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2);

//// non class/friend functions declaration ////
template<class S,class T>
const SpinStateTemplate<S,T> operator+(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2);
template<class S,class T>
const SpinStateTemplate<S,T> operator-(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2);
template<class S,class T>
const SpinStateTemplate<S,T> operator*(SpinStateTemplate<S,T> const &s, T const &x);
template<class S,class T>
const SpinStateTemplate<S,T> operator*(T const &x, SpinStateTemplate<S,T> const &s);
template<class S,class T>
const SpinStateTemplate<S,T> operator/(SpinStateTemplate<S,T> const &s, T const &x);

//// implementation ////
//// class functions ////

template<class S,class T>
SpinStateTemplate<S,T>::SpinStateTemplate() : stateCont(indexToRepresentative.size(),T(0)){
}
template<class S,class T>
SpinStateTemplate<S,T>::SpinStateTemplate(SpinStateTemplate const &s){
	this->stateCont=s.stateCont;
}
template<class S,class T>
SpinStateTemplate<S,T>::SpinStateTemplate(unsigned int const &index) : stateCont(indexToRepresentative.size(),T(0)){
	stateCont[index]=T(1);
}

template<class S,class T>
SpinStateTemplate<S,T>::SpinStateTemplate(unsigned int const &index, T const &value) : stateCont(indexToRepresentative.size(),T(0)){
	stateCont[index]=value;
}

template<class S,class T>
inline Doub SpinStateTemplate<S,T>::getNormalizationFactor(unsigned int const &index){
	return normalizationFactors[index];
}

template<class S,class T>
SpinStateTemplate<S,T>& SpinStateTemplate<S,T>::operator=(SpinStateTemplate<S,T> const &s){
	this->stateCont=s.stateCont;
	return *this;
}

template<class S,class T>
T SpinStateTemplate<S,T>::normQuad() const{
	// returns the square of the norm
	T x=0;
	for(unsigned int i=0;i<stateCont.size();i++){
		x+=conj(stateCont[i])*stateCont[i];
	}
	return x;
}

template<class S,class T>
inline unsigned int SpinStateTemplate<S,T>::getRepresentativeIndex(subStateInt const &stateA, subStateInt const &stateB, unsigned int &ix, unsigned int &iy, bool &spinFlip_i){
	// return -1 if the representative could not be found
	// 1) get representatives and indices of stateA and stateB
	unsigned int ra=stateToRepresentativeIndex[stateA];
	unsigned int rb=stateToRepresentativeIndex[stateB];
	unsigned int jax=stateToTranslationNumber_x[stateA];
	unsigned int jay=stateToTranslationNumber_y[stateA];
	unsigned int jbx=stateToTranslationNumber_x[stateB];
	unsigned int jby=stateToTranslationNumber_y[stateB];
	unsigned int ga=stateToOrbitNumber[stateA];
	unsigned int gb=stateToOrbitNumber[stateB];
	bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
	bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];

	// 2) get ix,iy,jx,jy

	if(repMethod==1){
		tuple6bool ij;
		if(ra<rb){
			//cout << "ra<rb\n";
			ij=orbitTranslationVec_lt[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b];
		}
		else{
			if(ra==rb){
				//cout << "ra=rb\n";
				ij=orbitTranslationVec_eq[4*(ga*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b];
			}
			else{
				//cout << "ra>rb\n";
				ij=orbitTranslationVec_gt[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b];
				unsigned int rc=ra;
				ra=rb;
				rb=rc;
			}
		}

		ix=get<0>(ij);
		iy=get<1>(ij);
		spinFlip_i=get<2>(ij);
		unsigned int jx=get<3>(ij);
		unsigned int jy=get<4>(ij);
		bool spinFlip_j=get<5>(ij);


		if((ix==-1) || (iy==-1) || (jx==-1) || (jy==-1)){
			return -1;
		}

		return sublatticeRepsToIndexVec[(size_t(rb)+size_t(allRepsSize)*ra-(size_t(ra)*(ra+1))/2)*xMax*yMax+jx*yMax*2+jy*2+spinFlip_j];
	}

	if(repMethod==2){
		tuple6bool ij=orbitTranslationVec_eq[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b];

		ix=get<0>(ij);
		iy=get<1>(ij);
		spinFlip_i=get<2>(ij);

		return sublatticeRepsToIndexVec[ra*allRepsSize*xMax*yMax*2+rb*xMax*yMax*2+get<3>(ij)*yMax*2+get<4>(ij)*2+get<5>(ij)];
	}

	cout << "ERROR: SublatticeCoding is not enabled, getRepresentativeIndex was called!\n";
	return -1;
}

template<class S,class T>
unsigned int SpinStateTemplate<S,T>::getStateToRepresentativeIndex(subStateInt const &state){
	return stateToRepresentativeIndex[state];
}
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::getStateToTranslationNumber_x(subStateInt const &state){
	return stateToTranslationNumber_x[state];
}
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::getStateToTranslationNumber_y(subStateInt const &state){
	return stateToTranslationNumber_y[state];
}
template<class S,class T>
unsigned int SpinStateTemplate<S,T>::getStateToTranslationNumber_spinFlip(subStateInt const &state){
	return stateToTranslationNumber_spinFlip[state];
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRandomState(){
	for(unsigned int i=0;i<stateCont.size();i++){
		stateCont[i]=T(rand())/T(RAND_MAX);
	}
}

template<class S,class T>
void SpinStateTemplate<S,T>::swap(SpinStateTemplate<S,T> &s2){
	stateCont.swap(s2.stateCont);
}

template<class S,class T>
void SpinStateTemplate<S,T>::clear(){
	vector<T>().swap(stateCont);
}

template<class S,class T>
unsigned int SpinStateTemplate<S,T>::size() const{
	return stateCont.size();
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentatives(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const& refy, bool const &useSzSymm, bool const &periodicBC, unsigned int const &useSublatticeCoding, bool const &useReflectionSymm){
	// this routine creates the representatives for a given Symetry-Sector
	// ONE site unit cell
	// sublattices are ordered in x-direction (this means that xMax must be even!)

	cout << "createRepresentatives\n";

	// 0) clear all vectors...
	indexToRepresentative.clear();
	sublatticeRepsToIndexVec.clear();
	stateToRepresentativeIndex.clear();
	stateToTranslationNumber_x.clear();
	stateToTranslationNumber_y.clear();
	stateToTranslationNumber_spinFlip.clear();
	stateToOrbitNumber.clear();
	orbitTranslationVec_lt.clear();
	orbitTranslationVec_eq.clear();
	orbitTranslationVec_gt.clear();
	normalizationFactors.clear();

	// call specific createRepresentative function
	if(periodicBC){
		switch(useSublatticeCoding){
			case 0:{
				createRepresentativesWithoutSublatticeCoding(Sz,kx,ky,spinFlip,useSzSymm);
				repMethod=5;
				break;
			}
			case 1:{
				createRepresentativesForSublatticeCodingOneSiteUC(Sz,kx,ky,spinFlip,useSzSymm);
				repMethod=1;
				break;
			}
			case 2:{
				createRepresentativesForSublatticeCodingTwoSiteUC(Sz,kx,ky,spinFlip,useSzSymm);
				repMethod=2;
				break;
			}
		}
	}
	else{
		if(useReflectionSymm){
			createRepresentativesReflectionSymmetry(Sz,refx,refy,useSzSymm);
			repMethod=6;
		}
		else{
			if(useSzSymm){
				if(useSublatticeCoding){
					createRepresentativesOnlySzLinTables(Sz);
					repMethod=8;
				}
				else{
					createRepresentativesOnlySz(Sz);
					repMethod=4;
				}
			}
			else{
				int x=totalS+1;
				if((x & (x-1))==0){ // this checks wether the local states uses ALL bits of the local hilbert space
					createRepresentativesNoSymm();
					repMethod=3;
				}
				else{
					createRepresentativesNoSymmPartialBits();
					repMethod=7;
				}
			}
		}
	}
	cout << "repMethod=" << repMethod << "\n";
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesForSublatticeCodingOneSiteUC(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm){
	cout << "createRepresentativesForSublatticeCodingOneSiteUC\n";
	cout << "step1\n";
	// 1) create ALL states (not given Sz!) of the sublattice (numberOfSites/2)
	// first of all determine the maximum integer number which can occur on a sublattice => numberOfSites/2 ones
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites/2);
	//number of upSpins
	mpq_class temp=Sz*2;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_ui();

	cout << "SzNum=" << SzNum << "\n";
	// resize vectors
	stateToRepresentativeIndex.resize(maxNumber);
	stateToTranslationNumber_x.resize(maxNumber);
	stateToTranslationNumber_y.resize(maxNumber);
	stateToTranslationNumber_spinFlip.resize(maxNumber);
	stateToOrbitNumber.resize(maxNumber);

	vector<subStateInt> allReps;
	vector<int> rSzVec;
	vector<bool> rValidVec;

	cout << "maxNumber=" << maxNumber << "\n";

	cout << "step2\n";
	// 2) calculate the translation numbers of the sublattice states and the corresponding representative (i.e. which translations are needed to transform this state to the corresponding representative)
	// in this step actual translations must be performed (=> LatticeSymmetryClass)
	orbitIndex=0;
	{
		list<pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > > > orbits;
		for(subStateInt i=0;i<maxNumber;i++){
			// 2a) create "full" state (zipper product with 0)
			S fullState=zipperProduct(S(i),S(0));

			// 2b) perform translation
			unsigned int jx,jy;
			bool spinFlipBool=false;
			if(spinFlip){
				spinFlipBool=true;
			}
			S repState=LatticeSymmetryClass::getRepresentativeAndTranslationNumbers(fullState,jx,jy,spinFlipBool,2); //TODO: this whole stuff could be improved by having a routine which directly translates sublatticeStates; at the moment i'm thinking that this is not the bottleneck, so whatever...

			// 2c) get sublattice representative from "full" representative
			subStateInt stateA,stateB;
			getSublatticeStates(repState,stateA,stateB);

			// find index of rep within allReps
			bool foundIndex=false;
			for(unsigned int j=0;j<allReps.size();j++){
				if(allReps[j]==stateA){
					foundIndex=true;
					stateToRepresentativeIndex[i]=j;
					break;
				}
			}
			if(!foundIndex){
				stateToRepresentativeIndex[i]=allReps.size();
				allReps.push_back(stateA);
				int Sz=0;
				rValidVec.push_back(calculateSz(stateA,numberOfSites/2,Sz));
				rSzVec.push_back(Sz);
			}

			stateToTranslationNumber_x[i]=jx;
			stateToTranslationNumber_y[i]=jy;
			stateToTranslationNumber_spinFlip[i]=spinFlipBool;
			// 2d) calculateOrbit number
			set<tuple<unsigned int,unsigned int,bool> > tempSet;
			LatticeSymmetryClass::getOrbitsOfState(repState,tempSet,spinFlip,2);

			bool foundOrbit=false;
			unsigned int actualOrbitIndex=0;
			list<pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > > >::iterator itList;
			for(itList=orbits.begin();itList!=orbits.end();++itList){
				if(itList->second==tempSet){
					foundOrbit=true;
					actualOrbitIndex=itList->first;
					break;
				}
			}
			if(!foundOrbit){
				orbits.push_back(pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > >(orbitIndex,tempSet));
				actualOrbitIndex=orbitIndex;
				orbitIndex++;
			}
			stateToOrbitNumber[i]=actualOrbitIndex;
		}
	}

	cout << "step3\n";
	// 3) calculate all needed combined states for gives Sz, kx and ky and the normalization factors
	// to know wether a state contributes or wether the norm is nonzero the translations must be performed, too

	xMax=LatticeSymmetryClass::getxMax();
	yMax=LatticeSymmetryClass::getyMax();

	subStateInt stateA,stateB;
	int SzTemp;
	unsigned int repIndex=0;

	// loop over all r
	//sublatticeRepsToIndex.clear();
	allRepsSize=allReps.size();
	cout << "allRepsSize=" << allRepsSize << "\n";
	cout << "xMax=" << xMax << "\n";
	cout << "yMax=" << yMax << "\n";
	//sublatticeRepsToIndexVec.resize(size_t(allRepsSize)*(allRepsSize)*(xMax)*yMax,-1);
	sublatticeRepsToIndexVec.resize(size_t(allRepsSize)*(allRepsSize+1)*(xMax/2)*yMax,-1);
	cout << "sublatticeRepsToIndexVec.size()=" << sublatticeRepsToIndexVec.size() << "\n";

	time_t startingTime = time(NULL);
	{
		for(unsigned int rIndex=0;rIndex<allReps.size();rIndex++){
			if(rIndex%1000==0){
				cout << rIndex << "/" << allReps.size() << "\n";
				cout << "repIndex=" << repIndex << "\n";
				cout << "sublatticeRepsToIndexVec.size()=" << sublatticeRepsToIndexVec.size() << "\n";
				cout << "indexToRepresentative.size()=" << indexToRepresentative.size() << "\n";
			}
			if(rValidVec[rIndex]){
				subStateInt r=allReps[rIndex];
				int rSz=rSzVec[rIndex];
				// loop over all r'
				for(unsigned int sIndex=rIndex;sIndex<allReps.size();sIndex++){ //only start at rIndex
					if(rValidVec[sIndex]){
						subStateInt s=allReps[sIndex];
						unsigned int minIndexForRepCheck=indexToRepresentative.size();
						if(minIndexForRepCheck>0){
							minIndexForRepCheck--;
						}
						//minIndexForRepCheck=0;
						// create all different T(jx,jy)|s>
						list<subStateInt> different_s;
						// loop over all jx,jy, spinFlip
						for(unsigned int jx=0;jx<xMax/2;jx++){
							for(unsigned int jy=0;jy<yMax;jy++){
								// translate r'
								S sFull=zipperProduct(S(s),S(0));
								S sFullTranslated=LatticeSymmetryClass::applySublatticeTranslationOperator(sFull,jx,jy,false,2);
								getSublatticeStates(sFullTranslated,stateA,stateB);
								if(calculateSz(stateA,numberOfSites/2,SzTemp)){
									if(!useSzSymm || ((rSz+SzTemp)==SzNum)){
										list<subStateInt>::iterator itList;
										for(itList=different_s.begin();itList!=different_s.end();++itList){
											if(*itList==stateA){
												break;
											}
										}
										if(itList==different_s.end()){
											different_s.push_back(stateA);
										}
									}
								}
								if(spinFlip){
									S sFull=zipperProduct(S(s),S(0));
									S sFullTranslated=LatticeSymmetryClass::applySublatticeTranslationOperator(sFull,jx,jy,true,2);
									getSublatticeStates(sFullTranslated,stateA,stateB);
									if(calculateSz(stateA,numberOfSites/2,SzTemp)){
										if(!useSzSymm || ((rSz+SzTemp)==SzNum)){
											list<subStateInt>::iterator itList;
											for(itList=different_s.begin();itList!=different_s.end();++itList){
												if(*itList==stateA){
													break;
												}
											}
											if(itList==different_s.end()){
												different_s.push_back(stateA);
											}
										}
									}
								}
							}
						}
						list<subStateInt>::iterator itList;
						for(itList=different_s.begin();itList!=different_s.end();++itList){
							stateA=*itList;
							// combine states
							S fullRep=zipperProduct(S(r),S(stateA));
							// calculate the norm of this state
							T normQuad;
							LatticeSymmetryClass::getNorm(fullRep,kx,ky,spinFlip,normQuad);
							if(myAbs(normQuad)>zeroEPS){
								// if the norm is nonzero then this state is a valid representative, add to indexToRepresentative
								// if the state of sublattice-a is invariant under a Translation T, then not all different states of sublattice-b are necessary, since some of this states could be created by translation of the full state. Therefore this is not a representative
								// to overcome this issue, check wether it is possible to reach this state through translation of an already existing representative
								bool repExistsAlready=false;
								for(unsigned int i=minIndexForRepCheck;i<indexToRepresentative.size();i++){
									// check if this representative is connected to the actual representative (fullRep)
									repExistsAlready=LatticeSymmetryClass::connectedThroughTranslation(indexToRepresentative[i],fullRep,spinFlip);
									if(repExistsAlready){
										break;
									}
								}
								if(!repExistsAlready){
									unsigned int jxIndex=stateToTranslationNumber_x[stateA];
									unsigned int jyIndex=stateToTranslationNumber_y[stateA];
									unsigned int jSpinFlipIndex=stateToTranslationNumber_spinFlip[stateA];
									indexToRepresentative.push_back(fullRep);
									normalizationFactors.push_back(1/sqrt(real(normQuad)));
									sublatticeRepsToIndexVec[(size_t(sIndex)+size_t(allRepsSize)*rIndex-(size_t(rIndex)*(rIndex+1))/2)*xMax*yMax+jxIndex*yMax*2+jyIndex*2+jSpinFlipIndex]=repIndex;
									repIndex++;
								}
							}
						}
					}
				}
			}
		}
	}
	cout << "Processing Time: ";
	printProcessingTime(startingTime);

	cout << "orbitSize=" << orbitIndex << "\n";

	xy=xMax*yMax;
	yxy=xy*yMax;
	xyxy=yxy*xMax;
	orbitxyxy=orbitIndex*xyxy;

	//orbitTranslationVec.resize(4*orbitIndex*orbitIndex*xMax*xMax*yMax*yMax);
	orbitTranslationVec_lt.resize(4*orbitIndex*orbitIndex*xMax*xMax*yMax*yMax,tuple6bool(-1,-1,false,-1,-1,false));
	orbitTranslationVec_eq.resize(4*orbitIndex*xMax*xMax*yMax*yMax,tuple6bool(-1,-1,false,-1,-1,false));
	orbitTranslationVec_gt.resize(4*orbitIndex*orbitIndex*xMax*xMax*yMax*yMax,tuple6bool(-1,-1,false,-1,-1,false));

	cout << "step4\n";
	startingTime = time(NULL);
	// 4) the map (ga,gb,jax,jay,jbx,jby) -> (ix,iy,jx,jy) must be created (ga,gb are the orbit numbers of the state)
	// 4a) loop over all ga,gb

	for(unsigned int index_a=0;index_a<allReps.size();index_a++){
		// 4b) take one representative for each of the orbit numbers
		// get representative for ga
		subStateInt ra=allReps[index_a];
		unsigned int ga=stateToOrbitNumber[ra];
		for(unsigned int index_b=0;index_b<allReps.size();index_b++){
			subStateInt rb=allReps[index_b];
			unsigned int gb=stateToOrbitNumber[rb];
			// get representative for gb

			if((ra<=rb)){ // with spinFlip symmetry the Sz-Check must be performed AFTER applying translation operators!
				S fullState_b=zipperProduct(S(rb),S(0));
				// 4c) loop in reverse order over i and j
				for(int jx=(xMax-1)/2;jx>=0;jx--){
					for(int jy=yMax-1;jy>=0;jy--){
						if(spinFlip){
							for(int ix=xMax-1;ix>=0;ix--){
								for(int iy=yMax-1;iy>=0;iy--){
									// j translation of rb
									{ //TODO: this loop structure and the spinFlip-ifs could be easily optimized
										S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,true,2);
										// getsublattice state of b
										subStateInt transA,transB;
										getSublatticeStates(fullStateTranslated_b,transB,transA);
										// i+j translation of full state
										S fullState=zipperProduct(S(ra),S(transB));
										if(calculateSz(fullState,numberOfSites,SzTemp)){
											if(!useSzSymm || (SzTemp==SzNum)){
												S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,true);
												// get sublattice states
												getSublatticeStates(fullStateTranslated,stateA,stateB);
												// 4d) look up exponents of resulting states |a> and |b>
												unsigned int jax=stateToTranslationNumber_x[stateA];
												unsigned int jay=stateToTranslationNumber_y[stateA];
												bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
												unsigned int jbx=stateToTranslationNumber_x[stateB];
												unsigned int jby=stateToTranslationNumber_y[stateB];
												bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
												unsigned int gStateA=stateToOrbitNumber[stateA];
												unsigned int gStateB=stateToOrbitNumber[stateB];

												// 4e) save this mapping
												tuple6bool tempTuple(ix,iy,true,jx,jy,true);

												// it is now important to check in which order the sublattice-representatives of stateA and stateB are
												unsigned int rStateAIndex=stateToRepresentativeIndex[stateA];
												unsigned int rStateBIndex=stateToRepresentativeIndex[stateB];
												if(rStateAIndex<rStateBIndex){
													orbitTranslationVec_lt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
												}
												else{
													if(rStateAIndex==rStateBIndex){
														orbitTranslationVec_eq[4*(gStateA*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
													}
													else{
														orbitTranslationVec_gt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
													}
												}
											}
										}
									}
									{
										S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,true,2);
										// getsublattice state of b
										subStateInt transA,transB;
										getSublatticeStates(fullStateTranslated_b,transB,transA);
										// i+j translation of full state
										S fullState=zipperProduct(S(ra),S(transB));
										if(calculateSz(fullState,numberOfSites,SzTemp)){
											if(!useSzSymm || (SzTemp==SzNum)){
												S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,false);
												// get sublattice states
												getSublatticeStates(fullStateTranslated,stateA,stateB);
												// 4d) look up exponents of resulting states |a> and |b>
												unsigned int jax=stateToTranslationNumber_x[stateA];
												unsigned int jay=stateToTranslationNumber_y[stateA];
												bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
												unsigned int jbx=stateToTranslationNumber_x[stateB];
												unsigned int jby=stateToTranslationNumber_y[stateB];
												bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
												unsigned int gStateA=stateToOrbitNumber[stateA];
												unsigned int gStateB=stateToOrbitNumber[stateB];

												// 4e) save this mapping
												tuple6bool tempTuple(ix,iy,false,jx,jy,true);
												// it is now important to check in which order the sublattice-representatives of stateA and stateB are
												unsigned int rStateAIndex=stateToRepresentativeIndex[stateA];
												unsigned int rStateBIndex=stateToRepresentativeIndex[stateB];
												if(rStateAIndex<rStateBIndex){
													orbitTranslationVec_lt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
												}
												else{
													if(rStateAIndex==rStateBIndex){
														orbitTranslationVec_eq[4*(gStateA*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
													}
													else{
														orbitTranslationVec_gt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
													}
												}
											}
										}
									}
								}
							}
						}
						for(int ix=xMax-1;ix>=0;ix--){
							for(int iy=yMax-1;iy>=0;iy--){
								// j translation of rb
								if(spinFlip){
									S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,false,2);
									// getsublattice state of b
									subStateInt transA,transB;
									getSublatticeStates(fullStateTranslated_b,transB,transA);
									// i+j translation of full state
									S fullState=zipperProduct(S(ra),S(transB));
									if(calculateSz(fullState,numberOfSites,SzTemp)){
										if(!useSzSymm || (SzTemp==SzNum)){
											S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,true);
											// get sublattice states
											getSublatticeStates(fullStateTranslated,stateA,stateB);

											// 4d) look up exponents of resulting states |a> and |b>
											unsigned int jax=stateToTranslationNumber_x[stateA];
											unsigned int jay=stateToTranslationNumber_y[stateA];
											bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
											unsigned int jbx=stateToTranslationNumber_x[stateB];
											unsigned int jby=stateToTranslationNumber_y[stateB];
											bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
											unsigned int gStateA=stateToOrbitNumber[stateA];
											unsigned int gStateB=stateToOrbitNumber[stateB];

											// 4e) save this mapping
											tuple6bool tempTuple(ix,iy,true,jx,jy,false);
											// it is now important to check in which order the sublattice-representatives of stateA and stateB are
											unsigned int rStateAIndex=stateToRepresentativeIndex[stateA];
											unsigned int rStateBIndex=stateToRepresentativeIndex[stateB];
											if(rStateAIndex<rStateBIndex){
												orbitTranslationVec_lt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
											}
											else{
												if(rStateAIndex==rStateBIndex){
													orbitTranslationVec_eq[4*(gStateA*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
												}
												else{
													orbitTranslationVec_gt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
												}
											}
										}
									}
								}
								S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,false,2);
								// getsublattice state of b
								subStateInt transA,transB;
								getSublatticeStates(fullStateTranslated_b,transB,transA);
								//cout << "transB=" << transB << "\n";
								// i+j translation of full state
								S fullState=zipperProduct(S(ra),S(transB));
								if(calculateSz(fullState,numberOfSites,SzTemp)){
									if(!useSzSymm || (SzTemp==SzNum)){
										S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,false);
										// get sublattice states
										getSublatticeStates(fullStateTranslated,stateA,stateB);
										// 4d) look up exponents of resulting states |a> and |b>
										unsigned int jax=stateToTranslationNumber_x[stateA];
										unsigned int jay=stateToTranslationNumber_y[stateA];
										bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
										unsigned int jbx=stateToTranslationNumber_x[stateB];
										unsigned int jby=stateToTranslationNumber_y[stateB];
										bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
										unsigned int gStateA=stateToOrbitNumber[stateA];
										unsigned int gStateB=stateToOrbitNumber[stateB];

										// 4e) save this mapping
										tuple6bool tempTuple(ix,iy,false,jx,jy,false);
										// it is now important to check in which order the sublattice-representatives of stateA and stateB are
										unsigned int rStateAIndex=stateToRepresentativeIndex[stateA];
										unsigned int rStateBIndex=stateToRepresentativeIndex[stateB];

										if(rStateAIndex<rStateBIndex){
											orbitTranslationVec_lt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
										}
										else{
											if(rStateAIndex==rStateBIndex){
												orbitTranslationVec_eq[4*(gStateA*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
											}
											else{
												orbitTranslationVec_gt[4*(gStateA*orbitxyxy+gStateB*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	cout << "Processing Time: ";
	printProcessingTime(startingTime);
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesForSublatticeCodingTwoSiteUC(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm){
	// this routine creates the representatives for a system with a TWO site unit cell
	cout << "createRepresentativesForSublatticeCodingTwoSiteUC\n";
	cout << "step1\n";
	// 1) create ALL states (not given Sz!) of the sublattice (numberOfSites/2)
	// first of all determine the maximum integer number which can occur on a sublattice => numberOfSites/2 ones
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites/2);
	// resize vectors
	stateToRepresentativeIndex.resize(maxNumber);
	stateToTranslationNumber_x.resize(maxNumber);
	stateToTranslationNumber_y.resize(maxNumber);
	stateToTranslationNumber_spinFlip.resize(maxNumber);
	stateToOrbitNumber.resize(maxNumber);

	vector<subStateInt> allReps;
	vector<int> rSzVec;
	vector<bool> rValidVec;
	mpq_class temp=Sz*2;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_si();

	cout << "maxNumber=" << maxNumber << "\n";

	int SzTemp;

	cout << "step2\n";
	// 2) calculate the translation numbers of the sublattice states and the corresponding representative (i.e. which translations are needed to transform this state to the corresponding representative)
	// in this step actual translations must be performed (=> LatticeSymmetryClass)
	orbitIndex=0;
	{
		list<pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > > > orbits;
		for(subStateInt i=0;i<maxNumber;i++){
			// 2a) create "full" state (zipper product with 0)
			S fullState=zipperProduct(S(i),S(0));

			// 2b) perform translation
			unsigned int jx,jy;
			bool spinFlipBool=false;
			if(spinFlip){
				spinFlipBool=true;
			}
			S repState=LatticeSymmetryClass::getRepresentativeAndTranslationNumbers(fullState,jx,jy,spinFlipBool,1); //TODO: this whole stuff could be improved by having a routine which directly translates sublatticeStates; at the moment i'm thinking that this is not the bottleneck, so whatever...

			// 2c) get sublattice representative from "full" representative
			subStateInt stateA,stateB;
			getSublatticeStates(repState,stateA,stateB);

			// find index of rep within allReps
			bool foundIndex=false;
			for(unsigned int j=0;j<allReps.size();j++){
				if(allReps[j]==stateA){
					foundIndex=true;
					stateToRepresentativeIndex[i]=j;
					break;
				}
			}
			if(!foundIndex){
				stateToRepresentativeIndex[i]=allReps.size();
				allReps.push_back(stateA);
				rValidVec.push_back(calculateSz(stateA,numberOfSites/2,SzTemp));
				rSzVec.push_back(SzTemp);
			}

			stateToTranslationNumber_x[i]=jx;
			stateToTranslationNumber_y[i]=jy;
			stateToTranslationNumber_spinFlip[i]=spinFlipBool;
			// 2d) calculateOrbit number
			set<tuple<unsigned int,unsigned int,bool> > tempSet;
			LatticeSymmetryClass::getOrbitsOfState(repState,tempSet,spinFlip,1);

			bool foundOrbit=false;
			unsigned int actualOrbitIndex=0;
			list<pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > > >::iterator itList;
			for(itList=orbits.begin();itList!=orbits.end();++itList){
				if(itList->second==tempSet){
					foundOrbit=true;
					actualOrbitIndex=itList->first;
					break;
				}
			}
			if(!foundOrbit){
				orbits.push_back(pair<unsigned int,set<tuple<unsigned int,unsigned int,bool> > >(orbitIndex,tempSet));
				actualOrbitIndex=orbitIndex;
				orbitIndex++;
			}
			stateToOrbitNumber[i]=actualOrbitIndex;
		}
	}

	cout << "step3\n";
	// 3) calculate all needed combined states for gives Sz, kx and ky and the normalization factors
	// to know wether a state contributes or wether the norm is nonzero the translations must be performed, too

	xMax=LatticeSymmetryClass::getxMax();
	yMax=LatticeSymmetryClass::getyMax();

	subStateInt stateA,stateB;
	unsigned int repIndex=0;

	// loop over all r
	allRepsSize=allReps.size();
	cout << "allRepsSize=" << allRepsSize << "\n";
	cout << "xMax=" << xMax << "\n";
	cout << "yMax=" << yMax << "\n";
	sublatticeRepsToIndexVec.resize(allRepsSize*allRepsSize*xMax*yMax*2,-1);
	cout << "sublatticeRepsToIndexVec.size()=" << sublatticeRepsToIndexVec.size() << "\n";
	cout << "orbitSize=" << orbitIndex << "\n";

	time_t startingTime = time(NULL);
	{
		for(unsigned int rIndex=0;rIndex<allReps.size();rIndex++){
			if(rIndex%1000==0){
				cout << rIndex << "/" << allReps.size() << "\n";
				cout << "repIndex=" << repIndex << "\n";
				cout << "sublatticeRepsToIndexVec.size()=" << sublatticeRepsToIndexVec.size() << "\n";
				cout << "indexToRepresentative.size()=" << indexToRepresentative.size() << "\n";
			}
			if(rValidVec[rIndex]){
				subStateInt r=allReps[rIndex];
				int rSz=rSzVec[rIndex];
				// loop over all r'
				for(unsigned int sIndex=0;sIndex<allReps.size();sIndex++){ //only start at rIndex
					if(rValidVec[sIndex]){
						subStateInt s=allReps[sIndex];
						unsigned int minIndexForRepCheck=indexToRepresentative.size();
						if(minIndexForRepCheck>0){
							minIndexForRepCheck--;
						}
						// create all different T(jx,jy)|s>
						list<subStateInt> different_s;
						// loop over all jx,jy, spinFlip
						for(unsigned int jx=0;jx<xMax;jx++){
							for(unsigned int jy=0;jy<yMax;jy++){
								// translate r'
								S sFull=zipperProduct(S(s),S(0));
								S sFullTranslated=LatticeSymmetryClass::applySublatticeTranslationOperator(sFull,jx,jy,false,1);
								getSublatticeStates(sFullTranslated,stateA,stateB);
								if(calculateSz(stateA,numberOfSites/2,SzTemp)){
									if(!useSzSymm || ((rSz+SzTemp)==SzNum)){
										list<subStateInt>::iterator itList;
										for(itList=different_s.begin();itList!=different_s.end();++itList){
											if(*itList==stateA){
												break;
											}
										}
										if(itList==different_s.end()){
											different_s.push_back(stateA);
										}
									}
									if(spinFlip){
										S sFull=zipperProduct(S(s),S(0));
										S sFullTranslated=LatticeSymmetryClass::applySublatticeTranslationOperator(sFull,jx,jy,true,1);
										getSublatticeStates(sFullTranslated,stateA,stateB);
										if(calculateSz(stateA,numberOfSites/2,SzTemp)){
											if(!useSzSymm || ((rSz+SzTemp)==SzNum)){
												list<subStateInt>::iterator itList;
												for(itList=different_s.begin();itList!=different_s.end();++itList){
													if(*itList==stateA){
														break;
													}
												}
												if(itList==different_s.end()){
													different_s.push_back(stateA);
												}
											}
										}
									}
								}
							}
						}
						list<subStateInt>::iterator itList;
						for(itList=different_s.begin();itList!=different_s.end();++itList){
							stateA=*itList;
							// combine states
							S fullRep=zipperProduct(S(r),S(stateA));
							// calculate the norm of this state
							T normQuad;
							LatticeSymmetryClass::getNorm(fullRep,kx,ky,spinFlip,normQuad);
							if(myAbs(normQuad)>zeroEPS){
								// if the norm is nonzero then this state is a valid representative, add to indexToRepresentative
								// if the state of sublattice-a is invariant under a Translation T, then not all different states of sublattice-b are necessary, since some of this states could be created by translation of the full state. Therefore this is not a representative
								// to overcome this issue, check wether it is possible to reach this state through translation of an already existing representative
								bool repExistsAlready=false;
								for(unsigned int i=minIndexForRepCheck;i<indexToRepresentative.size();i++){
									// check if this representative is connected to the actual representative (fullRep)
									repExistsAlready=LatticeSymmetryClass::connectedThroughTranslation(indexToRepresentative[i],fullRep,spinFlip);
									if(repExistsAlready){
										break;
									}
								}
								if(!repExistsAlready){
									unsigned int jxIndex=stateToTranslationNumber_x[stateA];
									unsigned int jyIndex=stateToTranslationNumber_y[stateA];
									unsigned int jSpinFlipIndex=stateToTranslationNumber_spinFlip[stateA];
									indexToRepresentative.push_back(fullRep);
									normalizationFactors.push_back(1/sqrt(real(normQuad)));
									sublatticeRepsToIndexVec[rIndex*allRepsSize*xMax*yMax*2+sIndex*xMax*yMax*2+jxIndex*yMax*2+jyIndex*2+jSpinFlipIndex]=repIndex;
									repIndex++;
								}
							}
						}
					}
				}
			}
		}
	}
	cout << "Processing Time: ";
	printProcessingTime(startingTime);

	cout << "orbitSize=" << orbitIndex << "\n";

	xy=xMax*yMax;
	yxy=xy*yMax;
	xyxy=yxy*xMax;
	orbitxyxy=orbitIndex*xyxy;

	orbitTranslationVec_eq.resize(4*orbitIndex*orbitIndex*xMax*xMax*yMax*yMax);

	cout << "step4\n";
	// 4) the map (ga,gb,jax,jay,jbx,jby) -> (ix,iy,jx,jy) must be created (ga,gb are the orbit numbers of the state)
	// 4a) loop over all ga,gb
	for(unsigned int ga=0;ga<orbitIndex;ga++){
		// 4b) take one representative for each of the orbit numbers
		// get representative for ga
		subStateInt ra=0;
		for(unsigned int i=0;i<stateToOrbitNumber.size();i++){
			if(stateToOrbitNumber[i]==ga){
				ra=allReps[stateToRepresentativeIndex[i]];
				break;
			}
		}
		for(unsigned int gb=0;gb<orbitIndex;gb++){
			// get representative for gb
			subStateInt rb=0;
			for(unsigned int i=0;i<stateToOrbitNumber.size();i++){
				if(stateToOrbitNumber[i]==gb){
					rb=allReps[stateToRepresentativeIndex[i]];
					break;
				}
			}
			S fullState_b=zipperProduct(S(rb),S(0));
			// 4c) loop in reverse order over i and j
			for(int jx=xMax-1;jx>=0;jx--){
				for(int jy=yMax-1;jy>=0;jy--){
					if(spinFlip){
						for(int ix=xMax-1;ix>=0;ix--){
							for(int iy=yMax-1;iy>=0;iy--){
								// j translation of rb
								{ //TODO: this loop structure and the spinFlip-ifs could be easily optimized
									S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,true,1);
									// getsublattice state of b
									subStateInt transA,transB;
									getSublatticeStates(fullStateTranslated_b,transB,transA);
									// i+j translation of full state
									S fullState=zipperProduct(S(ra),S(transB));
									S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,true);
									// get sublattice states
									getSublatticeStates(fullStateTranslated,stateA,stateB);
									// 4d) look up exponents of resulting states |a> and |b>
									unsigned int jax=stateToTranslationNumber_x[stateA];
									unsigned int jay=stateToTranslationNumber_y[stateA];
									bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
									unsigned int jbx=stateToTranslationNumber_x[stateB];
									unsigned int jby=stateToTranslationNumber_y[stateB];
									bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
									// 4e) save this mapping
									tuple6bool tempTuple(ix,iy,true,jx,jy,true);
									orbitTranslationVec_eq[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;}
								{
									S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,true,1);
									// getsublattice state of b
									subStateInt transA,transB;
									getSublatticeStates(fullStateTranslated_b,transB,transA);
									// i+j translation of full state
									S fullState=zipperProduct(S(ra),S(transB));
									S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,false);
									// get sublattice states
									getSublatticeStates(fullStateTranslated,stateA,stateB);
									// 4d) look up exponents of resulting states |a> and |b>
									unsigned int jax=stateToTranslationNumber_x[stateA];
									unsigned int jay=stateToTranslationNumber_y[stateA];
									bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
									unsigned int jbx=stateToTranslationNumber_x[stateB];
									unsigned int jby=stateToTranslationNumber_y[stateB];
									bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
									// 4e) save this mapping
									tuple6bool tempTuple(ix,iy,false,jx,jy,true);
									orbitTranslationVec_eq[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
								}
							}
						}
					}
					for(int ix=xMax-1;ix>=0;ix--){
						for(int iy=yMax-1;iy>=0;iy--){
							// j translation of rb
							if(spinFlip){
								S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,false,1);
								// getsublattice state of b
								subStateInt transA,transB;
								getSublatticeStates(fullStateTranslated_b,transB,transA);
								// i+j translation of full state
								S fullState=zipperProduct(S(ra),S(transB));
								S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,true);
								// get sublattice states
								getSublatticeStates(fullStateTranslated,stateA,stateB);
								// 4d) look up exponents of resulting states |a> and |b>
								unsigned int jax=stateToTranslationNumber_x[stateA];
								unsigned int jay=stateToTranslationNumber_y[stateA];
								bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
								unsigned int jbx=stateToTranslationNumber_x[stateB];
								unsigned int jby=stateToTranslationNumber_y[stateB];
								bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
								// 4e) save this mapping
								tuple6bool tempTuple(ix,iy,true,jx,jy,false);
								orbitTranslationVec_eq[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
							}
							S fullStateTranslated_b=LatticeSymmetryClass::applySublatticeTranslationOperator(fullState_b,jx,jy,false,1);
							// getsublattice state of b
							subStateInt transA,transB;
							getSublatticeStates(fullStateTranslated_b,transB,transA);
							// i+j translation of full state
							S fullState=zipperProduct(S(ra),S(transB));
							S fullStateTranslated=LatticeSymmetryClass::applyTranslationOperator(fullState,ix,iy,false);
							// get sublattice states
							getSublatticeStates(fullStateTranslated,stateA,stateB);
							// 4d) look up exponents of resulting states |a> and |b>
							unsigned int jax=stateToTranslationNumber_x[stateA];
							unsigned int jay=stateToTranslationNumber_y[stateA];
							bool spinFlip_a=stateToTranslationNumber_spinFlip[stateA];
							unsigned int jbx=stateToTranslationNumber_x[stateB];
							unsigned int jby=stateToTranslationNumber_y[stateB];
							bool spinFlip_b=stateToTranslationNumber_spinFlip[stateB];
							// 4e) save this mapping
							tuple6bool tempTuple(ix,iy,false,jx,jy,false);
							orbitTranslationVec_eq[4*(ga*orbitxyxy+gb*xyxy+jax*yxy+jay*xy+jbx*yMax+jby)+2*spinFlip_a+spinFlip_b]=tempTuple;
						}
					}
				}
			}
		}
	}

	cout << "Processing Time: ";
	printProcessingTime(startingTime);
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesWithoutSublatticeCoding(mpq_class const &Sz, int const &kx, int const &ky, int const &spinFlip, bool const &useSzSymm){
	cout << "createRepresentativesWithoutSublatticeCoding\n";
	// create all representatives by just saving everything within the vectors stateTo...
	// 1) create all states for given Sz
	// 1a) the maximum bitRepresentation for n sites
	bool spinFlipBool;
	if(spinFlip){
		spinFlipBool=true;
	}
	else{
		spinFlipBool=false;
	}

	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites);
	stateToRepresentativeIndex.resize(maxNumber,-1);
	stateToTranslationNumber_x.resize(maxNumber);
	stateToTranslationNumber_y.resize(maxNumber);
	stateToTranslationNumber_spinFlip.resize(maxNumber);
	// 1b) loop over all states and check Sz
	mpq_class temp=2*Sz;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_si();
	unsigned int repIndex=0;
	int SzTemp;
	for(subStateInt i=0;i<maxNumber;i++){
		if(calculateSz(i,numberOfSites,SzTemp)){
			if(!useSzSymm || (SzTemp==SzNum)){
				bool spinFlipRep=spinFlipBool;
				unsigned int jxRep,jyRep;
				T normQuad=0;
				LatticeSymmetryClass::getNorm(i,kx,ky,spinFlip,normQuad);
				if(myAbs(normQuad)>zeroEPS){
					subStateInt repState=LatticeSymmetryClass::getRepresentativeAndTranslationNumbers(i,jxRep,jyRep,spinFlipRep,1,false);
					if(repState==i){
						indexToRepresentative.push_back(i);
						normalizationFactors.push_back(1/sqrt(real(normQuad)));
						stateToRepresentativeIndex[i]=repIndex;
						repIndex++;
					}
					else{
						bool foundIndex=false;
						for(unsigned int j=0;j<indexToRepresentative.size();j++){
							if(indexToRepresentative[j]==repState){
								stateToRepresentativeIndex[i]=j;
								foundIndex=true;
								break;
							}
						}
						if(!foundIndex){
							cout << "ERROR: representative not found!\n";
						}

					}
					stateToTranslationNumber_x[i]=jxRep;
					stateToTranslationNumber_y[i]=jyRep;
					stateToTranslationNumber_spinFlip[i]=spinFlipRep;
				}
			}
		}
	}
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesReflectionSymmetry(mpq_class const &Sz, int const &refx, int const &refy, bool const &useSzSymm){
	cout << "createRepresentativesReflectionSymmetry\n";
	// create all representatives by just saving everything within the vectors stateTo...
	// 1) create all states for given Sz
	// 1a) the maximum bitRepresentation for n sites


	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites);
	stateToRepresentativeIndex.resize(maxNumber,-1);
	stateToTranslationNumber_x.resize(maxNumber);
	stateToTranslationNumber_y.resize(maxNumber);
	// 1b) loop over all states and check Sz
	mpq_class temp=2*Sz;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_si();
	unsigned int repIndex=0;
	int SzTemp;
	for(subStateInt i=0;i<maxNumber;i++){
		if(calculateSz(i,numberOfSites,SzTemp)){
			if(!useSzSymm || (SzTemp==SzNum)){
				unsigned int rxRep,ryRep;
				T normQuad=0;
				LatticeSymmetryClass::getNormForReflections(i,refx,refy,normQuad);
				//cout << "normQuad=" << normQuad << "\n";
				if(myAbs(normQuad)>zeroEPS){
					subStateInt repState=LatticeSymmetryClass::getRepresentativeAndReflectionNumbers(i,rxRep,ryRep);
					if(repState==i){
						indexToRepresentative.push_back(i);
						normalizationFactors.push_back(1/sqrt(real(normQuad)));
						stateToRepresentativeIndex[i]=repIndex;
						repIndex++;
					}
					else{
						bool foundIndex=false;
						for(unsigned int j=0;j<indexToRepresentative.size();j++){
							if(indexToRepresentative[j]==repState){
								stateToRepresentativeIndex[i]=j;
								foundIndex=true;
								break;
							}
						}
						if(!foundIndex){
							cout << "ERROR: representative not found!\n";
						}

					}
					stateToTranslationNumber_x[i]=rxRep;
					stateToTranslationNumber_y[i]=ryRep;
				}
			}
		}
	}
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesOnlySz(mpq_class const &Sz){
	cout << "createRepresentativesOnlySz\n";
	// 1) create all states for given Sz
	// 1a) the maximum bitRepresentation for n sites
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites);
	stateToRepresentativeIndex.resize(maxNumber);
	// 1b) loop over all states and check Sz
	mpq_class temp=2*Sz;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_si();
	unsigned int repIndex=0;
	int SzTemp;
	for(subStateInt i=0;i<maxNumber;i++){
		if(calculateSz(i,numberOfSites,SzTemp)){
			if(SzTemp==SzNum){
				indexToRepresentative.push_back(i);
				normalizationFactors.push_back(1);
				stateToRepresentativeIndex[i]=repIndex;
				repIndex++;
			}
		}
	}
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesOnlySzLinTables(mpq_class const &Sz){
	cout << "createRepresentativesOnlySzLinTables\n";
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites/2); // maximum number of sublattice state
	stateToTranslationNumber_x.resize(maxNumber);
	stateToTranslationNumber_y.resize(maxNumber);

	mpq_class temp=2*Sz;
	mpz_class temp2=temp.get_num();
	int SzNum=temp2.get_si();

	int Sza,Szb;
	unsigned int J=0;
	unsigned int subNumberOfSites=numberOfSites/2;
	for(subStateInt ia=0;ia<maxNumber;ia++){
		if(calculateSz(ia,subNumberOfSites,Sza)){
			unsigned int Ja=J;
			unsigned int Jb=0;
			for(subStateInt ib=0;ib<maxNumber;ib++){
				if(calculateSz(ib,subNumberOfSites,Szb)){
					if(Sza+Szb==SzNum){
						stateToTranslationNumber_x[ia]=Ja;
						stateToTranslationNumber_y[ib]=Jb;
						S fullState=zipperProduct(ia,ib);
						indexToRepresentative.push_back(fullState);
						normalizationFactors.push_back(1);
						Jb++;
						J++;
					}
				}
			}
		}
	}

	// test output
	cout << "Ja:\n";
	for(unsigned int i=0;i<stateToTranslationNumber_x.size();i++){
		cout << stateToTranslationNumber_x[i] << "\n";
	}
	cout << "Jb:\n";
	for(unsigned int i=0;i<stateToTranslationNumber_y.size();i++){
		cout << stateToTranslationNumber_y[i] << "\n";
	}

}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesNoSymm(){
	cout << "createRepresentativesNoSymm\n";
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites);
	indexToRepresentative.resize(maxNumber);
	for(subStateInt i=0;i<maxNumber;i++){
		indexToRepresentative[i]=i;
	}
	normalizationFactors.resize(maxNumber,1);
}

template<class S,class T>
void SpinStateTemplate<S,T>::createRepresentativesNoSymmPartialBits(){
	// if the Hilbert Space of a single Site only needs partial bits, then not all possible integer numbers are realized
	// the local hilbert space has 2*S+1 possible states
	// if this number is not a power of 2 then this method is called
	// to check if a state is "legal" use the calculateSz-function, since this function checks if the Sz-calculation is valid
	cout << "createRepresentativesNoSymmPartialBits\n";
	// 1) create all states for given Sz
	// 1a) the maximum bitRepresentation for n sites
	maxNumber=subStateInt(1)<<(localHilbertSpace*numberOfSites);
	//stateToRepresentativeIndex.resize(maxNumber);
	// 1b) loop over all states and check Sz
	unsigned int repIndex=0;
	int SzTemp;
	for(subStateInt i=0;i<maxNumber;i++){
		if(calculateSz(i,numberOfSites,SzTemp)){ // check if state is valid
			indexToRepresentative.push_back(i);
			//stateToRepresentativeIndex[i]=repIndex;
			//repIndex++;
		}
	}
	normalizationFactors.resize(indexToRepresentative.size(),1);
}

template<class S,class T>
void SpinStateTemplate<S,T>::applyLocalSpinHamiltonian(SpinStateTemplate<S,T> &outputState, vector<LocalSpinModelClassTemplate<S,T> > const &localSpinModel, list<pair<vector<unsigned short>,unsigned int> > const &siteCombinationsToReferenceSpinModel, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const &refy){
	cout << "applyH:\n";
	time_t startingTime = time(NULL);
	// applies the referenceSpinModels to the SpinHamiltonian

	// reassign zeros to outputState
	outputState.stateCont.assign(stateCont.size(),T(0));


	vector<vector<T> > newStateConts(numberOfThreads-1);
	for(unsigned int i=0;i<numberOfThreads-1;i++){
		newStateConts[i].resize(this->stateCont.size(),T(0));
	}


	#pragma omp parallel for
	for(unsigned int i=0;i<this->stateCont.size();i++){
		if(myAbs(this->stateCont[i])>zeroEPS){
			T prefactor=this->stateCont[i]/T(this->normalizationFactors[i]);
			list<pair<vector<unsigned short>,unsigned int> >::const_iterator itSiteComb;
			for(itSiteComb=siteCombinationsToReferenceSpinModel.begin();itSiteComb!=siteCombinationsToReferenceSpinModel.end();++itSiteComb){
				if(omp_get_thread_num()==0){
					localSpinModel[itSiteComb->second].applyLocalSpinModelOnBitStateOnSites(outputState.stateCont, this->indexToRepresentative[i], prefactor, itSiteComb->first, kx,ky,spinFlip,refx,refy,repMethod);
				}
				else{
					localSpinModel[itSiteComb->second].applyLocalSpinModelOnBitStateOnSites(newStateConts[omp_get_thread_num()-1], this->indexToRepresentative[i], prefactor, itSiteComb->first, kx,ky,spinFlip,refx,refy,repMethod);
				}
			}
		}
	}


	for(unsigned int i=0;i<numberOfThreads-1;i++){
		for(unsigned int j=0;j<newStateConts[0].size();j++){
			outputState.stateCont[j]+=newStateConts[i][j];
		}
	}

    cout << "Processing Time: ";
	printProcessingTime(startingTime);
}

template<class S,class T>
void SpinStateTemplate<S,T>::applyLocalOperator(SpinStateTemplate<S,T> &outputState, LocalSpinModelClassTemplate<S,T> const &localOperator, vector<vector<unsigned short> > const &siteCombinationsToOperatorModel, int const &kx, int const &ky, int const &spinFlip, int const &refx, int const &refy){

	// reassign zeros to outputState
	outputState.stateCont.assign(stateCont.size(),T(0));

	//applies the localOperator to the state


	vector<vector<T> > newStateConts(numberOfThreads-1);
	for(unsigned int i=0;i<numberOfThreads-1;i++){
		newStateConts[i].resize(this->stateCont.size(),T(0));
	}


	#pragma omp parallel for
	for(unsigned int i=0;i<this->stateCont.size();i++){
		if(myAbs(this->stateCont[i])>zeroEPS){
			T prefactor=this->stateCont[i]/T(this->normalizationFactors[i]);
			for(unsigned int siteCombIndex=0;siteCombIndex<siteCombinationsToOperatorModel.size();siteCombIndex++){
				if(omp_get_thread_num()==0){
					localOperator.applyLocalSpinModelOnBitStateOnSites(outputState.stateCont,this->indexToRepresentative[i],prefactor,siteCombinationsToOperatorModel[siteCombIndex],kx,ky,spinFlip,refx,refy,repMethod);
				}
				else{
					localOperator.applyLocalSpinModelOnBitStateOnSites(newStateConts[omp_get_thread_num()-1],this->indexToRepresentative[i],prefactor,siteCombinationsToOperatorModel[siteCombIndex],kx,ky,spinFlip,refx,refy,repMethod);
				}
			}
		}
	}


	for(unsigned int i=0;i<numberOfThreads-1;i++){
		for(unsigned int j=0;j<newStateConts[0].size();j++){
			outputState.stateCont[j]+=newStateConts[i][j];
		}
	}

}

template<class S,class T>
SpinStateTemplate<S,T>& SpinStateTemplate<S,T>::operator+=(SpinStateTemplate<S,T> const &s){
	if(stateCont.size()!=s.stateCont.size()){
		cout << "ERROR in operator+= : vectors are not the same size!\n";
		return *this;
	}
	for(unsigned int i=0;i<stateCont.size();i++){
		stateCont[i]+=s.stateCont[i];
	}
    return *this;
}
template<class S,class T>
SpinStateTemplate<S,T>& SpinStateTemplate<S,T>::operator-=(SpinStateTemplate<S,T> const &s){
	if(stateCont.size()!=s.stateCont.size()){
		cout << "ERROR in operator+= : vectors are not the same size!\n";
		return *this;
	}
	for(unsigned int i=0;i<stateCont.size();i++){
		stateCont[i]-=s.stateCont[i];
	}
    return *this;
}
template<class S,class T>
SpinStateTemplate<S,T>& SpinStateTemplate<S,T>::operator*=(T const &x){
	for(unsigned int i=0;i<stateCont.size();i++){
		stateCont[i]*=x;
	}
    return *this;
}
template<class S,class T>
SpinStateTemplate<S,T>& SpinStateTemplate<S,T>::operator/=(T const &x){
	if(myAbs(x)>zeroEPS){
		for(unsigned int i=0;i<stateCont.size();i++){
			stateCont[i]/=x;
		}
	}
	else{
		cout << "ERROR: StateTemplate<T>& StateTemplate<T>::operator/=(T const &x): Divison by Zero!\n";
	}
	return *this;
}

//// friend functions implementations ////
template<class S,class T>
ostream& operator<<(ostream &o, SpinStateTemplate<S,T> const &s){
	bool showposset=false;
	ios::fmtflags f;
	f = cout.flags();
	if(f & ios::showpos){
		showposset=true;
	}

	for(unsigned int i=0;i<s.stateCont.size();i++){
		o << showpos;
		o << s.stateCont[i] << "|" << noshowpos;
		o << s.indexToRepresentative[i];
		o << ">";
	}
	if(showposset){
		o << showpos;
	}

	return o;
}

template<class S,class T>
T operator*(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2){
	// calculates the scalar Product <s1|s2>
	T x=0;
	if((s1.stateCont.size()==s2.stateCont.size())){
		for(unsigned int i=0;i<s1.stateCont.size();i++){
			x+=conj(s1.stateCont[i])*s2.stateCont[i];
		}
	}

	return x;
}

template<class T,class S>
bool operator==(SpinStateTemplate<T,S> const& lhs, SpinStateTemplate<T,S> const& rhs){
	if(lhs.stateCont.size()!=rhs.stateCont.size()){
		return false;
	}
	else{
		for(unsigned int i=0;i<lhs.stateCont.size();i++){
			if(myAbs(lhs.stateCont[i]-rhs.stateCont[i])>zeroEPS){
				return false;
			}
		}
	}
	return true;
}

template<class T,class S>
bool operator != (SpinStateTemplate<T,S> const& lhs, SpinStateTemplate<T,S> const& rhs){
	return !(lhs==rhs);
}

//// non class/friend functions implementations ////
template<class S,class T>
const SpinStateTemplate<S,T> operator+(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2){
	SpinStateTemplate<S,T> s=s1;
    s+=s2;
    return s;
}
template<class S,class T>
const SpinStateTemplate<S,T> operator-(SpinStateTemplate<S,T> const &s1, SpinStateTemplate<S,T> const &s2){
	SpinStateTemplate<S,T> s=s1;
    s-=s2;
    return s;
}
template<class S,class T>
const SpinStateTemplate<S,T> operator*(SpinStateTemplate<S,T> const &s, T const &x){
	SpinStateTemplate<S,T> s2=s;
    s2*=x;
    return s2;
}
template<class S,class T>
const SpinStateTemplate<S,T> operator*(T const &x, SpinStateTemplate<S,T> const &s){
	SpinStateTemplate<S,T> s2=s;
    s2*=x;
    return s2;
}
template<class S,class T>
const SpinStateTemplate<S,T> operator/(SpinStateTemplate<S,T> const &s, T const &x){
	SpinStateTemplate<S,T> s2=s;
    s2/=x;
    return s2;
}


typedef SpinStateTemplate<stateInt,Complex> SpinStateComplex;

#endif /* SPINSTATETEMPLATE_H_ */
